(* Parse tree type and related signature *)

(* This must be on top level *)
(*  type Pos = int * int *)
    
signature ATHENA_PARSE_TREE =
  sig
    exception SympParseInternalError of string
    exception SympBug of string

    type Str = Str.Str

    datatype ParseTree = 
	Protocol of Pos.Pos * ParseTree * ParseTree list
      | Program of Pos.Pos * ParseTree list
      | Id of Pos.Pos * string
      | UniqueId of Pos.Pos * string * string
      | TypedId of Pos.Pos * string * string * ParseTree
      | Role of Pos.Pos * { name: ParseTree,
			    params: ParseTree list option,
			    body: ParseTree }
      | Parameter of Pos.Pos * (ParseTree list) * ParseTree
      | New of Pos.Pos * ParseTree
      | MessageType of Pos.Pos
      | NonceType of Pos.Pos
(*        | FreshNonceType of Pos.Pos *)
      | PrincipalType of Pos.Pos
(*        | SelfType of Pos.Pos *)
      | KeyType of Pos.Pos
      | PrivKeyType of Pos.Pos
      | PubKeyType of Pos.Pos
      | SymKeyType of Pos.Pos
      (* Tuple of messages *)
      | TupleType of Pos.Pos * (ParseTree list)
	(* EncryptedType(tp) is the type of an encrypted message of type tp.
	   It shouldn't match any other type except MessageType. *)
      | EncryptedType of Pos.Pos * ParseTree
      | Pub of Pos.Pos * ParseTree
      | Priv of Pos.Pos * ParseTree
      (* Secret symmetric key shared among the list of principals *)
      | SymKey of Pos.Pos * (ParseTree list)
	(* Inverse of the key *)
      | InvKey of Pos.Pos * ParseTree
      | RoleBody of Pos.Pos * ParseTree list
      | Labeled of ParseTree * ParseTree
      | Send of Pos.Pos * ParseTree
      | Receive of Pos.Pos * ParseTree
      | MessageTuple of Pos.Pos * ParseTree list
      | MessageConcur of Pos.Pos * ParseTree list
      | Crypt of Pos.Pos * ParseTree * ParseTree
      (* | Decrypt of Pos.Pos * ParseTree * ParseTree *)
      | Predicate of Pos.Pos * {name : ParseTree,
				params: ParseTree list option,
				def: ParseTree}
      | Theorem of Pos.Pos * {name : ParseTree,
			      params: ParseTree list option,
			      def: ParseTree}
      | PredicateContext of Pos.Pos * {name : ParseTree,
				params: ParseTree list option,
				def: ParseTree} * ParseTree list
      | TheoremContext of Pos.Pos * {name : ParseTree,
			      params: ParseTree list option,
			      def: ParseTree} * ParseTree list
      | Penetrator of Pos.Pos * ParseTree
      | Not of Pos.Pos * ParseTree
      | Or of Pos.Pos * ParseTree * ParseTree
      | And of Pos.Pos * ParseTree * ParseTree
      | Implies of Pos.Pos * ParseTree * ParseTree
      | Atomic of Pos.Pos * ParseTree list
      | FormulaInstance of Pos.Pos * ParseTree * ParseTree list option
      | RoleLabel of Pos.Pos * ParseTree * ParseTree list option * ParseTree
      | RoleStrand of Pos.Pos * ParseTree * ParseTree list option
      | Range of Pos.Pos * ParseTree * ParseTree

    val pos: ParseTree -> Pos.Pos
    val pos2string : Pos.Pos -> string

    (* Two pretty-printers.  pt2string does not expand roles/theorems,
       while pt2stringVerbose does.  pt2string calls pt2stringVerbose
       for top-level definitions.  *)
    val pt2string : ParseTree -> string
    val pt2stringVerbose : ParseTree -> string

    (* Equality up to pos fields. *)
    val ptEq : (ParseTree * ParseTree) -> bool

    (* Return the list of subtrees. *)
    val ptSubtrees: ParseTree -> ParseTree list
    (* Apply `f' to all the immediate children of `pt' and wrap it
       back into the same constructor.  Do not recurse any deeper.
       The `Context' version does NOT descend into contexts. *)
    val ptTransformOne : (ParseTree -> ParseTree) -> ParseTree -> ParseTree
    val ptTransformContextOne : (ParseTree -> ParseTree) -> ParseTree -> ParseTree

    (* ptTransform recursively applies a function to all of a parsetree,
       descending as needed.  ptTransformContext does the same, except
       it does NOT descend into contexts. *)
    val ptTransform : (ParseTree -> ParseTree) -> ParseTree -> ParseTree
    val ptTransformContext : (ParseTree -> ParseTree) -> ParseTree -> ParseTree

    (* instantiate instantiates a Role, Theorem, or Predicate with its
       actual parameters in place of its formals. *)
    val instantiate : (ParseTree * ParseTree list option * 
		       ParseTree list option) -> ParseTree

    (* expand replaces all FormulaInstances with their representations. *)
    val expand : (ParseTree list * ParseTree) -> ParseTree

    (* extractRoleParams extracts from a Role the role itself and its
       parameters, and raises an exception when given a non-Role argument. 
       extractSomeRoleParams does the same, except it "de-options" the
       parameters. *)

    val extractRoleParams : ParseTree -> (ParseTree * ParseTree list option)

    val extractSomeRoleParams : ParseTree -> (ParseTree * ParseTree list)

    (* subterm(m1, m2): Checks if m1 is a subterm of m2 *)
    val subterm: ParseTree * ParseTree -> bool

  end
