functor AthenaParseTreeFun(structure Pos: POS): ATHENA_PARSE_TREE =
  struct
    open Pos
    open Str
    open ListOps
    open SympBug

    exception SympParseInternalError of string

    datatype ParseTree = 
	Protocol of Pos * ParseTree * ParseTree list
      | Program of Pos * ParseTree list
      | Id of Pos * string
      | UniqueId of Pos * string * string
      | TypedId of Pos.Pos * string * string * ParseTree
      | Role of Pos.Pos * { name: ParseTree,
			    params: ParseTree list option,
			    body: ParseTree }
      | Parameter of Pos.Pos * (ParseTree list) * ParseTree
      | New of Pos.Pos * ParseTree
      | MessageType of Pos.Pos
      | NonceType of Pos.Pos
(*      | FreshNonceType of Pos.Pos *)
      | PrincipalType of Pos.Pos
(*      | SelfType of Pos.Pos *)
      | KeyType of Pos.Pos
      | PrivKeyType of Pos.Pos
      | PubKeyType of Pos.Pos
      | SymKeyType of Pos.Pos
      (* Tuple of messages *)
      | TupleType of Pos.Pos * (ParseTree list)
	(* EncryptedType(tp) is the type of an encrypted message of type tp.
	   It shouldn't match any other type except MessageType. *)
      | EncryptedType of Pos.Pos * ParseTree
      | Pub of Pos.Pos * ParseTree
      | Priv of Pos.Pos * ParseTree
      (* Secret symmetric key shared among the list of principals *)
      | SymKey of Pos.Pos * (ParseTree list)
	(* Inverse of the key *)
      | InvKey of Pos.Pos * ParseTree
      | RoleBody of Pos.Pos * ParseTree list
      | Send of Pos.Pos * ParseTree
      | Receive of Pos.Pos * ParseTree
      | Labeled of ParseTree * ParseTree
      | MessageTuple of Pos.Pos * ParseTree list
      | MessageConcur of Pos.Pos * ParseTree list
      | Crypt of Pos.Pos * ParseTree * ParseTree
      (* | Decrypt of Pos.Pos * ParseTree * ParseTree *)
      | Predicate of Pos.Pos * {name : ParseTree,
				params: ParseTree list option,
				def: ParseTree}
      | Theorem of Pos.Pos * {name : ParseTree,
			      params: ParseTree list option,
			      def: ParseTree}
      | PredicateContext of Pos.Pos * {name : ParseTree,
				params: ParseTree list option,
				def: ParseTree} * ParseTree list
      | TheoremContext of Pos.Pos * {name : ParseTree,
			      params: ParseTree list option,
			      def: ParseTree} * ParseTree list
      | Penetrator of Pos.Pos * ParseTree
      | Not of Pos.Pos * ParseTree
      | Or of Pos.Pos * ParseTree * ParseTree
      | And of Pos.Pos * ParseTree * ParseTree
      | Implies of Pos.Pos * ParseTree * ParseTree
      | Atomic of Pos.Pos * ParseTree list
      | FormulaInstance of Pos.Pos * ParseTree * ParseTree list option
      | RoleLabel of Pos.Pos * ParseTree * ParseTree list option * ParseTree
      | RoleStrand of Pos.Pos * ParseTree * ParseTree list option
      | Range of Pos.Pos * ParseTree * ParseTree

    fun pos (Protocol(p,_,_)) = p
      | pos (Program(p,_)) = p
      | pos (Id(p,_)) = p
      | pos (UniqueId(p,_,_)) = p
      | pos (TypedId(p,_,_,_)) = p
      | pos (Role(p,_)) = p
      | pos (Parameter(p,_,_)) = p
      | pos (New (p,_)) = p
      | pos (MessageType(p)) = p
      | pos (NonceType(p)) = p
(*        | pos (FreshNonceType(p)) = p *)
      | pos (PrincipalType(p)) = p
(*        | pos (SelfType(p)) = p *)
      | pos (KeyType(p)) = p
      | pos (PrivKeyType p) = p
      | pos (PubKeyType p) = p
      | pos (SymKeyType p) = p
      | pos (TupleType (p, _)) = p
      | pos (EncryptedType(p, _)) = p
      | pos (Pub(p,_)) = p
      | pos (Priv(p,_)) = p
      | pos (SymKey(p,_)) = p
      | pos (InvKey(p,_)) = p
      | pos (RoleBody(p,_)) = p
      | pos (Send(p,_)) = p
      | pos (Receive(p,_)) = p
      | pos (Labeled(i,_)) = pos i
      | pos (MessageTuple(p,_)) = p
      | pos (MessageConcur(p,_)) = p
      | pos (Crypt(p,_,_)) = p
      (* | pos (Decrypt(p,_,_)) = p *)
      | pos (Predicate (p,_)) = p
      | pos (Theorem (p,_)) = p
      | pos (PredicateContext (p,_,_)) = p
      | pos (TheoremContext (p,_,_)) = p
      | pos (Penetrator (p,_)) = p
      | pos (Atomic (p, _)) = p
      | pos (Not (p,_)) = p
      | pos (And (p,_,_)) = p
      | pos (Or (p,_,_)) = p
      | pos (Implies (p,_,_)) = p
      | pos (RoleLabel (p,_,_,_)) = p
      | pos (RoleStrand (p,_,_)) = p
      | pos (FormulaInstance (p,_,_)) = p
      | pos (Range (p,_,_)) = p

    fun pos2string (l,c) = (Int.toString l)^"."^(Int.toString c)
      
    (* fun pa2string [] = ""
      | pa2string [h] = 
      (pt2string h)
      | pa2string (h::t) =
      (pt2string h) ^ "; " ^ (pa2string t) *)

    fun params2string pa =
      "[" ^ (strlistmap2str "; " pt2string pa) ^ "]"

    (* Two pretty-printers.  pt2string does not expand roles/theorems,
       while pt2stringVerbose does.  pt2string calls pt2stringVerbose
       for top-level definitions.  *)  
    and pt2stringVerbose (Protocol (_, name, decls)) =
      "protocol " ^ (pt2string name) ^ " =\nbegin\n\n" ^
      (List.foldr (fn (x, y) => x ^ "\n\n" ^ y) "" 
       (List.map pt2stringVerbose decls))
      ^ "end"
      | pt2stringVerbose (Program (_, decls)) =
      List.foldr (fn (x, y) => x ^ "\n\n" ^ y) "" (List.map pt2string decls)
      | pt2stringVerbose (Id (_, id)) = id
      | pt2stringVerbose (UniqueId (_, uid, id)) = id ^ ("("^uid^")")
      | pt2stringVerbose (TypedId (_, uid, id, tp)) = 
          "TypedId("^id ^ "("^uid^"):"^ (pt2string tp)^")"
      | pt2stringVerbose (Role (_, {name=n,params=SOME pa,body=b})) =
      "role " ^ (pt2string n) ^ " " ^ (params2string pa) ^ " =\n" ^ 
      (pt2string b)
      | pt2stringVerbose (Role (_, {name=n,params=NONE,body=b})) =
      "role " ^ (pt2string n) ^ " =\n" ^ (pt2string b)
      | pt2stringVerbose (Parameter (_, l, tp)) = 
      "Parameter["^(strlist2str ", " (List.map pt2string l)) ^ "]: " ^ (pt2string tp)
      | pt2stringVerbose (NonceType _) = "Nonce"
(*        | pt2stringVerbose (FreshNonceType _) = "FreshNonce" *)
      | pt2stringVerbose (PrincipalType _) = "Principal"
(*        | pt2stringVerbose (SelfType _) = "Self" *)
      | pt2stringVerbose (KeyType _) = "Key"
      | pt2stringVerbose (PrivKeyType _) = "PrivKey"
      | pt2stringVerbose (PubKeyType _) = "PubKey"
      | pt2stringVerbose (SymKeyType _) = "SymKey"
      | pt2stringVerbose (TupleType(_, lst)) =
         "("^(strlistmap2str " * " pt2stringVerbose lst)^")"
      | pt2stringVerbose (EncryptedType(_, tp)) = "EncryptedType("^(pt2stringVerbose tp)^")"
      | pt2stringVerbose (New(_, pt)) = "New " ^ (pt2stringVerbose pt)
      | pt2stringVerbose (MessageType _) = "Message"
      | pt2stringVerbose (RoleBody (_, l)) = "begin\n" ^
      (List.foldr (fn (x, y) => x ^ "\n" ^ y) "")
       (List.map (fn s => "  " ^ s) (List.map pt2string l)) ^ "end"
      | pt2stringVerbose (Labeled (label, act)) = (pt2string label) ^ ": " ^
      (pt2string act)
      | pt2stringVerbose (Send (_, pt)) = "send " ^ (pt2string pt)
      | pt2stringVerbose (Receive (_, pt)) = "receive " ^ (pt2string pt)
      | pt2stringVerbose (Pub (_, pt)) = "(PK " ^ (pt2string pt) ^ ")"
      | pt2stringVerbose (Priv (_, pt)) = "(PVK " ^ (pt2string pt) ^ ")"
      | pt2stringVerbose (SymKey(_, lst)) =
         "SymKey("^(strlistmap2str ", " pt2stringVerbose lst)^")"
      | pt2stringVerbose (InvKey(_, pt)) = "(INV "^(pt2string pt)^")"
      | pt2stringVerbose (Crypt (_, m, k)) = "{" ^ (pt2string m) ^ "} " ^ 
      (pt2string k)
      (* | pt2stringVerbose (Decrypt (_, m, k)) = "decrypt (" ^ (pt2string m) ^ 
      ", " ^ (pt2string k) ^ ")" *)
      | pt2stringVerbose (MessageTuple (_, l)) = 
      "(" ^ strlist2str ", " (List.map pt2string l) ^ ")"
      | pt2stringVerbose (MessageConcur (_, l)) =
      (strlist2str "\n    and " (List.map pt2string l))
      | pt2stringVerbose (Theorem (_, {name=n,params=SOME pa,def=d})) =
      "theorem " ^ (pt2string n) ^ " " ^ (params2string pa) ^ " =\n  " ^
      (pt2string d)
      | pt2stringVerbose (Theorem (_, {name=n,params=NONE,def=d})) =
      "theorem " ^ (pt2string n) ^ " =\n  " ^ (pt2string d)
      | pt2stringVerbose (Predicate (_, {name=n,params=SOME pa,def=d})) =
      "predicate " ^ (pt2string n) ^ " " ^ (params2string pa) ^ " =\n  " ^
      (pt2string d)
      | pt2stringVerbose (Predicate (_, {name=n,params=NONE,def=d})) =
      "predicate " ^ (pt2string n) ^ " =\n  " ^ (pt2string d)
      | pt2stringVerbose (TheoremContext (_, {name=n,params=SOME pa,def=d}, c)) =
      "theorem " ^ (pt2string n) ^ " " ^ (params2string pa) ^ " =\n  " ^
      (pt2string d) 
      ^ "\nContext: " ^ strlist2str ", " (List.map pt2string c)
      | pt2stringVerbose (TheoremContext (_, {name=n,params=NONE,def=d}, c)) =
      "theorem " ^ (pt2string n) ^ " =\n  " ^ (pt2string d)
      ^ "\nContext: " ^ strlist2str ", " (List.map pt2string c)
      | pt2stringVerbose (PredicateContext (_, {name=n,params=SOME pa,def=d}, 
					    c)) =
      "predicate " ^ (pt2string n) ^ " " ^ (params2string pa) ^ " =\n  " ^
      (pt2string d)
      ^ "\nContext: " ^ strlist2str ", " (List.map pt2string c)
      | pt2stringVerbose (PredicateContext (_, {name=n,params=NONE,def=d}, c)) =
      "predicate " ^ (pt2string n) ^ " =\n  " ^ (pt2string d)
      ^ "\nContext: " ^ strlist2str ", " (List.map pt2string c)
      | pt2stringVerbose (Penetrator (_, id)) =
      "penetrator " ^ (pt2string id)
      | pt2stringVerbose (Not (_, pt)) = "not (" ^ (pt2string pt) ^ ")"
      | pt2stringVerbose (Or (_, pt1, pt2)) = "(" ^ (pt2string pt1) ^ 
      ") or (" ^ (pt2string pt2) ^ ")"
      | pt2stringVerbose (And (_, pt1, pt2)) = "(" ^ (pt2string pt1) ^ 
      ") and (" ^ (pt2string pt2) ^ ")"
      | pt2stringVerbose (Implies (_, pt1, pt2)) = "(" ^ (pt2string pt1) ^ 
      ") -> (" ^ (pt2string pt2) ^ ")"
      | pt2stringVerbose (Atomic (_, l)) = "{" ^ (strlist2str ", " 
					   (List.map pt2string l)) ^ "}"
      | pt2stringVerbose (RoleLabel (_, role, SOME opts, label)) =
      (pt2string role) ^ "[" ^ (strlist2str ", " (List.map pt2string opts)) ^ 
      "]." ^ (pt2string label)
      | pt2stringVerbose (RoleLabel (_, role, NONE, label)) =
      (pt2string role) ^ "." ^ (pt2string label)
      | pt2stringVerbose (RoleStrand (_, role, SOME opts)) =
      (pt2string role) ^ "[" ^ (strlist2str ", " (List.map pt2string opts)) ^
      "]"
      | pt2stringVerbose (RoleStrand (_, role, NONE)) =
      (pt2string role)
      | pt2stringVerbose (Range (_, l1, l2)) = (pt2string l1) ^ "-" ^
      (pt2string l2)
      | pt2stringVerbose (FormulaInstance (_, pt, SOME opts)) =
      (pt2string pt) ^ "[" ^ (strlist2str ", " (List.map pt2string opts)) ^
      "]"
      | pt2stringVerbose (FormulaInstance (_, pt, NONE)) =
      (pt2string pt)

    and pt2string (Protocol (_, name, decls)) =
      "protocol " ^ (pt2string name) ^ " =\nbegin\n\n" ^
      (List.foldr (fn (x, y) => x ^ "\n\n" ^ y) "" 
       (List.map pt2stringVerbose decls))
      ^ "end"
      | pt2string (Program (_, decls)) =
      (List.foldr (fn (x, y) => x ^ "\n\n" ^ y) "" 
       (List.map pt2string decls))
      | pt2string (Id (_, id)) = id
      | pt2string (UniqueId (_, uid, id)) = (* id ^  "("^uid^")" *) uid
      | pt2string (TypedId (_, uid, id, tp)) =
	 (* id ^ *) "("^uid^": " ^ (pt2string tp) ^")"
      | pt2string (Role (_, {name=n,params=SOME pa,body=b})) =
      (pt2string n)
      | pt2string (Role (_, {name=n,params=NONE,body=b})) =
      (pt2string n)
      | pt2string (Parameter (_, l, tp)) = 
      (strlist2str ", " (List.map pt2string l)) ^ ": " ^ (pt2string tp)
      | pt2string (NonceType _) = "Nonce"
(*        | pt2string (FreshNonceType _) = "FreshNonce" *)
      | pt2string (PrincipalType _) = "Principal"
(*        | pt2string (SelfType _) = "Self" *)
      | pt2string (KeyType _) = "Key"
      | pt2string (PrivKeyType _) = "PrivKey"
      | pt2string (PubKeyType _) = "PubKey"
      | pt2string (SymKeyType _) = "SymKey"
      | pt2string (TupleType(_, lst)) =
         "("^(strlistmap2str " * " pt2string lst)^")"
      | pt2string (EncryptedType(_, tp)) = "EncryptedType("^(pt2string tp)^")"
      | pt2string (New(_,pt)) = "New "^(pt2string pt)
      | pt2string (MessageType _) = "Message"
      | pt2string (RoleBody (_, l)) = "begin\n" ^
      (List.foldr (fn (x, y) => x ^ "\n" ^ y) "")
       (List.map (fn s => "  " ^ s) (List.map pt2string l)) ^ "end"
      | pt2string (Labeled (label, act)) = (pt2string label) ^ ": " ^
      (pt2string act)
      | pt2string (Send (_, pt)) = "send " ^ (pt2string pt)
      | pt2string (Receive (_, pt)) = "receive " ^ (pt2string pt)
      | pt2string (Pub (_, pt)) = "(PK " ^ (pt2string pt) ^ ")"
      | pt2string (Priv (_, pt)) = "(PVK " ^ (pt2string pt) ^ ")"
      | pt2string (SymKey(_, lst)) = "SymKey("^(strlistmap2str ", " pt2string lst)^")"
      | pt2string (InvKey (_, pt)) = "(INV " ^ (pt2string pt) ^ ")"
      | pt2string (Crypt (_, m, k)) = "{" ^ (pt2string m) ^ "} " ^ 
      (pt2string k)
      (* | pt2string (Decrypt (_, m, k)) = "decrypt (" ^ (pt2string m) ^ 
      ", " ^ (pt2string k) ^ ")" *)
      | pt2string (MessageTuple (_, l)) = 
      "(" ^ strlist2str ", " (List.map pt2string l) ^ ")"
      | pt2string (MessageConcur (_, l)) =
      (strlist2str "\n    and " (List.map pt2string l))
      | pt2string (Theorem (_, {name=n,params=SOME pa,def=d})) =
      (pt2string n)
      | pt2string (Theorem (_, {name=n,params=NONE,def=d})) =
      (pt2string n)
      | pt2string (Predicate (_, {name=n,params=SOME pa,def=d})) =
      (pt2string n)
      | pt2string (Predicate (_, {name=n,params=NONE,def=d})) =
      (pt2string n)
      | pt2string (TheoremContext (_, {name=n,params=SOME pa,def=d}, c)) =
      (pt2string n)
      | pt2string (TheoremContext (_, {name=n,params=NONE,def=d}, c)) =
      (pt2string n)
      | pt2string (PredicateContext (_, {name=n,params=SOME pa,def=d}, 
					    c)) =
      (pt2string n)
      | pt2string (PredicateContext (_, {name=n,params=NONE,def=d}, c)) =
      (pt2string n)
      | pt2string (Penetrator (_, id)) =
      "penetrator " ^ (pt2string id)
      | pt2string (Not (_, pt)) = "not (" ^ (pt2string pt) ^ ")"
      | pt2string (Or (_, pt1, pt2)) = "(" ^ (pt2string pt1) ^ ") or (" ^
      (pt2string pt2) ^ ")"
      | pt2string (And (_, pt1, pt2)) = "(" ^ (pt2string pt1) ^ ") and (" ^
      (pt2string pt2) ^ ")"
      | pt2string (Implies (_, pt1, pt2)) = "(" ^ (pt2string pt1) ^ ") -> (" ^
      (pt2string pt2) ^ ")"
      | pt2string (Atomic (_, l)) = "{" ^ (strlist2str ", " 
					   (List.map pt2string l)) ^ "}"
      | pt2string (RoleLabel (_, role, SOME opts, label)) =
      (pt2string role) ^ "[" ^ (strlist2str ", " (List.map pt2string opts)) ^ 
      "]." ^ (pt2string label)
      | pt2string (RoleLabel (_, role, NONE, label)) =
      (pt2string role) ^ "." ^ (pt2string label)
      | pt2string (RoleStrand (_, role, SOME opts)) =
      (pt2string role) ^ "[" ^ (strlist2str ", " (List.map pt2string opts)) ^
      "]"
      | pt2string (RoleStrand (_, role, NONE)) =
      (pt2string role)
      | pt2string (Range (_, l1, l2)) = (pt2string l1) ^ "-" ^
      (pt2string l2)
      | pt2string (FormulaInstance (_, pt, SOME opts)) =
      (pt2string pt) ^ "[" ^ (strlist2str ", " (List.map pt2string opts)) ^
      "]"
      | pt2string (FormulaInstance (_, pt, NONE)) =
      (pt2string pt)

    fun ptListEq ([], []) = true
      | ptListEq (h::t, h'::t') = (ptEq (h, h')) andalso (ptListEq (t, t'))
      | ptListEq _ = false

    and ptEq (Protocol (_, pt, ptl), Protocol (_, pt', ptl')) =
      (ptEq (pt, pt')) andalso (ptListEq (ptl, ptl'))
      | ptEq (Program (_, ptl), Program (_, ptl')) =
      ptListEq (ptl, ptl')
      | ptEq (Id (_, s), Id (_, s')) = (s = s')
      | ptEq (UniqueId (_, s1, s2), UniqueId (_, s1', s2')) = 
      (s1 = s1') andalso (s2 = s2')
      | ptEq (TypedId (_, s1, s2, t), TypedId (_, s1', s2', t')) = 
      (s1 = s1') andalso (s2 = s2') andalso (ptEq (t, t'))
      | ptEq (Role (_, {name = n, params = SOME ptl, body = pt}),
	      Role (_, {name = n', params = SOME ptl', body = pt'})) =
      (ptEq (n, n')) andalso (ptListEq (ptl, ptl')) andalso (ptEq (pt, pt'))
      | ptEq (Role (_, {name = n, params = NONE, body = pt}),
	      Role (_, {name = n', params = NONE, body = pt'})) =
      (ptEq (n, n')) andalso (ptEq (pt, pt'))
      | ptEq (Parameter (_, ptl, t), Parameter (_, ptl', t')) =
      (ptListEq (ptl, ptl')) andalso (ptEq (t, t'))
      | ptEq (New(_,t1), New(_,t2)) = ptEq(t1, t2)
      | ptEq (MessageType _, MessageType _) = true
      | ptEq (NonceType _, NonceType _) = true
(*        | ptEq (FreshNonceType _, FreshNonceType _) = true *)
      | ptEq (PrincipalType _, PrincipalType _) = true
(*        | ptEq (SelfType _, SelfType _) = true *)
      | ptEq (KeyType _, KeyType _) = true
      | ptEq (PrivKeyType _, PrivKeyType _) = true
      | ptEq (PubKeyType _, PubKeyType _) = true
      | ptEq (SymKeyType _, SymKeyType _) = true
      | ptEq (TupleType(_, l1), TupleType(_, l2)) = ptListEq(l1, l2)
      | ptEq (EncryptedType(_, t1), EncryptedType(_, t2)) = ptEq(t1, t2)
      | ptEq (Pub (_, pt), Pub (_, pt')) = ptEq (pt, pt')
      | ptEq (Priv (_, pt), Priv (_, pt')) = ptEq (pt, pt')
      | ptEq (SymKey(_, l1), SymKey(_, l2)) = ptListEq(l1, l2)
      | ptEq (InvKey(_, pt), InvKey(_, pt')) = ptEq (pt, pt')
      | ptEq (RoleBody (_, ptl), RoleBody (_, ptl')) = ptListEq (ptl, ptl')
      | ptEq (Labeled (pt1, pt2), Labeled (pt1', pt2')) = 
      (ptEq (pt1, pt1')) andalso (ptEq (pt2, pt2'))
      | ptEq (Send (_, pt), Send (_, pt')) = ptEq (pt, pt')
      | ptEq (Receive (_, pt), Receive (_, pt')) = ptEq (pt, pt')
      | ptEq (MessageTuple (_, ptl), MessageTuple (_, ptl')) = 
      ptListEq (ptl, ptl')
      | ptEq (MessageConcur (_, ptl), MessageConcur (_, ptl')) = 
      ptListEq (ptl, ptl')
      | ptEq (Crypt (_, pt1, pt2), Crypt (_, pt1', pt2')) = 
      (ptEq (pt1, pt1')) andalso (ptEq (pt2, pt2'))
      (* | ptEq (Decrypt (_, pt1, pt2), Decrypt (_, pt1', pt2')) = 
      (ptEq (pt1, pt1')) andalso (ptEq (pt2, pt2')) *)
      | ptEq (Predicate (_, {name = n, params = SOME ptl, def = pt}),
	      Predicate (_, {name = n', params = SOME ptl', def = pt'})) =
      (ptEq (n, n')) andalso (ptListEq (ptl, ptl')) andalso (ptEq (pt, pt'))
      | ptEq (Predicate (_, {name = n, params = NONE, def = pt}),
	      Predicate (_, {name = n', params = NONE, def = pt'})) =
      (ptEq (n, n')) andalso (ptEq (pt, pt'))
      | ptEq (Theorem (_, {name = n, params = SOME ptl, def = pt}),
	      Theorem (_, {name = n', params = SOME ptl', def = pt'})) =
      (ptEq (n, n')) andalso (ptListEq (ptl, ptl')) andalso (ptEq (pt, pt'))
      | ptEq (Theorem (_, {name = n, params = NONE, def = pt}),
	      Theorem (_, {name = n', params = NONE, def = pt'})) =
      (ptEq (n, n')) andalso (ptEq (pt, pt'))
      | ptEq (PredicateContext (_, {name = n, params = SOME ptl1, def = pt}, 
				ptl2),
	      PredicateContext (_, {name = n', params = SOME ptl1', 
				    def = pt'}, ptl2')) =
      (ptEq (n, n')) andalso (ptListEq (ptl1, ptl1')) andalso (ptEq (pt, pt'))
      andalso (ptListEq (ptl2, ptl2'))
      | ptEq (PredicateContext (_, {name = n, params = NONE, def = pt}, ptl),
	      PredicateContext (_, {name = n', params = NONE, def = pt'}, 
				ptl')) =
      (ptEq (n, n')) andalso (ptEq (pt, pt')) andalso (ptListEq (ptl, ptl'))
      | ptEq (TheoremContext (_, {name = n, params = SOME ptl1, def = pt}, 
				ptl2),
	      TheoremContext (_, {name = n', params = SOME ptl1', def = pt'}, 
			 ptl2')) =
      (ptEq (n, n')) andalso (ptListEq (ptl1, ptl1')) andalso (ptEq (pt, pt'))
      andalso (ptListEq (ptl2, ptl2'))
      | ptEq (TheoremContext (_, {name = n, params = NONE, def = pt}, ptl),
	      TheoremContext (_, {name = n', params = NONE, def = pt'}, 
				ptl')) =
      (ptEq (n, n')) andalso (ptEq (pt, pt')) andalso (ptListEq (ptl, ptl'))
      | ptEq (Penetrator (_, pt), Penetrator (_, pt')) = ptEq (pt, pt')
      | ptEq (Not (_, pt), Not (_, pt')) = ptEq (pt, pt')
      | ptEq (Or (_, pt1, pt2), Or (_, pt1', pt2')) =
      (ptEq (pt1, pt1')) andalso (ptEq (pt2, pt2'))
      | ptEq (And (_, pt1, pt2), And (_, pt1', pt2')) =
      (ptEq (pt1, pt1')) andalso (ptEq (pt2, pt2'))
      | ptEq (Implies (_, pt1, pt2), Implies (_, pt1', pt2')) =
      (ptEq (pt1, pt1')) andalso (ptEq (pt2, pt2'))
      | ptEq (Atomic (_, ptl), Atomic (_, ptl')) = 
      ptListEq (ptl, ptl')
      | ptEq (FormulaInstance (_, pt, SOME ptl), FormulaInstance 
	      (_, pt', SOME ptl')) = 
      (ptEq (pt, pt')) andalso (ptListEq (ptl, ptl'))
      | ptEq (FormulaInstance (_, pt, NONE), FormulaInstance 
	      (_, pt', NONE)) = 
      ptEq (pt, pt')
      | ptEq (RoleLabel (_, pt1, SOME ptl, pt2), RoleLabel 
	      (_, pt1', SOME ptl', pt2')) = 
      (ptEq (pt1, pt1')) andalso (ptListEq (ptl, ptl')) 
      andalso (ptEq (pt2, pt2'))
      | ptEq (RoleLabel (_, pt1, NONE, pt2), RoleLabel 
	      (_, pt1', NONE, pt2')) = 
      (ptEq (pt1, pt1')) andalso (ptEq (pt2, pt2'))
      | ptEq (RoleStrand (_, pt, SOME ptl), RoleStrand
	      (_, pt', SOME ptl')) = 
      (ptEq (pt, pt')) andalso (ptListEq (ptl, ptl'))
      | ptEq (RoleStrand (_, pt, NONE), RoleStrand
	      (_, pt', NONE)) = 
      ptEq (pt, pt')
      | ptEq (Range (_, pt1, pt2), Range (_, pt1', pt2')) =
      (ptEq (pt1, pt1')) andalso (ptEq (pt2, pt2'))
      | ptEq _ = false

    (* Return the list of subtrees. *)
    fun ptSubtrees pt =
	(case pt of
	   Protocol (p, pt, ptl) => pt::ptl
	 | Program (p, ptl) => ptl
	 | TypedId (p, uid, id, pt) => [pt]
	 | Role (p, {name = pt1, params = SOME ptl, body = pt2}) => pt1::pt2::ptl
	 | Role (p, {name = pt1, params = NONE, body = pt2}) => [pt1, pt2]
	 | Parameter (p, ptl, pt) => pt::ptl
	 | Pub (p, pt) => [pt]
	 | Priv (p, pt) => [pt]
	 | SymKey(_, l) => l
	 | InvKey(p, pt) => [pt]
	 | RoleBody (p, ptl) => ptl
	 | Labeled (pt1, pt2) => [pt1, pt2]
	 | Send (p, pt) => [pt]
	 | Receive (p, pt) => [pt]
	 | MessageTuple (p, ptl) => ptl
	 | MessageConcur (p, ptl) => ptl
	 | Crypt (p, pt1, pt2) => [pt1, pt2]
	 (* | Decrypt (p, pt1, pt2) => [pt1, pt2] *)
	 | Predicate (p, {name = pt1, params = SOME ptl, def = pt2}) => pt1::pt2::ptl
	 | Predicate (p, {name = pt1, params = NONE, def = pt2}) => [pt1, pt2]
	 | Theorem (p, {name = pt1, params = SOME ptl, def = pt2}) => pt1::pt2::ptl
	 | Theorem (p, {name = pt1, params = NONE, def = pt2}) => [pt1, pt2]
	 | PredicateContext (p, {name = pt1, params = SOME ptl1, def = pt2}, 
			     ptl2) => (pt1::ptl1)@(pt2::ptl2)
	 | PredicateContext (p, {name = pt1, params = NONE, def = pt2}, 
			     ptl) => pt1::pt2::ptl
	 | TheoremContext (p, {name = pt1, params = SOME ptl1, def = pt2}, 
			   ptl2) => (pt1::ptl1)@(pt2::ptl2)
	 | TheoremContext (p, {name = pt1, params = NONE, def = pt2}, 
			   ptl) => pt1::pt2::ptl
	 | Penetrator (p, pt) => [pt]
	 | Not (p, pt) => [pt]
	 | Or (p, pt1, pt2) => [pt1, pt2]
	 | And (p, pt1, pt2) => [pt1, pt2]
	 | Implies (p, pt1, pt2) => [pt1, pt2]
	 | Atomic (p, ptl) => ptl
	 | FormulaInstance (p, pt, SOME ptl) => pt::ptl
	 | FormulaInstance (p, pt, NONE) => [pt]
	 | RoleLabel (p, pt1, SOME ptl, pt2) => pt1::pt2::ptl
	 | RoleLabel (p, pt1, NONE, pt2) => [pt1, pt2]
	 | RoleStrand (p, pt, SOME ptl) => pt::ptl
	 | RoleStrand (p, pt, NONE) => [pt]
	 | Range (p, pt1, pt2) => [pt1, pt2]
	 | x => [])

    (* Apply `f' to all the immediate children of `pt' and wrap it
       back into the same constructor. *)
    fun ptTransformOne f pt =
	(case pt of
	   Protocol (p, pt, ptl) =>
	     Protocol (p, f pt, List.map f ptl)
	 | Program (p, ptl) =>
	     Program (p, List.map f ptl)
	 | TypedId (p, uid, id, pt) =>
	     TypedId (p, uid, id, f pt)
	 | Role (p, {name = pt1, params = SOME ptl, body = pt2}) =>
	     Role (p, {name = f pt1, params = SOME (List.map f ptl),
		       body = f pt2})
	 | Role (p, {name = pt1, params = NONE, body = pt2}) =>
	     Role (p, {name = f pt1, params = NONE, body = f pt2})
	 | Parameter (p, ptl, pt) =>
	     Parameter (p, List.map f ptl, f pt)
	 | Pub (p, pt) =>
	     Pub (p, f pt)
	 | Priv (p, pt) =>
	     Priv (p, f pt)
	 | SymKey(p, l) => SymKey(p, List.map f l)
	 | InvKey(p, pt) => InvKey(p, f pt)
	 | RoleBody (p, ptl) =>
	     RoleBody (p, List.map f ptl)
	 | Labeled (pt1, pt2) =>
	     Labeled (f pt1, f pt2)
	 | Send (p, pt) =>
	     Send (p, f pt)
	 | Receive (p, pt) =>
	     Receive (p, f pt)
	 | MessageTuple (p, ptl) =>
	     MessageTuple (p, List.map f ptl)
	 | MessageConcur (p, ptl) =>
	     MessageConcur (p, List.map f ptl)
	 | Crypt (p, pt1, pt2) =>
	     Crypt (p, f pt1, f pt2)
	 (* | Decrypt (p, pt1, pt2) =>
	     Decrypt (p, f pt1, f pt2) *)
	 | Predicate (p, {name = pt1, params = SOME ptl, def = pt2}) =>
	     Predicate (p, {name = f pt1, params = 
			    SOME (List.map f ptl), def = f pt2})
	 | Predicate (p, {name = pt1, params = NONE, def = pt2}) =>
	     Predicate (p, {name = f pt1, params = NONE, def = f pt2})
	 | Theorem (p, {name = pt1, params = SOME ptl, def = pt2}) =>
	     Theorem (p, {name = f pt1, params = 
			    SOME (List.map f ptl), def = f pt2})
	 | Theorem (p, {name = pt1, params = NONE, def = pt2}) =>
	     Theorem (p, {name = f pt1, params = NONE, def = f pt2})
	 | PredicateContext (p, {name = pt1, params = SOME ptl1, def = pt2}, 
			     ptl2) =>
	     PredicateContext (p, {name = f pt1, params = 
				   SOME (List.map f ptl1), 
				   def = f pt2},
			       List.map f ptl2)
	 | PredicateContext (p, {name = pt1, params = NONE, def = pt2}, 
			     ptl) =>
	     PredicateContext (p, {name = f pt1, params = NONE, 
				   def = f pt2}, List.map f ptl)
	 | TheoremContext (p, {name = pt1, params = SOME ptl1, def = pt2}, 
			   ptl2) =>
	     TheoremContext (p, {name = f pt1, params = 
				 SOME (List.map f ptl1), def = f pt2},
			     List.map f ptl2)
	 | TheoremContext (p, {name = pt1, params = NONE, def = pt2}, 
			   ptl) =>
	     TheoremContext (p, {name = f pt1, params = NONE, 
				 def = f pt2}, List.map f ptl)
	 | Penetrator (p, pt) =>
	     Penetrator (p, f pt)
	 | Not (p, pt) =>
	     Not (p, f pt)
	 | Or (p, pt1, pt2) =>
	     Or (p, f pt1, f pt2)
	 | And (p, pt1, pt2) =>
	     And (p, f pt1, f pt2)
	 | Implies (p, pt1, pt2) =>
	     Implies (p, f pt1, f pt2)
	 | Atomic (p, ptl) =>
	     Atomic (p, List.map f ptl)
	 | FormulaInstance (p, pt, SOME ptl) =>
	     FormulaInstance (p, f pt, SOME (List.map f ptl))
	 | FormulaInstance (p, pt, NONE) =>
	     FormulaInstance (p, f pt, NONE)
	 | RoleLabel (p, pt1, SOME ptl, pt2) =>
	     RoleLabel (p, f pt1, SOME (List.map f ptl), f pt2)
	 | RoleLabel (p, pt1, NONE, pt2) =>
	     RoleLabel (p, f pt1, NONE, f pt2)
	 | RoleStrand (p, pt, SOME ptl) =>
	     RoleStrand (p, f pt, SOME (List.map f ptl))
	 | RoleStrand (p, pt, NONE) =>
	     RoleStrand (p, f pt, NONE)
	 | Range (p, pt1, pt2) =>
	     Range (p, f pt1, f pt2)
	 | x => x)

    (* Apply `f' to all the immediate children of `pt' EXCEPT the
       context, and wrap it back into the same constructor. *)
    fun ptTransformContextOne f pt =
	(case pt of
	   Protocol (p, pt, ptl) =>
	     Protocol (p, f pt, List.map f ptl)
	 | Program (p, ptl) =>
	     Program (p, List.map f ptl)
	 | TypedId (p, uid, id, pt) =>
	     TypedId (p, uid, id, f pt)
	 | Role (p, {name = pt1, params = SOME ptl, body = pt2}) =>
	     Role (p, {name = f pt1, params = SOME (List.map f ptl),
		       body = f pt2})
	 | Role (p, {name = pt1, params = NONE, body = pt2}) =>
	     Role (p, {name = f pt1, params = NONE, body = f pt2})
	 | Parameter (p, ptl, pt) =>
	     Parameter (p, List.map f ptl, f pt)
	 | Pub (p, pt) =>
	     Pub (p, f pt)
	 | Priv (p, pt) =>
	     Priv (p, f pt)
	 | SymKey(p, ptl) => SymKey(p, List.map f ptl)
	 | InvKey (p, pt) => InvKey (p, f pt)
	 | RoleBody (p, ptl) =>
	     RoleBody (p, List.map f ptl)
	 | Labeled (pt1, pt2) =>
	     Labeled (f pt1, f pt2)
	 | Send (p, pt) =>
	     Send (p, f pt)
	 | Receive (p, pt) =>
	     Receive (p, f pt)
	 | MessageTuple (p, ptl) =>
	     MessageTuple (p, List.map f ptl)
	 | MessageConcur (p, ptl) =>
	     MessageConcur (p, List.map f ptl)
	 | Crypt (p, pt1, pt2) =>
	     Crypt (p, f pt1, f pt2)
	 (* | Decrypt (p, pt1, pt2) =>
	     Decrypt (p, f pt1, f pt2) *)
	 | Predicate (p, {name = pt1, params = SOME ptl, def = pt2}) =>
	     Predicate (p, {name = f pt1, params = 
			    SOME (List.map f ptl), def = f pt2})
	 | Predicate (p, {name = pt1, params = NONE, def = pt2}) =>
	     Predicate (p, {name = f pt1, params = NONE, def = f pt2})
	 | Theorem (p, {name = pt1, params = SOME ptl, def = pt2}) =>
	     Theorem (p, {name = f pt1, params = 
			    SOME (List.map f ptl), def = f pt2})
	 | Theorem (p, {name = pt1, params = NONE, def = pt2}) =>
	     Theorem (p, {name = f pt1, params = NONE, def = f pt2})
	 | PredicateContext (p, {name = pt1, params = SOME ptl1, def = pt2}, 
			     ptl2) =>
	     PredicateContext (p, {name = f pt1, params = 
				   SOME (List.map f ptl1), 
				   def = f pt2},
			       ptl2)
	 | PredicateContext (p, {name = pt1, params = NONE, def = pt2}, 
			     ptl) =>
	     PredicateContext (p, {name = f pt1, params = NONE, 
				   def = f pt2}, ptl)
	 | TheoremContext (p, {name = pt1, params = SOME ptl1, def = pt2}, 
			   ptl2) =>
	     TheoremContext (p, {name = f pt1, params = 
				 SOME (List.map f ptl1), def = f pt2},
			     ptl2)
	 | TheoremContext (p, {name = pt1, params = NONE, def = pt2}, 
			   ptl) =>
	     TheoremContext (p, {name = f pt1, params = NONE, 
				 def = f pt2}, ptl)
	 | Penetrator (p, pt) =>
	     Penetrator (p, f pt)
	 | Not (p, pt) =>
	     Not (p, f pt)
	 | Or (p, pt1, pt2) =>
	     Or (p, f pt1, f pt2)
	 | And (p, pt1, pt2) =>
	     And (p, f pt1, f pt2)
	 | Implies (p, pt1, pt2) =>
	     Implies (p, f pt1, f pt2)
	 | Atomic (p, ptl) =>
	     Atomic (p, List.map f ptl)
	 | FormulaInstance (p, pt, SOME ptl) =>
	     FormulaInstance (p, f pt, SOME (List.map f ptl))
	 | FormulaInstance (p, pt, NONE) =>
	     FormulaInstance (p, f pt, NONE)
	 | RoleLabel (p, pt1, SOME ptl, pt2) =>
	     RoleLabel (p, f pt1, SOME (List.map f ptl), f pt2)
	 | RoleLabel (p, pt1, NONE, pt2) =>
	     RoleLabel (p, f pt1, NONE, f pt2)
	 | RoleStrand (p, pt, SOME ptl) =>
	     RoleStrand (p, f pt, SOME (List.map f ptl))
	 | RoleStrand (p, pt, NONE) =>
	     RoleStrand (p, f pt, NONE)
	 | Range (p, pt1, pt2) =>
	     Range (p, f pt1, f pt2)
	 | x => x)

    (* Apply `f' to `pt', then apply it recursively to its children
       all the way down.  Comes in two flavors: modifying the context
       and not modifying it, resp. *)

    fun ptTransform f pt = ptTransformOne (ptTransform f) (f pt)
    fun ptTransformContext f pt = ptTransformContextOne (ptTransform f) (f pt)

    fun removeParamTypes [] = [] 
      | removeParamTypes (Parameter (_, ptl, _)::t) =
      ptl @ (removeParamTypes t)
      | removeParamTypes (h::t) =
      raise SympBug ("ParseTree.removeParamTypes: "
		     ^(pt2string h) ^ ": unexpected non-parameter!")

    fun subFun(formals, actuals) =
	let val pairs =
	        (case zipOpt(formals, actuals) of
		     SOME pairs => pairs
		   | NONE => raise SympBug
		      ("ParseTreeStruct.subFun: parameters don't match:\n  formals: ["
		       ^(strlistmap2str ", " pt2stringVerbose formals)
		       ^"],\n  actuals: ["
		       ^(strlistmap2str ", " pt2stringVerbose actuals)^"]"))
	    fun getUid(UniqueId(_, uid, _)) = SOME uid
	      | getUid(TypedId(_, uid, _, _)) = SOME uid
	      | getUid _ = NONE
	    fun subOne(formal, actual) pt =
		if (getUid pt) = (getUid formal) then actual
		else ptTransformOne(subOne(formal, actual)) pt
	    fun subst pt = List.foldr(fn(pair, acc)=> subOne pair acc) pt pairs
	in  (* Return the substitution function *)
	    subst
	end

(*    fun subFun ([], []) =
      (fn x => x)
      | subFun (TypedId(_,uid,_,tp)::t, UniqueId (p, suid, sid)::t') =
      (fn u as UniqueId(_,uid',_) => 
       if (uid = uid') then 
	 TypedId(p,suid,sid,tp)
       else
	 (subFun (t, t')) u |
       u as TypedId(_,uid',_,_) => 
       if (uid = uid') then 
	 TypedId(p,suid,sid,tp)
       else
	 (subFun (t, t')) u |
	 x => x)
      | subFun (TypedId(_,uid,_,tp)::t, (h' as TypedId (_,_,_,_))::t') =
      (fn u as UniqueId(_,uid',_) => 
       if (uid = uid') then 
	 h'
       else
	 (subFun (t, t')) u |
       u as TypedId(_,uid',_,_) => 
       if (uid = uid') then 
	 h'
       else
	 (subFun (t, t')) u |
	 x => x)
      | subFun (lst as h::t, []) = 
      raise SympBug ("Parameter lists not of equal size in subFun (orig.): [" ^
		     (strlistmap2str ", " pt2string lst)^"]")
      | subFun ([], lst as h::t) =
      raise SympBug ("Parameter lists not of equal size in subFun (sub.): " ^
		     (strlistmap2str ", " pt2string lst)^"]")
      | subFun (h1::t, h2::_) =
      raise SympBug ("subFun: One of those is not a TypedId: " ^
		     (pt2stringVerbose h1)^", or "^(pt2stringVerbose h2))
*)
    fun instantiate (pt, SOME parameters, SOME sparameters) =
      ptTransform (subFun (removeParamTypes parameters, sparameters)) pt
      | instantiate (pt, NONE, NONE) =
      pt
      | instantiate (pt, _, _) =
      raise SympBug ("Instantiation failed because parameters cannot match.")

    fun expandLoop defined (exp as FormulaInstance 
			    (p, UniqueId(p1,uid,id), sparams)) =
      (case
	 (List.find (fn (PredicateContext (_,{name=UniqueId(_,x,_),...},_)) => 
		     (x=uid) |
		     (TheoremContext (_, {name=UniqueId(_,x,_), ...}, _)) => 
		       (x=uid) |
		       _ => false) defined) of
	 SOME (PredicateContext (p,{params=pa,def=d,...},_)) =>
	   instantiate (d, pa, sparams)
       | SOME (TheoremContext (p,{params=pa,def=d,...},_)) =>
	   instantiate (d, pa, sparams)
       | SOME _ => raise SympBug
	   "This cannot happen (expandLoop in athena-parsetree.sml)."
       | NONE => raise SympBug ((pt2string exp) ^ 
				": reference to undefined predicate/theorem!"))
      | expandLoop _ pt = pt

    fun expand (defined, pt) =
      let
	val iter = ptTransformContext (expandLoop defined) pt
      in
	if not (iter = pt) then
	  expand (defined, iter)
	else
	  iter
      end

    fun extractRoleParams (role as (Role (_, {params=pa,...}))) =
      (role, pa)
      | extractRoleParams _ = raise SympBug
      "Attempted to extract role and parameters from non-Role!"

    fun extractSomeRoleParams (role as (Role (_, {params=SOME pa,...}))) =
      (role, pa)
      | extractSomeRoleParams _ = raise SympBug
      "Attempted to extract role and parameters from non-Role (or empty arguments)!"

    (* Message subterm relation (non-strict version) *)
    fun subterm(m1, m2) = ptEq(m1, m2) orelse
	  List.exists(fn t=> subterm(m1, t)) (ptSubtrees m2)

  end
