(* Defines the list of proof rules.  The actual definitions of proof
   rules have their own module for the code clarity. *)

signature PROOF_RULES_ATHENA =
  sig
    structure Sequent: SEQUENT
    type options = 
      Sequent.SequentCommon.SequentBase.InputLanguage.Options.options    
    type ProofSystemTactic
    type ProofSystemTacticState

    type InferenceRule = Sequent.InferenceRule
    type InferenceRuleArgument = Sequent.InferenceRuleArgument
    type InferenceRuleArgumentSpec = Sequent.InferenceRuleArgumentSpec
    type InferenceRuleResult = Sequent.InferenceRuleResult
    type ProofSystemCommand = Sequent.ProofSystemCommand
    type ProofSystemCommandResult = Sequent.ProofSystemCommandResult
    type Sequent = Sequent.Sequent
    type SequentHints = Sequent.SequentHints
    type ProgramContext = Sequent.ProgramContext
    type Substitution = Sequent.Substitution

    val proofRules: unit -> Sequent.InferenceRule list
    val proverCommands: unit -> Sequent.ProofSystemCommand list
    val getTacticName: ProofSystemTactic -> string
    val getTacticHelp: ProofSystemTactic -> string
    (* Extracting argument "types" from a tactic *)
    val getTacticArgs: ProofSystemTactic -> InferenceRuleArgumentSpec list

    (* "Typecheck" the rule's arguments and return SOME error message
       if they don't match.  Otherwise return NONE. *)
    val checkRuleArgs: InferenceRuleArgumentSpec list -> 
      InferenceRuleArgument list -> string option

    (* See if the sequent matches the conclusion of the rule, and if
       it does, compute the necessary parameters for the rule. `SOME'
       doesn't guarantee yet that the rule applies, but `NONE'
       guarantees that it doesn't. *)

    val matchRule: (Sequent * ProgramContext)
	  -> (InferenceRule * (InferenceRuleArgument list))
	  -> Substitution option

    (* Try to apply the rule and get a list of new sequents as
       premisses.  NONE means the rule doesn't apply.  May take a
       (partial) substitution obtained by the matchRule funciton.
       The first argument is a printing function to send messages to
       the user while applying the rule. *)
    val applyRule: (string -> unit) -> (Sequent * ProgramContext)
	        -> InferenceRule * (InferenceRuleArgument list) * (Substitution option)
	        -> InferenceRuleResult

    (* Run the proof sytem command on the sequent.  Commands normally
       do not change the sequent or the proof tree, and are only used
       to display some information or test various conditions for the
       sequent. *)

    val applyCommand: (string -> unit) -> (Sequent * ProgramContext)
	        -> ProofSystemCommand * (InferenceRuleArgument list)
                -> ProofSystemCommandResult option

    (* Get the list of rules that match a given sequent, together with
       the lists of rules' arguments *)
    val matchAllRules: Sequent -> (InferenceRule * (InferenceRuleArgument list)) list

    (* applyTactic printFun (tactic, args, context, stateOpt) (seq, hintsOpt) continuation

       Applies `tactic' to the sequent `seq'.  This can be an
       intermediate step within the tactic from an optional state
       `stateOpt'.  Actions may also depend on the optional sequent
       hints.  If any rule has been applied to the sequent, the
       `continuation' is called which will install the rule in the
       proof tree and will call `applyTactic' again with the new
       state given to it. *)

    val applyTactic: bool -> (string -> unit)
	        -> (ProofSystemTactic * (InferenceRuleArgument list) * 
		    ProgramContext
		    * (ProofSystemTacticState option))
                -> (Sequent * (SequentHints option))
	        (* Result *)
	        -> (string * (InferenceRuleArgument list)
		    * ((Sequent * (SequentHints option)) list)
		    * (ProofSystemTacticState option)) option

    (* Each rule has a unique string identifier *)
    val getRuleName: InferenceRule -> string
    val getCommandName: ProofSystemCommand -> string

    (* Each rule has a set of arguments *)
    val getRuleArgs: InferenceRule -> InferenceRuleArgumentSpec list
    val getCommandArgs: ProofSystemCommand -> InferenceRuleArgumentSpec list

    (* Help on the rule *)
    val getRuleHelp: InferenceRule -> string
    val getCommandHelp: ProofSystemCommand -> string

  end
