functor SequentAthenaFun(structure Interface: INTERFACE
			 structure Hash: HASH) =
  struct

    open Str
    open Interface

    structure SequentBase =
      struct
	open Hash

	open SympBug
	open ListOps
	open Pos

	structure InputLanguage =
	    InputLanguageAthenaFun(structure Interface = Interface
				   structure Hash = Hash)
	open InputLanguage

	open ParserAthena
	open ParseTreeStruct

	(* Tests if h1 is a subset of h2 by seeing if every element in h1
	   is also in h2, and with the same value. *)

	fun hashSubset (h1, h2, eq) =
	  let
	    val hl = hash2any (fn x => x) h1
	  in
	    List.all (fn (k, v) => case (findHash (h2, k)) of
		      SOME v' => eq (v, v')
		    | NONE => false) hl
	  end

	(* Two hash tables are equivalent if each is a subset of the
	   other.  A more efficient version: if the sizes are the
	   same, inclusion in one direction is sufficient. *)

	fun hashEq (h, h', eq) =
	    (hashSize h) = (hashSize h') andalso (hashSubset (h, h', eq))

	(* A node is simply a particular "line" from an instantiated
	   role (the `action').  We refer to it by the unique ID of
	   the strand, the line #, and its ParseTree (this last for
	   convenience and to avoid lookups). *)

	type Node = {strandId: string,
		     index: int,
		     roleId: string, (* The unique ID of the role *)
		     (* Instantiation of the role *)
		     roleParams: ParseTree list option, 
		     action: ParseTree}

	(* Access functions *)
	fun nodeStrand({strandId=id, ...}: Node) = id
	fun nodeIndex({index=i, ...}: Node) = i
	fun nodeRole({roleId=id, ...}: Node) = id
	fun nodeRoleParams({roleParams=p, ...}: Node) = p
	fun nodeAction({action=a, ...}: Node) = a

	fun makeNode(strandId, index, roleId, params, action) =
	    {strandId=strandId,
	     index=index,
	     roleId=roleId,
	     roleParams=params,
	     action=action}: Node

	(* Node equality needs only test the unique ID and line #. *)

	fun nodeEq (n1, n2) =
	     (nodeStrand n1) = (nodeStrand n2)
	     andalso (nodeIndex n1) = (nodeIndex n2)

	(* A MessageExtr is used to indicate a subterm of the message
	   sent in a certain node.  It is the type for goals (and for
	   the origin bindings. *)

	datatype MessageExtr = ExtractMessage of Node |
	  ExtractTuple of (MessageExtr * int) |
	  ExtractConcur of (MessageExtr * int)
	  (* | MTuple of MessageExtr list *)

	datatype Goal = Goal of MessageExtr * ParseTree
	datatype Binding = Binding of MessageExtr * ParseTree

	(* Goal equality is simply recursive equality down to the node
	   level, at which point node equality is used. *)

	fun messageExtrEq (ExtractMessage n, ExtractMessage n') = 
	  nodeEq (n, n')
	  | messageExtrEq (ExtractTuple (m, i), ExtractTuple (m', i')) =
	  (messageExtrEq (m, m')) andalso (i = i')
	  | messageExtrEq (ExtractConcur (m, i), ExtractConcur (m', i')) =
	  (messageExtrEq (m, m')) andalso (i = i')
	  (* | messageExtrEq (MTuple [], MTuple []) = true
	  | messageExtrEq (MTuple (h::t), MTuple (h'::t')) = 
	  (messageExtrEq (h, h')) andalso (messageExtrEq (MTuple t, MTuple t')) *)
	  | messageExtrEq (_, _) = false 

	fun goalEq (Goal(me, pt), Goal(me', pt')) = (messageExtrEq (me, me')) 
	  andalso (ptEq (pt, pt'))

	fun bindingEq (Binding(me, pt), Binding(me', pt')) = (messageExtrEq (me, me')) 
	  andalso (ptEq (pt, pt'))

	type Var = string

	datatype VarType = Key | Principal | Message | Nonce | FreshNonce |
	  VTuple of VarType list

	type NodeHash = (Node, unit) Hash

	(* A SemiBundle is a set of nodes which should be backwards closed
	   under =>. *)

	type SemiBundle = ((string, ParseTree) Hash * NodeHash)

	fun semiBundleEq ((roleHash, nodeHash), (roleHash', nodeHash')) =
	  (hashEq (roleHash, roleHash', ptEq)) 
	  andalso (hashEq (nodeHash, nodeHash', op=))

	fun getNodes((_, nodeHash): SemiBundle) = nodeHash
	fun getRoles((roleHash, _): SemiBundle) = roleHash

	type StrandSet = SemiBundle
	  
	fun makeStrandSet(roleHash, nodeHash) = (roleHash, nodeHash): StrandSet

	(* A state is as defined in the Athena paper: a semibundle, a
	   goalBinding (mapping a goal to its binding), a set `goals'
	   of unbound goals, and a mapping to types (currently unused,
	   present for future expansion to tuple matching). *)

	type State = {semibundle : SemiBundle,
		      goalBinding : (Goal, Binding) Hash,
		      (* List is less efficient, but more predictable in the ordering,
		         which is important for more robust proofs that refer to goals *)
		      goals : Goal list,
		      (* List of constraints of the form t \interm t'
		         used in term unification *)
		      intermConstr: (ParseTree * ParseTree) list,
		      typeMap : (Var, VarType) Hash}
	  
	fun stateEq ({semibundle = sb, goalBinding = goalBinding,
		      goals = g, intermConstr=C, typeMap = map}: State,
		     {semibundle = sb', goalBinding = goalBinding',
		      goals = g', intermConstr=C', typeMap = map'}: State) =
	  (semiBundleEq (sb, sb')) andalso (hashEq (goalBinding, goalBinding',
						    bindingEq))
	  andalso (listEq goalEq (g, g')) andalso (hashEq (map, map', op=))
	  andalso (listEq (fn((t1,t2), (t1', t2'))=>
			   ptEq(t1, t1') andalso ptEq(t2,t2')) (C, C'))

	type SequentHints = unit

        (* Static protocol parameters used by pruning theorems *) 
	type PruneParams = { (* max encr. depth, if the protocol is eligible for 
			      the `Decr' pruning theorem *)
			     Emax: int option,
			     (* If any regular strand sends a secret key of a certain kind *)
			     symKeySent: bool,
			     privKeySent: bool,
			     (* If any regular strand signs a message *)
			     signedSent: bool,
			     (* The ID of the penetrator protocol *)
			     penetrator: string }

	(* The sequent is then either an initial sequent of a protocol and
	   formula or a state sequent of a state, a delta, and a context.
	   The sequents are as defined in the Athena paper. *)

	datatype Sequent = 
	  StateSequent of {l : State,
			   pruneParams: PruneParams,
			   delta : StrandSet,
			   context : ParseTree list}
	| InitSequent of { formula: ParseTree,
			   pruneParams: PruneParams }

	datatype SequentPart = GammaPart | DeltaPart | GoalPart | AnyPart

	(* Some functions used to give a hash function for nodes, 
	   etc.) *)
	  
	fun nodeString (n: Node) =
	      ((nodeStrand n) ^ "." ^ (Int.toString(nodeIndex n)))
	fun nodeStringVerbose (n: Node) =
	      "["^((nodeStrand n) ^ "." ^ (Int.toString(nodeIndex n)))
	      ^": "^(pt2string(nodeAction n))^"]"
	  
	fun nodeStringDebug (n: Node) =
	      "Node{strandId="^(nodeStrand n)
	      ^",\n     index=" ^ (Int.toString(nodeIndex n))
	      ^",\n     roleId="^(nodeRole n)
	      ^",\n     roleParams="^(case (nodeRoleParams n) of
					 SOME lst => "["^(strlistmap2str ", " pt2string lst)^"]"
				       | NONE => "NONE")
	      ^",\n     action="^(pt2string(nodeAction n))^"}"
	  
	fun ExtractMessageString (ExtractMessage node) = (nodeString node)
	  | ExtractMessageString (ExtractTuple (m, n)) = 
	  (ExtractMessageString m) ^ "." ^ (Int.toString n)
	  | ExtractMessageString (ExtractConcur (m, n)) = 
	  (ExtractMessageString m) ^ "." ^ (Int.toString n)
	  (* | ExtractMessageString (MTuple l) = 
	  strlistmap2str "," (fn m => ExtractMessageString m) l *)

	fun goalString(Goal(em, pt)) =
	      "Goal["^(pt2string pt)^" = "^(ExtractMessageString em)^"]"
	  
	fun bindingString(Binding(em, pt)) =
	      "Binding["^(pt2string pt)^" = "^(ExtractMessageString em)^"]"
	  
	fun varString v = v

	fun formulaInd2str(part, nOpt) =
	    let fun part2str GammaPart = "-"
		  | part2str DeltaPart = "+"
		  | part2str GoalPart = "g"
		  | part2str AnyPart = "*"
	    in
		(part2str part)
		^(case nOpt of
		      NONE => ""
		    | SOME n => Int.toString n)
	    end

	fun op *(x,y) = Conc(x,y)

	(* Extract the nodes to which the binding belongs *)
	fun Binding2Node(Binding(E, _)) =
	    let fun loop(ExtractMessage n) = n
		  | loop(ExtractTuple(E, _)) = loop E
		  | loop(ExtractConcur(E, _)) = loop E
		  (* | loop (MTuple lst) = union nodeEq (List.map loop lst) *)
	    in loop E
	    end

	(* Extract the nodes to which the goal belongs *)
	fun Goal2Node(Goal(E, _)) =
	    let fun loop(ExtractMessage n) = n
		  | loop(ExtractTuple(E, _)) = loop E
		  | loop(ExtractConcur(E, _)) = loop E
		  (* | loop (MTuple lst) = union nodeEq (List.map loop lst) *)
	    in loop E
	    end

	(* Total ordering on nodes *)
	fun nodeLess(n1:Node, n2:Node) =
	    let val (r1, r2) = (nodeRole n1, nodeRole n2)
		val (s1, s2) = (nodeStrand n1, nodeStrand n2)
		val (i1, i2) = (nodeIndex n1, nodeIndex n2)
	    in
		String.<(r1, r2) orelse
		(r1 = r2 andalso
		 (String.<(s1, s2) orelse 
		  (s1 = s2 andalso i1 < i2)))
	    end

	(* Total ordering on goals for sorting  *)
	fun goalLess(Goal(em1, _), Goal(em2, _)) =
	    let fun em2list(ExtractMessage node, acc) = (node, acc)
		  | em2list(ExtractTuple(m, i), acc) = em2list(m, i::acc)
		  | em2list(ExtractConcur(m, i), acc) = em2list(m, i::acc)
		fun lexOrder([], []) = false (* Two elements are equal *)
		  | lexOrder([], _) = true
		  | lexOrder(_, []) = false
		  | lexOrder(i1::l1, i2::l2) =
		      (i1 < i2) orelse (i1 = i2 andalso lexOrder(l1, l2))
		val (n1, l1) = em2list(em1, [])
		val (n2, l2) = em2list(em2, [])
	    in  (* Order goals by reverse order of nodes, to put the last nodes first.
		   This should speed up goal binding. *)
		nodeLess(n2, n1) orelse (nodeEq(n1, n2) andalso lexOrder(l1, l2))
	    end

	fun goalBindingStr(goalBinding, C) =
	    let val Cstring = (Str "[")
		    *(Strlistmap2Str ",\n  "
		      (fn(t1,t2)=>Str((pt2string t1)^" << "^(pt2string t2))) C)
		    *(Str "]\n")
		fun pair2str(g, b) =
		(Str ("("^(goalString g)^"\n   <<- "^(bindingString b)^")"))
		val pairs = sortList(fn((x,_),(y,_))=> goalLess(x,y))
		                    (hash2any(fn x=>x) goalBinding)
	    in
		(Str "Goal Bindings = [\n  ")
		*(Strlist2Str "\n  " (List.map pair2str pairs))
		*(Str "]\n\nInterm Constraints:\n ")
		*Cstring
	    end

	fun state2strCommon (nodeStr, verbose) (state: State) =
	    let val {semibundle=sb,
		     goals=g,
		     goalBinding=goalBinding,
		     intermConstr = C, ...} = state
		fun h2a h = List.map nodeStr (sortList nodeLess (hash2any #1 h))
		val gamma = getNodes sb
		val gammaList = mapEnumerate(fn(i, s)=>(Str ("[-"^(Int.toString(i+1))^"] "^s)))
		                            (h2a gamma)
		val gList = mapEnumerate(fn(i, g)=>(Str ("[g"^(Int.toString(i+1))^"] "
							 ^(goalString g)))) g
		val gammaStrings = Strlist2Str "\n" gammaList
		val gStrings = Strlist2Str "\n" gList
		val numGoals = List.length gList
	    in 
		(if verbose then goalBindingStr(goalBinding, C) else (Str ""))
		*gStrings
		* (Str ("\n----- "^(Int.toString numGoals)^" UNBOUND GOAL"
			^(case numGoals of
			      1 => ""
			    | _ => "S")^" -----\n"))
		* gammaStrings
	    end  

	fun pruneParams2string(pp: PruneParams) =
	    let val { Emax=Emax, symKeySent=symKeySent,
		      privKeySent=privKeySent,
		      signedSent=signedSent, penetrator=penetrator } = pp
	    in
		"\n === Penetrator = "^penetrator
		^(case Emax of
		      NONE => ""
		    | SOME n => ", Max. Encryption Depth = "^(Int.toString n))
		^" ===\n === roles "^(if symKeySent then "" else "do not ")
		^"send symmetric keys, "^(if privKeySent then "" else "do not ")
		^"send private keys and "^(if signedSent then "" else "do not ")
		^"send signed messages ==="
	    end
	  
	fun sequent2strCommon (_, verbose) (InitSequent{formula=pt, pruneParams=pp}) =
	    let fun printPt(TheoremContext(_, {name=n, params=pa, def=d}, _)) =
		    Str("theorem = "^(pt2string n)
			^(case pa of
			      SOME lst => "["^(strlistmap2str ", " pt2string lst)^"]"
			    | NONE => "")
			^" =\n  "^(pt2string d)^"\n")
		  | printPt pt = Str((pt2string pt)^"\n")
	    in (printPt pt)*(Str ((if verbose then (pruneParams2string pp) else "")^"\n"))
	    end
	  | sequent2strCommon (parms as (nodeStr, verbose)) (StateSequent stateSeq) =
	  let 
	    val { l = state,
		  delta = (_, delta),
		  pruneParams = pp, ...} = stateSeq
	    fun h2a h = List.map nodeStr (sortList nodeLess (hash2any #1 h))
	    val stateString = state2strCommon parms state
	    val deltaList = mapEnumerate(fn(i, s)=>(Str ("["^(Int.toString(i+1))^"] "^s)))
		                        (h2a delta)
	    val deltaStrings = Strlist2Str "\n" deltaList
	  in
	    stateString
	    * (Str "\n|------------->\n")
	    * deltaStrings * (Str "\n")
	    * (Str ((if verbose then (pruneParams2string pp) else "")^"\n"))
	    
	  end

	val sequent2str = sequent2strCommon (nodeStringVerbose, false)
	val sequent2strDebug = sequent2strCommon (nodeStringDebug, true)

	fun sequent2string seq = Str2string(sequent2str seq)
	fun sequent2stringDebug seq = Str2string(sequent2strDebug seq)

	val state2str = state2strCommon (nodeStringVerbose, false)
	val state2strDebug = state2strCommon (nodeStringDebug, true)
	  
	fun state2string seq = Str2string(state2str seq)
	fun state2stringDebug seq = Str2string(state2strDebug seq)

	fun binding2string(goalBinding, C) = Str2string(goalBindingStr(goalBinding, C))

	fun hints2string hints = ""
      end

    (* Open them here, before SML figures out their signatures and constrains them *)
    open SequentBase
    open InputLanguage

    (* Some common types and functions *)
    structure SequentCommon = SequentCommonFun(structure Interface = Interface
					       structure SequentBase = SequentBase)

    open SequentCommon

    open SympBug
    open Str
    open ListOps

    type Substitution = unit

    type RuleArg = InferenceRuleArgument

    type InferenceRuleArgumentSpec =
	  { name: string,
	    Type: InferenceRuleArgumentType }

    type ProofSystemCommandResult = unit

    type ProofSystemCommand = { name: string,
			       help: string,
			       params: InferenceRuleArgumentSpec list,
			       apply: (string -> unit)
			       -> Sequent * (RuleArg list)
			       -> ProofSystemCommandResult option }

    type InferenceRule = { name : string,
			   help : string, (* Help text *)
			   params : InferenceRuleArgumentSpec list,
			   match: (Sequent * ProgramContext * (RuleArg list)
                                   * (Substitution option) -> 
				   Substitution option) option,
			   apply: ((string -> unit)
                                   -> Sequent * ProgramContext * (RuleArg list) 
				   * Substitution
                                   ->  InferenceRuleResult) option }

    fun IRArgSpecName({name=name,...}: InferenceRuleArgumentSpec) = name
    fun IRArgSpecType({Type=tp,...}: InferenceRuleArgumentSpec) = tp

    (* Functions needed for pruning theorems *)

    fun typeComplexity(MessageType _) = 1
      | typeComplexity(PrincipalType _) = 1
      | typeComplexity(NonceType _) = 1
      | typeComplexity(KeyType _) = 1
      | typeComplexity(SymKeyType _) = 1
      | typeComplexity(PrivKeyType _) = 1
      | typeComplexity(PubKeyType _) = 1
      | typeComplexity(TupleType(_, lst)) = List.foldr(op +) 1 (List.map typeComplexity lst)
      | typeComplexity(EncryptedType(_, tp)) = 1 + (typeComplexity tp)
      | typeComplexity x = raise SympBug("typeComplexity: not a type: "^(pt2string x))

    datatype IntInf = IntInfinity | Int of int

    fun InfPlus(Int x, Int y) = Int(x+y)
      | InfPlus _ = IntInfinity

    fun InfLess(IntInfinity, IntInfinity) = false
      | InfLess(Int x, Int y) = x < y
      | InfLess(Int _, IntInfinity) = true
      | InfLess(IntInfinity, Int _) = false

    fun InfString IntInfinity = "Inf"
      | InfString (Int n) = Int.toString n

    fun typeMaxComplexity(MessageType _) = IntInfinity
      | typeMaxComplexity(PrincipalType _) = Int 1
      | typeMaxComplexity(NonceType _) = Int 1
      | typeMaxComplexity(KeyType _) = Int 1
      | typeMaxComplexity(SymKeyType _) = Int 1
      | typeMaxComplexity(PrivKeyType _) = Int 1
      | typeMaxComplexity(PubKeyType _) = Int 1
      | typeMaxComplexity(TupleType(_, lst)) =
	  List.foldr InfPlus (Int 1) (List.map typeMaxComplexity lst)
      | typeMaxComplexity(EncryptedType(_, tp)) = InfPlus(Int 1, typeMaxComplexity tp)
      | typeMaxComplexity x = raise SympBug("typeMaxComplexity: not a type: "^(pt2string x))

    (* Max depth of encryption nesting in `pt'.
       Takes into account the set of interm constraints (list of pairs), if given. *)
    fun maxEncryptionDepth Copt pt =
	let val funName = "maxEncryptionDepth"
	    fun C2str(SOME C) =
		"["^(strlistmap2str ",\n "
		     (fn(t,t')=>"("^(pt2string t)^" << "^(pt2string t')^")") C)
		^"]"
	      | C2str NONE = "NONE"
	    val _ = pushFunStackLazy(funName, fn()=>(pt2string pt)^"\n C = "
				     ^(C2str Copt))
	    fun findInterms(C, pt) =
	        List.mapPartial(fn (t1, t2)=> if subterm(pt, t2) then SOME t1 else NONE) C
	    fun checkC(NONE, pt) = 0
	      | checkC(SOME C, pt) =
	        let val interms = findInterms(C, pt)
		in
		    maxList(op <, 0) (List.map (maxEncryptionDepth Copt) interms)
		end
	    fun loop(Crypt(_, pt, _)) = 1+(loop pt)
	      | loop(MessageTuple(_, lst)) = maxList(op <, 0) (List.map loop lst)
	      | loop pt = checkC(Copt, pt)
		
	    val res = loop pt
	    val _ = popFunStackLazy(funName, fn()=>Int.toString res)
	in
	    res
	end

  (* Checks whether the protocol is eligible for the pruning theorem
     that catches the cycles in `Decr' role, and collects some info if
     it is.  Takes the list of roles (full definitions) and the
     penetrator name, and returns Emax (max. depth of nested
     encryption) or NONE. *)

    fun getProtocolPruneParams(pHash: ProtocolHash) =
	let val funName = "getProtocolPruneParams"
	    val _ = pushFunStackLazy(funName, fn()=>ProtocolHash2str pHash)
	    val penetr = #penetratorProtocol pHash
	    val roles = hash2any #2 (#roles pHash)
	    val debug = lazyVerbDebug(getOptions()) funName
	    fun newVars lst =
	        let fun param2lst(Parameter(_, lst, _)) = lst
		      | param2lst x = [x]
		in List.foldr(op @) [] (List.map param2lst lst)
		end
	    (* Extract terms that the role sends *)
	    fun roleSentTerms role =
		let fun extractBody (Role (_, {body = RoleBody (_, b),...})) = b
		      | extractBody _ = raise SympBug 
		         (funName^": Attempted to extract body from non-Role!")
		    fun getSentTerms(Labeled(_, a)) = getSentTerms a
		      | getSentTerms(Send(_, m)) = getSentTerms m
		      | getSentTerms(Receive _) = []
		      | getSentTerms(MessageConcur(_, lst)) = lst
		      | getSentTerms m = [m]
		    val body = extractBody role
		in
		    List.foldr(op @) [] (List.map getSentTerms body)
		end

	    fun Emax pt = maxEncryptionDepth NONE pt
	    fun Cmax pt =
		let val tp = messageType pt
		    fun loop(EncryptedType(_, tp)) = [typeMaxComplexity tp]
		      | loop(TupleType(_, lst)) = List.foldr(fn(t, lst)=>(loop t)@lst) [] lst
		      | loop _ = [Int 1]
		    val res = maxList(InfLess, Int 0) (loop tp)
		    val _ = debug(fn()=>funName^": Cmax("^(pt2string pt)^") = "
				  ^(InfString res)^"\n")
		in
		    res
		end
	    (* See if the term contains a retrievable key *)
	    fun hasSymKey (MessageTuple(_, lst)) = List.exists hasSymKey lst
	      | hasSymKey (Crypt(_, m, _)) = hasSymKey m
	      | hasSymKey pt = isSubtype(messageType pt, SymKeyType dp)
	    fun hasPrivKey (MessageTuple(_, lst)) = List.exists hasPrivKey lst
	      | hasPrivKey (Crypt(_, m, _)) = hasPrivKey m
	      | hasPrivKey pt = isSubtype(messageType pt, PrivKeyType dp)
	    fun hasSigned(MessageTuple(_, lst)) = List.exists hasSigned lst
	      | hasSigned(Crypt(_, _, Priv _)) = true
	      | hasSigned(Crypt(_, m, Pub _)) = hasSigned m
	      | hasSigned(Crypt(_, m, k)) =
		(* We don't consider symmetric keys for signatures *)
		if isSubtype(messageType k, SymKeyType dp) then hasSigned m
		(* But If key is unknown, it could potentially be someone's private key *)
		else true
	      | hasSigned _ = false
	    val sentTerms = List.foldr(op @) [] (List.map roleSentTerms roles)
	    val C = maxList (InfLess, Int 0) (List.map Cmax sentTerms)
	    val _ = debug(fn()=>funName^": C = "^(InfString C)^"\n")

	    val res = { Emax = if penetr <> DefaultPenetrator.penetratorName then NONE
			       else if InfLess(C, IntInfinity) then
				   SOME(maxList(op <, 0)(List.map Emax sentTerms))
			       else NONE,
		        penetrator = penetr,
			symKeySent = List.exists hasSymKey sentTerms,
			privKeySent = List.exists hasPrivKey sentTerms,
			signedSent = List.exists hasSigned sentTerms }: PruneParams

	    val _ = popFunStack(funName, pruneParams2string res)
	in
	    res
	end



    fun makeSequent context (Object (theorem)) =
	let val pHash = (case context of
				    Local(_, (_, phash)) => phash
				  | Global _ => raise SympBug
				     ("makeSequent: wrong context (Global instead of Local):\n"
				      ^(ProgramContext2str context)))
	in InitSequent { formula = theorem,
			 pruneParams=getProtocolPruneParams pHash }
	end
      | makeSequent context exp = 
      raise SympBug "Expected a theorem object when making initial sequent!"
					
    fun sequentEq(InitSequent pt, InitSequent pt') = ptEq (pt, pt')
      | sequentEq(StateSequent {l = l, delta = delta, context = context,...},
		  StateSequent {l = l', delta = delta', context = context',...}) =
      (stateEq (l, l')) andalso (semiBundleEq (delta, delta'))
      | sequentEq _ = false
    fun sequentSubsumes(pair as (_, _)) = sequentEq pair
    fun hintsEq(_, _) = true

    (* Extract sequent part from a UIArg specification *)
    fun UIArg2FormulaIndex(UInumber n) =
	  if n >= 0 then (DeltaPart, SOME n)
	  else (GammaPart, SOME(~n))
      | UIArg2FormulaIndex(UIstring str) =
	  (case String.explode str of
	       [#"*"] => (AnyPart, NONE)
	     | [#"-"] => (GammaPart, NONE)
	     | [#"+"] => (DeltaPart, NONE)
	     | [#"g"] => (GoalPart, NONE)
	     | #"*"::digits => 
		   ((case Int.fromString(String.implode digits) of
			 NONE => raise ProverError("Bad formula number: "^str)
		       | SOME n => (AnyPart, SOME n))
			 handle _ => raise ProverError
			     ("Too large formula number: "^str))
	     | #"-"::digits => 
		   ((case Int.fromString(String.implode digits) of
			 NONE => raise ProverError("Bad formula number: "^str)
		       | SOME n => (GammaPart, SOME n))
			 handle _ => raise ProverError
			     ("Too large formula number: "^str))
	     | #"g"::digits => 
		   ((case Int.fromString(String.implode digits) of
			 NONE => raise ProverError("Bad formula number: "^str)
		       | SOME n => (GoalPart, SOME n))
			 handle _ => raise ProverError
			     ("Too large formula number: "^str))
	     | digits as (_::_) => 
		   ((case Int.fromString(String.implode digits) of
			 NONE => raise ProverError("Bad formula number: "^str)
		       | SOME n => (DeltaPart, SOME n))
			 handle _ => raise ProverError
			     ("Too large formula number: "^str))
	     | _ =>  raise ProverError
			 ("Bad formula number: \""^str^"\""))
      | UIArg2FormulaIndex arg = raise ProverError
	       ("Bad formula number: "^(UIArg2string arg))

    fun FormulaIndex2UIArg part = UIstring(formulaInd2str part)

    fun InfRuleSpec2str({name=name,
			 Type=t}: InferenceRuleArgumentSpec) =
	let val NAME = String.implode(List.map Char.toUpper 
				      (String.explode name))
	in
	    (NAME^": "^(InfRuleArgType2string t))
	end

    fun formatComm command =
	let val {interface=UI,...} = getOptions()
	in formatProverCommand UI command
	end

    fun checkRuleArgs specs args = NONE: string option

    fun matchRule(seq, context) (rule, args) = (SOME()): Substitution option

    fun applyRule printFun (seq, context) (rule, args, substOpt) = 
      InferenceRuleFailure

    fun applyCommand printFun (seq, context) (command, args) =
	   NONE: ProofSystemCommandResult option

    fun matchAllRules seq = raise SympBug
	  ("SequentAthena.matchAllRules: sorry, I'm not implemented yet.")

    fun string2hints str = ()

    fun getSequentInfo seq request = "<sequent info>"

end

