(* Typechecking without overloading of identifiers *)

functor AthenaTypecheckFun(structure Interface: INTERFACE
			   structure AthenaTypes: ATHENA_TYPES): 
  ATHENA_TYPECHECK =
  struct
    structure Interface = Interface
    structure AthenaTypes = AthenaTypes

    open AthenaTypes
    open Str
    open Pos
    open Interface
    open ListOps
    open ParseTreeStruct
    open Hash
    open Options

    structure DefaultPenetrator =
	DefaultPenetratorFun(structure ParseTreeStruct = ParseTreeStruct)

    structure EmptyPenetrator =
	EmptyPenetratorFun(structure ParseTreeStruct = ParseTreeStruct)


    (* Just a dummy for now. *)
    fun typeCheckExpr options expr _ =
      expr

    fun isKey(Priv _) = true
      | isKey(Pub _) = true
      | isKey(SymKey _) = true
      | isKey(InvKey(_, k)) = isKey k
      | isKey(Id _) = true
      | isKey _ = false

    (* Converts keys (and keys only) to strings for some errors. *)
    fun key2string k = 
	(if not(isKey k) then "[NOT A KEY] = " else "")
        ^(pt2string k)

    (* Extract message type.  If we cannot determine it, return the
       default `defType'.  If the hash is given, try to find the types
       of untyped IDs in the hash. *)

    fun messageTypeHash _ (TypedId(_,_,_,tp)) = tp
      | messageTypeHash (SOME hash) (Id(p, id)) =
	  (case findHash(hash, id) of
	       SOME (_, tp) => tp
	     | NONE => MessageType p)
      | messageTypeHash _ (Parameter(_, _, tp)) = tp
      | messageTypeHash _ (Pub _) = PubKeyType dp
      | messageTypeHash _ (Priv _) = PrivKeyType dp
      | messageTypeHash _ (SymKey _) = SymKeyType dp
      | messageTypeHash hash (m as InvKey(_,k)) =
	  (case messageTypeHash hash k of
	       PubKeyType p => PrivKeyType p
	     | PrivKeyType p => PubKeyType p
	     | SymKeyType p => SymKeyType p
	     | KeyType p => KeyType p
	     | tp => raise SympBug
		 ("Typecheck.messageType: not a key type in InvKey: "
		  ^(pt2string tp)^", expr = "^(pt2string m)))
      | messageTypeHash hash (MessageTuple(p, lst)) =
	  TupleType(p, List.map (messageTypeHash hash) lst)
      | messageTypeHash hash (Crypt(p, m, _)) =
	  EncryptedType(p, messageTypeHash hash m)
      | messageTypeHash _ _ = MessageType dp

    fun messageType pt = messageTypeHash NONE pt

    (* A more generic approach to checking the same relationship,
       used elsewhere. *)

    fun isSubtype (TupleType(_, l1), TupleType(_, l2)) =
	  (case zipOpt(l1, l2) of
	       SOME pairs => List.all isSubtype pairs
	     | NONE => false)
      | isSubtype (PrivKeyType _, KeyType _) = true
      | isSubtype (PubKeyType _, KeyType _) = true
      | isSubtype (SymKeyType _, KeyType _) = true
      | isSubtype (EncryptedType(_, t1), EncryptedType(_, t2)) = isSubtype(t1, t2)
      | isSubtype (_, MessageType _) = true
      | isSubtype (New(_, t1), t2) = isSubtype(t1, t2)
      | isSubtype (t1, New(_, t2)) = isSubtype(t1, t2)
      | isSubtype (a, b) = ptEq (a, b)

    (* Returns inverse of a given key for decryption via pattern-matching. *)
    fun keyInvHash _ (Priv(p,id)) = (Pub(p,id))
      | keyInvHash _ (Pub(p,id)) = (Priv(p,id))
      | keyInvHash _ (InvKey(p, x)) = x 
      | keyInvHash hashOpt x =
	if isSubtype(messageTypeHash hashOpt x, SymKeyType dp) then x
	else InvKey(pos x, x)

    fun keyInv k = keyInvHash NONE k

    (* Unwrap formal parameter list from `Parameter' and make it a
       normal flat list of pairs (id, tp). *)
    fun flattenParams params =
	let fun param2lst(Parameter(_, lst, tp)) = List.map (fn x=>(x,tp)) lst
	      | param2lst x = raise SympBug
	    ("Typecheck.flattenParams: parameter is not a `Parameter': "
	     ^(pt2stringVerbose x))
	in List.foldr(op @) [] (List.map param2lst params)
	end
    
    (* Tells if the value given is available in knownVals. *)
    fun knowVal(k, knownVals, pHash) =
	let val funName = "knowVal"
	    val _ = pushFunStackLazy(funName, fn()=>
				     (pt2string k)^", ["
				     ^(strlistmap2str ", " pt2string knownVals)^"]")
	    fun getId(Id(_, id)) = id
	      | getId(TypedId(_, _, id, _)) = id
	      | getId(UniqueId(_, _, id)) = id
	      | getId x = raise SympBug
		  ("Typecheck.knowVal/getId: not an ID: "^(pt2string x))

	    fun isId(Id _) = true
	      | isId(TypedId _) = true
	      | isId(UniqueId _) = true
	      | isId _ = false

	    val (k', tp) =
		if isId k then
		    (case findHash(pHash, getId k) of
			 SOME pair => pair
		       | NONE => raise SympBug
			     ("Typecheck.knowVal: ID is not a parameter: "
			      ^(pt2string k)^" at "^(pos2string(pos k))))
		else (k, messageType k)

	    fun collapseKey(InvKey(p, k)) = keyInvHash(SOME pHash)(collapseKey k)
	      | collapseKey(k as Id(p, id)) =
	         (case findHash(pHash, id) of
		      SOME (k',_) => k'
		    | NONE => raise SympBug
		       ("Typecheck.knowVal: key ID is not a parameter: "
			^(pt2string k)^" at "^(pos2string p)))
	      | collapseKey k = k
	    fun sameKey(Priv(_, p1)) (Priv(_, p2)) = (getId p1) = (getId p2)
	      | sameKey(Pub(_, p1)) (Pub(_, p2)) = (getId p1) = (getId p2)
	      | sameKey(InvKey(_, k1)) (InvKey(_, k2)) = (getId k1) = (getId k2)
	      | sameKey k1 k2 =
		  isId k1 andalso isId k2 andalso (getId k1) = (getId k2)
	    fun findSelf lst =
		let fun findOne(Id(p, id)) =
		          (case findHash(pHash, id) of
			       SOME(_, New(_, PrincipalType _)) => true
			     | SOME(_, PrincipalType _) => false
			     | SOME(x, tp) => raise TypeError
				 ((pos2string p)
				  ^": SymKey parameter is not of Principal type:\n  "
				  ^(pt2string x)^", type = "^(pt2string tp))
			     | NONE => raise TypeError
				 ((pos2string p)^": Undeclared identifier: "^id))
		      | findOne x = raise TypeError
			  ((pos2string(pos x))^": SymKey parameter is not ID: "
			   ^(pt2string x))
		in
		    List.exists findOne lst
		end
	    val k = collapseKey k
	    (* Assume that public keys and principal names are always known.
	       Check only private and symm. keys *)
	    val res = (case (k, tp) of
			   (Pub _, _) => true
			 | (SymKey(_, lst as [_, _]), _) => findSelf lst
			 | (SymKey(p, lst), _) => raise TypeError
			     ((pos2string p)
			      ^": SymKey currently must take exactly two parameters.")
			 | (_, PrincipalType _) => true
			 | _ => List.exists (sameKey k) knownVals)
	    val _ = popFunStack(funName, if res then "true" else "false")
	in
	    res
	end

    (* Makes a hash of the parameter types for a role or theorem definition, and
       establishes the initial knownVals list based on types. *)
    fun paramTypes (NONE, masterName, hash, knownVals) = (hash, knownVals)
      | paramTypes (SOME lst, masterName, hash, knownVals) =
	let fun getId(Id(_, id)) = id
	      | getId(TypedId(_, _, id, _)) = id
	      | getId(UniqueId(_, _, id)) = id
	      | getId exp = raise SympBug
	         ((pos2string (pos exp)) 
		  ^ ": non-ID in parameter list in TypeCheck.paramTypes:\n  "
		  ^(pt2stringVerbose exp))
	    fun checkOne(expr, tp) =
		(case findHash (hash, getId expr) of
		     NONE => 
			 if (getId expr) = masterName then raise TypeError
			     ((pos2string (pos expr)) ^": parameter " 
			      ^ (pt2string expr)
			      ^ " is the same as the role/theorem name.")
			 else (insertHashDestructive (hash, getId expr, tp); ())
		   | SOME _ => raise TypeError 
			 ((pos2string (pos expr)) ^": " ^ (pt2string expr) ^ 
			  " appears twice in parameter list."))
	    fun addOne(param, tp) =
		(case tp of
		     PrincipalType _ => Priv(dp, param)
		   | New(_, PrincipalType _) => Priv(dp, param)
		   | _ => param)
	    fun isNew(_, New _) = true
	      | isNew _ = false
	    val params = flattenParams lst
	    val newParams = List.filter isNew params
	    val _ = List.app checkOne params
	    val newKnown = (List.map addOne newParams)
		handle Subscript => raise SympBug
		    ("Typecheck.paramTypes: some newIndex is out of range for "
		     ^masterName)
	in
	    (hash, union ptEq [newKnown, knownVals])
	end

    (* Check a key to be well-formed *)
    fun checkKey(pHash, k) =
	let fun checkId(i as Id(p,id)) =
	        (case findHash (pHash, id) of
		     NONE => raise TypeError ((pos2string (pos i)) ^
					      ": undeclared identifier ("
					      ^ id ^ ").")
		   | SOME (_, PrincipalType(_)) => ()
		   | SOME(_, New(_, PrincipalType _)) => ()
		   | SOME (_, tp) => raise TypeError
		       ((pos2string p) ^
			": attempts to extract key from non-principal ("^
			id ^ ") of type "^(pt2string tp)))
	      | checkId m = raise TypeError
		((pos2string (pos m)) ^
		 ": attempts to extract key from non-principal ("^
		 (pt2string m) ^ ").")
	    fun loop(Pub(p, id)) = checkId id
	      | loop(Priv(p, id)) = checkId id
	      | loop(SymKey(p, lst as [_,_])) = List.app checkId lst
	      | loop(SymKey(p, lst)) = raise TypeError
		  ((pos2string p)
		   ^": SymKey currently must take exactly two parameters.")
	      | loop(InvKey(p, k)) = loop k
	      | loop(Id(p,id)) =
		(case findHash(pHash, id) of
		     NONE => raise TypeError
			 ((pos2string p)^": undeclared identifier ("^id^").")
		   | SOME (_, tp) =>
			 if isSubtype(tp, KeyType dp) then ()
			 else raise TypeError
			 ((pos2string p)^"Expected ID of a key type, but found "
			  ^id^" of type "^(pt2string tp)))
	      | loop x = raise TypeError
		     ("Expected a key, found: "^(pt2string x))
	in
	    loop k
	end

    (* Typecheck a message term against the type `tp'.  Do not bother
       about other info than the types and that all the IDs are
       defined. *)

    fun checkMessage defined (m, tp) =
	let val funName = "checkMessage"
	    val _ = pushFunStackLazy(funName, fn()=>(pt2string m)^", "^(pt2string tp)
				     ^",\n defined = ["
				     ^(strlistmap2str ", " pt2stringVerbose defined)^"]")
	    fun isIn id (Parameter(_, lst, _)) = List.exists (isIn id) lst
	      | isIn id (TypedId(_, _, x, _)) = (x = id)
	      | isIn _ _ = false
	    fun findType(p, id) =
		(case List.find (isIn id) defined of
		     SOME(Parameter(_, _, tp)) => tp
		   | SOME(TypedId(_,_,_,tp)) => tp
		   | SOME x => raise SympBug
			 (funName^": bad parameter: "^(pt2stringVerbose x))
		   | NONE => raise TypeError ((pos2string p) ^ 
				  ": undeclared identifier (" ^
				  id ^ ")."))
	    (* fun checkId(p, id) =
	        (case List.find(fn TypedId(_, _, id', _) => id = id' | _ => false) definedList of
		     SOME pt => pt
		   | NONE => raise TypeError
		      ((pos2string p)^": undeclared identifier ("^id^").")) *)
	    fun loop(m as Id(p, id), tp) =
		let val tp' = findType(p, id)
		in
		    if isSubtype(tp', tp) then (m, tp')
		    else raise TypeError((pos2string p)^": type mismatch: ID ("^(pt2string m)
					 ^") of type "^(pt2string tp')
					 ^", expected type: "^(pt2string tp))
		end
	      | loop(Pub(p, m), tp) =
		if isSubtype(PubKeyType p, tp) then
		  let val (m', tp') = loop(m, PrincipalType p)
		  in
		      (Pub(p, m'), PubKeyType p)
		  end
		else raise TypeError
		    ((pos2string p)^": type mismatch: (PK _) has type PubKey, expected type: "
		     ^(pt2string tp))
	      | loop(Priv(p, m), tp) =
		if isSubtype(PrivKeyType p, tp) then
		  let val (m', tp') = loop(m, PrincipalType p)
		  in
		      (Priv(p, m'), PrivKeyType p)
		  end
		else raise TypeError
		    ((pos2string p)^": type mismatch: (PVK _) has type PrivKey, expected type: "
		     ^(pt2string tp))
	      | loop(m as SymKey(p, lst), tp) =
		if isSubtype(SymKeyType p, tp) then
		  let val _ = if (List.length lst) <> 2 then raise TypeError
		               ((pos2string p)
				^": SymKey currently must take exactly two parameters.")
			      else ()
		      val pairs = List.map(fn m=>loop(m, PrincipalType p)) lst
		  in
		      (SymKey(p, List.map #1 pairs), SymKeyType p)
		  end
		else raise TypeError
		    ((pos2string p)^": type mismatch: (SymKey _) has type SymKey, expected type: "
		     ^(pt2string tp))
	      | loop(m as InvKey(p, k), tp) =
		let val (k', tp') = loop(m, KeyType p)
		in
		    (case tp' of
			 KeyType _ => (keyInv k', tp')
		       | SymKeyType _ => (keyInv k', tp')
		       | PrivKeyType p => (keyInv k', PubKeyType p)
		       | PubKeyType p => (keyInv k', PrivKeyType p)
		       | x => raise SympBug
			  ("Typechecker.checkMessage: Non-key type in InvKey: "
			   ^(pt2string x)^", expr = "^(pt2stringVerbose m)
			   ^"\nTypechecked to: "^(pt2stringVerbose (InvKey(p, k')))))
		end
	      | loop(m as Crypt(p, m1, k), tp) =
		let val (k', ktp) = loop(k, KeyType p)
		    val (m1', tp1') = loop(m1, MessageType p)
		    val tp' = EncryptedType(p, tp1')
		in
		    if isSubtype(tp', tp) then (Crypt(p, m1', k'), tp')
		    else raise TypeError
			((pos2string p)^": type mismatch. Derived type: "
			 ^(pt2string tp')^",\n expected type: "^(pt2string tp)
			 ^"\n in message: "^(pt2string m))
		end
	      | loop(m as MessageTuple(p, lst), tp) =
		let val pairs = List.map (fn m=>loop (m, MessageType dp)) lst
		    val tp' = TupleType(p, List.map #2 pairs)
		in
		    if isSubtype(tp', tp) then
			(MessageTuple(p, List.map #1 pairs), tp')
		    else raise TypeError
			((pos2string p)^": type mismatch. Derived type: "
			 ^(pt2string tp')^",\n expected type: "^(pt2string tp)
			 ^"\n in message: "^(pt2string m))
		end
	      | loop (m,_) = raise SympBug
		  ("Typechecker.checkMessage: Not a message expression: "
		   ^(pt2stringVerbose m))
	    val res as (pt, tp) = loop(m, tp)
	    val _ = popFunStackLazy(funName, fn()=>(pt2string pt)^": "^(pt2string tp))
	in
	    res
	end

    (* Checks a Send action for proper use of keys, types. *)
    fun checkSend (pHash, m, initialized, knownVals) =
	let val funName = "checkSend"
	    val _ = pushFunStackLazy(funName, fn()=>(pt2string m)
				     ^",\n initialized = ["
				     ^(strlistmap2str ", " pt2string initialized)
				     ^"],\n knownVals = ["
				     ^(strlistmap2str ", " pt2string knownVals)
				     ^"]")
	    fun loop(Crypt (p, m, k)) =
	          if (not (knowVal (k, knownVals, pHash))) then
		      raise TypeError
			  ((pos2string p) 
			   ^": attempts to encrypt with unavailable key ("
			   ^(key2string k) ^ ").")
		  else loop m
	      | loop (i as Id (p, id)) =
		  (case findHash (pHash, id) of
		       NONE => raise TypeError
			   ((pos2string p) ^ ": undeclared identifier ("
			    ^ id ^ ").")
		     | SOME (_, New _) => union ptEq [[i], initialized]
		     | SOME _ => initialized)
	      | loop (MessageTuple (_, mlist)) =
		  List.foldr (fn (m, init) => checkSend (pHash, m, init, knownVals))
		             initialized mlist
	      | loop (MessageConcur (_, mlist)) =
		  List.foldr (fn (m, init) => checkSend (pHash, m, init, knownVals))
		             initialized mlist
	      | loop (key as Pub (p, k)) =
		  if (not (knowVal (key, knownVals, pHash))) then
		      raise TypeError ((pos2string p) ^
				       ": attempts to send unavailable key (" ^
				       key2string (key) ^ ").")
		  else initialized
	      | loop (key as Priv (p, k)) =
		  if (not (knowVal (key, knownVals, pHash))) then
		      raise TypeError ((pos2string p) ^
				       ": attempts to send unavailable key (" ^
				       key2string (key) ^ ").")
		  else initialized
	      | loop (key as SymKey(p, _)) =
		  if (not (knowVal (key, knownVals, pHash))) then
		      raise TypeError ((pos2string p) ^
				       ": attempts to send unavailable key (" ^
				       key2string (key) ^ ").")
		  else initialized
	      | loop (key as InvKey(p, k)) =
		  if (not (knowVal (keyInvHash(SOME pHash) k, knownVals, pHash))) then
		      raise TypeError ((pos2string p) ^
				       ": attempts to send unavailable key (" ^
				       key2string (key) ^ ").")
		  else (checkKey(pHash, key); initialized)
	      | loop m = initialized
	    val res = loop m
	    val _ = popFunStackLazy(funName, fn()=>" initialized = ["
				    ^(strlistmap2str ", " pt2string res)^"]")
	in
	    res
	end

    (* Checks a Receive action for proper use of keys, types, freshNonces. *)
    fun checkReceive (pHash, m, initialized, knownVals) =
	let val funName = "checkReceive"
	    val _ = pushFunStackLazy(funName, fn()=>(pt2string m)
				     ^",\n initialized = ["
				     ^(strlistmap2str ", " pt2string initialized)
				     ^"],\n knownVals = ["
				     ^(strlistmap2str ", " pt2string knownVals)
				     ^"]")
	    fun loop (Crypt (p, m, k)) =
	          if (not (knowVal (keyInvHash(SOME pHash) k, knownVals, pHash))) then
		      raise TypeError
			  ((pos2string p) ^
			   ": attempts to decrypt with unavailable key (" ^
			   (key2string (keyInvHash(SOME pHash) k)) ^ ").")
		  else loop m
	      | loop (i as Id (p, id)) =
		  (case findHash (pHash, id) of
		       NONE => raise TypeError ((pos2string p) ^
						": undeclared identifier (" ^
						id ^ ").")
		     | SOME (_, New _) =>
			   if (not (List.exists (fn Id(_, x) => x = id |
						 _ => false) initialized)) then
			       raise TypeError 
				   ((pos2string p) ^ 
				    ": tries to receive uncreated new value (" ^
				    id ^ ").")
			   else knownVals
		     | SOME(_, tp) => 
			   if isSubtype(tp, KeyType dp) andalso
			       not(knowVal(i, knownVals, pHash)) then i::knownVals
			   else knownVals)
	      | loop (MessageTuple (_, mlist)) =
		  List.foldr (fn (m, known) => checkReceive (pHash, m, initialized, 
							     known))
		             knownVals mlist
	      | loop (MessageConcur (_, mlist)) =
		  List.foldr (fn (m, known) => checkReceive (pHash, m, initialized, 
							     known))
		             knownVals mlist
	      | loop (pvk as Priv (p, i as Id(_,id))) =
		  (case findHash (pHash, id) of
		       NONE => raise TypeError ((pos2string (pos i)) ^
						": undeclared identifier ("
						^ id ^ ").")
		     | SOME (_, PrincipalType(_)) =>
			   if (not (knowVal (pvk, knownVals, pHash))) then
			       (pvk :: knownVals)
			   else knownVals
		     | SOME _ => 
			   raise TypeError 
			       ((pos2string p) ^
				": attempts to extract key from non-principal ("^
				id ^ ")."))
	      | loop (pvk as Priv (p, m)) =
		  raise TypeError ((pos2string (pos m)) ^
				   ": attempts to extract key from non-principal ("^
				   (pt2string m) ^ ").")
	      | loop (k as Pub _) = (checkKey(pHash, k); knownVals)
	      | loop (k as SymKey(p, _)) =
		let val _ = checkKey(pHash, k)
		in
		    if not(knowVal(k, knownVals, pHash)) then k::knownVals
		    else knownVals
		end
	      | loop (k as InvKey(p, k')) =
		let val _ = checkKey(pHash, k')
		in
		    if not(knowVal(keyInvHash(SOME pHash) k', knownVals, pHash)) then k::knownVals
		    else knownVals
		end
	      | loop m = knownVals
	    val res = loop m
	    val _ = popFunStackLazy(funName, fn()=>" knownVals = ["
				    ^(strlistmap2str ", " pt2string res)^"]")
	in
	    res
	end

    (* Checks the body of a role, tracking labels as well. *)
    fun checkBody (pHash, [], _, _, labels) = labels
      | checkBody (pHash, Send (p, m) :: rest, initialized, knownVals, labels) =
	  checkBody (pHash, rest, checkSend (pHash, m, initialized, knownVals),
		     knownVals, labels)
      | checkBody (pHash, Receive (p, m) :: rest, initialized, knownVals, labels) =
	  checkBody (pHash, rest, initialized,
		     checkReceive (pHash, m, initialized, knownVals), labels)
      | checkBody (pHash, Labeled (lid as Id(p,id), act) :: rest, initialized, 
		   knownVals, labels) =
	  if (not (List.exists (fn Id(_, x) => x = id |
				_ => false) labels)) then
	      checkBody (pHash, act :: rest, initialized, knownVals, 
			 (UniqueId (p, genSym id, id)) :: labels)
	  else
	      raise TypeError ((pos2string p) ^ ": duplicated label in role body.")
      | checkBody (pHash, h :: _, initialized, knownVals, labels) =
	  raise SympBug ((pos2string (pos h)) ^ 
			 ": non send/receive in protocol body.")

    (* Creates a hash mapping ids (string) to pairs
       (typecheckedParam, tp) for the formal parameters.
       The `tp' can be `new(tp)', but the `new' flag is stripped in the TypedId. *)
    fun TIDFormalParams NONE = makeHashDefault(op=, fn x=>x)
      | TIDFormalParams (SOME paramList) =
	let val hash = makeHashDefault(op=, fn x=>x)
	    fun stripNew(New(_, tp)) = tp
	      | stripNew tp = tp
	    fun addToHash (Parameter (_, list, tp)) =
		 List.app (fn Id(p,id) => 
			       (insertHashDestructive
				(hash, id, (TypedId (p, genSym id, id, stripNew tp), tp));
				())
			    | exp => raise SympBug
			       ((pos2string (pos exp) ^": formal parameter is not an ID!")))
		 list
	      | addToHash exp = raise SympBug 
		   ((pos2string (pos exp) ^ ": non-parameter in parameters!"))
	in
	    (List.app addToHash paramList; hash)
	end

    (* Return the role "closure" for the given expression, which in
       the case of a role is essentially unchanged.  First return
       value is the Role, second is the parameters (for easy access
       in typechecking) and the third is a list of labels in
       reverse order. *)
    fun typeCheckRoleExpr options 
      (role as (Role (p, {name=Id(_,id), params=pa, body=RoleBody(_,b)})))  =
      let
	val (paramTypeHash, knownVals : ParseTree list) = 
	  paramTypes (pa, id, makeHashDefault(op=, fn x => x), [])
	val paramHash = TIDFormalParams pa
	val labels = checkBody (paramHash, b, [], knownVals, [])
	fun TIDSubst (Id(pos, id)) =
	  (case findHash(paramHash, id) of
	     SOME (param, _) => param
	   | _ => UniqueId(pos, genSym id, id))
	  | TIDSubst (Labeled(Id(p,id),act)) =
	  (case (List.find (fn (UniqueId(_,_,x)) => x = id |
			    _ => false) labels) of
	     SOME (UniqueId(_,uid,_)) =>
	       Labeled (UniqueId(p,uid,id),act)
	   | SOME exp =>
	       raise SympBug ((pos2string (pos exp)) ^ 
			      ": this case should not occur (TIDSubst)!")
	   | NONE => 
	       raise SympBug ((pos2string p) ^ 
			      ": label somehow failed to be checked."))
	  | TIDSubst x = x
      in
	(ptTransform TIDSubst role, pa, labels)
      end
      | typeCheckRoleExpr options r = 
      raise SympBug ((pos2string (pos r)) ^ 
		     ": proper role expression expected.")

    (* Checks that ranges are in correct order and exist in the
       given role. *)
    fun checkLabelOrder (rid, p, label1, false, label2, true, [], l2) =
      raise TypeError ((pos2string p) ^ ": " ^ label1 ^ 
		       " not defined in role " ^ rid ^ 
		       " in ranged strand expression.")
      | checkLabelOrder (rid, p, label1, true, label2, false, [], l2) =
      raise TypeError ((pos2string p) ^ ": " ^ label2 ^ 
		       " not defined in role " ^ rid ^ 
		       " in ranged strand expression.")
      | checkLabelOrder (rid, p, label1, false, label2, false, [], l2) =
      raise TypeError ((pos2string p) ^ ": " ^ label1 ^ 
		       " not defined in role " ^ rid ^ 
		       " in ranged strand expression.")
      | checkLabelOrder (rid, p, label1, false, label2, false, 
			 UniqueId(_,uid,h)::t, l2) =
      if (h = label2) then
	if (not (label1 = label2)) then
	  checkLabelOrder (rid, p, label1, false, label2, true, t, uid)
	else
	  (uid, uid)
      else
	if (h = label1) then
	  checkLabelOrder (rid, p, label1, true, label2, false, t, l2)
	else
	  checkLabelOrder (rid, p, label1, false, label2, false, t, l2)
      | checkLabelOrder (rid, p, label1, true, label2,false,
			 UniqueId(_,uid,h)::t, l2) =
      if (h = label2) then
	raise TypeError ((pos2string p) ^ ": " ^ label2 ^ " precedes " ^
			 label1 ^ 
			 " in role, but not in ranged strand expression.")
      else
	checkLabelOrder (rid, p, label1, true, label2, false, t, l2)
      | checkLabelOrder (rid, p, label1, false, label2,true, 
			 UniqueId(_,uid,h)::t, l2) =
      if (h = label1) then
	(uid, l2)
      else
	checkLabelOrder (rid, p, label1, false, label2, true, t, l2)
      | checkLabelOrder (rid, p, label1, _, label2, _, _, _) =
      raise SympBug 
	((pos2string p) ^ 
	 ": label listing contains non-label, or true, true given!")

    fun getParamsTypes [] = []
      | getParamsTypes ((Parameter(_, lst, tp))::params) =
	  (List.map(fn _=> tp) lst)@(getParamsTypes params)
      | getParamsTypes (x::_) = raise SympBug
	  ("Typecheck.getParamsTypes: not a Parameter: "^(pt2stringVerbose x))

    fun checkParameters(p1, r, definedParams, SOME tparams, SOME sparameters) =
	let val tps = getParamsTypes tparams
	    val paramPairs =
		(case zipOpt(sparameters, tps) of
				  SOME lst => lst
				| NONE => raise TypeError
				      ((pos2string p1)
				       ^": The number of actual parameters ("
				       ^(Int.toString(List.length sparameters))
				       ^") doesn't match that of formal parameters ("
				       ^(Int.toString(List.length tps))
				       ^"\n in expression: "^(pt2string r)))
	    val definedList = (case definedParams of
				   SOME l => l
				 | NONE => [])
	    val tcPairs = List.map (checkMessage definedList) paramPairs
	in 
	    SOME(List.map #1 tcPairs)
	end
      | checkParameters(_, _, _, NONE, NONE) = NONE
      | checkParameters(p, r, _, NONE, SOME sparams) = raise TypeError
	  ((pos2string p) ^ ": object given parameters, expects none.")
      | checkParameters(p, r, _, SOME tparams, NONE) = raise TypeError
	  ((pos2string p) ^ ": object expects parameters, given none.")
	     

    (* Typechecks the strands in an atomic formula. *)
    fun checkStrands ([], _, _, new) = new
      | checkStrands ((r as RoleLabel (p1, Id(p2,rid), sparameters, 
				       Id(p3,label))) :: t, defined, params,
		      new) =
      (case (List.find (fn (Role (_, {name=UniqueId(_,_,x),...}), _, _) => 
			x = rid |
			_ => false) defined) of
	 NONE => raise TypeError ((pos2string p1) ^
				  ": undefined role (" ^
				  rid ^ ").")
       | SOME (role as Role(_,{name=UniqueId(_,uid,_),...}),tparams,labels) => 
	   let val sparameters = checkParameters(p1, r, params, tparams, sparameters)
	   in
	    (case (List.find (fn UniqueId(_,_,x) => x = label |
		   _ => false) labels) of
	       SOME (UniqueId(_,lid,_)) =>
		 (checkStrands (t, defined, params, 
				(RoleLabel (p1, UniqueId(p2,uid,rid),
					    sparameters, 
					    UniqueId(p3,lid,label))) :: new))
	     | SOME exp =>
		 raise SympBug ((pos2string (pos exp)) ^
				": should be a UniqueId (checkStrands)!")
	     | NONE =>
		 raise TypeError 
		   ((pos2string p2) ^
		    ": label " ^ label ^ " not defined in role " ^
		    rid ^ "."))
	   end
      | SOME (exp, _, _) =>
	   raise SympBug ((pos2string (pos exp)) ^ 
			  ": this case should not occur (checkStrands)!"))
      | checkStrands ((r as RoleLabel (p1, Id(p2,rid), sparameters, 
				       Range(p3,Id(p4,label1),Id(p5,label2)))) 
		      :: t, defined, params, new) =
      (case (List.find (fn (Role (_,{name=UniqueId(_,_,x),...}),_,_) => x=rid |
			_ => false)
	     defined) of
	 NONE => raise TypeError ((pos2string p1) ^
				  ": undefined role (" ^
				  rid ^ ").")
       | SOME (role as Role(_,{name=UniqueId(_,uid,_),...}),
	       tparams, labels) => 
	   let val sparameters = checkParameters(p1, r, params, tparams, sparameters)
	       val (uid1, uid2) =
		   checkLabelOrder (rid,p1,label1,false,label2,false,labels,"")
	   in
	      checkStrands (t, defined, params, 
			    RoleLabel (p1,UniqueId(p2,uid,rid),sparameters,
				       Range(p3,UniqueId(p4,uid1,label1),
					     UniqueId(p5,uid2,label2))) :: new)
	   end
       | SOME (exp, _, _) =>
	   raise SympBug ((pos2string (pos exp)) ^ 
			  ": this case should not occur (checkStrands)!"))
      | checkStrands ((r as RoleStrand (p, Id(_,rid), sparameters)) :: t,
		      defined, params, new) =
         (case (List.find (fn (Role (_,{name=UniqueId(_,_,x),...}),_,_) => x=rid |
			   _ => false) defined) of
	      NONE => raise TypeError
		  ((pos2string p) ^ ": undefined role (" ^ rid ^ ").")
	    | SOME (role as Role(_,{name=UniqueId(_,uid,_),...}), tparams, _) => 
		  let val sparameters = checkParameters(p, r, params, tparams, sparameters)
		  in
		      checkStrands (t, defined, params, 
				    RoleStrand(p,UniqueId(p,uid,rid),sparameters) :: new)
		  end
	    | SOME (exp, _, _) =>
		  raise SympBug ((pos2string (pos exp)) ^ 
				 ": this case should not occur (checkStrands)!"))
      | checkStrands (expr::t, _, _, _) =
	 raise SympBug ((pos2string (pos expr)) ^ ": not a strand expression!")

    (* Recurses into a formula structure, expanding theorems
       and performing typecheck on parameters. *)
    fun checkFormulaDown (Atomic (p, slist), defined, params: ParseTree list option) =
      let
	val formula = checkStrands (slist, defined, params, [])
      in
	Atomic (p, formula)
      end
      | checkFormulaDown (Not (p, f), defined, params) =
      let
	val formula = checkFormulaDown (f, defined, params)
      in
	Not (p, formula)
      end
      | checkFormulaDown (And (p, f1, f2), defined, params) =
      let
	val formula1 = checkFormulaDown (f1, defined, params)
	val formula2 = checkFormulaDown (f2, defined, params)
      in
	And (p, formula1, formula2)
      end
      | checkFormulaDown (Or (p, f1, f2), defined, params) =
      let
	val formula1 = checkFormulaDown (f1, defined, params)
	val formula2 = checkFormulaDown (f2, defined, params)
      in
	Or (p, formula1, formula2)
      end
      | checkFormulaDown (Implies (p, f1, f2), defined, params) =
      let
	val formula1 = checkFormulaDown (f1, defined, params)
	val formula2 = checkFormulaDown (f2, defined, params)
      in
	Implies (p, formula1, formula2)
      end
      | checkFormulaDown (FormulaInstance (p, Id(_,id), sparameters), 
			  defined, params) =
      (case (List.find (fn (PredicateContext (_,{name=UniqueId(_,_,x),...},_),
			    _, _) => x = id |
			(TheoremContext (_, {name=UniqueId(_,_,x),...},_),_,_) =>
			  x = id |
			_ => false) defined) of
	 NONE => (case (List.find(fn(Role(_,{name=UniqueId(_,_,x),...}),_,_) =>
				  x = id |
				  _ => false) defined) of
		    NONE => raise TypeError 
		      ((pos2string p) ^
		       ": undefined role, predicate or theorem (" ^
		       id ^ ").")
		    | SOME _ =>
		      let
			val formula = 
			  checkStrands ([RoleStrand(p,Id(p,id), sparameters)],
					defined, params, [])
		      in
			Atomic (p, formula)
		      end)
       | SOME (pred as PredicateContext 
	       (_,{name=UniqueId(_,uid,_),...},_), tparams, _) => 
	   let val sparameters = checkParameters(p, pred, params, tparams, sparameters)
	   in FormulaInstance (p, UniqueId(p,uid,id), sparameters)
	   end
       | SOME (theorem as TheoremContext 
	       (_,{name=UniqueId(_,uid,_),...},_), tparams, _) => 
	   let val sparameters = checkParameters(p, theorem, params, tparams, sparameters)
	   in FormulaInstance (p, UniqueId(p,uid,id), sparameters)
	   end
      | SOME (exp, _, _) => raise 
	 SympBug ((pos2string (pos exp)) ^ 
		  ": this case should not occur (checkFormulaDown)!"))
      | checkFormulaDown (expr, _, _) =
      raise SympBug ((pos2string (pos expr)) ^ 
		     ": not a formula!")

    fun findFresh params =
	let (* `Self' parameters are allowed to repeat for the same role *)
	    fun isNew(_, New(_, PrincipalType _)) = false
	      | isNew(_, New _) = true 
	      | isNew _ = false
	    fun isSelf(_, New(_, PrincipalType _)) = true
	      | isSelf _ = false
	    val params = flattenParams params
	in
	    (List.map isNew params, List.map isSelf params)
	     
	end

    fun binds bindHash defined r =
      let
	fun checkStrand(rid, sparams) =
	    let val (role, pa) = extractSomeRoleParams
		      (case 
			   (List.find 
			    (fn (Role (_, {name=UniqueId(_,x,_),...}), _, _) => 
			     x = rid |
			     _ => false) defined) of
			   SOME (r, _, _) => r
			 | NONE => 
			       raise SympBug 
				   ((pt2string r) ^ 
				    ": typechecker failed to detect undefined role reference!"))
		val (fresh, self) = findFresh pa
		fun checkRID(p, p', rid', sparams', id, self) =
		    (if not ((rid' = rid) andalso
			     (self orelse (ListPair.all ptEq (sparams, sparams'))))
			 then raise TypeError 
			     ((pos2string p) ^
			      ": " ^ id ^ " already bound at " ^
			      (pos2string p') ^ ".")
		     else ())
		fun checkParam self (TypedId(p,uid,id,_), true) =
		    (case findHash (bindHash, uid) of
			 SOME(RoleStrand (p', UniqueId (_, rid', _), SOME sparams')) =>
			     checkRID(p, p', rid', sparams', id, self)
		       | SOME (RoleLabel (p', UniqueId (_, rid', _), SOME sparams', _)) => 
			     checkRID(p, p', rid', sparams', id, self)
		       | SOME x => raise SympBug 
			   ("TypeCheck.binds: not a RoleStrand or RoleLabel in bindHash: "
			    ^(pt2stringVerbose x))
		       | NONE => (insertHashDestructive (bindHash, uid, r); ()))
		  | checkParam _ _ = ()
	    in
		(ListPair.app (checkParam false) (sparams, fresh);
		 ListPair.app (checkParam true) (sparams, self))		 
	    end
	fun loop(RoleStrand (p, UniqueId (_, rid, _), SOME sparams)) = checkStrand(rid, sparams)
	  | loop(RoleLabel (p, UniqueId (_, rid, _), SOME sparams, _)) = checkStrand(rid, sparams)
	  | loop pt = ()
      in
	  loop r
      end

    (* Check that we do not bind `new' parameters in different strands to the same value.
       Return *)

    fun bindCheck (theorem, defined) =
      let
	(* val expanded = expand (List.map #1 defined, theorem) *)
	val bindHash = (makeHashDefault(op=, (fn x => x)))
	fun loop pt = (binds bindHash defined pt;
		       List.app loop (ptSubtrees pt))
      in
	  loop theorem
      end
			      
    (* Performs typechecking/expansion on a theorem. *)
    fun checkTheorem options (Id(p,n), pa, def, ffContext) defined =
      let val funName = "checkTheorem"
	  val debug = lazyVerbDebug options funName
	val TIDFP = TIDFormalParams pa
	val _ = debug(fn()=>funName^": TIDFP = ["
		      ^(strlist2str ",\n " (hash2any(fn(id, (p,_))=>"("^id^" -> "
						     ^(pt2stringVerbose p)^")") TIDFP))
		      ^"]\n")
	fun TIDSubst (Id(pos, id)) =
	  (case findHash(TIDFP, id) of
	     SOME (param, _) => param
	   | _ => UniqueId(pos, genSym id, id))
	  | TIDSubst x = ptTransformContextOne TIDSubst x
	val pa = Option.map (List.map TIDSubst) pa
	val _ = debug(fn()=>funName^": params = "
		      ^(case pa of
			    SOME lst => "["^(strlistmap2str ", " pt2stringVerbose lst)^"]"
			  | NONE => "NONE")^"\n")
	val formula = (paramTypes(pa, n, makeHashDefault(op=, fn x => x), []);
		       checkFormulaDown (def, defined, pa))
	val theorem = TIDSubst(ffContext (p, {name=Id(p,n), params=pa, def=formula},
					  List.map #1 defined))
      in 
	(bindCheck (theorem, defined);
	 (theorem, pa, []))
      end
      | checkTheorem options (expr, _, _, _) _ =
      raise SympBug ((pos2string (pos expr)) ^ 
		     ": name of predicate/theorem is not an ID!"
		     ^(pt2stringVerbose expr))
	

    (* Typecheck a theorem expression.  A theorem is either defined
       or referred to by its name.  Expands any roles/theorems
       referred to in the formula.  Return is the same as for 
       typeCheckRoleExpr, except that third item is always empty. *)
    fun typeCheckTheoremInternal options (Theorem (_, {name=n, params=pa,
					       def=d})) defined =
      (checkTheorem options (n, pa, d, TheoremContext) defined)
    | typeCheckTheoremInternal options (Predicate (_, {name=n, params=pa,
					       def=d})) defined =
      (checkTheorem options (n, pa, d, PredicateContext) defined)
    | typeCheckTheoremInternal options expr _ =
      raise SympBug 
	((pos2string (pos expr)) ^ 
	 ": typeCheckTheoremInternal called on non-Predicate/Theorem!")
	

    local
	fun getId(Id(_, id)) = id
	  | getId(TypedId(_, _, id, _)) = id
	  | getId(UniqueId(_, _, id)) = id
	  | getId exp = raise SympBug
	      ("Typecheck.namePos: not a name: "^(pt2string exp))
    in
	(* Returns the name and position in a definition, for checking
	   against redefinition. *)
	fun namePos (Role (p,{name=id,...})) = (getId id, p)
	  | namePos (Predicate (p,{name=id,...})) = (getId id, p)
	  | namePos (Theorem (p,{name=id,...})) = (getId id, p)
	  | namePos (PredicateContext (p,{name=id,...},_)) = (getId id, p)
	  | namePos (TheoremContext (p,{name=id,...},_)) = (getId id, p)
	  | namePos x = (getId x, pos x)
    end				     

    (* Checks to make sure a name isn't being redefined.  We use a 
       single namespace for roles, predicates, and theorems to avoid
       confusion. *)
    fun checkRedefined (thing, []) = ()
      | checkRedefined (thing, ((h,_,_)::t)) =
      let
	val ((n1, p1), (n2, p2)) = (namePos thing, namePos h)
      in
	if (n1 = n2) then
	  raise TypeError ((pos2string p1) ^ ": redefinition of " ^ n1 ^
			   ", originally defined at " ^ (pos2string p2) ^ ".")
	else
	  checkRedefined (thing, t)
      end
      

    (* Typecheck and add the penetrator `id' (string name) to the
       current protocol named `n' to the hashes `protocols' and
       `protocolRecord'.

       Both regular and penetrator roles from the penetrator protocol
       are added.  This allows one to write extensions of existing
       penetrators. *)

    fun addPenetrator(p, n, id, protocols, protocolRecord: ProtocolHash, penetrRef) =
	let val funName = "addPenetrator"
	    val _ = pushFunStackLazy(funName, fn()=>"protocol: "^(pt2string n)
				     ^", penetrator: "^id)
	    val res = 
	    case findHash (protocols, id) of
		SOME (_, penetratorHash: ProtocolHash) =>
		    let val roleMap = (hash2any (fn x=>x) (#roles penetratorHash))
			             @(hash2any (fn x=>x) (#penetrator penetratorHash))
			val roleUidMap = (hash2any (fn x=>x) (#rolesUID penetratorHash))
			                @(hash2any (fn x=>x) (#penetratorUID penetratorHash))
			val newParamsMap = hash2any(fn x=>x) (#newParams penetratorHash)
			val selfParamsMap = hash2any(fn x=>x) (#selfParams penetratorHash)
			val roles = List.map #2 roleMap
			val newDefined = List.map 
			    (fn role as Role(_,{params=pa,body=RoleBody(_,b),...}) =>
			     (role, pa, 
			      List.rev
			      (List.map
			       (fn Labeled(uid,_) => uid |
				exp =>
				    raise SympBug 
					((pt2string exp) ^ 
					 " this should not occur (typeCheckAll, labels)!"))
			       (List.filter (fn Labeled(_,_) => true |
					     _ => false) b))) |
			     exp => raise SympBug 
				 ((pt2string exp) ^ 
				  " only roles should be imported!"))
			    roles
			fun addRole(rid, role) =
			    (insertHashDestructive (#penetrator protocolRecord, rid, role);
			     ())
			fun addRoleUID(rUid, role) =
			    (insertHashDestructive (#penetratorUID protocolRecord, rUid, role);
			     ())
			fun addNewParam(rUid, lst) =
			    (insertHashDestructive (#newParams protocolRecord, rUid, lst);
			     ())
			fun addSelfParam(rUid, lst) =
			    (insertHashDestructive (#selfParams protocolRecord, rUid, lst);
			     ())
			val _ = List.app addRole roleMap
			val _ = List.app addRoleUID roleUidMap
			val _ = List.app addNewParam newParamsMap
			val _ = List.app addSelfParam selfParamsMap
			val _ = penetrRef := id
		    in
			(newDefined, roles)
		    end
	      | NONE => raise TypeError ((pos2string p) ^ 
					 ": undefined penetrator (" ^ id ^ ").")
	    val _ = popFunStack(funName, "exiting")
	in
	    res
	end

    (* Recurses into definitions, accumulating them for use in later
       definitions, performing typechecking and expanding definitions when
       they are used. *)
    fun typeCheckAll options (Protocol (_, n, []), _, _, _, _) = []
      | typeCheckAll options (Protocol (_, n, (role as Role (p,
							     {name = Id(_,rn),
							      ...})) :: t), 
			      defs, protocols : (string, ParseTree *
						 ProtocolHash) Hash,
			      protocolRecord : ProtocolHash, penetrRef) =
      let
	val newDefined as (role, paramsOpt, _) = typeCheckRoleExpr options role
	fun getRoleUID(Role(_, {name=UniqueId(_, uid, _), ...})) = uid
	  | getRoleUID x = raise SympBug
	     ("typeCheckAll: Role name is not UniqueId: "^(pt2stringVerbose x))
	fun getNewIndices paramsOpt = 
	    let val funName = "getNewIndices"
		val debug = lazyVerbDebug(getOptions()) funName
		fun pair2str(pt, tp) = (pt2string pt)^": "^(pt2string tp)
		val _ = pushFunStackLazy(funName, fn()=>
					 (case paramsOpt of
					      SOME params => strlistmap2str ", " pair2str params
					    | NONE => "NONE"))
		(* all but new principals *)
		fun filterIndex(i, (_, New(_, PrincipalType _))) = NONE
		  | filterIndex(i, (_, New _)) = SOME i
		  | filterIndex _ = NONE
		(* Only the new principals (Self parameters) *)
		fun filterSelfIndex(i, (_, New(_, PrincipalType _))) = SOME i
		  | filterSelfIndex _ = NONE
		val res =
		    (case paramsOpt of
			 SOME params =>
			     (List.mapPartial (fn x=>x) (mapEnumerate filterIndex params),
			      List.mapPartial (fn x=>x) (mapEnumerate filterSelfIndex params))
			   | NONE => ([],[]))
		val _ = popFunStackLazy(funName, fn()=>
					"self = ["^(strlistmap2str ", " Int.toString (#2 res))
					^"], others = ["
					^(strlistmap2str ", " Int.toString (#1 res))^"]")
	    in
		res
	    end
	val (newParams, selfParams) = getNewIndices(Option.map flattenParams paramsOpt)

      in
	(checkRedefined (role, defs);
	 insertHashDestructive (#roles protocolRecord, rn, role);
	 insertHashDestructive (#rolesUID protocolRecord, getRoleUID role, role);
	 insertHashDestructive (#newParams protocolRecord, getRoleUID role, newParams);
	 insertHashDestructive (#selfParams protocolRecord, getRoleUID role, selfParams);
	 role :: (typeCheckAll options (Protocol (p,n,t), 
					newDefined :: defs, protocols,
					protocolRecord, penetrRef)))
      end
      | typeCheckAll options 
      (Protocol (_, n, ((pred as Predicate (p,{name = Id(_,pn),
					       ...})) :: t)), defs, protocols,
       protocolRecord, penetrRef) =
      let
	val newDefined as (pred, _, _) = 
	  typeCheckTheoremInternal options pred defs
      in
	(checkRedefined (pred, defs);
	 insertHashDestructive (#predicates protocolRecord, pn, pred); 
	 pred :: (typeCheckAll options (Protocol (p,n,t), 
					newDefined :: defs, protocols,
					protocolRecord, penetrRef)))
      end
      | typeCheckAll options 
      (Protocol (_, n, ((theorem as Theorem (p,{name = Id(_,tn),
						...})) :: t)), defs, protocols,
       protocolRecord, penetrRef) =  
      let
	val newDefined as (theorem, _, _) = 
	  typeCheckTheoremInternal options theorem defs
      in
	(checkRedefined (theorem, defs);
	 insertHashDestructive (#theorems protocolRecord, tn, theorem); 
	 theorem :: (typeCheckAll options (Protocol (p,n,t), 
					   newDefined :: defs, protocols,
					   protocolRecord, penetrRef)))
      end
      | typeCheckAll options
      (Protocol (_, n, (penetrator as Penetrator (p,Id(_,id))) :: t), defs,
       protocols, protocolRecord, penetrRef) =
         let val (newDefined, roles) = addPenetrator(p, n, id, protocols, protocolRecord,
						     penetrRef)
	 in roles @ (typeCheckAll options (Protocol (p,n,t),
					   newDefined @ defs,
					   protocols,
					   protocolRecord, penetrRef))
	 end
      
      | typeCheckAll options (Protocol (p, n, h::t), _, _, _, _) = 
      raise SympBug ((pos2string p) ^ 
		     ": role, penetrator, predicate, or theorem expected (" ^
		     (pt2string h) ^ ").")
      | typeCheckAll options (expr, _, _, _, _) =
      raise SympBug ((pos2string (pos expr)) ^ 
		     ": protocol expected (in typeCheckAll)!")

    (* Return the Protocol, with all roles and theorems checked and
       (for theorems) expanded fully.  Raises TypeError if there is
       any problem. *)
    fun typeCheckProtocol options program 
      (expr as Protocol (p, name as Id(_,n), _))=
      let val funName = "Athena:typeCheckProtocol"
	  val _ = pushFunStackLazy(funName, fn()=>pt2string expr)
	  val penetratorProtocolRef = ref ""
	val protocolRecord as {roles=roles,
			       rolesUID=rolesUID,
			       penetrator=penetrator,
			       penetratorUID=penetratorUID,
			       newParams=newParams,
			       selfParams=selfParams,
			       predicates=predicates,
			       theorems=theorems,
			       ...}
	      = {roles = makeHashDefault (op=, fn x=>x),
		 rolesUID = makeHashDefault (op=, fn x=>x),
		 penetratorProtocol = "",
		 penetrator = makeHashDefault (op=, fn x=>x),
		 penetratorUID = makeHashDefault (op=, fn x=>x),
		 newParams=makeHashDefault(op=, fn x=>x),
		 selfParams=makeHashDefault(op=, fn x=>x),
		 predicates = makeHashDefault (op=, fn x=>x),
		 theorems = makeHashDefault (op=, fn x=>x)} : ProtocolHash

	val checked = typeCheckAll options (expr, [], program, protocolRecord,
					    penetratorProtocolRef)
	val newProtocol = {roles=roles,
			   rolesUID=rolesUID,
			   penetrator=penetrator,
			   penetratorUID=penetratorUID,
			   newParams=newParams,
			   selfParams=selfParams,
			   predicates=predicates,
			   theorems=theorems,
			   penetratorProtocol = !penetratorProtocolRef}
	val res =
	    case findHash (program, n) of 
		NONE =>
		    (insertHashDestructive (program, n, (Protocol (p, name, checked),
							 newProtocol));
		     Protocol (p, name, checked))
	      | SOME (Protocol (p', _, _), _) =>
		    (raise TypeError ((pos2string p) ^ ": redefinition of " ^ n ^
				      ", originally defined at " ^ (pos2string p')
				      ^ "."))
	      | SOME _ => raise SympBug "This cannot happen (typeCheckProtocol)."
	val _ = popFunStack(funName, "exiting")
      in res
      end
      | typeCheckProtocol options _ expr =
      raise SympBug ((pos2string (pos expr)) ^ 
		     ": protocol expected (in typeCheckProtocol)!")

    (* Return the Program, with all roles and theorems checked and
       (for theorems) expanded fully.  Raises TypeError if there is
       any problem. *)
    fun typeCheckProgram options (pt as Program (p, decls)) =
      let val funName = "Athena:typeCheckProgram"
	  val _ = pushFunStackLazy(funName, fn()=> pt2string pt)
	val program = makeHashDefault (op=, fn x=>x)
	val debug = verbDebug options funName
	  "Starting typechecking.\n"
	(* Typecheching the default penetrator first *)
	val defPenetrator = typeCheckProtocol options program DefaultPenetrator.penetrator
	val emptyPenetrator = typeCheckProtocol options program EmptyPenetrator.penetrator
	val newDecls = List.map (typeCheckProtocol options program) decls
	val res = Global program
	val _ = popFunStackLazy(funName, fn()=>ProgramContext2str res)
      in
	 res
      end
      | typeCheckProgram options expr =
      raise SympBug ((pos2string (pos expr)) ^ 
		     ": non-program given to typeCheckProgram!")

    (* First parsetree should be a context, second should
       be an ID.  Returns the associated Theorem, if one exists.  
       Otherwise raises TypeError. *) 
    fun typeCheckTheorem options (Local (pname, (_, context))) (Id(_,name)) =
      (case (findHash (#theorems context, name)) of
	 SOME theorem => (Object theorem)
       | NONE => raise TypeError ("The theorem " ^
				  name ^ " is not defined in this context ("
				  ^ pname ^ ")."))
      | typeCheckTheorem _ _ _ =
      raise SympBug ("Improper context or theorem in Athena:typeCheckTheorem.")

    (* First parsetree should be the full context, second should
       be an ID.  Returns the associated local context, if one exists.  
       Otherwise raises TypeError. *)
    fun typeCheckContext options (Global protocolHash) (Id(_,name)) =
      (case (findHash (protocolHash, name)) of
	 SOME protocol => Local (name, protocol)
       | NONE => raise TypeError ("The protocol " ^
				  name ^ " is not defined in the program."))
      | typeCheckContext _ _ _ =
      raise SympBug ("Improper context or theorem in Athena:typeCheckContext.")

end
