(* SYMP parser in SML'97 *)

structure AthenaParseTreeStruct = ParseTreeStruct
structure ParseError = ParseError
open ParseTreeStruct
open ParseError

fun tuple2list (TupleType(_,lst)) = lst
  | tuple2list x = [x]

%%

%name Athena

%verbose

%pure

%header (functor AthenaLrValsFun(structure Token : TOKEN
			       structure ParseTreeStruct: ATHENA_PARSE_TREE
			       structure ParseError: PARSE_ERROR)
 : sig structure ParserData : PARSER_DATA
            where type result = ParseTreeStruct.ParseTree;
       structure Tokens : Athena_TOKENS
   end)

%eop EOF
%pos int
%noshift EOF

%term PROTOCOL of Pos.Pos
| ROLE of Pos.Pos
| BEGIN of Pos.Pos
| END of Pos.Pos
| RP of Pos.Pos
| LP of Pos.Pos
| LB of Pos.Pos
| RB of Pos.Pos
| LCB of Pos.Pos
| RCB of Pos.Pos
| SEND of Pos.Pos
| RECEIVE of Pos.Pos
| NONCE of Pos.Pos
| FRESHNONCE of Pos.Pos
| NEW of Pos.Pos
| MESSAGE of Pos.Pos
| SYMKEY of Pos.Pos
| PRIVKEY of Pos.Pos
| PUBKEY of Pos.Pos
| ENCRYPTED of Pos.Pos
| TIMES of Pos.Pos
| INV of Pos.Pos
| DOT of Pos.Pos
| EQ of Pos.Pos
| NOTEQ of Pos.Pos
| IMPLIES of Pos.Pos
| ARROW of Pos.Pos
| DARROW of Pos.Pos
| BAR of Pos.Pos
| DBAR of Pos.Pos
| COMMA of Pos.Pos
| COLON of Pos.Pos
| SEMI of Pos.Pos
| AND of Pos.Pos
| OR of Pos.Pos
| NOT of Pos.Pos
| IFF of Pos.Pos
| EOF of Pos.Pos
| EQDEF of Pos.Pos
| PUB of Pos.Pos
| PRIV of Pos.Pos
| KEY of Pos.Pos
| DECRYPT of Pos.Pos
| PENETRATOR of Pos.Pos
| PREDICATE of Pos.Pos
| PRINCIPAL of Pos.Pos
| THEOREM of Pos.Pos
| QUOTE of Pos.Pos * string
| ID of Pos.Pos * string
| MINUS of Pos.Pos
| SELF of Pos.Pos

 (* %prefer SEMI *)

%left SEMI
%left COMMA
%nonassoc BEGIN END
%nonassoc PROTOCOL
%nonassoc ROLE PENETRATOR PREDICATE THEOREM
%nonassoc SEND RECEIVE
%left IFF
%right  IMPLIES ARROW DARROW
%left OR
%left AND
%nonassoc COLON
%nonassoc NOT
%left MINUS
%left TIMES
%nonassoc NEW
%nonassoc LP RP LB RB SELF NONCE MESSAGE FRESHNONCE
%nonassoc SYMKEY

%nonterm protocol of ParseTree
| semi
| comma
| program of ParseTree
| role of ParseTree
| programdefs of ParseTree list
| declarations of ParseTree list
| declaration of ParseTree
| parameters of ParseTree list option
| roleparameters of ParseTree list option
| sparameters of ParseTree list option
| roleexpr of ParseTree
| ids of ParseTree list
| params of ParseTree list
| param of ParseTree
| roleparams of ParseTree list
| roleparam of ParseTree
| sparams of ParseTree list
| sparam of ParseTree
| mexpr of ParseTree
| mtexpr of ParseTree
| mtuple of ParseTree list
| mconcur of ParseTree list
| mconcur2 of ParseTree list
| pactions of ParseTree list
| paction of ParseTree
| atomtype of ParseTree
| tupletype of ParseTree list
| typeexpr of ParseTree
| predicates of ParseTree
| predicate of ParseTree
| penetrator of ParseTree
| theorems of ParseTree
| theorem of ParseTree
| formula of ParseTree
| atomicformula of ParseTree
| label of ParseTree
| labeling of ParseTree
| strand of ParseTree
| strands of ParseTree list

%start program

%%

program 	: programdefs (Program(pos(List.hd(programdefs)),programdefs))
programdefs     : protocol semi ([protocol])
                | protocol semi programdefs (protocol::programdefs)
protocol	: PROTOCOL ID EQ BEGIN declarations END
	            (Protocol(PROTOCOL,Id(ID),
                              (Penetrator(BEGIN, Id(BEGIN, "DefaultPenetrator")))::declarations))
	        | PROTOCOL ID EQ BEGIN penetrator declarations END
                    (Protocol(PROTOCOL,Id(ID),penetrator::declarations))
semi		: SEMI () | ()
comma		: COMMA () | ()
mtexpr		: mexpr (mexpr)
		| mconcur (MessageConcur(pos(List.hd(mconcur)),mconcur))
mexpr		: ID (Id(ID))
		| LP mtuple RP ((case mtuple of
                                    [x] => x
                                  | _ => MessageTuple(LP,mtuple)))
		| LCB mexpr RCB mexpr (Crypt (LCB, mexpr1, mexpr2))
		| PUB mexpr (Pub (PUB, mexpr))
		| PRIV mexpr (Priv (PRIV, mexpr))
                | INV mexpr (InvKey(INV, mexpr))
                | SYMKEY LP mtuple RP (SymKey(SYMKEY, mtuple))
mconcur		: mexpr AND mconcur2 (mexpr::mconcur2)
mconcur2	: mexpr ([mexpr])
		| mexpr AND mconcur2 (mexpr::mconcur2)
mtuple		: mexpr ([mexpr])
		| mexpr COMMA mtuple (mexpr::mtuple)
declarations	: ([])
                | declaration semi declarations (declaration::declarations)
declaration	: role (role)
		| predicate (predicate)
		| theorem (theorem)
paction		: SEND mtexpr (Send(SEND,mtexpr))
		| RECEIVE mtexpr (Receive(RECEIVE,mtexpr))
		| ID COLON paction (Labeled(Id(ID), paction))
pactions	: paction semi ([paction])
                | paction semi pactions (paction::pactions)
roleexpr	: BEGIN pactions END (RoleBody(BEGIN,pactions))
role		: ROLE ID roleparameters EQ roleexpr
                  (Role(ROLE,
			{ name = Id(ID),
			  params = roleparameters,
			  body = roleexpr }))
predicate	: PREDICATE ID parameters EQ formula
                  (Predicate(PREDICATE,
			{ name = Id(ID),
			  params = parameters,
			  def = formula }))
penetrator	: PENETRATOR ID (Penetrator (PENETRATOR,Id(ID)))
theorem		: THEOREM ID parameters EQ formula
                  (Theorem(THEOREM,
			{ name = Id(ID), 
			  params = parameters,
			  def = formula }))
formula		: atomicformula (atomicformula)
		| formula OR formula (Or(pos(formula1), formula1, formula2))
		| formula AND formula (And(pos(formula1), formula1, formula2))
		| formula IMPLIES formula (Implies(pos(formula1), formula1, formula2))
		| formula ARROW formula (Implies(pos(formula1), formula1, formula2))
		| formula DARROW formula (Implies(pos(formula1), formula1, formula2))
		| NOT formula (Not(NOT, formula))
                | LP formula RP (formula)
atomicformula	: LCB strands RCB (Atomic(pos(List.hd strands), strands))
		| ID sparameters (FormulaInstance(pos(Id(ID)), Id(ID), sparameters))
		| ID sparameters DOT labeling
                     (Atomic(pos(Id(ID)),
                             [RoleLabel (pos(Id(ID)), Id(ID), sparameters, labeling)]))
strands		: strand ([strand])
		| strand COMMA strands (strand::strands)
strand		: ID sparameters (RoleStrand (pos(Id(ID)), Id(ID), sparameters))
		| ID sparameters DOT labeling (RoleLabel (pos(Id(ID)), Id(ID), sparameters, labeling))
labeling	: ID (Id (ID))
		| ID MINUS label (Range (pos(Id(ID)), Id(ID), label))
label		: ID (Id (ID))
atomtype	: MESSAGE (MessageType(MESSAGE))
		| NONCE (NonceType(NONCE))
		| PRINCIPAL (PrincipalType(PRINCIPAL))
		| KEY (KeyType(KEY))
                | PRIVKEY (PrivKeyType(PRIVKEY))
                | PUBKEY (PubKeyType(PUBKEY))
                | SYMKEY (SymKeyType(SYMKEY))
                | ENCRYPTED LB typeexpr RB (EncryptedType(ENCRYPTED, typeexpr))
                | LP typeexpr RP (typeexpr)
typeexpr        : tupletype (case tupletype of
                                [x] => x
                              | lst as h::_ => TupleType(pos h, lst)
                              | [] => raise ParseError("Empty tuple type (this shouldn't happen)"))
tupletype       : atomtype ([atomtype])
                | atomtype TIMES tupletype (atomtype::tupletype)
sparameters	: (NONE)
		| LB sparams RB (SOME(sparams))
sparam		: mexpr (mexpr)
sparams		: sparam ([sparam])
		| sparam COMMA sparams (sparam::sparams)
parameters	: (NONE)
		| LB params RB (SOME(params))
roleparameters	: (NONE)
		| LB roleparams RB (SOME(roleparams))
ids		: ID comma ([Id(ID)])
		| ID comma ids ((Id(ID))::ids)
param		: ids COLON typeexpr (Parameter(COLON, ids, typeexpr))
params		: param semi ([param])
		| param semi params (param::params)
roleparam	: ids COLON typeexpr (Parameter(COLON, ids, typeexpr))
                | ids COLON NEW typeexpr (Parameter(COLON, ids, New(NEW, typeexpr)))
                | ids COLON FRESHNONCE
                    (Parameter(COLON, ids, New(FRESHNONCE, NonceType FRESHNONCE)))
                | ids COLON SELF (Parameter(COLON, ids, New(SELF, PrincipalType SELF)))
roleparams	: roleparam semi ([roleparam])
		| roleparam semi roleparams (roleparam::roleparams)
