(* Implementation of the AbstractSplit rule.

   The rule splits cases on different interpretations of uninterpreted
   (Skolem and abstract) constants.  The result will contain only
   abstract constants, and if the model otherwise finite, it will be
   suitable for model checking.

 *)

functor AbstractSplitFun(structure Hash: HASH
			 structure ProofRulesCommon: PROOF_RULES_DEFAULT_COMMON): ABSTRACT_SPLIT =
  struct

    structure ProofRulesCommon = ProofRulesCommon
(*      structure UnionFind = UnionFindFun(structure Hash = Hash) *)

    (* Local exception to abort the rule *)
    exception Quit of string

    open ProofRulesCommon.Sequent.SequentCommon
    open ProofRulesCommon
    open Sequent
    open SequentBase
    open Trans.Abstract
    open InputLanguage
    open Typecheck.Evaluate.ParserDefault.Interface
    open Typecheck.Evaluate.ParserDefault.ParseTreeStruct
    open Hash
    open ListOps
    open Pos

    datatype Group =
	Group of { skolem: ParseTree list,
		   abstract: ParseTree list,
		   interpreted: ParseTree list }

    (* Check if the expression is a constant.  If `sk' = true, then
       consider Skolem constants as constants, otherwise not. *)

    fun isConstCommon sk (SkolemConst _) = sk
      | isConstCommon sk (AbstractConst _) = true
      | isConstCommon sk (StaticFormalConst _) = sk
      | isConstCommon sk (Appl(_, x, y)) = isConstCommon sk x andalso isConstCommon sk y
      | isConstCommon sk (TupleExpr(_, lst)) = List.all (isConstCommon sk) lst
      | isConstCommon sk (RecordExpr(_, lst)) =
	  List.all(fn (RecordAsst(_,_,e))=>isConstCommon sk e | _ => false) lst
      | isConstCommon sk (ExtractAppl(_, e)) = isConstCommon sk e
      | isConstCommon sk (ExtractTuple(_, e)) = isConstCommon sk e
      | isConstCommon sk (ExtractRecord(_, e)) = isConstCommon sk e
      | isConstCommon sk (ExtractIndex e) = isConstCommon sk e
      | isConstCommon sk  x = isValue x

    val isConst = isConstCommon true

    (* `stable' here means it doesn't alias with anything but,
        possibly, Skolem constants.  In other words, exclude Skolem
        constants from it. *)

    val isStableConst = isConstCommon false

    (* Check for a state var expression *)
    fun isStateVar (StateVar _) = true
      | isStateVar (DynPatternFormal _) = true
      | isStateVar (ExtractTuple(_,e)) = isStateVar e
      | isStateVar (ExtractRecord(_,e)) = isStateVar e
      | isStateVar (ExtractAppl(_,e)) = isStateVar e
      | isStateVar (ExtractIndex(e)) = isStateVar e
      (* We are only interested in statically known components of a state var *)
      | isStateVar (Appl(_,e,arg)) = isStateVar e andalso isStableConst arg
      | isStateVar _ = false

    (* Collect all the Skolem constants in the list of formulas.
       Uninterpreted static parameters are treated as Skolem constants. *)
    fun collectSkolem lst =
	let val hash = makeHashDefault(ptEq, pt2stringDebug)
	    fun isSkolemFunction(SkolemConst _) = true
	      | isSkolemFunction(StaticFormalConst _) = true
	      | isSkolemFunction(Appl(_, ff, arg)) =
		  isSkolemFunction ff andalso isConst arg
	      | isSkolemFunction _ = false
	    fun loop(c as SkolemConst _) = 
		  (insertHashDestructive(hash, c, ()); c)
	      | loop(c as StaticFormalConst{value=NONE,...}) =
		  (insertHashDestructive(hash, c, ()); c)
	      | loop(c as Appl _) = 
		  if isSkolemFunction c then 
		      (insertHashDestructive(hash, c, ()); c)
		  else c
	      | loop x = ptTransform loop x
	in
	    (List.map loop lst; hash2any(fn (x,_)=>x) hash)
	end

    (* Collect all the abstract constants in the list of formulas *)
    fun collectAbstract lst =
	let val hash = makeHashDefault(ptEq, pt2stringDebug)
	    fun loop(c as AbstractConst _) = 
		  (insertHashDestructive(hash, c, ()); c)
	      | loop x = ptTransform loop x
	in
	    (List.map loop lst; hash2any(fn (x,_)=>x) hash)
	end

    (* Collect all the interpreted constants in the list of formulas *)
    fun collectInterpreted lst =
	let val hash = makeHashDefault(ptEq, pt2stringDebug)
	    fun loop c = if isValue c then (insertHashDestructive(hash, c, ()); c)
			 else ptTransform loop c
	in
	    (List.map loop lst; hash2any(fn (x,_)=>x) hash)
	end

    (* For a constant `c' return the list of its primitive-type
       components with their types, and a fully expanded version of
       the constant into its primitive expressions. *)

    fun splitTypes findObject c =
	let fun wrapExtrRecord c (RecordField{name=name, Type=tp}) = (ExtractRecord(name, c), tp)
	      | wrapExtrRecord c x = raise SympBug
	         ("AbstractSplit.splitTypes/wrapExtrRecord: not a RecordField: "
		  ^(pt2string x))
	    fun wrapExtrTuple c (i, tp) = (ExtractTuple(i, c), tp)
	    fun wrapRecord([], acc) = RecordExpr(dp, acc)
	      | wrapRecord((c as ExtractRecord(name, _))::lst, acc) = 
		  wrapRecord(lst, (RecordAsst(dp, name, c))::acc)
	      | wrapRecord (c::_, _) = raise SympBug
		  ("AbstractSplit.splitTypes/wrapRecord: not an ExtractRecord: "
		  ^(pt2string c))
	    fun wrapTuple lst = TupleExpr(dp, lst)
	    fun loop(c: ParseTree, RecordType(_, lst)) =
	          let val fields = List.map(wrapExtrRecord c) lst
		      val resList = List.map loop fields
		      val pairs = List.foldr(op @) [] (List.map(fn(x,_)=>x) resList)
		      val subexprs = List.map(fn(_,x)=>x) resList
		  in (pairs, wrapRecord(subexprs, []))
		  end
	      | loop(c, TupleType(_, lst)) =
	          let val num = List.tabulate(List.length lst, fn i=>i)
		      val fields = List.map(wrapExtrTuple c) (zip(num,lst))
		      val resList = List.map loop fields
		      val pairs = List.foldr(op @) [] (List.map #1 resList)
		      val subexprs = List.map #2 resList
		  in (pairs, wrapTuple subexprs)
		  end
	      | loop(c, tp as EnumType _) = raise SympBug
		  ("Abstraction for datatypes is not implemented yet:\n  "
		   ^(pt2string tp))
	      | loop(c, tp as FunType _) = raise ProverError
		  ("Cannot abstract a bare functional constant: "
		   ^(pt2string(TypedExpr(dp, c, tp)))
		   ^"\n\nEvery uninterpreted function must be applied to something\n"
		   ^"before it can be abstracted.")
	      | loop(c, tp as Uid _) = raise SympBug
		  ("Abstraction for UID (that's most likely a datatype) "
		   ^"is not implemented yet:\n  "^(pt2string tp))
	      | loop(c, tp) = ([(c, tp)], c)
	in loop(c, getExprType findObject c): ((ParseTree * ParseTree) list) * ParseTree
	end

    (* Collect all the constants from the list of formulas (filtered
       by `ff'), split them into elementary types, and return three
       hashes: one with all elementary constants mapping them to their
       types, the second mapping types to the list of constants, and
       the third mapping constants to their elementary components with
       types. *)

    fun collectSplitCommon(findObject, ff) lst =
	let val constRaw = ff lst
	    val constHash = makeHashDefault(ptEq, pt2string)
	    val typeHash = makeHashDefault(ptEq, pt2string)
	    (* Hash mapping each constant into the list of its primitive components *)
	    val primHash = makeHashDefault(ptEq, pt2string)
	    (* Mapping each const. into its fully expanded version *)
	    val expHash = makeHashDefault(ptEq, pt2string)
	    fun splitTypesRecord c =
		let val (pairs, expandedC) = splitTypes findObject c
		    val _ = insertHashDestructive(primHash, c, pairs)
		    val _ = insertHashDestructive(expHash, c, expandedC)
		in pairs
		end
	    val constAtomicLists = List.map splitTypesRecord constRaw
	    fun addConst(c, tp) = 
		(insertHashDestructive(constHash, c, tp);
		 case findHash(typeHash, tp) of
		     NONE => insertHashDestructive(typeHash, tp, [c])
		   | SOME lst => insertHashDestructive(typeHash, tp, c::lst);
		 ())
	    val _ = List.app (List.app addConst) constAtomicLists
	in
	    (constHash, typeHash, primHash, expHash)
	end

    fun collectSplitSkolem findObject lst = 
	  collectSplitCommon(findObject, collectSkolem) lst
    fun collectSplitAbstract findObject lst = 
	  collectSplitCommon(findObject, collectAbstract) lst
    fun collectSplitInterpreted findObject lst =
	  collectSplitCommon(findObject, collectInterpreted) lst

    (* The final rule specs *)
    val specs = [{name="constants",
		  Type=ProverArgListType ProverArgObjectType,
		  default=DefaultRequired}]
    val concTemp = { model = ValueTemp "model",
		     init = ValueTemp "init",
		     context = ProgramContextTemp "context",
		     invar = [ListFormulaTemp("invar",NONE)],
		     assumptions = [ListFormulaTemp("assump",NONE)],
		     conclusions = [ListFormulaTemp("conc",NONE)] }

    (* Take an expanded constant and try to fold it back into as
       compact form as possible *)

    fun collapseExp c =
	let (* Check that all fields are extracted from the same expression *)
	    fun checkRecord'(e, (RecordAsst(_, n, ExtractRecord(n', e'))::fields)) =
		ptEq(n,n') andalso ptEq(e, e') andalso checkRecord'(e, fields)
	      | checkRecord'(e, _::_) = false
	      | checkRecord'(e, []) = true
	    fun checkRecord [] = NONE
	      | checkRecord((ExtractRecord(_, e))::fields) =
		if checkRecord'(e, fields) then SOME e
		else NONE
	      | checkRecord(_::_) = NONE
	    (* Same for tuples *)
	    fun checkTuple'(e, i, (ExtractTuple(j, e'))::lst) =
		i = j andalso ptEq(e, e') andalso checkTuple'(e, i+1, lst)
	      | checkTuple'(e, _, _::_) = false
	      | checkTuple'(_, _, []) = true
	    fun checkTuple [] = NONE
	      | checkTuple((ExtractTuple(0, e))::lst) =
		if checkTuple'(e, 1, lst) then SOME e
		else NONE
	      | checkTuple(_::_) = NONE
	    (* The main recursive descent *)
	    fun loop(RecordExpr(p, lst)) =
		let val newList = List.map loop lst
		in 
		    case checkRecord newList of
			SOME e => e
		      | NONE => RecordExpr(p, newList)
		end
	      | loop(TupleExpr(p, lst)) =
		let val newList = List.map loop lst
		in
		    case checkTuple newList of
			SOME e => e
		      | NONE => TupleExpr(p, lst)
		end
	      | loop x = ptTransform loop x
	in loop c
	end

    fun apply printFun (seq: Sequent, context: ProgramContext, args: RuleArg list, sub) =
	let val funName = "ProofRules.AbstractSplit.apply"
	    val debug = lazyVerbDebug(getOptions()) funName
	    val objHash = ProgramContext2ObjectHash context
	    val contextPt = ProgramContext2pt context
	    val findObject = Typecheck.makeFindObject objHash
	    val (_, termHash) = sub
	    (* The hash of all newly generated unames (strings)
	       mapping them to the constants *)
	    val newNamesHash = makeHashDefault(op =, fn x=>x)
	    (* Abstract mapping of primitive type constants to abstract constants
	       generated so far *)
	    val absHash = makeHashDefault(ptEq, pt2string)

	    fun pair2str(x, lst) = 
		"("^(pt2stringDebug x)^", ["^(ptlist2strDebug "," lst)^"])"
	    fun pairs2str pairs = 
		"["^(strlist2str ", " (List.map pair2str pairs))^"]"
	    fun pairsList2str pairslst =
		"["^(strlist2str ",\n " (List.map pairs2str pairslst))^"]"

	    fun augmentSeqContext(seq: Sequent, newNames) =
		let val {model=model, init=init, invar=invar,
			 context=pc, assumptions=assump,
			 conclusions=conc} = seq
		    fun loop(SequentClosure{names=names, parent=p}) =
			  SequentClosure{names=(List.rev newNames)@names, parent=p}
		      | loop c = SequentClosure{names=(List.rev newNames), parent=c}
		    val (cxt, hash) = (ProgramContext2pt pc, ProgramContext2ObjectHash pc)
		    val newContext = makeProgramContext(loop cxt, hash)
		in
		    {model=model, init=init, invar=invar,
		     context=newContext, assumptions=assump,
		     conclusions=conc}: Sequent
		end

	    (* The argument `sub' here is phony, just to avoid applying this
	       function when the sequent doesn't match *)
	    fun getConstantsValue sub =
		(case findArgListValue args "constants" of
		     SOME lst => lst
		   | NONE => raise SympBug
			     ("AbstractSplit.apply: "
			      ^"`constants' is not in the arguments:\n["
			      ^(strlist2str ", " (List.map RuleArg2string args))^"]"))
	    fun val2const(ArgObject obj) = ProgramObject2pt obj
	      | val2const v = raise SympBug
		  ("AbstractSplit.apply/val2const: Not an ArgObject: "
		   ^(RuleArgValue2string v))

	    (* Separate collected contstants into Skolem, Abstract, and
	       Interpreted, split them into primitive types, and group into same-type
	       lists for future case splitting. *)

	    fun const2groups lst =
		let val (skolemHash, skolemTypeHash, skolemPrim, skolemExp) = 
		          collectSplitSkolem findObject lst
		    val (abstractHash, abstractTypeHash, abstractPrim, absExp) =
			  collectSplitAbstract findObject lst
		    val (interpretedHash, interpretedTypeHash, interpretedPrim, intExp) =
			  collectSplitInterpreted findObject lst
		    fun hash2keys hash = hash2any(fn (x,_)=>x) hash

		    (* Go through all the types of Skolems and join
		       them with corresponding abstract and
		       interpreted ones *)

		    fun joinType tp =
			let fun flatten(SOME lst) = lst
			      | flatten NONE = []
			in Group{ skolem = flatten(findHash(skolemTypeHash, tp)),
				  abstract = flatten(findHash(abstractTypeHash, tp)),
				  interpreted = flatten(findHash(interpretedTypeHash, tp)) }
			end

		    val groups = List.map joinType (hash2keys skolemTypeHash)
		in
		    case groups of
			[] => (printFun "\nThere are no Skolem constants to split cases on.\n";
			      NONE)
		      | _ => SOME{result=groups,
				  skolemPrim=skolemPrim,
				  abstractPrim=abstractPrim,
				  interpretedPrim=interpretedPrim,
				  skolemExp=skolemExp}
		end

	    (* Find or generate a new abstract constant for the given
	       constant.  It no longer descends into the extractors,
	       but just cooks up some sensible new name and makes a
	       new constant of the corresponding primitive type. *)

	    fun const2abstract(c, unique) =
		let val funName = "ProofRules.AbstractSplit.const2abstract"
		    val _ = pushFunStackLazy(funName, fn()=>pt2stringDebug c)
		    fun makeNew(c, name) =
		        let val n = newName()
			    val tp = getExprType findObject c
			    val uid = Uid n
			    val absName = (case GetObjectName c of
					       SOME n => n
					     | NONE => uid)
			    val absConst = AbstractConst{name=absName, uname=uid,
							 Type=tp, unique=unique}
			    val _ = insertHashDestructive(newNamesHash, n, absConst)
			    val _ = insertHashDestructive(absHash, c, absConst)
			    val _ = addObject(makeProgramObject(absConst, objHash), SOME uid)
    			in absConst
			end
		    val res = (case (findHash(absHash, c), c) of
				   (SOME a, _) => a
				 (* If it is already abstract, don't do anything *)
				 | (_, AbstractConst _) => c
				 (* New name will be the same constant.  
				  Too bad, I guess, but what the heck... *)
				 | (NONE, _) => makeNew(c, c))
		    val _ = popFunStackLazy(funName, fn()=>pt2stringDebug res)
		in res
		end

(*		    fun loop(ExtractRecord(name, c)) = ExtractRecord(name, loop c)
		      | loop(ExtractTuple(i, c)) = ExtractTuple(i, loop c)
		      | loop(ExtractIndex c) = ExtractIndex(loop c)
		      | loop(ExtractAppl(constr, c)) = ExtractAppl(constr, loop c)
		      | loop(SkolemConst{name=name, Type=tp, ...}) =
			let val n = newName()
			    val absConst = AbstractConst{name=name, uname=Uid n, Type=tp}
			    val _ = insertHashDestructive(newNamesHash, n, absConst)
			in absConst
			end
		      | loop(StaticFormalConst{name=name, Type=tp, ...}) =
		          let val n = newName()
			      val absConst = AbstractConst{name=name, uname=Uid n, Type=tp}
			      val _ = insertHashDestructive(newNamesHash, n, absConst)
		          in absConst
			  end
		      | loop x = raise SympBug
			  ("AbstractSplit.const2abstract: not an extractor or SkolemConst:\n  "
			   ^(pt2string x))
		in loop c
		end
*)

 	    fun findNewNames pairs =
		let val funName = funName^"/findNewNames"
		    fun pairs2str pairs =
			"["^(strlist2str ", " 
			     (List.map(fn(x,y)=>"("^(pt2stringDebug x)
				       ^","^(pt2stringDebug y)^")") pairs))
			^"]"
		    val _ = pushFunStackLazy(funName, fn()=> pairs2str pairs)
		    val debug = lazyVerbDebug(getOptions()) funName
		    (* Hash the new vars by their Uid's to avoid repetitions *)
		    val hash = makeHashDefault(op =, fn x=>x)
		    fun hash2str hash = 
			let fun keyval2str(k, v) = "("^k^","^v^")"
			    val keyvals = hash2any(fn (x,y)=>(x, pt2stringDebug y)) hash
			in
			    "["^(strlist2str "," (List.map keyval2str keyvals))^"]"
			end
		    val _ = debug(fn()=>funName^": newNamesHash = "^(hash2str newNamesHash)
				  ^"\n")
		    fun checkConst(ExtractRecord(_, c)) = checkConst c
		      | checkConst(ExtractTuple(_, c)) = checkConst c
		      | checkConst(ExtractIndex c) = checkConst c
		      | checkConst(ExtractAppl(_, c)) = checkConst c
		      | checkConst(AbstractConst{uname=Uid n, ...}) =
			  (case findHash(newNamesHash, n) of
			       SOME v => (insertHashDestructive(hash, n, v); ())
			     | NONE => ())
		      | checkConst _ = ()
		    fun doPair(old, new) = checkConst new
		    val _ = List.app doPair pairs
		    val res = hash2any(fn (_,x)=>x) hash
		    val _ = popFunStackLazy(funName, fn()=>ptlist2str ", " res)
		in
		    res
		end

	    (* Generate an abstract interpretation for a type from the
	       list of constants.  The input is the list of pairs
	       (const, equiv_class).  The abstraction is built using
	       the first elements of the pairs, and this first element
	       is replaced by a new abstracted constant.

	       Returns the TypeAbstraction and the new list of pairs
	       (absConst, equiv_class).

	       Assumption: all the constants in the list have the same type. *)

	    fun makeTypeAbs(lst as hd::_) =
		let val tp = let val (c, _) = hd in getExprType findObject c end
		    (* New uid for the abstract type *)
		    val tpuid = Uid(newName())
		    (* Hash for the abstraction mapping *)
		    val mapHash = makeHashDefault(ptEq, pt2string)
		    (* Generate an abstract constant for the actual constant *)
		    fun makeAbsConst (pt, rest) =
			let val res = const2abstract(pt, true)
			    (* Local abstraction mapping, just for this type *)
			    val _ = insertHashDestructive(mapHash, pt, res)
			in  (* Add the original representative
			       constant to the set of "other"
			       elements, so it'll be replaced in the
			       model by the new abstract constant. *)
			    (res, pt::rest)
			end
		    val other = const2abstract(TypedExpr(dp, Id(dp, "other"), tpuid), false)
		    val newPairs = List.map makeAbsConst lst
		    val abstp = AbstractType((List.map #1 newPairs)@[other])
		    val _ = addType(makeExpressionType(abstp, objHash), SOME tpuid)
		    val trueVal = Builtin{name=True dp, Type=BoolType dp}
		    val falseVal = Builtin{name=False dp, Type=BoolType dp}
		    (* The abstraction mapping *)
		    fun mapping x =
			(case findHash(mapHash, x) of
			     SOME res => res
			   | NONE => other)
		    (* Abstract interpretation of equality *)
		    val absEqObject =
			let val formal = newName()
			    (* ParseTree code for
			       fn(x,y) => if x = Other & y = Other then (true | false)
				          else x = y endif *)
			    val pat = TuplePattern(dp, [Id(dp, "x"), Id(dp, "y")])
			    val xEq = Appl(dp, Eq dp, TupleExpr(dp, [Id(dp, "x"), other]))
			    val yEq = Appl(dp, Eq dp, TupleExpr(dp, [Id(dp, "y"), other]))
			    val c = Appl(dp, And dp, TupleExpr(dp, [xEq, yEq]))
			    val e1 = NondetExpr(dp, [True dp, False dp])
			    val e2 = Appl(dp, Eq dp, TupleExpr(dp, [Id(dp, "x"), Id(dp, "y")]))
			    val expr = IfExpr(dp, [CondExpr(dp, c, e1)], e2)
			    val raw = Fn(dp, [ChoiceExpr(dp, pat, expr)])
			    val initType = FunType(dp, TupleType(dp, [abstp, abstp]), BoolType dp)
			    val cxt = makeProgramContext(contextPt, objHash)
			    val (obj, _) = typeCheckExpr(getOptions()) cxt (raw, initType)
				handle TypeError str => raise SympBug
				  ("makeTypeAbs: TypeError while typechecking "
				   ^"the abstract `=' function:\n  "^str)
			in obj
			end
		    (* Replace BoolType with BoolType2 in `pt' *)
		    fun replaceBool(BoolType _) = BoolType2
		      | replaceBool pt = ptTransform replaceBool pt
		    (* Extract the parse tree and `fix' the boolean type *)
		    (* val absEq = replaceBool(ProgramObject2pt absEqObject) *)
		    val absEq = ProgramObject2pt absEqObject
		    val typeAbs = makeTypeAbstraction(tp, abstp, mapping, [(Eq dp, absEq)])
		in (typeAbs, newPairs)
		end
	      | makeTypeAbs [] = raise SympBug("AbstractSplit.makeTypeAbs: empty list")

	    (* Take a group and generate all possible interpetations
	       for Skolem constants.  Returns a list of
	       equiv. relations; each equiv. rel. is a pair: 
	       (absType, pairs), where `pairs' is a list of pairs
	       (representativeAbstractedConst, ListOfEquivConsts). *)

 	    fun splitGroup(Group{skolem = skolem, abstract = abs, interpreted = interp, ...}) =
		let val funName = funName ^ "/splitGroup"
		    val debug = lazyVerbDebug(getOptions()) funName
		    val _ = pushFunStackLazy(funName, fn()=>
					     "{skolem=["^(ptlist2str "," skolem)
					     ^"], abstract=["^(ptlist2str "," abs)
					     ^"], interpreted=["^(ptlist2str "," interp)
					     ^"]}")
		    val nonSkolem = abs@interp
		    (* Create a hash of skolems for faster lookup at the end *)
		    val skolemHash = makeHashDefault(ptEq, pt2string)
		    val _ = List.map(fn s => insertHashDestructive(skolemHash, s, ())) skolem

		    (* Generate all possible equiv. relations for the
		       list of skolem constants and other completely
		       or partially interpreted constants. *)

		    fun genEquiv(skolem, others) =
			let fun pickOne([], acc) = []
			      | pickOne (h::tl, acc) = (h,tl@acc)::(pickOne(tl, h::acc))
			    fun injectToRel x (classes, used) = 
				let val picks = pickOne(classes, [])
				    fun makeRel(class, rest) = (rest, (x::class)::used)
				in 
				    (classes, [x]::used)::(List.map makeRel picks)
				end
			    fun inject(x, rels) =
				  List.foldr(op @) [] (List.map(injectToRel x) rels)
			    val skolemRels = List.map(fn x=>(x,[]))(genEquivClasses skolem)
			    val relsPairs = List.foldr inject skolemRels others
			    val rels = List.map(fn(x,y)=>x@y) relsPairs
			in rels
			end

		    (* Convert an equiv. relation into pairs
		       (representativeNonSkolem, ListOfEqualConsts) *)
		    fun rel2pairs sets =
			let (* set2pair(vars, accum) *)
			    fun set2pair ([], []) = raise SympBug
                                 ("AbstractSplit.apply/splitGroup/coll2pairs: "
                                  ^"no constants in the set.")
				(* Only Skolem const. in the
				eq. class.  Convert one of them into
				an abstract const. and make it a
				representative. *)
			      | set2pair ([], c::acc) = 
				let val newAbs = const2abstract(c, true)
				in
				    (* The original Skolem const must also be replaced by the new
				       abstract one, so put it back in the list. *)
				    (newAbs, c::acc)
				end
			      | set2pair(v::vars, acc) =
				 (case findHash(skolemHash, v) of
				      SOME _ => set2pair(vars, v::acc)
				    | NONE => (v, acc@vars))
			    val pairs = List.map(fn set=>set2pair(set, [])) sets
			in
			    makeTypeAbs pairs
			end

		    val rels = genEquiv(skolem, nonSkolem)
		    val res = List.map rel2pairs rels
		    val _ = popFunStackLazy(funName, fn()=> pairsList2str(List.map #2 res))
		in
		    res
		end 

	    (* Computes the Cartesian product of the sets of cases in the case vector.

	       The "vector" is a list of type groups.  Each group is a
	       list of pairs (absType, equivRel), each of which
	       provides one possible abstract interpretation for the
	       type of the group.

	       Produce the list of all possible combinations of these
	       interpretations (take one from each type group); each
	       combination is (absTypeList, pairs).  Here `pairs' is
	       the flattenned list of `equivRel' pairs merged from all
	       the groups. *)

	    fun caseProduct casesVector =
		let val combinations = comblist casesVector
		    (* Collect all the type abstractions in one list
		       and merge all the pairs from all picked type
		       groups *)
		    fun flattenLoop([], tps, acc) = (tps, acc)
		      | flattenLoop((tp, lst)::lsts, tps, acc) =
			  flattenLoop(lsts, tp::tps, lst@acc)
		    fun flatten lst = flattenLoop(lst, [], [])
		in
		    List.map flatten combinations
		end

	    (* Filter the `result' with `ff' and keep the global
	       hashes of primitive components mappings *)
	    fun doFun ff {result=result,
			  skolemPrim=sp,
			  abstractPrim=ap,
			  interpretedPrim=ip,
			  skolemExp=se} =
		{result=ff result,
		 skolemPrim=sp, abstractPrim=ap,
		 interpretedPrim=ip, skolemExp=se}

	    (* Re-assemble Skolem vars from the pieces and prepare the substitution
	     as a list of pairs (skolemVar, itsNewValue) *)

  	    fun assembleSubst(pairs, skolemExpList) =
		let fun substOnePair((_, []), e) = e
		      | substOnePair((new, (old::lst)), e) = 
			  substOnePair((new, lst), (replaceFormula false (e, old, new)))
		    fun substOne pairs e = (e, collapseExp(List.foldr substOnePair e pairs))
		    (* List of pairs (old, new) of assembled Skolem constants *)
		in
		    List.map(substOne pairs) skolemExpList
		end

	    (* Generate one new sequent for a single case of equiv. rel.
	       A case is a pair (typeAbsList, pairs). *)

	    fun generateSequent{result=(typeAbsList, pairs), 
				skolemPrim=sp,
				abstractPrim=ap,
				interpretedPrim=ip,
				skolemExp=se} =
		let val funName = funName^"/generateSequent"
		    val debug = lazyVerbDebug(getOptions()) funName
		    val _ = pushFunStackLazy(funName, fn()=>pairs2str pairs)
		    val skolemExpList = hash2any(fn (_,x)=>x) se
		    val substList = assembleSubst(pairs, skolemExpList)
		    (* val _ = debug(fn()=>"\n"^funName^": substList = [") *)
		    val newNames = findNewNames substList
		    val _ = debug(fn()=>funName^": newNames = ["
				  ^(ptlist2strDebug "," newNames)^"]\n")
		    val flags = {prop=false, initOnly=false, RHSonly=false}
		    fun substOne((old, new), seq) = replaceSeq flags (seq, old, new)
		    val seq = List.foldr substOne seq substList
		    val seq = List.foldr addSeqTypeAbstraction seq typeAbsList
		    val res as {context=pc,...} = augmentSeqContext(seq, newNames)
		    val _ = popFunStackLazy(funName, fn()=>"\n"^(sequent2string res)
					    ^"\n  in context: \n  "
					    ^(pt2stringDebug(ProgramContext2pt pc)))
		in
		    res
		end

	    fun generateSequents{result=cases, 
				 skolemPrim=sp,
				 abstractPrim=ap,
				 interpretedPrim=ip,
				 skolemExp=se}  =
		let fun genSeq oneCase = generateSequent{result=oneCase, 
							 skolemPrim=sp,
							 abstractPrim=ap,
							 interpretedPrim=ip,
							 skolemExp=se}
		    val seqs = List.map genSeq cases
		in List.map (fn s=> (s, NONE: SequentHints option)) seqs
		end
		
	    fun checkSeq(seq: Sequent) =
		let val {invar=invar, assumptions=assump, conclusions=conc,...} = seq
		    fun checkLeft f = isEMu f orelse isLTMu f
		    fun checkRight f = isAMu f orelse isLTMu f
		in List.all checkLeft invar
		    andalso List.all checkLeft assump
		    andalso List.all checkRight conc
		end
	    val _ = if checkSeq seq then ()
		    else raise Quit("The sequent does not qualify as being ACTL or LTL.") 
	    val seqTerm = sequent2term(seq, SOME termHash)
	    val subOpt = matchSeq sub (applySeqArgs(concTemp, args), NONE, seqTerm)
	    val constsOpt = Option.map((List.map val2const) o getConstantsValue) subOpt
	    val groupsOpt = Option.mapPartial const2groups constsOpt
	    val casesVectorOpt = Option.map(doFun(List.map splitGroup)) groupsOpt
	    val casesOpt = Option.map(doFun caseProduct) casesVectorOpt
	    val newSeqsOpt = Option.map generateSequents casesOpt
	in
	    case newSeqsOpt of
		SOME seqs => InferenceRuleSuccess seqs
	      | NONE => InferenceRuleFailure
	end handle Quit str => (printFun("\n"^str^"\n"); InferenceRuleFailure)

    val abstractSplitRule =
	{ name = "abstract_split",
	  help = ("Split cases on equalities between Skolem constants present in the\n"
		  ^"current sequent and convert these Skolem constants into abstract\n"
		  ^"constants.\n\n"
		  ^"The rule is invertible when Skolem constants are only compared for\n"
		  ^"equality or used as indices in array state variables.  Otherwise,\n"
		  ^"invertibility is not guaranteed, meaning that you may get a spurious\n"
		  ^"counterexample by model checking the resulting sequents."),
	  params = specs,
	  conclusion = SOME concTemp,
	  premisses = NONE,
	  match = NONE,
	  apply = SOME apply }: InferenceRule

    (* Rule AbstractVars: abstracting away some state variables *)
    local
	val specs = [{name="vars",
		      Type=ProverArgListType ProverArgObjectType,
		      default=DefaultRequired}]
	fun apply printFun (seq: Sequent, context: ProgramContext, args: RuleArg list, sub) =
	    let val funName = "ProofRules.AbstractVars.apply"
		val debug = lazyVerbDebug(getOptions()) funName
		val findObject = Typecheck.makeFindObject(ProgramContext2ObjectHash context)
		val (_, termHash) = sub
		fun getObject(ArgObject obj) =
		    let val pt = ProgramObject2pt obj
		    in
		      if isStateVar pt then pt
		      else raise Quit
			  ("Argument not a state variable: "^(pt2string pt))
		    end
		  | getObject x = raise SympBug
		     (funName^": argument not an object: "^(RuleArgValue2string x))
		(* Get the full state var as it declared in Symp *)
		fun getAtomicVar(ExtractTuple(_, v)) = getAtomicVar v
		  | getAtomicVar(ExtractRecord(_, v)) = getAtomicVar v
		  | getAtomicVar(ExtractIndex v) = getAtomicVar v
		  (* So far we count (var c) as an atomic var, since
		     splitting an array might involve messing with
		     infinity *)
		  | getAtomicVar(v as Appl _) = v
		  | getAtomicVar v = v
		(* Get the full expansion of the variable *)
		fun expandVar v =
		    let val (pairs, ev) = splitTypes findObject v
		    in ev
		    end
		fun subst(v, old, new) = collapseExp(replaceFormula false (v, old, new))

		(* For a variable, make a pair (old, new) to use as a
		   substitution in the formulas and the model. *)
		fun makeSubstPair v =
		    let val av = getAtomicVar v
			val ev = expandVar av
			val tp = getExprType findObject v
			val anyvalue = Builtin{name=Anyvalue dp,
					       Type=tp}
			val newV = subst(ev, v, anyvalue)
		    in (av, newV)
		    end
		val flags = {prop=false, initOnly=false, RHSonly=true}
		fun substOne((old, new), seq) = replaceSeq flags (seq, old, new)
		val vars = (case findArgListValue args "vars" of
				SOME lst => List.map getObject lst
			      | NONE => raise SympBug
				 (funName^": VARS is not in the list of arguments."))
		fun checkSeq(seq: Sequent) =
		    let val {invar=invar, assumptions=assump, conclusions=conc,...} = seq
			fun checkLeft f = isEMu f orelse isLTMu f
			fun checkRight f = isAMu f orelse isLTMu f
		    in List.all checkLeft invar
			andalso List.all checkLeft assump
			andalso List.all checkRight conc
		    end
		val _ = if checkSeq seq then ()
			else raise Quit("The sequent does not qualify as being ACTL or LTL.") 
		val substPairs = List.map makeSubstPair vars
		val newSeq = List.foldr substOne seq substPairs
	    in
		InferenceRuleSuccess[(newSeq, NONE)]
	    end handle Quit str => (printFun("\n"^str^"\n"); InferenceRuleFailure)
    in
	val abstractVarsRule =
	    { name = "abstract_vars",
	      help = ("Removes the specified state variables from the model by replacing\n"
		      ^"them with completely nondeterministic values (anyvalue)."),
	      params = specs,
	      conclusion = NONE,
	      premisses = NONE,
	      match = NONE,
	      apply = SOME apply }: InferenceRule
    end

  end
