(* Conversion of the default input language custom interface to the
   uniform one.  We do not constrain the functor with the
   INPUT_LANGUAGE signature here, since we want to use the
   implementation types in the proof system implementation.  These
   details will be hidden in the proof system inteface. *)

functor InputLanguageDefaultFun(structure Interface: INTERFACE
				structure Hash: HASH): INPUT_LANGUAGE_DEFAULT =
  struct
    structure Interface = Interface
    structure ParserDefault = ParserDefaultFun(structure Interface = Interface)
    structure Evaluate
        = EvaluateFun(structure ParserDefault = ParserDefault)

    structure Typecheck
	= TypecheckFun(structure Hash = Hash
		       structure Evaluate = Evaluate)

    open SympBug
    open ParserDefault
    open Typecheck
    open Evaluate
    open ParseTreeStruct

    (* Typechecked program, expression, declaration, whatever, which
       is the result of typechecking a ParseTree. *)
    type ProgramObject = {obj: ParseTree, objectHash: ObjectHash}
    (* Type or an expression*)
    type ExpressionType = {Type: ParseTree, objectHash: ObjectHash}
    (* Context for typechecking *)
    type ProgramContext = {context: ParseTree, objectHash: ObjectHash}

    datatype SelectObject =
	SelectObject of ProgramObject
      | SelectType of ExpressionType
      | SelectContext of ProgramContext

    fun ProgramObject2pt({obj=pt, ...}: ProgramObject) = pt
    fun ProgramObject2ObjectHash({objectHash=hash, ...}: ProgramObject) = hash
    fun ProgramContext2pt({context=cxt, ...}: ProgramContext) = cxt
    fun ProgramContext2ObjectHash({objectHash=hash, ...}: ProgramContext) = hash
    fun makeProgramObject(pt, hash) = {objectHash=hash, obj=pt}: ProgramObject
    fun makeProgramContext(pt, hash) = {objectHash=hash, context=pt}: ProgramContext
    fun makeExpressionType(pt, hash) = {objectHash=hash, Type=pt}: ExpressionType

    val (contextKeyword, theoremKeyword, expressionKeyword, typeKeyword) =
	("module", "theorem", "val", "type")

    (* typeCheckProgram is the same as in Typecheck *)
    fun typeCheckProgram options program =
	let val _ = resetTypeCheckStat()
	    val (prog, hash) = Typecheck.typeCheckProgram options program
	    val debug = lazyVerbDebugStr options
	    val _ = debug "common" (fn () => (Conc(Str "\n\nType checked program:\n\n",
						   pt2str prog)))
            val _ = debug "commonDebug" (fn () => (Conc(Str "\n\nType checked program:\n\n",
							pt2strDebug prog)))
            val _ = lazyVerbDebug options "stat" (typeCheckStat2string)
	in 
	    {context=prog, objectHash=hash}: ProgramContext
	end

    (* FIX the typechecker to initialize `findObject' with the given one in the ProgramContext *)
    fun typeCheckExpr options ({context=context, objectHash=hash}: ProgramContext) (expr,tp) =
	let val _ = resetTypeCheckStat()
	    val (newE, newTp, newHash) =
		  Typecheck.typeCheckExpr options (context, hash) (expr,tp)
	in ({obj=newE, objectHash=newHash}: ProgramObject,
	    {Type=newTp, objectHash=newHash}: ExpressionType)
	end handle EvalError str => raise TypeError("[Evaluator] "^str)

    fun typeCheckContext options ({context=context, objectHash=hash}: ProgramContext) tree = 
	let val _ = resetTypeCheckStat()
	    val (closure, newHash) = 
		  Typecheck.typeCheckModuleExpr options (context, hash) tree
	in
	    {context=closure, objectHash=newHash}: ProgramContext
	end

    fun typeCheckTheorem options ({context=context, objectHash=hash}: ProgramContext) thm =
	let val _ = resetTypeCheckStat()
	    val (thm', newHash) = Typecheck.typeCheckTheorem options (context, hash) thm
	in
	    {obj=thm', objectHash=newHash}: ProgramObject
	end handle EvalError str => raise TypeError("[Evaluator] "^str)

    fun addObject({obj=obj, objectHash=hash}: ProgramObject, uidOpt) =
	  updateObjectHash(hash, obj, uidOpt)

    fun addContext({context=obj, objectHash=hash}: ProgramContext, uidOpt) =
	  updateObjectHash(hash, obj, uidOpt)

    fun addType({Type=obj, objectHash=hash}: ExpressionType, uidOpt) =
	  updateObjectHash(hash, obj, uidOpt)

    (* Classification of the ParseTree values into Objects, Context, and Types *)
    fun isType(Tvar _) = true
      | isType(TypeInst _) = true
      | isType(TypeClosure _) = true
      | isType(StaticFormalType _) = true
      | isType(PType _) = true
      | isType(Type _) = true
      | isType(Datatype _) = true
      | isType(BoolType _) = true
      | isType BoolType2 = true
      | isType(NatType _) = true
      | isType(IntType _) = true
      | isType(RangeType _) = true
      | isType(FunType _) = true
      | isType(ArrayType _) = true
      | isType(TupleType _) = true
      | isType(RecordType _) = true
      | isType(EnumType _) = true
      | isType(AbstractType _) = true
      | isType _ = false

    fun isContext(TopClosure _) = true
      | isContext(Module _) = true
      | isContext(ModuleClosure _) = true
      | isContext(LetClosure _) = true
      | isContext(LetAsstClosure _) = true
      | isContext(ChoiceClosure _) = true
      | isContext(ChoiceAsstClosure _) = true
      | isContext(ChooseClosure _) = true
      | isContext(ChooseAsstClosure _) = true
      | isContext(ForeachAsstClosure _) = true
      | isContext(SyncClosure _) = true
      | isContext(AsyncClosure _) = true
      | isContext(SequentClosure _) = true
      | isContext _ = false

    (* Objects that can also be contexts *)
    fun maybeContext(FunClosure _) = true
      | maybeContext(ForallClosure _) = true
      | maybeContext(ExistsClosure _) = true
      | maybeContext(MuClosure _) = true
      | maybeContext(NuClosure _) = true
      | maybeContext _ = false

    (* Fetch an object from the ObjectHash and classify it *)
    fun findSelectObject(hash, uid as Uid _) =
	let fun select pt =
		if isType pt then SelectType{Type=pt, objectHash=hash}
		else if isContext pt then SelectContext{context=pt, objectHash=hash}
		     else SelectObject{obj=pt, objectHash=hash}
	in
	    Option.map select (makeFindObject hash uid)
	end
      | findSelectObject(_, x) = raise SympBug
	 ("findSelectObject: not a Uid: "^(pt2string x))

    fun findProgramContext(hash, uid as Uid _) =
	let fun select pt =
		if isContext pt orelse maybeContext pt then
		    SOME({context=pt, objectHash=hash}: ProgramContext)
		else NONE
	in
	    Option.mapPartial select (makeFindObject hash uid)
	end
      | findProgramContext(_, x) = raise SympBug
	 ("findProgramContext: not a Uid: "^(pt2string x))

    fun findProgramObject(hash, uid as Uid _) =
	let fun select pt =
		if isContext pt orelse isType pt then NONE
		else SOME({obj=pt, objectHash=hash}: ProgramObject)
	in
	    Option.mapPartial select (makeFindObject hash uid)
	end
      | findProgramObject(_, x) = raise SympBug
	 ("findProgramObject: not a Uid: "^(pt2string x))

    fun findExpressionType(hash, uid as Uid _) =
	let fun select pt =
		if isContext pt orelse maybeContext pt then
		    SOME({Type=pt, objectHash=hash}: ExpressionType)
		else NONE
	in
	    Option.mapPartial select (makeFindObject hash uid)
	end
      | findExpressionType(_, x) = raise SympBug
	 ("findExpressionType: not a Uid: "^(pt2string x))

    (* Equality on parse trees and objects *)
    (* ptEq is the same *)
    fun poEq ({obj=x,...}: ProgramObject, {obj=y,...}: ProgramObject) = ptEq(x,y)
    fun pcEq ({context=x,...}: ProgramContext, {context=y,...}: ProgramContext) = ptEq(x,y)
  end
