(* The default implementation of the PROOF_SYSTEM abstract interface *)

functor ProofSystemDefaultFun(structure Hash: HASH
			      structure Interface: INTERFACE): PROOF_SYSTEM =
  struct
    structure ProofRules = 
	ProofRulesDefaultFun(structure Hash = Hash
			     structure Interface = Interface)

    open Interface
    open ProofRules
    open Sequent
    open SequentCommon
    open SequentBase
    open Hash
    open SympBug
    open Str

    local
	val rules = makeHashDefault(op =, fn x=>x): (string, InferenceRule) Hash
	val commands = makeHashDefault(op =, fn x=>x): (string, ProofSystemCommand) Hash
	val tactics = makeHashDefault(op =, fn x=>x): (string, ProofSystemTactic) Hash
    in
	fun findRuleByName str = findHash(rules, str)
	fun findCommandByName str = findHash(commands, str)
	fun findTacticByName str = findHash(tactics, str)
	fun addNewRule r = (insertHashDestructive(rules, getRuleName r, r); ())
	fun addNewCommand c = (insertHashDestructive(commands, getCommandName c, c); ())
	fun addNewTactic t = (insertHashDestructive(tactics, getTacticName t, t); ())

	(* Initialize the set of inference rules from the ProofRules module *)
	fun initProofSystem () = 
	      (List.app addNewRule (proofRules());
	       List.app addNewCommand (proverCommands()))
	fun getAllRules() = hash2any(fn (_,x) => x) rules
	fun getAllCommands() = hash2any(fn (_,x) => x) commands
	fun getAllTactics() = hash2any(fn (_,x) => x) tactics

	fun getInitRule() = 
	    (case findRuleByName "init" of
		 SOME r => r
	       | NONE => raise SympBug
		     ("ProofSystemDefault/getInitRule: no \"init\" rule"))

	(* Apply the rule, and check if the resulting sequents are trivially
	   true by applying the init rule.  All sequents that are proven this way
	   are discarded from the result. *)

	fun applyRule printFun (seq, context) (rule, args, substOpt) =
	    let val debug = verbDebug(getOptions()) "applyRule"
		val res = ProofRules.applyRule printFun (seq, context) (rule, args, substOpt)
		val _ = debug("\napplyRule: the `raw' result of applying the rule "
			      ^(getRuleName rule)^":\n  "
			      ^(InfRuleResult2string res)^"\n")
		val _ = debug("\napplyRule: checking for trivially true sequents\n")
		(* Annihilate the sequent if it's trivially true *)
		fun checkSeq (pair as (seq, _)) =
		    (case ProofRules.applyRule printFun (seq, context) (getInitRule(), [], NONE) of
			 InferenceRuleSuccess [] => 
			     (debug("\napplyRule: this sequent is trivially true:\n  "
				    ^(sequent2string seq));
			      NONE)
		       | _ => SOME pair)
		(* Clean up all trivial sequents *)
		fun checkSeqs(InferenceRuleSuccess seqs) =
		      InferenceRuleSuccess(List.mapPartial checkSeq seqs)
		  | checkSeqs x = x
	    in
		checkSeqs res
	    end
    end

    val debugNames = 
	[(* Input language *)
	 "ptEqCommon",
	 (* Sequent module *)
	 "matchTerm",
	 "matchTerm/loop",
	 "matchTerms",
	 "matchTerms/loop",
	 "matchTerms/matchFilter",
	 "matchTerms/matchFirst",
	 "matchSeq",
	 "replaceFormula",
	 "replaceFormula/subst",
	 "skolemMatch",
	 "skolemApply",
	 "skolemApply/skolem",
	 "updateSubstitution",
	 "substTermID",
	 "SequentDefault.subst",
	 "SequentDefault.subst/loop",
	 "SequentDefault.subst/constrValue",
	 "substSequent",
	 "updateSubstitutionRaw",
	 (* Default Proof System *)
	 "SequentDefault.applyRule",
	 "completeRuleArgs",
	 "compileArgs",
	 "replaceApply",
	 "ProofRules.split.apply",
	 "ProofRules.split.apply/splitPair",
	 "ProofRules.AbstractSplit.apply",
	 "ProofRules.AbstractSplit.apply/splitGroup",
	 "ProofRules.AbstractSplit.apply/generateSequent",
	 "ProofRules.AbstractSplit.apply/findNewNames",
	 "ProofRules.AbstractSplit.const2abstract",
	 "ProofRules.splitLabel.apply",
	 (* Typechecking in the default proof system *)
	 "stat",
	 "Typecheck.findName",
	 "findinContextCommon",
	 "findinContextCommonDebug",
	 "finalizeExpr",
	 "finalizeExpr/loop",
	 "expandType",
	 "expandType/loopName",
	 "instantiateModule",
	 "instantiateModule/resolveParam",
	 "instantiateModule/loop",
	 "unifyTypes",
	 "coerce",
	 "tcName",
	 "tcNameDebug",
	 "tcObject",
	 "tcObjectDebug",
	 "tc",
	 "tcDebug",
	 "typeCheckPattern",
	 "tcPattern/loop",
	 "typeCheckModuleExpr",
	 "typeCheckModuleExprDebug",
	 "tcModuleName",
	 "tcModuleInst",
	 "tcModuleExpr",
	 "typeCheckStatParam",
	 "typeCheckType",
	 "typeCheckType/loop",
	 "evaluateExpr",
	 "evaluateExpr/recurCall",
	 "Evaluate.subs",
	 "Evaluate.substitute",
	 "Evaluate.applyFun",
	 "Evaluate.theSame",
	 (* Abstraction *)
	 "getTypeSize",
	 "getTypeSize/loop",
	 "getTypeValues",
	 "getTypeValues/loop",
	 (* Translation: common functions *)
	 "addToCone",
	 "updateConeDestructive",
	 "inConeOfCommon",
	 "getVarConeDirect",
	 "getVarCone",
	 "getVarConeHashCommon",
	 "getVarConeHashCommon/checkCycle",
	 "getVarConeHashCommon/loop",
	 "getVarCone1",
	 "getVarCone1Delayed",
	 "getVarCone1Delayed/split",
	 "getVarCone1Delayed/lookup",
	 "TransCommon.goodStateVar",
	 (* Translation: preparation phase (translating to AsstVarsTree) *)
	 "splitEType",
	 "splitEType/split",
	 "varsFromDelayed",
	 "varsFromDelayed/loop",
	 "collectAsstVars",
	 "splitType",
	 "splitType/loop",
	 "extractAsstTree",
	 "balanceAsstTree",
	 "rebuildAsstVars",
	 "rebuildAsstVars/collect",
	 (* Specialization phase (specializing trans. rel. to the property, cone of influence) *)
	 "customizeAsstVars",
	 "customizeAsstVars0",
	 "customizeAsstVars0/trim",
	 "customizeAsstVars0/trim/loop",
	 "coneFromVar",
	 "coneFromVar/loop",
	 (* Generation phase (translation to boolean formulas) *)
	 "TransGen.appendLists",
	 "substVarValuesPairs",
	 "TransGen.matches",
	 "TransGen.matchesDebug",
	 "normalizeDefCond",
	 "normalizeDefAppl",
	 "normalizeDefEq",
	 "normalizeDefOp",
	 "normAnd",
	 "normOr",
	 "exprDef",
	 "exprDef/exprSplit",
	 "exprDef/split",
	 "exprDef/doFun",
	 "quantifierDef",
	 "builtinDef",
	 "generateTrans",
	 "walkTree",
	 "asstGen",
	 "asstGen1",
	 "asstGen1/loop",
	 "asstGenDebug",
	 "minimizeDefVarsExpr",
	 "minimizeDefVarsAsst",
	 "boolExpr",
	 "generateModelTrans",
	 (* Generating SMV code *)
	 "SMVcode",
	 "TransSMV.addMap",
	 "TransSMV.addTVMap",
	 "TransSMV.genParts",
	 "TransSMV.genAux",
	 "TransSMV.genAux/doCommon",
	 "TransSMV.genAux/doDefine",
	 "TransSMV.genAux/doVar",
	 "TransSMV.genVarDecl",
	 "formulaStr",
	 "formulaStr/formStr",
	 "TransSMV.getMap",
	 "TransSMV.getTVMap"]

  end
