(* A specialization of the top-level signature.sig.  Must be
   compatible with it at all times. *)

signature SEQUENT_DEFAULT =
  sig
(*      structure SequentBase: SEQUENT_BASE_DEFAULT *)
    structure SequentCommon: SEQUENT_COMMON_Default
    structure Abstract: ABSTRACT
    structure Hash: HASH

    type ParseTree = SequentCommon.SequentBase.InputLanguage.ParseTree
    type ProgramContext = SequentCommon.SequentBase.InputLanguage.ProgramContext

    type Model = SequentCommon.SequentBase.Trans.Model
    type ('a,'b) Hash = ('a,'b) Hash.Hash

    type InferenceRuleArgumentValue = SequentCommon.InferenceRuleArgumentValue
    type InferenceRuleArgument = SequentCommon.InferenceRuleArgument
    type InferenceRuleArgumentType = SequentCommon.InferenceRuleArgumentType
    type InferenceRuleResult = SequentCommon.InferenceRuleResult
    type UserInputArgument = SequentCommon.Interface.UserInputArgument

    type Sequent = SequentCommon.SequentBase.Sequent
    type SequentPart = SequentCommon.SequentBase.SequentPart
    type SequentHints = SequentCommon.SequentBase.SequentHints

    type ProgramObject = SequentCommon.SequentBase.InputLanguage.ProgramObject

    type TermInd = int (* Term index *)
	    
    (* Sequent template types *)

    datatype Template =
      (* Formula variable *)
	FormulaTemp of string
      (* Template var matching a list of formulas which possibly match a common template *)
      | ListFormulaTemp of string * (Template option)
      (* Built-in function (explicitly specified) *)
      | BuiltinTemp of ParseTree
      (* Arbitrary value of any type, unspecified functions, etc. *)
      | ValueTemp of string
      (* Program context variable *)
      | ProgramContextTemp of string
      (* Same as ValueTemp, only the expression is thrown away *)
      | WildCardTemp
      (* Same as ListValueTemp, only the expressions are thrown away *)
      | ListWildCardTemp
      (* List of object values *)
      | ListValueTemp of string * (Template option)
      | TupleTemp of Template list
      | ApplTemp of Template * Template
	(* QuantifierTemp(vars,body,context) *)
      | ForallTemp of Template list * Template * Template
      | ExistsTemp of Template list * Template * Template
      (* Context variable C[T] that contains an object matching T inside *)
      | ContextTemp of string * Template
      (* T1 : T2 - an object matching T1 has a type matching T2 *)
      | TypedTemp of Template * Template
      (* T :: pi - a template constrained to certain objects *)
      | ConstrainedTemp of Template * TempConstrain
      (* A new Skolem constant; doesn't match anything, only used for instantiation *)
      | SkolemTemp
      (* Matches or instantiates this very constant *)
      | ConstantTemp of TemplateValue
      (* Assigns a value to a string that matches a template *)
      | NamedTemp of string * Template

    (* Extra constraints to the templates (possibly taken from the rule's parameters) *)
    and TempConstrain =
	(* No restriction *)
	AnyConstr
      (* Model value *)
      | ModelConstr
      (* Program Context value *)
      | ProgramContextConstr
      (* Initial predicate value *)
      | InitConstr
	(* The template must be a formula matching a specification if given *)
      | FormulaConstr of (SequentPart * (int option)) option
	(* Template is a bound variable possibly matching a template *)
      | BoundConstr of Template option
	(* Any object, possibly equal to the given one *)
      | ObjectConstr of ParseTree option
      (* Value is a parse tree constant (for special cases like parent closures, etc.) *)
      | ParseTreeConstr
      (* The value must be a context *) 
      | ContextConstr
      (* Disjunction of several constraints *)
      | SomeConstr of TempConstrain list
	(* List of elements constrained to TempConstrain *)
      | ListConstr of TempConstrain

    (* Intermediate representation of expressions (ParseTree) using some sort of indexing *)

    (* Defines the term and all of its parents (terms that contain it
       as an immediate subterm).  The parents are used to determine
       quickly all the term's occurrences. *)
    and TermID = TermID of { ind: TermInd, parents: TermID list ref }

    and Term = (* The actual representation of terms *)
	ConstTerm of ParseTree
      | TupleTerm of TermID list
      | ApplTerm of TermID * TermID
      | TypedTerm of TermID * TermID
      | ForallTerm of { names: TermID list,
		        body: TermID,
			parent: ParseTree }
      | ExistsTerm of { names: TermID list,
		        body: TermID,
			parent: ParseTree }

    (* Values associated with template variables *)
    and TemplateValue =
	ModelValue of Model
      (* Set of initial states of the model, if any *)
      | InitValue of ParseTree option
      (* Context, usually of the current sequent *)
      | ProgramContextValue of ProgramContext
      (* Object can be a formula, or a constant, but not a bound var. *)
      | ObjectValue of TermID
      (* Special constants like parent closures, etc. *)
      | ParseTreeValue of ParseTree
      (* Bound variable *)
      | BoundValue of TermID
      | NamedValue of string * TemplateValue
      (* Value of list variables (ListVarTemp and BoundVarList) *)
      | ListValue of TemplateValue list
      | SubgoalValue of Sequent
      | ContextValue of TemplateValue -> TemplateValue
      | ConstrainedValue of TemplateValue * TempConstrain

    type SequentTemplate = { model: Template,
			     init: Template,
			     context: Template,
			     (* Allow several templates in the formula lists.
			        The order shouldn't be important. *)
			     invar: Template list,
			     assumptions: Template list,
			     conclusions: Template list }

    (* Hashes that connect indices with terms in both directions *)
    type TermHash = { ind2termHash: (TermInd, Term * TermID) Hash,
		      term2idHash: (Term, TermID) Hash,
		      (* Next free term index for this hash *)
		      nextInd: TermInd ref,
		      (* Unique ID for each hash *)
		      hashID: int }

    (* Term representation of a sequent *)
    type SequentTerm = { model: Model,
			 init: ParseTree option,
			 context: ProgramContext,
			 invar: TermID list,
			 assumptions: TermID list,
			 conclusions: TermID list,
			 hash: TermHash }

    (* Associate template variable names with values *)
    type Substitution = (string, TemplateValue) Hash * TermHash

    (* Default value of the inf. rule argument *)
    datatype DefaultArgValue =
	(* Match the template with the sequent, if it makes sense.
	   E.g. formula number doesn't need a template, and then it's NONE. *)
	DefaultMatch of Template option
	(* Use this static value *)
      | DefaultValue of InferenceRuleArgumentValue
	(* Compute the default value dynamically using this function *)
      | DefaultValueFun of unit -> InferenceRuleArgumentValue
	(* No default value, the user is required to supply the value *)
      | DefaultRequired

    type InferenceRuleArgumentSpec =
	{ name: string,
	  Type: InferenceRuleArgumentType,
	  (* When default is NONE, the argument is required, otherwise it's optional *)
	  default: DefaultArgValue }

    (* Internal representation of compiled rule arguments *)
    datatype RuleArgValue =
	ArgString of string
      | ArgNumber of int
      | ArgSubgoal of string
      | ArgFormulaNum of SequentPart * (int option)
      | ArgObject of ProgramObject
      | ArgList of RuleArgValue list

    type RuleArg = {name: string, value: RuleArgValue}

    (* The `conclusion' is a sequent template that may contain free variables, some of them
       from `params', and they are instantiated to match the actual sequent to which this
       rule is applied.  `match' complements `conclusion' or replaces it.  Similarly,
       `premisses' are new sequent templates that are instantiated with values that
       variables in `conclusion' receive.  `apply' function, if present, complements or
       replaces `premisses'. *)

    type InferenceRule = { name: string,
			   help: string, (* Help text *)
			   params: InferenceRuleArgumentSpec list,
			   conclusion: SequentTemplate option,
			   premisses: SequentTemplate list option,
			   (* Given a concrete sequent, try to match it with the conclusion.
			      May apply some other rule-specific tests.
			      Without this function, match the conclusion in the standard way. *)
			   match: (Sequent * ProgramContext * (RuleArg list)
				   * (Substitution option) -> Substitution option) option,
			   (* Apply the rule to a concrete sequent with the list of arguments and
			      the substitution computed by the match function.  Return NONE if
			      failed, or a list of new sequents.

			      The first argument is a function to print info as the rule
			      is being applied (progress report, comments, etc.) *)
			   apply: ((string -> unit)
				   -> Sequent * ProgramContext * (RuleArg list) * Substitution
				   ->  InferenceRuleResult) option }

    type ProofSystemCommandResult = unit

    type ProofSystemCommand = { name: string,
			        help: string,
				params: InferenceRuleArgumentSpec list,
				apply: (string -> unit)
                                         -> Sequent * (RuleArg list)
                                         -> ProofSystemCommandResult option }

    (* Values from top-level sequent.sig *)

    val IRArgSpecName: InferenceRuleArgumentSpec -> string
    val IRArgSpecType: InferenceRuleArgumentSpec -> InferenceRuleArgumentType

    (* Create a new sequent from a theorem (which may include a model) *)
    val makeSequent: ProgramContext -> ProgramObject -> Sequent
    val sequentEq: Sequent * Sequent -> bool
    (* Check if the first sequent subsumes the second.
       That is, proving the first one also proves the second one.
       This function can be soundly approximated with sequentEq. *)
    val sequentSubsumes: Sequent * Sequent -> bool
(*      val getHints: Sequent -> SequentHints option *)
    val hintsEq: SequentHints * SequentHints -> bool

    val UIArg2FormulaIndex: UserInputArgument -> SequentPart * (int option)
    val FormulaIndex2UIArg: SequentPart * (int option) -> UserInputArgument

    (* Operations on inference rules *)

    val string2hints: string -> SequentHints

    (* Getting extra information about a sequent.  The type of request
       is specified by a string, the output is also a string which the
       user will see. *)

    val getSequentInfo: Sequent -> string -> string

    (* Values from our own SequentDefault *)

    val RuleArgValue2string: RuleArgValue -> string
    val RuleArg2string: RuleArg -> string

    val termHashID: TermHash -> int
    val termEq: Term * Term -> bool
    val termIDeq: TermID * TermID -> bool
    val termID2str: TermID -> string
    val termInd2str: TermInd -> string
    val term2strHash: Term -> string
    val temp2str: Template -> string
    val constr2str: TempConstrain -> string
    val value2str: TemplateValue -> string
    val termHash2str: TermHash -> string
    val subst2str: Substitution -> string
    val seqTemp2str: SequentTemplate -> string
    val makeTermHash: unit -> TermHash
    val copyTermHash: TermHash -> TermHash

    val subtermIDs: Term -> TermID list

    val term2pt: TermHash -> Term -> ParseTree
    val termID2pt: TermHash -> TermID -> ParseTree
    val termInd2termID: TermHash -> TermInd -> TermID
    val termInd2pt: TermHash -> TermInd -> ParseTree
    val termInd2term: TermHash -> TermInd -> Term
    val termID2term: TermHash -> TermID -> Term

    (* Check if the term is already in the hash and return its ID; otherwise
       generate a new ID and insert it into the hash *)
    val term2termID: TermHash -> Term -> TermID

    (* Convert ParseTree to the Term representation, DESTRUCTIVELY
       modifying the given TermHash and returning the termID.  If the
       term is not yet in the hash, its list of parents is empty. *)
    val pt2termID: TermHash -> ParseTree -> TermID

    (* Substitution of term IDs: replace all occurances of oldID by
       newID in the term represented by the `id', and return the new
       term ID.  No existing IDs or terms are replaced. *) 
    val substTermID: TermHash -> (TermID * TermID) -> TermID -> TermID

    (* Convert Sequent into SequentTerm representation *)
    val sequent2term: Sequent * (TermHash option) -> SequentTerm

    val getRuleName: InferenceRule -> string
    val getCommandName: ProofSystemCommand -> string
    val getRuleHelp: InferenceRule -> string
    val getCommandHelp: ProofSystemCommand -> string
    val getRuleArgs: InferenceRule -> InferenceRuleArgumentSpec list
    val getCommandArgs: ProofSystemCommand -> InferenceRuleArgumentSpec list

    val InfRuleSpec2str: InferenceRuleArgumentSpec -> string
    val makeSubstitution: TermHash -> Substitution

    val TemplateEq: Template * Template -> bool
    val TempConstrEq: TempConstrain * TempConstrain -> bool
    val valueEq: TemplateValue * TemplateValue -> bool
    val TempConstrLE: TempConstrain * TempConstrain -> bool
    val TempConstrMin: TempConstrain * TempConstrain -> TempConstrain option

    (* NOTE: termHash is NOT copied (it doesn't make sense to do so);
       it is expected that the set of terms is fixed at the time of
       generating a substitution. *)
    val copySubstitution: Substitution -> Substitution
    (* Raw update to the substitution; nothing is checked, be careful with it. *)
    val updateSubstitutionRaw: string * TemplateValue -> Substitution -> Substitution

    (* Adds a new name substitution to the current `sub', updates the
       given substitution DESTRUCTIVELY, and returns it.  If the name
       already defined, it must be exactly the same as `v'; if not,
       return NONE. *)
    val updateSubstitution: string * TemplateValue -> Substitution -> Substitution option

    (* Update a list variable: if it exists, it's value must be
       ListValue, and the list of new values is added to the list. *)
    val updateListSubstitution: string * (TemplateValue list)
           -> Substitution -> Substitution option

    (* Applies a substitution to a template, obtaining a TemplateValue.  All vars must be
       mapped to values. *)
    val subst: Substitution -> TempConstrain -> Template -> TemplateValue option

    (* "Typecheck" the rule's arguments (`InferenceRuleArgument') and
       return SOME error message if they don't match.  Otherwise
       return NONE. *)
    val checkRuleArgs: InferenceRuleArgumentSpec list
	   -> InferenceRuleArgument list -> string option

    (* Goes through the arguments and instantiates the defaults for
       missing ones.  The only default args it does NOT instantiate
       are template matches.  Those must be done in the sequent match.
       Returns the list of new arguments sans the missing ones with
       template defauts. *)
    val completeRuleArgs: InferenceRuleArgumentSpec list
	   -> InferenceRuleArgument list -> InferenceRuleArgument list

    (* Transforms a TemplateValue into a ParseTree if possible, or returns NONE. *)
    val TemplateValue2ParseTree: TermHash -> TemplateValue -> ParseTree option
    (* Transforms a TemplateValue into a ParseTree list.  Consider a
       single ParseTree element as a one-element list. *)
    val TemplateValue2ParseTreeList: TermHash -> TemplateValue -> ParseTree list option
    val TemplateValue2Model: TemplateValue -> Model option
    val TemplateValue2ProgramContext: TemplateValue -> ProgramContext option

    (* Takes a sequent template and a substitution and returns a sequent.
       All the variables in the template must have values in the substitution.
       Upon failure return NONE. *)
    val substSequent: Substitution -> SequentTemplate -> Sequent option

    (* Match a term (given as TermID and its formula's position in the
       sequent) with a scalar template under a constraint.  Return
       either the new substitution (without altering the original), or
       NONE.  Implements T::pi|-e~>sigma.

       This implements a direct matching - no search for a "good"
       choice is performed for the templates that occur several times. *)
    val matchTerm: Template * TempConstrain
	   -> TermID * (SequentPart * int option)
	   -> Substitution -> Substitution option

    (* `terms' is a list of pairs (termID, formulaInd) *)
    val matchTerms: Template list * TempConstrain
	  -> (TermID * (SequentPart * int option)) list
	  -> Substitution -> Substitution option

    val matchModel: Template * Model -> Substitution -> Substitution option
    val matchInit: Template * (ParseTree option) -> Substitution -> Substitution option
    val matchProgramContext: Template * ProgramContext -> Substitution
	    -> Substitution option
    val matchSeq: Substitution -> SequentTemplate * (SequentHints option) * SequentTerm
	    -> Substitution option

    (* Converts arguments from the general to our own representation - RuleArg,
       also add ArgObject values to the substitution. 
       compileArgs(sub, context, args) *)
    val compileArgs: Substitution * ProgramContext * (InferenceRuleArgument list) -> RuleArg list

    (* Sort through the rule's arguments and the conclusion template,
       and annotate the vars with the constraints from the arguments *)
    val applyArgs: Template * (RuleArg list) -> Template
    (* Same as above, only for the sequent template. *)
    val applySeqArgs: SequentTemplate * (RuleArg list) -> SequentTemplate

    (* See if the sequent matches the conclusion of the rule, and if it does,
       compute the substitution for it.  Otherwise return NONE.  *)
    val matchRuleSubst: Substitution -> Sequent * ProgramContext * SequentTerm
	   -> InferenceRule * (RuleArg list) -> Substitution option

    val matchRule: Sequent * ProgramContext
	   -> InferenceRule * (InferenceRuleArgument list) -> Substitution option

    val applyRule: (string -> unit) -> (Sequent * ProgramContext)
	        -> InferenceRule * (InferenceRuleArgument list) * (Substitution option)
	        -> InferenceRuleResult

    val applyCommand: (string -> unit) -> (Sequent * ProgramContext)
	        -> ProofSystemCommand * (InferenceRuleArgument list)
	        -> ProofSystemCommandResult option

    (* Get the list of rules that match a given sequent, together with
       the lists of rules' arguments *)
    val matchAllRules: Sequent -> (InferenceRule * (InferenceRuleArgument list)) list

     (* Reimplementation of SequentCommon functions for our internal representation *)
    val ArgStringValue: RuleArgValue -> string option
    val ArgNumberValue: RuleArgValue -> int option
    val ArgSubgoalValue: RuleArgValue -> string option
    val ArgFormulaNumValue: RuleArgValue -> (SequentPart * (int option)) option
    val ArgObjectValue: RuleArgValue -> ProgramObject option
    val ArgListValue: RuleArgValue -> RuleArgValue list option

    val findArg: RuleArg list -> string -> RuleArgValue option
	    
    val findArgValue: RuleArg list -> string -> RuleArgValue option
    val findArgStringValue: RuleArg list -> string -> string option
    val findArgNumberValue:RuleArg list -> string -> int option
    val findArgSubgoalValue: RuleArg list -> string -> string option
    val findArgFormulaNumValue: RuleArg list -> string -> (SequentPart * (int option)) option
    val findArgObjectValue: RuleArg list -> string -> ProgramObject option
    val findArgListValue: RuleArg list -> string -> RuleArgValue list option

  end
  
