signature TRANS =
  sig

    (* "Proof terms" for keeping sets of assigned variables in
    different parts of assignments.  Corresponds to the sequent C;N;L;I |- A. *)

    structure Abstract: ABSTRACT
    structure Typecheck: TYPECHECK
    sharing Typecheck.Evaluate.ParserDefault = Abstract.ParserDefault
    type Abstraction = Abstract.Abstraction
    type TypeAbstraction = Abstract.TypeAbstraction
    type ParseTree = Typecheck.Evaluate.ParserDefault.ParseTreeStruct.ParseTree
    type Str = Str.Str
    type options = Typecheck.Evaluate.ParserDefault.Interface.Options.options

    type Cone (* Cone of influence *)

    (* correspondence b/w SyMP state vars and primitive vars in the translation *)
    type PrimitiveVars

    datatype ConeVar =
	PVar of ParseTree
      | DelayedVar of ParseTree (* Delayed computation of cone, DelayedVar(expr) *)

    datatype VarType =
	NormalVar of ParseTree
      | NextVar of ParseTree
      | InitVar of ParseTree

    type AsstVars = {norm: VarType list,
		     next: VarType list,
		     init: VarType list}

    (* Representation of sequent N;L;I |- A1;...;An *)
    (* type AsstSeq = AsstVars * (ParseTree list) *)

    datatype AsstVarsTree =
        (* vars |- varName := expr *)
	NormalAsstTree of AsstVars * ParseTree * ParseTree
      | NextAsstTree of AsstVars * ParseTree * ParseTree
      | InitAsstTree of AsstVars * ParseTree * ParseTree
      | NopAsstTree of AsstVars
      | ListAsstTree of AsstVars * (AsstVarsTree list)
        (* vars |- let [defs] in <asst> end *)
      | LetAsstTree of AsstVars * (ParseTree list) * AsstVarsTree
        (* vars |- case selExpr of [(uname1, pat1, asst1), ..., (uname1, patN, asstN)] endcase *)
      | CaseAsstTree of AsstVars * ParseTree * ((ParseTree * ParseTree * AsstVarsTree) list)
        (* vars |- if c1 then asst1 elsif c2 then asst2 .... else asstN *)
      | IfAsstTree of AsstVars * ((ParseTree * AsstVarsTree) list) * AsstVarsTree
      | ChooseAsstTree of AsstVars * (ParseTree list option) * ((ParseTree * AsstVarsTree) list)
      | ForeachAsstTree of AsstVars * (ParseTree list) * AsstVarsTree
      | LabeledAsstTree of AsstVars * ParseTree * AsstVarsTree

    (* Transition relation of a leaf module (not a composition of any other modules) *)
    datatype AtomicModel =
	AtomicModel of { name: ParseTree option,
			 uname: ParseTree,
			 assts: AsstVarsTree,
			 cone: Cone,
			 pvars: PrimitiveVars,
			 (* We might not need this list, as we have the master list *)
			 (* stateVars: ParseTree list *)
			 (* Abstraction modules nested into the current module.
			    These modules may have further abstractions inside them. *)
			 absModules: AtomicModel list
		       }

    (* Transition relation for the model representaiton *)
    datatype TransRel = 
	(* Parallel composition of two modules *)
	TransSync2 of TransRel * TransRel
      | TransAsync2 of TransRel * TransRel
	(* Parallel composition over the parameters' ranges *)
      | TransSync of { names: ParseTree list,
		       body: TransRel,
		       parent: ParseTree }
      | TransAsync of { names: ParseTree list,
		        body: TransRel,
			parent: ParseTree }
      | TransAtomic of AtomicModel

    (* Representation of a model in the sequent *)
    type Model = { trans: TransRel,
		   objectHash: Typecheck.ObjectHash,
		   (* List of StateVar objects *)
		   stateVars: ParseTree list,
		   (* The united set of primitive variables *)
		   pvars: PrimitiveVars,
		   abs: Abstraction }

    (* Representation of a low-level transition relation (the body of the def-structure) *)
    datatype DefExprBody =
	(* Atomic expression *)
	DefExprBodyAtomic of ParseTree
	(* Sequential list of conditionals (no guarded conditionals for expressions) *)
      | DefSeqList of (DefExprBody * DefExprBody) list
	(* array indexed by the first components (functional type) *)
      | DefArray of {pairs: (ParseTree * DefExprBody) list, Type: ParseTree}
      | DefTuple of DefExprBody list
      | DefRecord of (ParseTree * DefExprBody) list
	(* Datatype values are represented as 
	   (indexValue, [(constructor, argValue), ...]) *)
      | DefEnum of DefExprBody * ((ParseTree * DefExprBody) list)

    (* The def-structure for expressions *)
    type DefsType

    datatype DefExpr =
	DefExprDecl of {defs: DefsType,
			body: DefExprBody}
      | DefExprAtomic of DefExprBody

    (* The def-structure for assignments (allows recursive nesting of def-structures) *)
    datatype DefAsstBody =
	(* triple [norm, next, init] (or [W, F, T], in the hacker's manual) *)
	DefAsstBodyAtomic of {norm: DefExprBody, next: DefExprBody, init: DefExprBody}
	(* Sequential conditionals *)
      | DefAsstSeqList of (DefExprBody * DefAsstBody) list
	(* Guarded-nondeterministic conditionals *)
      | DefAsstGuardList of (DefExprBody * DefAsstBody) list
	(* Boolean connectives over assingments *)
      | DefAsstAnd of DefAsstBody list
      | DefAsstOr of DefAsstBody list

    datatype DefAsst =
	DefAsstAtomic of DefAsstBody
      | DefAsstDecl of {defs: DefsType,
			body: DefAsstBody}

    type rebuildAsstOptions

    val def2Str: DefExpr -> Str.Str
    val defasst2Str: DefAsst -> Str.Str
    val def2str: DefExpr -> string
    val defasst2str: DefAsst -> string
    val getDefs: DefExpr -> (ParseTree * (DefExprBody list)) list
    val getAsstDefs: DefAsst -> (ParseTree * (DefExprBody list)) list
    val getBody: DefExpr -> DefExprBody
    val getAsstBody: DefAsst -> DefAsstBody

    val vtName: VarType -> ParseTree

    (* val getAsstSeq: AsstVarsTree -> AsstSeq *)
    (* val getPredTrees: AsstVarsTree -> AsstVarsTree list *)
    val getAsstVars: AsstVarsTree -> AsstVars
    (* Set new state variables to the top-level node *)
    val setAsstVars: AsstVars -> AsstVarsTree -> AsstVarsTree

    val AsstVars2str: AsstVars -> string
    val avt2Str : AsstVarsTree -> Str
    val avt2str : AsstVarsTree -> string
    val varCone2Str : VarType * (ConeVar list) -> Str
    val varCone2str : VarType * (ConeVar list) -> string
    val cone2Str : Cone  -> Str
    val cone2str : Cone  -> string
    val prim2Str : PrimitiveVars -> ParseTree -> Str
    val prim2str : PrimitiveVars -> ParseTree -> string
    val primVars2Str : PrimitiveVars -> Str
    val primVars2str : PrimitiveVars -> string
    (* Return the list of vars in the cone of a given var.
       getVarConeN collects only variables up to a given depth. 
       (~1 = `entire cone') *)
    val getVarConeN: options * (ParseTree -> ParseTree option) * PrimitiveVars * int
	  * int * Cone * Abstraction -> VarType -> ConeVar list
    val getVarCone: options * (ParseTree -> ParseTree option) * PrimitiveVars * int
	  * Cone * Abstraction -> VarType -> ConeVar list
    val inConeOf: options * (ParseTree -> ParseTree option) * PrimitiveVars * int
	  * Cone * Abstraction -> VarType * ConeVar -> bool option
    (* Make a clean copy of the cone for later destructive modifications *)
    val copyCone: Cone -> Cone

    (* Operations on PrimitiveVars *)
    val makePrimitiveVars: unit -> PrimitiveVars
    val mergePrimitiveVars: PrimitiveVars * PrimitiveVars -> PrimitiveVars

    (* substAVT options findObject pvars lim (tree, old, new) 
       Substitute `new' for `old' in the `tree' and return the new tree. *)
    val substAVT: options * (ParseTree -> ParseTree option) * PrimitiveVars
           * int -> (AsstVarsTree * ParseTree * ParseTree) -> AsstVarsTree

    (* For a list of assignments generate an AsstVarsTree.  On the
       way, check for circular and multiple assignments, and if found
       any, raise TransError. *)

    (* We are not using Abstraction at this stage yet.  We generate
       the most general representation of the transition relation and
       reuse it later. *)

    val collectAsstVars: options * (ParseTree -> ParseTree option) * int * Abstraction ->
      (ParseTree list * ParseTree list) -> (AsstVarsTree * Cone * PrimitiveVars)

    (* Customize an AsstVarsTree to a particular property under a
       particular abstraction *)

    val customizeAsstVars: options * (ParseTree -> ParseTree option)
          * PrimitiveVars * int * Cone * Abstraction
         -> ParseTree list -> AsstVarsTree -> AsstVarsTree

    (* Traverse the AsstTree, extract the ground vars from the
       assignments and add them to the tree.  The resulting tree is
       then balanced.  On the way, construct the new cone and
       PrimitiveVars database.  Returns (asstTree, cone, pvars).

       This is a handy function to apply to an abstracted or
       COI-filtered assignment tree.

       rebuildAsstVars(options, findObject, pvars, lim, abs) (stateVars, asstTree) *)

    val rebuildAsstVars: options * (ParseTree -> ParseTree option)
                       * PrimitiveVars * int * Abstraction
          -> (ParseTree list option * AsstVarsTree * rebuildAsstOptions)
          -> (AsstVarsTree * Cone * PrimitiveVars)

    val rebuildAsstOptionsDefault: rebuildAsstOptions
    val rebuildAsstOptionsCheckAssts: rebuildAsstOptions -> rebuildAsstOptions
    val rebuildAsstOptionsBalance: rebuildAsstOptions -> rebuildAsstOptions
    val rebuildAsstOptionsPreserveVars: rebuildAsstOptions -> rebuildAsstOptions

    (* Rebuild the transition relation of the model relative to the
       current abstraction in the model. *)
    val rebuildModel: rebuildAsstOptions -> Model -> Model
	
    (* Removes all the `init' assignments from the model *)
    val removeInit: Model -> Model

    (* Perform COI reduction on the model, along with other necessary
       required transformations before generating the transition
       relation *)
    val specializeModel: options -> Model -> ParseTree list -> Model

    (* Set the new abstraction for the model 

       Warning: changing abstraction requires rebuilding the model *)
    val updateAbstraction: Model * Abstraction -> Model
    val addTypeAbstraction: Model * TypeAbstraction -> Model

    (* Generate a transition relation def-structure for a Model *)
    val generateModelTrans: Model -> DefAsst

    (* boolExpr(findObject, abs) expr
       Translate a boolean expression into a boolean def-structure suitable for 
       the use in external model checkers *)
    val boolExpr: (ParseTree -> ParseTree option) * Abstraction * (ParseTree list option)
           -> ParseTree -> DefExpr

    (* Translation of an arbitrary expression into a def-structure.
       Sequent e ~> F *)
    val expressionDef: (ParseTree -> ParseTree option) * Abstraction * (ParseTree list option)
            -> ParseTree -> DefExpr   

    val getModuleAssts: ParseTree -> ParseTree list

    (* makeModel (context, objectHash) model *)
    val makeModel: ParseTree * Typecheck.ObjectHash -> ParseTree -> Model
    val modelEq: Model * Model -> bool
    val model2str: Model -> string
    (* Later we either extend the functions above, or include more
       functions, to compute an abstracted transition relation,
       relative to a provided `Abstraction'. *)

    (* Some functions to manipulate expression def-structures: add
       extractors and compute equality. *)
    val normalizeDefET: (ParseTree -> ParseTree option) -> int * DefExpr -> DefExpr
    val normalizeDefER: (ParseTree -> ParseTree option) -> ParseTree * DefExpr -> DefExpr
    val normalizeDefEC: (ParseTree -> ParseTree option) -> ParseTree * DefExpr -> DefExpr
    val normalizeDefEI: (ParseTree -> ParseTree option) -> DefExpr -> DefExpr
    val normalizeDefEq: (ParseTree -> ParseTree option) -> DefExpr * DefExpr -> DefExpr
    (* normalizeDefAppl(findObject, abs, stateVarsOpt) (E, G) *)
    val normalizeDefAppl: (ParseTree -> ParseTree option) * Abstraction * (ParseTree list option)
         -> DefExpr * DefExpr -> DefExpr

    (* Creating a def-structure from lists of defs and a body *)
    val makeDef: (ParseTree * (DefExprBody list)) list list * DefExprBody -> DefExpr
    val makeAsstDef: (ParseTree * (DefExprBody list)) list list * DefAsstBody -> DefAsst

  end
