signature TRANS_GEN =
  sig
    structure TransCommon: TRANS_COMMON

    type Abstraction = TransCommon.Abstract.Abstraction
    type ParseTree = TransCommon.ParseTree
    type options = TransCommon.options
    type AsstVarsTree = TransCommon.AsstVarsTree
    type PrimitiveVars = TransCommon.PrimitiveVars
    type Model = TransCommon.Model

    (* Representation of a low-level transition relation (the body of the def-structure) *)
    datatype DefExprBody =
	(* Atomic expression *)
	DefExprBodyAtomic of ParseTree
	(* Sequential list of conditionals (no guarded conditionals for expressions) *)
      | DefSeqList of (DefExprBody * DefExprBody) list
	(* array indexed by the first components (functional type) *)
      | DefArray of {pairs: (ParseTree * DefExprBody) list, Type: ParseTree}
      | DefTuple of DefExprBody list
      | DefRecord of (ParseTree * DefExprBody) list
	(* Datatype values are represented as 
	   (indexValue, [(constructor, argValue), ...]) *)
      | DefEnum of DefExprBody * ((ParseTree * DefExprBody) list)

    (* The def-structure for expressions *)
    type DefsType

    datatype DefExpr =
	DefExprDecl of {defs: DefsType,
			body: DefExprBody}
      | DefExprAtomic of DefExprBody

    (* The def-structure for assignments (allows recursive nesting of def-structures) *)
    datatype DefAsstBody =
	(* triple [norm, next, init] (or [W, F, T], in the hacker's manual) *)
	DefAsstBodyAtomic of {norm: DefExprBody, next: DefExprBody, init: DefExprBody}
	(* Sequential conditionals *)
      | DefAsstSeqList of (DefExprBody * DefAsstBody) list
	(* Guarded-nondeterministic conditionals *)
      | DefAsstGuardList of (DefExprBody * DefAsstBody) list
	(* Boolean connectives over assingments *)
      | DefAsstAnd of DefAsstBody list
      | DefAsstOr of DefAsstBody list

    datatype DefAsst =
	DefAsstAtomic of DefAsstBody
      | DefAsstDecl of {defs: DefsType,
			body: DefAsstBody}
    val def2Str: DefExpr -> Str.Str
    val defasst2Str: DefAsst -> Str.Str
    val def2str: DefExpr -> string
    val defasst2str: DefAsst -> string
    val getDefs: DefExpr -> (ParseTree * (DefExprBody list)) list
    val getAsstDefs: DefAsst -> (ParseTree * (DefExprBody list)) list
    val getBody: DefExpr -> DefExprBody
    val getAsstBody: DefAsst -> DefAsstBody

    (* substAVT options findObject pvars lim (tree, old, new) 
       Substitute `new' for `old' in the `tree' and return the new tree. *)
    val substAVT: options * (ParseTree -> ParseTree option) * PrimitiveVars
           * int -> (AsstVarsTree * ParseTree * ParseTree) -> AsstVarsTree

    (* Generate a transition relation def-structure for a Model *)
    val generateModelTrans: Model -> DefAsst

    (* boolExpr(findObject, abs, stateVarsOpt) expr
       Translate a boolean expression into a boolean def-structure suitable for 
       the use in external model checkers *)
    val boolExpr: (ParseTree -> ParseTree option) * Abstraction * (ParseTree list option)
           -> ParseTree -> DefExpr

    (* Translation of an arbitrary expression into a def-structure.
       Sequent e ~> F *)
    val expressionDef: (ParseTree -> ParseTree option) * Abstraction * (ParseTree list option)
            -> ParseTree -> DefExpr   

    (* Some functions to manipulate expression def-structures: add
       extractors and compute equality. *)
    val normalizeDefET: (ParseTree -> ParseTree option) -> int * DefExpr -> DefExpr
    val normalizeDefER: (ParseTree -> ParseTree option) -> ParseTree * DefExpr -> DefExpr
    val normalizeDefEC: (ParseTree -> ParseTree option) -> ParseTree * DefExpr -> DefExpr
    val normalizeDefEI: (ParseTree -> ParseTree option) -> DefExpr -> DefExpr
    (* normalizeDefEq useAbstraction (findObject, abs, stateVarsOpt) (F1, F2) *)
    val normalizeDefEq: bool
            -> (ParseTree -> ParseTree option) * Abstraction * (ParseTree list option)
            -> DefExpr * DefExpr -> DefExpr

    (* normalizeDefAppl useAbstraction (findObject, abs, stateVarsOpt) (E, G) *)
    val normalizeDefAppl: bool
         -> (ParseTree -> ParseTree option) * Abstraction * (ParseTree list option)
         -> DefExpr * DefExpr -> DefExpr

    (* Creating a def-structure from lists of defs and a body *)
    val makeDef: (ParseTree * (DefExprBody list)) list list * DefExprBody -> DefExpr
    val makeAsstDef: (ParseTree * (DefExprBody list)) list list * DefAsstBody -> DefAsst

  end
