functor TransGenFun(structure TransPrep: TRANS_PREP): TRANS_GEN =

  struct
    structure TransPrep = TransPrep

    open TransPrep
    open TransCommon

    open ListOps
    open Hash
    open Pos
    open Str
    open Typecheck
    open Evaluate
    open Abstract
    open ParserDefault
    open ParseTreeStruct
    open Interface
    open Options
    open TimingStat

    (* Hash of results for heavily called functions to avoid
       exponential blowup of repeated calls (dynamic programming) *)

    type Key = string * (ParseTree list)

    (* Representation of a low-level transition relation 
       (the body of the def-structure). *)
    datatype DefExprBody =
	(* Atomic expression *)
	DefExprBodyAtomic of ParseTree
	(* Sequential list of conditionals (no guarded conditionals for expressions) *)
      | DefSeqList of (DefExprBody * DefExprBody) list
	(* array indexed by the first components (functional type),
	   and the type of the whole thing *)
      | DefArray of {pairs: (ParseTree * DefExprBody) list, Type: ParseTree}
      | DefTuple of DefExprBody list
      | DefRecord of (ParseTree * DefExprBody) list
	(* Datatype values are represented as 
	   (indexValue, [(constructor, argValue), ...]) *)
      | DefEnum of DefExprBody * ((ParseTree * DefExprBody) list)

    type DefsType = (ParseTree, (DefExprBody list)) Hash

    (* The def-structure for expressions *)
    datatype DefExpr =
	DefExprDecl of {defs: DefsType,
			body: DefExprBody}
      | DefExprAtomic of DefExprBody

    (* The def-structure for assignments (allows recursive nesting of def-structures) *)
    datatype DefAsstBody =
	(* triple [norm, next, init] (or [W, F, T], in the hacker's manual) *)
	DefAsstBodyAtomic of {norm: DefExprBody, next: DefExprBody, init: DefExprBody}
	(* Sequential conditionals *)
      | DefAsstSeqList of (DefExprBody * DefAsstBody) list
	(* Guarded-nondeterministic conditionals *)
      | DefAsstGuardList of (DefExprBody * DefAsstBody) list
	(* Boolean connectives over assingments *)
      | DefAsstAnd of DefAsstBody list
      | DefAsstOr of DefAsstBody list

    datatype DefAsst =
	DefAsstAtomic of DefAsstBody
      | DefAsstDecl of {defs: DefsType,
			body: DefAsstBody}

    (* Debugging function: call (op @), but time it. *)
    fun appendLists(l1, l2) =
	let val funName = "TransGen.appendLists"
	    val timer = startTimer funName
	in (l1@l2) before 
	    (updateStat timer; 
	     verbDebug (getOptions()) funName ("\n"^funName^": "^(statTimer2string timer)))
	end
    val op @ = appendLists

    (* This is the global defs that everyone will share, for efficiency *)
    val defsHash = ref(makeHashDefault(ptEq, pt2string): DefsType)

(*
    fun keyEq((fname1, ptList1), (fname2, ptList2)) =
	  fname1 = fname2 andalso (case zipOpt(ptList1, ptList2) of
				       SOME pairs => List.all ptEq pairs
				     | NONE => false)
    fun key2str(fname, ptList) = "@@"^fname^"@@ "^(ptlist2str "," ptList)
    val resultHash = ref(makeHashDefault(keyEq, key2str): (Key, ParseTree) Hash)
    fun newResultHash() = (resultHash := makeHashDefault(keyEq, key2str))

    fun lookupResult(key as (fname, _)) =
	let val res = findHash(!resultHash, key)
	    val _ = (case res of
			 SOME _ => lazyVerbDebug(getOptions()) fname 
			     (fn()=>fname^": lookup succeeds\n")
		       | NONE => ())
	in res
	end
    (* Record result in the hash and return it back *)
    fun saveResult(fname, args, v) = 
	  (insertHashDestructive(!resultHash, (fname, args), v); v)
*)
    fun isBoolFun (Iff _) = true
      | isBoolFun (And _) = true
      | isBoolFun (Or _) = true
      | isBoolFun (Implies _) = true
      | isBoolFun (Eq _) = true
      | isBoolFun (NotEq _) = true
      | isBoolFun (Not _) = true
      | isBoolFun (Ag _) = true
      | isBoolFun (Af _) = true
      | isBoolFun (Eg _) = true
      | isBoolFun (Ef _) = true
      | isBoolFun (Ax _) = true
      | isBoolFun (Ex _) = true
      | isBoolFun (Au _) = true
      | isBoolFun (Eu _) = true
      | isBoolFun (Ar _) = true
      | isBoolFun (Er _) = true
      | isBoolFun (Lt _) = true
      | isBoolFun (Gt _) = true
      | isBoolFun (Le _) = true
      | isBoolFun (Ge _) = true
      | isBoolFun _ = false

    (* Check whether the parse tree has temporal operators inside *)
    fun hasTemporal pt = if isTemporal pt then true
			 else List.exists hasTemporal (ptSubtrees pt)

    (* 2-valued temporal operators *)
    local
	val b1tp = FunType(dp,BoolType2,BoolType2)
	val b2tp = FunType(dp,TupleType(dp,[BoolType2,BoolType2]),BoolType2)
    in
        val Ag2 = Object{name = Ag dp,
		     uname=Id(dp,newName()),
		     Type = b1tp,
		     def = Builtin{name=Ag dp, Type=b1tp}}
	val Af2 = Object{name = Af dp,
			 uname=Id(dp,newName()),
			 Type = b1tp,
			 def = Builtin{name=Af dp, Type=b1tp}}
	val Eg2 = Object{name = Eg dp,
			 uname=Id(dp,newName()),
			 Type = b1tp,
			 def = Builtin{name=Eg dp, Type=b1tp}}
	val Ef2 = Object{name = Ef dp,
			 uname=Id(dp,newName()),
			 Type = b1tp,
			 def = Builtin{name=Ef dp, Type=b1tp}}
	val Ax2 = Object{name = Ax dp,
			 uname=Id(dp,newName()),
			 Type = b1tp,
			 def = Builtin{name=Ax dp, Type=b1tp}}
	val Ex2 = Object{name = Ex dp,
			 uname=Id(dp,newName()),
			 Type = b1tp,
			 def = Builtin{name=Ex dp, Type=b1tp}}

	val Au2 = Object{name = Au dp,
			 uname=Id(dp,newName()),
			 Type = b2tp,
			 def = Builtin{name=Au dp, Type=b2tp}}
	val Eu2 = Object{name = Eu dp,
			 uname=Id(dp,newName()),
			 Type = b2tp,
			 def = Builtin{name=Eu dp, Type=b2tp}}
	val Ar2 = Object{name = Ar dp,
			 uname=Id(dp,newName()),
			 Type = b2tp,
			 def = Builtin{name=Ar dp, Type=b2tp}}
	val Er2 = Object{name = Er dp,
			 uname=Id(dp,newName()),
			 Type = b2tp,
			 def = Builtin{name=Er dp, Type=b2tp}}
	val Globally2 = Object{name = Globally dp,
			       uname=Id(dp,newName()),
			       Type = b1tp,
			       def = Builtin{name=Globally dp, Type=b1tp}}
	val Eventually2 = Object{name = Eventually dp,
				 uname=Id(dp,newName()),
				 Type = b1tp,
				 def = Builtin{name=Globally dp, Type=b1tp}}
	val NextTime2 = Object{name = NextTime dp,
			       uname=Id(dp,newName()),
			       Type = b1tp,
			       def = Builtin{name=NextTime dp, Type=b1tp}}
	val Until2 = Object{name = Until dp,
			    uname=Id(dp,newName()),
			    Type = b2tp,
			    def = Builtin{name=Until dp, Type=b2tp}}
	val Releases2 = Object{name = Releases dp,
			       uname=Id(dp,newName()),
			       Type = b2tp,
			       def = Builtin{name=Releases dp, Type=b2tp}}
    end


    (* Convert a temporal operator to 2-valued boolean one *)
    fun mapTemporal(Ag _) = Ag2
      | mapTemporal(Af _) = Af2
      | mapTemporal(Eg _) = Eg2
      | mapTemporal(Ef _) = Ef2
      | mapTemporal(Ax _) = Ax2
      | mapTemporal(Ex _) = Ex2
      | mapTemporal(Au _) = Au2
      | mapTemporal(Eu _) = Eu2
      | mapTemporal(Ar _) = Ar2
      | mapTemporal(Er _) = Er2
      | mapTemporal(Globally _) = Globally2
      | mapTemporal(Eventually _) = Eventually2
      | mapTemporal(Until _) = Until2
      | mapTemporal(Releases _) = Releases2
      | mapTemporal(NextTime _) = NextTime2
      | mapTemporal(Builtin{name=Op, ...}) = mapTemporal Op
      | mapTemporal x = x

    fun tryAbstract options abs tp =
	(case abstract options abs tp of
	     SOME ta => getAbsType ta
	   | NONE => tp)

    (* Sequent `(delta <=~ gamma) => d'.  Find an index for an array
       state var `gamma' that is used in `delta' (the first arg).  If
       `delta' doesn't have `gamma' at all, return NONE.  Assumption:
       `delta' is primitive. *)

    fun findArrayIndex (Appl(_,delta,d), gamma) =
	  if svSimilar(delta,gamma) then SOME(d)
	  else findArrayIndex(delta,gamma)
      | findArrayIndex (ExtractRecord(_,delta), gamma) = findArrayIndex(delta,gamma)
      | findArrayIndex (ExtractTuple(_,delta), gamma) = findArrayIndex(delta,gamma)
      | findArrayIndex (ExtractAppl(_,delta), gamma) = findArrayIndex(delta,gamma)
      | findArrayIndex (ExtractIndex delta, gamma) = findArrayIndex(delta, gamma)
      (* delta is atomic *)
      | findArrayIndex (_,_) = NONE

    (* Sequent `V |- gamma => D'.  Split V into subsets that are parts
       of (gamma d) for each `d'.  The set D is a set of pairs (d, V_d). *)

    fun splitByArrayIndex ([],_) = []
      | splitByArrayIndex (delta::V, gamma) =
	let val D = splitByArrayIndex(V, gamma)
	    fun insert((d1,Vd)::lst, d, delta) =
		  if ptEq(d1,d) then (d,delta::Vd)::lst
		  else (d1,Vd)::(insert(lst, d, delta))
	      | insert([], d, delta) = [(d,[delta])]
	in 
	    (case findArrayIndex(vtName delta, gamma) of
		 NONE => D
	       | SOME d => insert(D,d,delta))
	end

    val true3Val = Builtin{name=True dp, Type=BoolType dp}
    val false3Val = Builtin{name=False dp, Type=BoolType dp}
    val trueVal = True2
    val falseVal = False2
    val trueDef = DefExprAtomic(DefExprBodyAtomic trueVal)
    val falseDef = DefExprAtomic(DefExprBodyAtomic falseVal)
    val true3Def = DefExprAtomic(DefExprBodyAtomic true3Val)
    val false3Def = DefExprAtomic(DefExprBodyAtomic false3Val)
    (* 2-valued propositional connectives *)
    val notVal = Builtin{name=Not dp, Type=FunType(dp,BoolType2, BoolType2)}
    val andVal = Builtin{name=And dp, 
			 Type=FunType(dp,TupleType(dp,[BoolType2, BoolType2]),
				      BoolType2)}
    val orVal = Builtin{name=Or dp, 
			Type=FunType(dp,TupleType(dp,[BoolType2, BoolType2]),
				     BoolType2)}
    val impliesVal = Builtin{name=Implies dp, 
			     Type=FunType(dp,TupleType(dp,[BoolType2, BoolType2]),
					  BoolType2)}
    val iffVal = Builtin{name=Iff dp, 
			 Type=FunType(dp,TupleType(dp,[BoolType2, BoolType2]),
				     BoolType2)}
    (* Create an `undefined' value of type `tp' *)
    fun bot tp = Builtin{name=Undefined dp, Type=tp}
    (* Create an `Anyvalue' expression of type `tp' *)
    fun anyvalue tp = Builtin{name=Anyvalue dp, Type=tp}

    val botBoolDef = DefExprAtomic(DefExprBodyAtomic(bot(BoolType dp)))
    fun negFormula f = Appl(dp, notVal, f)
    fun impliesFormula(f,g) = Appl(dp, impliesVal, TupleExpr(dp,[f,g]))
    fun iffFormula(f,g) = Appl(dp, iffVal, TupleExpr(dp,[f,g]))
    fun foldOp (Op,default) lst =
	let fun foldFun(f1,f2) = Appl(dp,Op,TupleExpr(dp,[f1,f2]))
	    fun doFold [] = default
	      | doFold (f::lst) = List.foldr foldFun f lst
	in doFold(List.rev lst)
	end
    val foldAnd = foldOp(andVal, trueVal)
    val foldOr = foldOp(orVal, falseVal)
    (* Formula for `if c then tcase else fcase endif' *)
    fun caseFormula(c, tcase, fcase) =
	foldAnd[impliesFormula(c,tcase),impliesFormula(negFormula c, fcase)]
    (* Create a formula `e1 = e2' of type BoolType2 *)
    fun eqFormula findObject (e1,e2) = 
	let val tp = getExprType findObject e1
	in Appl(dp,Builtin{name=Eq dp,
			   Type=FunType(dp,TupleType(dp,[tp,tp]),BoolType2)},
		TupleExpr(dp,[e1,e2]))
	end
    (* Substitute `v' for `e' in a ParseTree `x'. 
       Checks for name shadowing in FunClosure and ChoiceClosure. *)
    fun subst(v,e) (body as FunClosure{formals=f,...}) = 
	if eqUNames(v,f) then body else substRec(v,e) body
      | subst(v,e) (body as ChoiceClosure{uname=u,...}) = 
	    if eqUNames(v,u) then body else substRec(v,e) body
      | subst(v,e) body =
		if eqUNames(v,body) then e else substRec(v,e) body
    and substRec(v,e) x = ptTransform (subst (v, e)) x
    fun foldCase ([],Fbot) = Fbot
      | foldCase ((c,f)::pairs, Fbot) =
	caseFormula(c,f,foldCase(pairs,Fbot))

    (* Checks if the expression is a constant literal (an expression
       that has a concrete interpreted value) *)

    fun isLiteral (Builtin{name=Anyvalue _, ...}) = false
      | isLiteral (Builtin _) = true
      | isLiteral (FunClosure _) = true
      | isLiteral (Number _) = true
      | isLiteral (TypeConstr _) = true
      | isLiteral (ExtractTuple(_,e)) = isLiteral e
      | isLiteral (ExtractRecord(_,e)) = isLiteral e
      | isLiteral (ExtractAppl(_,e)) = isLiteral e
      | isLiteral (ExtractIndex(e)) = isLiteral e
      | isLiteral (Object{def=e,...}) = isLiteral e
      | isLiteral (Appl(_,e,e')) = isLiteral e andalso isLiteral e'
      | isLiteral _ = false
    (* Check if the expression is atomic (a state var or a constant with only constant
       expressions in the array indices) *)
    fun isAtomic (StateVar _) = true
      | isAtomic (Next _) = true
      | isAtomic (DynPatternFormal _) = true
      | isAtomic (ExtractTuple(_,e)) = isAtomic e
      | isAtomic (ExtractRecord(_,e)) = isAtomic e
      | isAtomic (ExtractAppl(_,e)) = isAtomic e
      | isAtomic (ExtractIndex(e)) = isAtomic e
      | isAtomic (Appl(_,e,e')) = 
	  isAtomic e andalso not(isTemporal e) andalso isConst e'
      | isAtomic (DefVar _) = true
      | isAtomic e = isConst e

    fun isBuiltin (Builtin _) = true
      | isBuiltin (Object{def=e,...}) = isBuiltin e
      | isBuiltin (ObjectInst{obj=e,...}) = isBuiltin e
      | isBuiltin _ = false

    fun getBuiltinName (Builtin{name=n,...}) = SOME n
      | getBuiltinName (Object{def=e,...}) = getBuiltinName e
      | getBuiltinName (ObjectInst{obj=e,...}) = getBuiltinName e
      | getBuiltinName _ = NONE

    fun getBuiltinCore (e as Builtin _) = SOME e
      | getBuiltinCore (Object{def=e,...}) = getBuiltinCore e
      | getBuiltinCore (ObjectInst{obj=e,...}) = getBuiltinCore e
      | getBuiltinCore _ = NONE

    (* Wraps `e' into an extractor *)
    fun wrapExtr(ExtractAppl(c,_), e) = ExtractAppl(c,e)
      | wrapExtr(ExtractTuple(n,_), e) = ExtractTuple(n,e)
      | wrapExtr(ExtractRecord(f,_), e) = ExtractRecord(f,e)
      | wrapExtr(ExtractIndex _, e) = ExtractIndex e
      | wrapExtr(Appl(p,_,d), e) = Appl(p,e,d)
      | wrapExtr(x,_) = raise SympBug
	("wrapExtr: not an extractor: "^(pt2string x))
    fun wrapExtrs lst expr = List.foldr wrapExtr expr (List.rev lst)

    fun extractIndex options findObject pvars e =
	(* if isStateVar e then 
	  (case getPrimitiveIndexVar pvars e of
		   SOME x => x
		 | NONE => raise SympBug
		       ("generateTrans/extractIndex: state var doesn't have index var:\n  "
			^(pt2string e)) )
	else *) evaluateExpr options findObject (ExtractIndex e)

    (* Get all the values of a given type after trying to abstract it. *)
    fun getTypeValuesAbs(options, findObject, abs) tp =
	let val atp = tryAbstract options abs tp
	    val typeSize = (case getTypeSize(options, findObject) atp of
				s as FiniteSize n => if n <= (get_limit options) then s
						     else InfiniteSize
			      | InfiniteSize => InfiniteSize)
	in
	    case typeSize of
		InfiniteSize => NONE
	      | FiniteSize _ => getTypeValues(options, findObject) atp
	end

    (* If the expression is a constant, and its type is abstracted,
       find the corresponding abstract value for it.  Otherwise return
       the expression unchanged. *)

    fun tryAbsMap(parms as (options, findObject, abs)) e =
	let (* Define what we mean here by a constant *)
	    fun isConst(SkolemConst _) = true
	      | isConst(AbstractConst _) = true
	      | isConst(StaticFormalConst _) = true
	      | isConst(Appl(_, x, y)) = isConst x andalso isConst y
	      | isConst(TupleExpr(_, lst)) = List.all isConst lst
	      | isConst(RecordExpr(_, lst)) =
		  List.all(fn (RecordAsst(_,_,e))=>isConst e | _ => false) lst
	      | isConst(ExtractAppl(_, e)) = isConst e
	      | isConst(ExtractTuple(_, e)) = isConst e
	      | isConst(ExtractRecord(_, e)) = isConst e
	      | isConst(ExtractIndex e) = isConst e
	      | isConst x = isValue x

(*	    fun isStateVar(StateVar _) = true
	      | isStateVar(ExtractAppl(_, e)) = isStateVar e
	      | isStateVar(ExtractTuple(_, e)) = isStateVar e
	      | isStateVar(ExtractRecord(_, e)) = isStateVar e
	      | isStateVar(ExtractIndex e) = isStateVar e
	      | isStateVar _ = false
*)
	    fun doMap(SOME typeAbs) = getAbsMap typeAbs e
	      | doMap NONE = e
	    fun doAbs() =
		let val tp = getExprType findObject e
		in
		    doMap(abstract options abs tp)
		end
	in
	    if isConst e then doAbs() else e
	end

    (* Compute all the vectors of values for variables in `names' *)
    fun allVarValues(options, findObject, lim, abs) names =
	let val tps = List.map(fn x=>tryAbstract options abs (getExprType findObject x)) names
	    val typeSize = 
		(case getTypeSize(options, findObject) (TupleType(dp,tps)) of
		     FiniteSize n => if n <= lim then FiniteSize n 
				     else InfiniteSize
		   | InfiniteSize => InfiniteSize)
	    val _ = (case typeSize of
			 InfiniteSize => raise TransError
			     ("Quantifier expansion will be too large for the type:\n  "
			      ^(pt2string (TupleType(dp,tps))))
		       | _ => ())
	    fun getTpVals tp = 
		(case getTypeValues(options, findObject) tp of
		     SOME lst => lst
		   (* This, actually, shouldn't happen already *)
		   | NONE => raise TransError
			 ("type of a quantified variable is too large "
			  ^"to list all its values:\n  "
			  ^(pt2string tp)
			  ^"\nin expressions\n  "^(ptlist2str "\n  " names)))
	    val ds = comblist(List.map getTpVals tps)
	in ds
	end
	    
	    
    (* Return the list of expressions e_d obtained by substituting all
       possible combinations of values for `names' into `e'.  Each new
       expression is statically evaluated. *)

    fun substVarValues(parms as (options, findObject, lim, abs)) (names, e) =
	let val evalExpr = evaluateExpr options findObject
	    val ds = allVarValues parms names
	    fun doOne dvec =
		List.foldr(fn((v,d),e) => evalExpr(subst(v,d) e)) e (zip(names, dvec))
	in List.map doOne ds
	end

    fun substAVT(options, findObject, pvars, lim) (tree, old, new) =
	let fun substExpr e = evaluateExpr options findObject (subst(old, new) e)
	    fun substVar(NormalVar v) = NormalVar(substExpr v)
	      | substVar(NextVar v) = NextVar(substExpr v)
	      | substVar(InitVar v) = InitVar(substExpr v)
	    fun substVars{init=initV, next=nextV, norm=normV} =
		let val initV = List.map substVar initV
		    val nextV = List.map substVar nextV
		    val normV = List.map substVar normV
		in {init=initV, next=nextV, norm=normV}
		end			  
	    fun loop(t as NormalAsstTree(vars, v, e)) =
		NormalAsstTree(substVars vars, substExpr v, substExpr e)
	      | loop(t as NextAsstTree(vars, v, e)) =
		NextAsstTree(substVars vars, substExpr v, substExpr e)
	      | loop(t as InitAsstTree(vars, v, e)) = 
		InitAsstTree(substVars vars, substExpr v, substExpr e)
	      | loop(NopAsstTree vars) = NopAsstTree(substVars vars)
	      | loop(ListAsstTree(vars, lst)) =
		ListAsstTree(substVars vars, List.map loop lst)
	      | loop(LetAsstTree(vars, decls, t)) =
		LetAsstTree(substVars vars, decls, loop t)
	      | loop(CaseAsstTree(vars, sel, pairs)) =
		let fun ff(u, p, t) = (substExpr u, substExpr p, loop t)
		in CaseAsstTree(substVars vars, substExpr sel, List.map ff pairs)
		end
	      | loop(IfAsstTree(vars, pairs, t)) =
		let fun ff(c, t) = (substExpr c, loop t)
		in IfAsstTree(substVars vars, List.map ff pairs, loop t)
		end
	      | loop(ChooseAsstTree(vars, paramsOpt, pairs)) =
		let fun ff(c, t) = (substExpr c, loop t)
		in ChooseAsstTree(substVars vars, paramsOpt, List.map ff pairs)
		end
	      | loop(ForeachAsstTree(vars, params, t)) =
		ForeachAsstTree(substVars vars, params, loop t)
	      | loop(LabeledAsstTree(vars, label, t)) =
		LabeledAsstTree(substVars vars, label, loop t)
	    val tree = loop tree

	in tree
	end

    (* Same as above, only do it for pairs (expr, asstTree) *)

    fun substVarValuesPairs(options, findObject, pvars, lim, abs) names pairs =
	let val funName = "substVarValuesPairs"
	    fun pairs2str pairs = "[\n  "
	        ^(strlist2str "\n  " (List.map(fn (c,t)=>
					       (pt2string c)^" => "^(avt2str t)) pairs))
		^"]"
	    fun argsFn() =
		"names = ["
		^(ptlist2str ", " names)^"],\n   pairs = "
		^(pairs2str pairs)
	    val _ = pushFunStackLazy(funName, argsFn)
	    val evalExpr = evaluateExpr options findObject
	    val substAVT = substAVT(options, findObject, pvars, lim)
	    fun doExpr(e, dvec) =
		List.foldr(fn((v,d),e) => evalExpr(subst(v,d) e)) e (zip(names, dvec))
	    fun doAVT(t, dvec) = List.foldr(fn((v,d),t) => substAVT(t,v,d)) t (zip(names, dvec))
	    fun doPair [] (_, acc) = acc
	      | doPair(dvec::ds) (pair as (e1, t1), acc) =
		  doPair ds (pair, (doExpr(e1, dvec), doAVT(t1, dvec))::acc)
	    val ds = allVarValues(options, findObject, lim, abs) names
	    val res = List.foldr (doPair ds) [] pairs
	    val _ = popFunStackLazy(funName, fn()=> pairs2str res)
	in 
	    res
	end

    fun isEq(Eq _) = true
      | isEq(Builtin{name=n,...}) = isEq n
      | isEq _ = false

    (* If we have `x = anyvalue', translate that into (true | false) *)
    fun checkEqAnyvalue(TupleExpr(_, [x, y])) = isAnyvalue x orelse isAnyvalue y
      | checkEqAnyvalue _ = false


    (* Return new defs for newly created def-structures *)
    fun newDefsRaw() = makeHashDefault(ptEq, pt2string)
    (* Return shared defs *)
    fun sharedDefsRaw() = !defsHash

    (* The normalization operators for the expression def-structure. *)

    fun getDefs(DefExprDecl{defs=defs,...}) = hash2any(fn x=>x) defs
      | getDefs _ = []

    fun getDefsRaw(DefExprDecl{defs=defs,...}) = defs
      | getDefsRaw _ = newDefsRaw()

    fun getBody(DefExprDecl{body=body,...}) = body
      | getBody(DefExprAtomic body) = body
	    
    fun getAsstDefs(DefAsstDecl{defs=defs,...}) = hash2any(fn x=>x) defs
      | getAsstDefs _ = []

    fun getAsstDefsRaw(DefAsstDecl{defs=defs,...}) = defs
      | getAsstDefsRaw _ = newDefsRaw()

    fun getAsstBody(DefAsstDecl{body=body,...}) = body
      | getAsstBody(DefAsstAtomic body) = body
	    
    (* Build the minimal def-structure from lists of definitions and
       body.  It takes lists of lists of individual definitions
       because it's often necessary to merge defs from different
       places. *)

    fun makeDef(defsLists, body) =
	let val empty = ref true
	    val defsHash = newDefsRaw()
	    fun addList lst =
		let fun addOne(v, Es) =
		        (insertHashDestructive(defsHash, v, Es);
			 empty := false; ())
		in List.app addOne lst
		end
	    val _ = List.app addList defsLists
	in 
	    if !empty then DefExprAtomic body
	    else DefExprDecl{defs=defsHash, body=body}
	end

    (* Similar constructor for assignment def-structures *)
    fun makeAsstDef(defsLists, body) =
	let val empty = ref true
	    (* Grab the global defsHash *)
	    val defsHash = newDefsRaw()
	    fun addList lst =
		let fun addOne(v, Es) =
		        (insertHashDestructive(defsHash, v, Es);
			 empty := false; ())
		in List.app addOne lst
		end
	    val _ = List.app addList defsLists
	in 
	    if !empty then DefAsstAtomic body
	    else DefAsstDecl{defs=defsHash, body=body}
	end

    (* Use it as if it generates a new set of defs.  In reality, it just
       adds new defs destructively to the same common hash. *)

    fun addDefs(defs: DefsType, newDefs) =
	let fun addOne(v, Es) = (insertHashDestructive(defs, v, Es); ())
	in (List.app addOne newDefs; defs)
	end

    (* This is a wrapper hiding the fact that we are sharing defs.  If
       some hashes are new, they'll be merged into the shared one
       unless NONE of them are shared.  In the latter case, a new hash
       is constructed. *)

    fun mergeDefs(lst : DefsType list) =
	let val theHash = (case List.find(fn h=>sameHash(h, !defsHash)) lst of
			       SOME _ => !defsHash
			     | NONE => newDefsRaw())				   
	    fun doOne hash =
	        if sameHash(theHash, hash) then ()
		else List.app(fn(k,v)=>(insertHashDestructive(theHash, k, v); ()))
		             (hash2any(fn x=>x) hash)
	in 
	    (List.app doOne lst; theHash)
	end

    fun makeDefRaw(defs: DefsType, body) = DefExprDecl{defs=defs, body=body}
    fun makeAsstDefRaw(defs: DefsType, body) = DefAsstDecl{defs=defs, body=body}

    local fun op *(x, y) = Conc(x,y)
    in
	fun defBody2Str b =
	    let fun body2str(DefExprBodyAtomic pt) = pt2str pt
		  | body2str(DefSeqList lst) =
		      (Str "[")*(Strlistmap2Str ", " cond2str lst)*(Str "]")
		  | body2str(DefArray{pairs=lst, Type=tp}) =
		      (Str "Array[")*(Strlistmap2Str ", " (pair2str ", ") lst)*(Str ": ")
		      *(pt2str tp)*(Str "]")
		  | body2str(DefTuple lst) = (Str "(")*(Strlistmap2Str ", " body2str lst)*(Str ")")
		  | body2str(DefRecord lst) = 
		      (Str "{")*(Strlistmap2Str ", " (pair2str "=") lst)*(Str "}")
		  | body2str(DefEnum(ind, lst)) = 
		      (Str "Enum(")*(body2str ind)
		      *(case lst of
			    [] => Str ""
			  | _ => (Str ", [")*(Strlistmap2Str ", " (pair2str ", ") lst)*(Str "]"))
		      *(Str ")")
		and pair2str sep (c, b) = (Str "(")*(pt2str c)*(Str sep)*(body2str b)*(Str ")")
		and cond2str(c, b) = (Str "(")*(body2str c)*(Str " => ")*(body2str b)*(Str ")")
	    in body2str b
	    end
	    
	fun def2Str d =
	    let fun defpair2str(v, d) = (pt2str v)*(Str " = [")
		      *(Strlistmap2Str ", " defBody2Str d)*(Str "]")
		fun loop(d as DefExprDecl _) =
		    (Str "def{ ")*(Strlistmap2Str "\n   " defpair2str(getDefs d))
		    *(Str "\nin ")*(defBody2Str(getBody d))*(Str "\n}end")
		  | loop(DefExprAtomic body) = defBody2Str body
	    in
		loop d
	    end

	fun defAsstBody2Str b =
	    let fun pair2str(c, b) = (Str "(")*(defBody2Str c)*(Str ", ")*(body2str b)*(Str ")")
		and body2str(DefAsstBodyAtomic{norm=W, next=F, init=T}) =
		    (Str "[ INIT\n  ")*(defBody2Str T)
		    *(Str "\n  INVAR\n  ")*(defBody2Str W)
		    *(Str "\n  TRANS\n  ")*(defBody2Str F)*(Str "]")
		  | body2str(DefAsstSeqList lst) =
		    (Str "(")*(Strlistmap2Str ", " pair2str lst)*(Str ")")
		  | body2str(DefAsstGuardList lst) =
		    (Str "(")*(Strlistmap2Str " | " pair2str lst)*(Str ")")
		  | body2str(DefAsstAnd lst) =
		    (Str "(")*(Strlistmap2Str " and " body2str lst)*(Str ")")
		  | body2str(DefAsstOr lst) =
		    (Str "(")*(Strlistmap2Str " or " body2str lst)*(Str ")")
	    in body2str b
	    end

	fun defasst2Str d =
	    let fun defpair2str(v, d) =
		      (pt2str v)*(Str " = [")*(Strlistmap2Str ", " defBody2Str d)*(Str "]")
		fun loop(DefAsstAtomic body) = defAsstBody2Str body
		  | loop(d as DefAsstDecl _) =
		    (Str "def{ ")*(Strlistmap2Str "\n   " defpair2str(getAsstDefs d))
		    *(Str "\nin ")*(defAsstBody2Str(getAsstBody d))*(Str "\n}end")
	    in
		loop d
	    end
    end (* local *)

    fun def2str F = Str2string(def2Str F)
    fun defBody2str b = Str2string(defBody2Str b)

    fun defasst2str F = Str2string(defasst2Str F)
    fun defAsstBody2str b = Str2string(defAsstBody2Str b)

    fun getDefBodyType findObject (DefExprBodyAtomic pt) = getExprType findObject pt
      | getDefBodyType findObject (DefSeqList((_, b)::_)) = getDefBodyType findObject b
      | getDefBodyType findObject (DefSeqList[]) = raise SympBug
	  ("TransGen/getDefBodyType: no elements in DefSeqList")
      | getDefBodyType findObject (DefArray{Type=tp, ...}) = tp
      | getDefBodyType findObject (DefTuple lst) =
	let val tps = List.map (getDefBodyType findObject) lst
	in
	    TupleType(dp, tps)
	end
      | getDefBodyType findObject (DefRecord pairs) =
	let fun doPair(f, e) =
	        let val tp = getDefBodyType findObject e
		in RecordField{name=f, Type=tp}
		end
	in RecordType(dp, List.map doPair pairs)
	end
      | getDefBodyType findObject (F as DefEnum(ind, pairs)) =
	let fun find c = Option.map #2 (List.find(fn(c', E)=> ptEq(c, c')) pairs)
	    fun makeOne c =
		(case find c of
		     NONE => c
		   | SOME E => Of(dp, c, getDefBodyType findObject E))
	    fun ripExtr(ExtractIndex x) = x
	      | ripExtr x = raise SympBug
		  ("TransGen.getDefBodyType: type element of DefEnum's index"
		   ^" is not ExtractIndex:\n  "
		   ^(pt2string x)^"\n  expr: "^(defBody2str F))
	    val cs = (case getDefBodyType findObject ind of
			  AbstractType lst => List.map ripExtr lst
			| t => raise SympBug
			 ("TransGen.getDefBodyType: index of DefEnum is not of AbstractType:\n  "
			  ^"type: "^(pt2string t)
			  ^"\n  expr: "^(def2str(DefExprAtomic F))))
	in EnumType(dp, List.map makeOne cs)
	end

    fun getDefType findObject (DefExprAtomic body) = getDefBodyType findObject body
      | getDefType findObject (DefExprDecl{body=body,...}) = getDefBodyType findObject body

    (* Normalizes a condition list ((C_1, F_1), ..., (C_n, F_n)) into
       components of a valid def-structure.  Here in the list, both
       C_i's and F_i's can be (valid) def-structures. *)

    local
	(* Type of body returned by `ff' and what to do with it *)
	datatype 'a BodyType = None (* Trash this pair *)
                             | NoDefs of 'a (* Throw away all its defs *)
                             | Last of 'a (* The cond. was true, so it's the last pair *)
                             | Normal of 'a (* Keep the defs *)
	(* Return (newDefs, (var, body)) *)
	fun translateDefPairs(findObject, ff) ((C, F)::rest) = 
	    let val recur = translateDefPairs(findObject, ff)
		val (cDef, cBody) =
		(case getBody C of
		     DefExprBodyAtomic d => ([], d)
		   (* Elements other than DefSeqList can
		    occur in the array where the
		    "condition" is really an array index,
		    which potentially could be a tuple, a
		    record, or another array. *)
		   | b =>
			 let val n = newName()
			     val v = DefVar{name=n, Type=getDefBodyType findObject b}
			 in ([(v, [b])], v)
			 end)
		fun makePair None = recur rest
		  | makePair(NoDefs cBody) = (getDefsRaw F, (cBody, getBody F))::(recur rest)
		  | makePair(Last cBody) = [(getDefsRaw F, (cBody, getBody F))]
		  | makePair(Normal cBody) = 
		      (addDefs(mergeDefs[getDefsRaw C, getDefsRaw F], cDef),
		       (cBody, getBody F))::(recur rest)
	    in makePair (ff cBody)
	    end
	  | translateDefPairs _ [] = []
    in	    
	fun normalizeDefCond findObject condList =
	    let val funName = "normalizeDefCond"
		fun pair2str(C, F) = "("^(def2str C)^", "^(def2str F)^")"
		fun argsFn() = "["^(strlistmap2str ", " pair2str condList)^"]"
		val _ = pushFunStackLazy(funName, argsFn)
		val eval = evaluateExpr (getOptions()) findObject
		fun ff b =
		    let val bb = eval b
		    in 
			if ptEq(bb, False2) then None
			else if ptEq(bb, True2) then Last (getBody trueDef)
			else Normal(DefExprBodyAtomic bb)
		    end
		val res = translateDefPairs(findObject, ff) condList
		val defs = mergeDefs(List.map #1 res)
		val body = DefSeqList(List.map #2 res)
		val res = makeDefRaw(defs, body)
		val _ = popFunStackLazy(funName, fn()=>def2str res)
	    in
		res
	    end

	(* Same but for the array (the first element of the pair must be ParseTree) *)
	fun normalizeDefArray findObject (condList, tp) =
	    let val defPairs = translateDefPairs(findObject, fn x=> Normal x)
		val res = defPairs condList
		val defs = mergeDefs(List.map #1 res)
		val body = DefArray{pairs=List.map #2 res, Type=tp}
	    in
		makeDefRaw(defs, body)
	    end
    end

    (* Normalization of assignments *)
    local
	(* Return (newDefs, (var, body)) *)
	fun translateDefPair findObject (C, F) = 
	    let val (cDef, cBody) =
		(case getBody C of
		     b as DefExprBodyAtomic _ => ([], b)
		   (* Elements other than DefSeqList can
		    occur in the array where the
		    "condition" is really an array index,
		    which potentially could be a tuple, a
		    record, or another array. *)
		   | b =>
			 let val n = newName()
			     val v = DefVar{name=n, Type=getDefBodyType findObject b}
			 in ([(v, [b])], DefExprBodyAtomic v)
			 end)
	    in (addDefs(mergeDefs[getDefsRaw C, getAsstDefsRaw F], cDef), (cBody, getAsstBody F))
	    end
    in	    
	fun normalizeDefAsstCond findObject condList =
	    let val res = List.map (translateDefPair findObject) condList
		val defs = mergeDefs(List.map #1 res)
		val body = DefAsstSeqList(List.map #2 res)
	    in
		makeAsstDefRaw(defs, body)
	    end

	(* Same but for the array (the first element of the pair must be ParseTree) *)
	fun normalizeDefAsstGuard findObject condList =
	    let val res = List.map (translateDefPair findObject) condList
		val defs = mergeDefs(List.map #1 res)
		val body = DefAsstGuardList(List.map #2 res)
	    in
		makeAsstDefRaw(defs, body)
	    end

	fun normAsstAnd findObject lst =
	    let val defs = mergeDefs(List.map getAsstDefsRaw lst)
	    in makeAsstDefRaw(defs, DefAsstAnd(List.map getAsstBody lst))
	    end

	fun normAsstOr findObject lst =
	    let val defs = mergeDefs(List.map getAsstDefsRaw lst)
	    in makeAsstDefRaw(defs, DefAsstOr(List.map getAsstBody lst))
	    end

    end

    (* ff will map a list of atomic expressions to a valid
       def-structure for the operator.  The list will be the same
       length and in the same order as the provided list of
       def-structures. *)

    fun normalizeDefOp findObject ff lst =
	let val funName = "normalizeDefOp"
	    val _ = pushFunStackLazy(funName, fn()=>(strlistmap2str ", " def2str lst))
	    (* For each argument, compute its definitions and the body
	       (an atomic det. expression) *)
	    fun doOne(F as DefExprAtomic(DefExprBodyAtomic e)) = (getDefsRaw F, e)
	      | doOne F =
		let val n = newName()
		    val v = DefVar{name=n, Type=getDefType findObject F}
		in (addDefs(getDefsRaw F, [(v, [getBody F])]), v)
		end
	    val pairs = List.map doOne lst
	    val args = List.map #2 pairs
	    val F = ff args
	    val defs = mergeDefs((getDefsRaw F)::(List.map #1 pairs))
	    val res = makeDefRaw(defs, getBody F)
	    val _ = popFunStackLazy(funName, fn()=>def2str res)
	in
	    res
	end

    fun normAnd findObject lst =
	let val funName = "normAnd"
	    val _ = pushFunStackLazy(funName, fn()=>(strlistmap2str ", " def2str lst))
	    val eval = evaluateExpr (getOptions()) findObject
	    fun ff lst = DefExprAtomic(DefExprBodyAtomic(eval(foldAnd lst)))
	    val res = normalizeDefOp findObject ff lst
	    val _ = popFunStackLazy(funName, fn()=>def2str res)
	in res
	end

    fun normOr findObject lst =
	let val funName = "normOr"
	    val _ = pushFunStackLazy(funName, fn()=>(strlistmap2str ", " def2str lst))
	    val eval = evaluateExpr (getOptions()) findObject
	    fun ff lst = DefExprAtomic(DefExprBodyAtomic(eval(foldOr lst)))
	    val res = normalizeDefOp findObject ff lst
	    val _ = popFunStackLazy(funName, fn()=>def2str res)
	in res
	end

    fun normImpl findObject(E1, E2) =
	let val funName = "normImpl"
	    val _ = pushFunStackLazy(funName, fn()=>(strlistmap2str ", " def2str [E1, E2]))
	    val eval = evaluateExpr (getOptions()) findObject
	    fun ff[x,y] = DefExprAtomic(DefExprBodyAtomic(eval(impliesFormula(x,y))))
	      | ff lst = raise SympBug
		  ("TransGen.normImpl: wrong number of arguments to implication: "
		   ^(Int.toString(List.length lst))^"\n  args: ["
		   ^(strlistmap2str ", " pt2string lst)^"]")
	    val res = normalizeDefOp findObject ff [E1, E2]
	    val _ = popFunStackLazy(funName, fn()=>def2str res)
	in res
	end

    fun normIff findObject(E1, E2) =
	let val funName = "normIff"
	    val _ = pushFunStackLazy(funName, fn()=>(strlistmap2str ", " def2str [E1, E2]))
	    val eval = evaluateExpr (getOptions()) findObject
	    fun ff[x,y] = DefExprAtomic(DefExprBodyAtomic(eval(iffFormula(x,y))))
	      | ff lst = raise SympBug
		  ("TransGen.normIff: wrong number of arguments to implication: "
		   ^(Int.toString(List.length lst))^"\n  args: ["
		   ^(strlistmap2str ", " pt2string lst)^"]")
	    val res = normalizeDefOp findObject ff [E1, E2]
	    val _ = popFunStackLazy(funName, fn()=>def2str res)
	in res
	end

    fun normNot findObject E =
	let val funName = "normNot"
	    val _ = pushFunStackLazy(funName, fn()=>(def2str E))
	    val eval = evaluateExpr (getOptions()) findObject
	    fun ff[x] = DefExprAtomic(DefExprBodyAtomic(eval(negFormula x)))
	      | ff lst = raise SympBug
		  ("TransGen.normNot: wrong number of arguments to implication: "
		   ^(Int.toString(List.length lst))^"\n  args: ["
		   ^(strlistmap2str ", " pt2string lst)^"]")
	    val res = normalizeDefOp findObject ff [E]
	    val _ = popFunStackLazy(funName, fn()=>def2str res)
	in res
	end 

    (* The heart of the normalization operator is really a bunch of
       specialized functions that normalize def-structures for
       particular operators.  Some of them are mutually dependent, so
       we define most of them here as a highly confusing and
       obfuscating knot of mutually recursive functions.  So, here it
       comes... *)

    (* Extract a tuple element `n' from `E' *)
    fun normalizeDefET findObject (n, E) =
	let val eval = evaluateExpr(getOptions()) findObject
	    fun loopBody(DefExprBodyAtomic e) = DefExprBodyAtomic(eval(ExtractTuple(n, e)))
	      | loopBody(DefSeqList lst) =
		  DefSeqList(List.map(fn(c,e)=>(c, loopBody e)) lst)
	      | loopBody(b as DefTuple lst) = 
		  (List.nth(lst, n)
		   handle _ => raise SympBug
		       ("TransGen/normalizeDefET/loopBody: tuple index is larger than the length:\n  "
			^"n = "^(Int.toString n)
			^"\n  expr: "^(def2str(DefExprAtomic b))))
	      | loopBody b = raise SympBug
		  ("TransGen/normalizeDefET/loopBody: not a tuple:\n  "
		   ^(def2str(DefExprAtomic b)))
	    fun loop(DefExprDecl{defs=defs, body=b}) =
		  DefExprDecl{defs=defs, body= loopBody b}
	      | loop(DefExprAtomic b) = DefExprAtomic(loopBody b)
	in
	    loop E
	end

    (* Extract a record field `f' from `E' *)
    fun normalizeDefER findObject (f, E) =
	let val eval = evaluateExpr(getOptions()) findObject
	    fun loopBody(DefExprBodyAtomic e) = DefExprBodyAtomic(eval(ExtractRecord(f, e)))
	      | loopBody(DefSeqList lst) =
		  DefSeqList(List.map(fn(c,e)=>(c, loopBody e)) lst)
	      | loopBody(b as DefRecord lst) =
		  (case List.find(fn (x,_)=> ptEq(f,x)) lst of
		       SOME(_, b) => b
		     | NONE => raise SympBug
			   ("TransGen/normalizeDefER/loopBody: no field in the record:\n  "
			    ^"field: "^(pt2string f)
			    ^"\n  expr: "^(def2str (DefExprAtomic b))))
	      | loopBody b = raise SympBug
		  ("TransGen/normalizeDefER/loopBody: not a record:\n  "
		   ^(def2str(DefExprAtomic b)))
	    fun loop(DefExprDecl{defs=defs, body=b}) =
		  DefExprDecl{defs=defs, body= loopBody b}
	      | loop(DefExprAtomic b) = DefExprAtomic(loopBody b)
	in
	    loop E
	end

    (* Extract the argument of a datatype constructor `c' from `E' *)
    fun normalizeDefEC findObject (c, E) =
	let val eval = evaluateExpr(getOptions()) findObject
	    fun loopBody(DefExprBodyAtomic e) = DefExprBodyAtomic(eval(ExtractAppl(c, e)))
	      | loopBody(DefSeqList lst) =
		  DefSeqList(List.map(fn(c,e)=>(c, loopBody e)) lst)
	      | loopBody b = raise SympBug
		  ("TransGen/normalizeDefEC/loopBody: not a datatype:\n  "
		   ^(def2str(DefExprAtomic b)))
	    fun loop(DefExprDecl{defs=defs, body=b}) =
		  DefExprDecl{defs=defs, body= loopBody b}
	      | loop(DefExprAtomic b) = DefExprAtomic(loopBody b)
	in
	    loop E
	end

    (* Extract the index of a datatype value from `E' *)
    fun normalizeDefEI findObject E =
	let val eval = evaluateExpr(getOptions()) findObject
	    fun loopBody(DefExprBodyAtomic e) = DefExprBodyAtomic(eval(ExtractIndex e))
	      | loopBody(DefSeqList lst) =
		  DefSeqList(List.map(fn(c,e)=>(c, loopBody e)) lst)
	      | loopBody(DefEnum(b, _)) = b
	      | loopBody b = raise SympBug
		  ("TransGen/normalizeDefEI/loopBody: not a datatype:\n  "
		   ^(def2str(DefExprAtomic b)))
	    fun loop(DefExprDecl{defs=defs, body=b}) =
		  DefExprDecl{defs=defs, body= loopBody b}
	      | loop(DefExprAtomic b) = DefExprAtomic(loopBody b)
	in
	    loop E
	end

    fun normalizeDefTuple lst = makeDefRaw(mergeDefs(List.map getDefsRaw lst),
					   DefTuple(List.map getBody lst))

    fun normalizeDefRecord pairs =
	let val defs = mergeDefs(List.map (fn(f, E)=>getDefsRaw E) pairs)
	    val body = DefRecord(List.map(fn(f, E)=>(f, getBody E)) pairs)
	in makeDefRaw(defs, body)
	end

    (* Given the def-structures for the index and (c, E) pairs for the
       type constructures that require arguments, construct a single
       normalized def-structure. *)

    fun normalizeDefEnum(index, pairs) =
	let val defs = mergeDefs((getDefsRaw index)::(List.map (fn(c, E)=>getDefsRaw E) pairs))
	    val body = DefEnum(getBody index, List.map(fn(f, E)=>(f, getBody E)) pairs)
	in makeDefRaw(defs, body)
	end

    (* Extract the result type from a function type *)
    fun getResType(FunType(_, _, tp)) = SOME tp
      | getResType(ArrayType(_, _, tp)) = SOME tp
      | getResType(TypeClosure{def=tp,...}) = getResType tp
      | getResType(StaticFormalType{value=SOME tp,...}) = getResType tp
      | getResType x = NONE		    

    (* Take an expression which is supposed to be an application of an
       abstractly interpreted function to atomic expressions of
       primitive types, and translate it into a def-structure.

       We assume that the expression consists of if-statements,
       equalities for primitive types and atomic expressions, and
       atomic expressions or nondeterministic operators over them are
       allowed as a result of the entire expression. *)

    fun genAbsDef findObject (expr as IfExpr(_, lst, last)) =
	let val genAbsDef = genAbsDef findObject
	    fun doCond(CondExpr(_, c, e)) = (genAbsDef c, genAbsDef e)
	      | doCond x = raise SympBug
	         ("TransGen/genAbsDef: not a CondExpr in IfExpr:\n  cond: "
		  ^(pt2string x)^"\n  expr: "^(pt2string expr))
	    val lastPair = (DefExprAtomic(DefExprBodyAtomic trueVal), genAbsDef last)
	    val pairs = (List.map doCond lst)@[lastPair]
	in
	    normalizeDefCond findObject pairs
	end
      | genAbsDef fo (expr as NondetExpr(_, lst)) =
	let val Es = List.map (genAbsDef fo) lst
	    val n = newName()
	    val v = DefVar{name=n, Type=getExprType fo expr}
	    val defs = addDefs(mergeDefs(List.map getDefsRaw Es), [(v, List.map getBody Es)])
	in 
	    makeDefRaw(defs, DefExprBodyAtomic v)
	end
      | genAbsDef fo (expr as Appl(_, Builtin{name=Eq _, Type=tp}, arg)) =
	let val normCond = normalizeDefCond fo
	    val restp = (case getResType tp of
			     SOME tp => tp
			   | NONE => raise SympBug
			      ("TransGen/genAbsDef: equality is not of FunType:\n  "
			       ^(pt2string tp)^"\n in expression: "^(pt2string expr)))
	    (* Raw equality will be bool2.  Make sure we get the right type back. *)
	    fun doEq BoolType2 = makeDefRaw(sharedDefsRaw(), DefExprBodyAtomic expr)
	      | doEq (BoolType _) =
		 normCond[(makeDefRaw(sharedDefsRaw(), DefExprBodyAtomic expr), true3Def),
			  (trueDef, false3Def)]
	      | doEq tp = raise SympBug
		 ("TransGen/genAbsDef: equality result is not of BoolType:\n  "
		  ^(pt2string tp)^"\n in expression: "^(pt2string expr))
	in
	    doEq restp
	end
      (* Otherwise the expression must be atomic *)
      | genAbsDef _ e = DefExprAtomic(DefExprBodyAtomic e)

    (* Auxiliary printing function *)
    fun ptlistopt2str(prefix, alt) NONE = alt
      | ptlistopt2str(prefix, alt) (SOME lst) =
	  prefix^"["^(strlistmap2str ", " pt2string lst)^"]"

    (* Application (E G).  Split cases on E. *)
    fun normalizeDefAppl useAbstraction (parms as (findObject, abs, stateVarsOpt)) (E, G) =
	let val funName = "normalizeDefAppl"
	    val _ = pushFunStackLazy(funName, fn()=>(def2str E)^", "^(def2str G)
				     ^(ptlistopt2str(", stateVars = ", "") stateVarsOpt))
	    val options = getOptions()
	    val exprDef = exprDef parms NONE
	    val normEq = normalizeDefEq useAbstraction parms
	    val normCond = normalizeDefCond findObject
	    val normOr = normOr findObject
	    val goodStateVar = goodStateVar(findObject, stateVarsOpt)
	    val typeVals = getTypeValuesAbs(options, findObject, abs)
	    fun isAbstractType(AbstractType _) = true
	      | isAbstractType t = isSome(abstract options abs t)
	    (* Compute a condition for comparing d and G *)
	    fun makeCond G d = normEq(G, makeDefRaw(sharedDefsRaw(), DefExprBodyAtomic d))
	    fun loop(d as DefExprDecl{body=E, ...}, G) =
	        let val F = loop(DefExprAtomic E, G)
		in makeDefRaw(mergeDefs[(getDefsRaw d), (getDefsRaw F)], getBody F)
		end
	      | loop(DefExprAtomic(DefSeqList pairs), G) =
		let fun doPair(c, E) = (DefExprAtomic c, loop(DefExprAtomic E, G))
		in
		    normCond (List.map doPair pairs)
		end
	      | loop(DefExprAtomic(A as DefArray{pairs=pairs, Type=tp}), G) =
		let fun doPair(d, E) = (makeCond G d, DefExprAtomic E)
		    val tp = (case getResType(getDefType findObject(DefExprAtomic A)) of
				  SOME t => t
				| NONE => raise SympBug
				      ("TransGen."^funName^": DefArray is not of FunType:\n  "
				       ^(def2str(DefExprAtomic A))
				       ^"\n  Applied to: "^(def2str G)))
		    val Ebot = exprDef(NondetExpr(dp, [anyvalue tp, bot tp]))
		in
		    normCond ((List.map doPair pairs)@[(trueDef, Ebot)])
		end
	      | loop(DefExprAtomic(DefExprBodyAtomic e), G) =
		let fun def2pt E =
		        let val defs = getDefsRaw E
			in
			    (case getBody E of
				 DefExprBodyAtomic e => (defs, e)
			       | b => let val n = newName()
					  val v = DefVar{name=n, Type=getDefType findObject E}
				      in (addDefs(defs, [(v, [b])]), v)
				      end)
			end
		    fun genAppl(E, G) =
			let val ((Edefs, e), (Gdefs, g)) = (def2pt E, def2pt G)
			in 
			    makeDefRaw(mergeDefs[Edefs, Gdefs], DefExprBodyAtomic(Appl(dp, e, g)))
			end
		    fun genPartialAppl(tp, E, G) =
			let val isBad = not o (goodStateVar e)
			    val ds = (case typeVals tp of
					  SOME ds => ds
					| NONE => raise TransError
					      ("Even though abstracted, the type is still "
					       ^"too big to list its values:\n  "
					       ^(pt2string tp)))
			    val badArgs = List.filter isBad ds
			    val Cbot = normOr(List.map(makeCond G) badArgs)
			    val n = newName()
			    val v = DefVar{name=n, Type=tp}
			    val Ebot = exprDef(NondetExpr(dp, [anyvalue tp, bot tp]))
			    val E = normCond[(Cbot, Ebot), (trueDef, genAppl(E, G))]
			in loop(makeDefRaw(addDefs(mergeDefs[getDefsRaw Ebot, getDefsRaw E],
						   [(v, [getBody Ebot])]),
					   getBody E), G)
			end
		in
		    if isStateVar e then
			let val tp = (getExprType findObject e)
			in
			    if isAbstractType tp then genPartialAppl(tp, E, G)
			    else genAppl(E, G)
			end
		    else genAppl(E, G)
		end					 
	      | loop(d as DefExprAtomic _, G) = raise SympBug
		 ("TransGen.normalizeDefAppl: not a function in application (E G):\n  "
		  ^"E: "^(def2str d)^"\n  G: "^(def2str G))
	    val res = loop(E,G)
	    val _ = popFunStackLazy(funName, fn()=>def2str res)
	in res
	end

    (* Convert a def-structure of any boolean type into bool2 *)
    and BoolToBool2 useAbstraction (parms as (findObject, abs, stateVarsOpt)) F =
	let val normEq = normalizeDefEq useAbstraction parms
	    val tp = getDefType findObject F
	in
	    case tp of
		BoolType _ => normEq(F, true3Def)
	      | BoolType2 => F
	      | _ => raise SympBug
		 ("TransGen.BoolToBool2: not a boolean type expression: "^(pt2string tp)
		  ^"\n  expr: "^(def2str F))
	end

    (* Convert a def-structure of any boolean type into bool2 *)
    and BoolToBool3 useAbstraction (parms as (findObject, abs, stateVarsOpt)) F =
	let val normEq = normalizeDefEq useAbstraction parms
	    val normCond = normalizeDefCond findObject
	    val tp = getDefType findObject F
	    fun toBool3 F = normCond[(F, true3Def), (trueDef, false3Def)]
	in
	    case tp of
		BoolType2 => toBool3 F
	      | BoolType _ => F
	      | _ => raise SympBug
		 ("TransGen.BoolToBool3: not a boolean type expression: "^(pt2string tp)
		  ^"\n  expr: "^(def2str F))
	end

    (* 2-valued equality, F1 = F2.  We split expressions by types.

       The flag `useAbstraction' indicates whether we should use
       abstract interpretation of equality when applicable or not. *)

    and normalizeDefEq useAbstraction (parms as (findObject, abs, stateVarsOpt))
			              (F1: DefExpr, F2: DefExpr) =
	let val funName = "normalizeDefEq"
	    val _ = pushFunStackLazy(funName, fn()=>(def2str F1)^", "^(def2str F2)
				     ^(ptlistopt2str(", stateVars = ", "") stateVarsOpt))
	    val normEq = normalizeDefEq useAbstraction parms
	    val normApp = normalizeDefAppl useAbstraction parms
	    val normET = normalizeDefET findObject
	    val normER = normalizeDefER findObject
	    val normEC = normalizeDefEC findObject
	    val normEI = normalizeDefEI findObject
	    val normAnd = normAnd findObject
	    val normOp = normalizeDefOp findObject
	    val options = getOptions()
	    val eval = evaluateExpr options findObject
	    val genAbsDef = genAbsDef findObject
	    val typeVals = getTypeValuesAbs(options, findObject, abs)
	    val BoolToBool2 = BoolToBool2 useAbstraction parms 
	    fun normEqPrim(F1, F2) =
		let fun ff[e1, e2] =
		          DefExprAtomic(DefExprBodyAtomic(eqFormula findObject (e1, e2)))
		      | ff _ = raise SympBug
			  ("TransGen/normalizeDefEq/normEqPrim: this can't happen")
		in normOp ff [F1, F2]
		end
	    fun normAbsEq(typeAbs, F1, F2) =
		if useAbstraction then
		    (case getAbsInt(getOptions()) typeAbs (Eq dp) of
			 SOME abs =>
			     let fun ff args = genAbsDef(eval(Appl(dp, abs, TupleExpr(dp, args))))
				 val F = normOp ff [F1, F2]
			     in
				 BoolToBool2 F
			     end
		       | NONE => normEqPrim(F1, F2))
		else normEqPrim(F1, F2)
	    fun doFun t1 =
	        let val ds =
		        (case typeVals t1 of
			     SOME ds => ds
			   | NONE => raise TransError
			    ("Array index type is too large (or infinite) to list all its values: "
			     ^(pt2string t1)^"\n  in equality translation: "
			     ^(def2str F1)^" = "^(def2str F2)))
		    fun doOne d =
			let val D = makeDefRaw(sharedDefsRaw(), DefExprBodyAtomic d)
			in normEq(normApp(F1, D), normApp(F2, D))
			end
		in normAnd(List.map doOne ds)
		end
	    (* Split by the type *)
	    fun split(TupleType(_, lst)) =
		let fun loop(n, i) = 
		        if i >= n then []
			else (normEq(normET(i, F1), normET(i, F2)))::(loop(n, i+1))
		in normAnd(loop(List.length lst, 0))
		end
	      | split(RecordType(_, lst)) =
		let fun doOne(RecordField{name=f, ...}) = normEq(normER(f, F1), normER(f, F2))
		      | doOne x = raise SympBug
		         ("TransGen/normalizeDefEq/split: RecordType's field is not RecordField:\n  "
			  ^(pt2string x))
		in normAnd(List.map doOne lst)
		end
	      | split(EnumType(_, lst)) =
		let fun doOne(c as TypeConstr{Type=FunType _,...}) =
		          normEq(normEC(c, F1), normEC(c, F2))
		      | doOne(c as TypeConstr _) = DefExprAtomic(DefExprBodyAtomic trueVal)
		      | doOne x = raise SympBug
		         ("TransGen/normalizeDefEq/split: EnumType's field is not TypeConstr:\n  "
			  ^(pt2string x))
		    val C = normEq(normEI F1, normEI F2)
		in normAnd(C::(List.map doOne lst))
		end
	      | split(FunType(_, t1, t2)) = doFun t1
	      | split(ArrayType(_, t1, t2)) = doFun t1
	      | split(StaticFormalType{value=SOME t,...}) = split t
	      | split(t as StaticFormalType{value=NONE,...}) =
		(case abstract(getOptions()) abs t of
		     SOME typeAbs => normAbsEq(typeAbs, F1, F2)
		   | NONE => raise SympBug
			 ("TransGen/normalizeDefEq/split: cannot translate "
			  ^"an uninterpreted type:\n  "^(pt2string t)
			  ^"\nAbstraction:\n"^(abs2str abs)))
	      | split(tp as TypeClosure{recursive=false, def=t,...}) = split t
	      | split(tp as TypeClosure{recursive=true, def=t,...}) =
		(case abstract(getOptions()) abs t of
		     SOME typeAbs => normAbsEq(typeAbs, F1, F2)
		   | NONE => raise SympBug
			 ("TransGen/normalizeDefEq/split: cannot translate"
			  ^" an (infinite) recursive type:\n  "
			  ^(pt2string tp)^" = "^(pt2string t)
			  ^"\nAbstraction:\n"^(abs2str abs)))
	      | split(t as Uid _) =
		(case findObject t of
		     SOME tp => split tp
		   | NONE => raise SympBug
		      ("TransGen/normalizeDefEq/split:  Uid type is not in object hash: "
		       ^(pt2string t)))
	      (* The type must be primitive *)
	      | split tp =
		(case abstract(getOptions()) abs tp of
		     SOME typeAbs => normAbsEq(typeAbs, F1, F2)
		   | NONE => normEqPrim(F1, F2))
	    val res = split(getDefType findObject F1)
	    val _ = popFunStackLazy(funName, fn()=>def2str res)
	in res
	end

    and normalizeDefNotEq useAbstraction (parms as (findObject, abs, _)) (F1, F2) =
	let val C = normalizeDefEq useAbstraction parms (F1, F2)
	in normalizeDefCond findObject [(C, falseDef), (trueDef, trueDef)]
	end

    (* 3-valued equality *)
    and normalizeDefEq3(parms as (findObject, abs, _)) (F1, F2) =
	let val C = normalizeDefEq true parms (F1, F2)
	in BoolToBool3 true parms C
	end

    and normalizeDefNotEq3(parms as (findObject, abs, _)) (F1, F2) =
	let val C = normalizeDefEq true parms (F1, F2)
	in BoolToBool3 true parms C
	end

    (* Translate pattern matching.
       Sequent: matches(e, P) ~> F *)
    and matches(parms as (findObject, abs, _)) (expr, pat) = 
	let val options = getOptions()
	    val debug = lazyVerbDebug options
	    val funName = "TransGen.matches"
	    val _ = pushFunStackLazy(funName, (fn()=>(pt2string expr)^", "^(pt2string pat)))
	    (* val builtinFun = builtinFun options findObject allVars abs pvars lim *)
	    val normET = normalizeDefET findObject
	    val normER = normalizeDefER findObject
	    val normEC = normalizeDefEC findObject
	    val normEI = normalizeDefEI findObject
	    val normAnd = normAnd findObject
	    val normEq = normalizeDefEq true parms
	    val evalExpr = evaluateExpr options findObject
	    val matches = matches parms
	    val exprDef = exprDef parms NONE
	    (* Make an atomic def-structure out of an expression *)
	    fun makeAtomDef e = DefExprAtomic(DefExprBodyAtomic e)
	    fun matches' (_, Underscore _) = makeAtomDef trueVal
	      | matches' (_, PatternFormal _) = makeAtomDef trueVal
	      | matches' (e, AsPattern(_,_,pat)) = matches(e, pat)
	      | matches' (e, TypedPattern(_,pat,_)) = matches(e, pat)
	      | matches' (e, TuplePattern(_,lst)) = 
		 let val es = List.map(fn n=>evalExpr (ExtractTuple(n,e)))
		                      (List.tabulate(List.length lst, fn i=>i))
		 in normAnd(List.map matches (zip(es,lst)))
		 end
	      | matches' (e, RecordPattern(_,lst)) =
		 let fun doField (RecordAsst(_, f, pat)) = 
		          matches(evalExpr (ExtractRecord(f,e)), pat)
		       | doField x = raise SympBug
			  ("generateTrans/matches: not a RecordAsst: "
			   ^(pt2string x))
		 in normAnd(List.map doField lst)
		 end
	      | matches' (e, ApplPattern(_, c as TypeConstr _, pat)) =
		 (let val eI = evalExpr(ExtractIndex e)
		     val ec = evalExpr (ExtractAppl(c,e))
		     val C = normEq(exprDef eI, exprDef c)
		 in normAnd[C, matches(ec, pat)]
		 end handle EvalError _ => falseDef)
	      (* `d' must be a contstant or TypeConstr *)
	      | matches' (e, d) = normEq(exprDef e, exprDef d)
	    val _ = debug "TransGen.matchesDebug" 
		     (fn()=>"\nexpr = "^(pt2stringDebug expr)^"\npat = "^(pt2stringDebug pat)^"\n")
	    (* First check whether it can be computed statically *)
	    val seluname = Uid(newName())
	    val ee = evalExpr expr
	    val matchexpr = 
		if isConst ee andalso isConst pat then
		    eqFormula findObject (ee,pat)
		else CaseExpr(dp,expr,
			      [ChoiceClosure{pattern=pat,
					     uname=seluname,
					     names=[],
					     body=trueVal,
					     parent=Fake},
			       ChoiceClosure{pattern=Underscore dp,
					     uname=seluname,
					     names=[],
					     body=falseVal,
					     parent=Fake}])
	    val matchRes = evalExpr matchexpr
	    val F = if isConst matchRes then makeAtomDef matchRes else matches'(expr, pat)
	    val _ = popFunStackLazy(funName,(fn()=> def2str F))
	in F
	end

    (* Translate application of built-in operators into a def-structure *)
    and builtinDef(parms as (findObject, abs, stateVarsOpt)) (f, e) =
	let val funName = "builtinDef"
	    val _ = pushFunStackLazy(funName, fn()=>(pt2string f)^", "^(pt2string e)
				     ^(ptlistopt2str(", stateVars = ", "") stateVarsOpt))
	    val options = getOptions()
	    val exprDef = exprDef parms NONE
	    val normCond = normalizeDefCond findObject
	    val normET = normalizeDefET findObject
	    val normAnd = normAnd findObject
	    val normOr = normOr findObject
	    val normImpl = normImpl findObject
	    val normNot = normNot findObject
	    val normIff = normIff findObject
	    val normEq2 = normalizeDefEq true parms
	    val normNotEq2 = normalizeDefNotEq true parms
	    val normEq3 = normalizeDefEq3 parms
	    val BoolToBool2 = BoolToBool2 true parms
	    val normNotEq3 = normalizeDefNotEq3 parms
	    val normApp = normalizeDefAppl true parms
	    val evalExpr = evaluateExpr options findObject
	    val exprType = getExprType findObject
	    val getDefType = getDefType findObject
	    val abstract = abstract options abs
	    val tryAbs = tryAbstract options abs
	    val genAbsDef = genAbsDef findObject
	    val E = exprDef e

	    (* Boolean functions: if any of the arguments is of
	       2-valued boolean type, convert the entire expression
	       into 2-valued bool.  This normally happens when an
	       argument has temporal operators inside. *)

	    fun doImplies() =
		let val E1 = normET(0, E)
		    val E2 = normET(1, E)
		    val (tp1, tp2) = (getDefType E1, getDefType E2)
		    val (C1t, C1f, tp1is3) =
			  (case tp1 of
			       BoolType2 => (E1, normNot E1, false)
			     | BoolType _ => (normEq2(E1, true3Def), normEq2(E1, false3Def),
					      true)
			     | _ => raise SympBug
				(funName^"/doImplies: type tp1 is not boolean: "
				 ^(pt2stringDebug tp1)))
		    val (C2t, C2f, tp2is3) =
			  (case tp2 of
			       (* Negation will not be used in bool2 case *)
			       BoolType2 => (E2, E2, false)
			     | BoolType _ => (normEq2(E2, true3Def), normEq2(E2, false3Def), true)
			     | _ => raise SympBug
				(funName^"/doImplies: type tp2 is not boolean: "
				 ^(pt2stringDebug tp2)))
		in 
		    if tp1is3 andalso tp2is3 then 
			normCond[(normAnd[C1t, C2f], false3Def),
				 (C1f, true3Def),
				 (C2t, true3Def),
				 (trueDef, botBoolDef)]
		    (* Conservative approximation: return false if an important arg. is bottom *)
		    else normOr[C1f, C2t]
		end

	    (* The operators that preserve `undefined' value and not
	       require any special translation and are handled
	       uniformly.  The list `args' is a list of
	       def-structures. *)

	    fun doOther(Op, tp, E) =
		let val resTp = (case getResType tp of
				     SOME t => t
				   | NONE => raise SympBug
		         ("TransGen.builtinDef/doOther/getResType: Op has wrong type:\n  "
			  ^(pt2string tp)^"\n  Op: "^(pt2string Op)))
		    val botDef = DefExprAtomic(DefExprBodyAtomic(bot resTp))
		    val fDef = DefExprAtomic(DefExprBodyAtomic Op)
		    (* Create a condition checking the argument for
		       being `undefined'.  Also see if there is an
		       abstract interpretation of the `Op' on that
		       argument. *)
		    fun makeOneCond E =
			let val tp = getDefType E
			    val atp = tryAbs tp
			    val botArg = DefExprAtomic(DefExprBodyAtomic(bot atp))
			    (* Abstract interpretation of the operator *)
			    val absOp = Option.mapPartial(fn at=>getAbsInt options at Op)
				                         (abstract tp)
			in ((normEq2(E, botArg), botDef), absOp)
			end
		    fun makeCond E n = makeOneCond(normET(n, E))
		    (* Finish application taking into account abstract interpretation *)
		    fun applyAbs(NONE, arity, E) = normApp(fDef, E)
		      | applyAbs(SOME absOp, arity, E) =
			let (* Translate an argument def-structure
			       into an atomic argument for the `absOp' *)
			    fun makeOneArg(DefExprAtomic(DefExprBodyAtomic e)) =
				  (sharedDefsRaw(), e)
			      | makeOneArg E =
				let val v = DefVar{name=newName(), Type=getDefType E}
				in (addDefs(getDefsRaw E, [(v, [getBody E])]), v)
				end
			    (* Take the arity and prepare the appropriate ParseTree arg *)
			    fun makeArg 1 =
				let val (defs, v) = makeOneArg E
				in ([defs], v)
				end
			      | makeArg n =
				let val args = List.map(fn i=>makeOneArg(normET(i, E)))
				                       (List.tabulate(n, fn x=>x))
				    val defs = List.map #1 args
				in
				    (defs, TupleExpr(dp, List.map #2 args))
				end
			    val (defs, arg) = makeArg arity
			    val F = genAbsDef(evalExpr(Appl(dp, absOp, arg)))
			in
			    makeDefRaw(mergeDefs((getDefsRaw F)::defs), getBody F)
			end

		    (* The expression's def-structure `E' must be an n-tuple. *)
		    fun makeDefTuple(n, E) =
			let val pairs = List.map(makeCond E)(List.tabulate(n, fn x=>x))
			    val conds = List.map #1 pairs
			    val absOps = List.mapPartial #2 pairs


			    val absOp = (case absOps of
					     [] => NONE
					   | aOp::_ => SOME aOp)
			    val applDef = applyAbs(absOp, n, E)
			in normCond(conds@[(trueDef, applDef)])
			end
		in
		    (case arityOf Op of
			 0 => raise SympBug
			     ("TransGen.builtinDef: Builtin constant (0-arity) "
			      ^"applied to an argument:\n  "^(pt2string(Appl(dp, f, e))))
		       | 1 => let val (cond, absOp) = makeOneCond E
				  val res = (trueDef, applyAbs(absOp, 1, E))
			      in normCond([cond, res])
			      end
		       | n => makeDefTuple(n, E))
		end

	    (* Idea: translate all the arguments into (arg = true2). *)
	    fun doTemporal Op =
		let val n = arityOf Op
		    fun doOne i = BoolToBool2(normET(i, E))
		    val arg = (case n of
				    1 => BoolToBool2 E
				  | _ => normalizeDefTuple
					  (List.map doOne (List.tabulate(n, fn x=>x))))
		    val opDef = DefExprAtomic(DefExprBodyAtomic(mapTemporal Op))
		in normApp(opDef, arg)
		end

	    (* The main case split *)
	    fun loop(f as Builtin{name=Eq _, Type=tp}) =
		let val resTp = exprType(Appl(dp, f, TupleExpr(dp, [True2, True2])))
		in
		    case resTp of
			BoolType2 => normEq2(normET(0, E), normET(1, E))
		      | BoolType _ => normEq3(normET(0, E), normET(1, E))
		      | _ => raise SympBug
			 ("TransGen.builtinDef/loop: Eq has a non-boolean result type:  "
			  ^(pt2string resTp))
		end
	      | loop(f as Builtin{name=NotEq _, ...}) =
		let val resTp = exprType(Appl(dp, f, TupleExpr(dp, [True2, True2])))
		in
		    case resTp of
			BoolType2 => normNotEq2(normET(0, E), normET(1, E))
		      | BoolType _ =>  normNotEq3(normET(0, E), normET(1, E))
		      | _ => raise SympBug
			 ("TransGen.builtinDef/loop: NotEq has a non-boolean result type:  "
			  ^(pt2string resTp))
		end
	      | loop(f as Builtin{name=And _, ...}) =
		let val E1 = normET(0, E)
		    val E2 = normET(1, E)
		    val (tp1, tp2) = (getDefType E1, getDefType E2)
		    val (C1t, C1f, tp1is3) =
			  (case tp1 of
			       (* Negation will not be used in bool2 case *)
			       BoolType2 => (E1, E1, false)
			     | BoolType _ => (normEq2(E1, true3Def), normEq2(E1, false3Def), true)
			     | _ => raise SympBug
				(funName^"/And: type tp1 is not boolean: "
				 ^(pt2stringDebug tp1)))
		    val (C2t, C2f, tp2is3) =
			  (case tp2 of
			       (* Negation will not be used in bool2 case *)
			       BoolType2 => (E2, E2, false)
			     | BoolType _ => (normEq2(E2, true3Def), normEq2(E2, false3Def), true)
			     | _ => raise SympBug
				(funName^"/And: type tp2 is not boolean: "
				 ^(pt2stringDebug tp2)))
		in 		    
		    if tp1is3 andalso tp2is3 then 
			 normCond[(normAnd[C1t, C2t], true3Def),
				  (C1f, false3Def),
				  (C2f, false3Def),
				  (trueDef, botBoolDef)]
		    else normAnd[C1t, C2t]
		end
	      | loop(f as Builtin{name=Or _, ...}) =
		let val E1 = exprDef(evalExpr(ExtractTuple(0, e)))
		    val E2 = exprDef(evalExpr(ExtractTuple(1, e)))
		    val (tp1, tp2) = (getDefType E1, getDefType E2)
		    val (C1t, C1f, tp1is3) =
			  (case tp1 of
			       (* Negation will not be used in bool2 case *)
			       BoolType2 => (E1, E1, false)
			     | BoolType _ => (normEq2(E1, true3Def), normEq2(E1, false3Def), true)
			     | _ => raise SympBug
				(funName^"/Or: type tp1 is not boolean: "
				 ^(pt2stringDebug tp1)))
		    val (C2t, C2f, tp2is3) =
			  (case tp2 of
			       (* Negation will not be used in bool2 case *)
			       BoolType2 => (E2, E2, false)
			     | BoolType _ => (normEq2(E2, true3Def), normEq2(E2, false3Def), true)
			     | _ => raise SympBug
				(funName^"/Or: type tp2 is not boolean: "
				 ^(pt2stringDebug tp2)))
		in 
		    if tp1is3 andalso tp2is3 then 
			 normCond[(normAnd[C1f, C2f], false3Def),
				  (C1t, true3Def),
				  (C2t, true3Def),
				  (trueDef, botBoolDef)]
		    else normOr[C1t, C2t]
		end
	      | loop(f as Builtin{name=Implies _, ...}) = doImplies()
	      | loop(f as Builtin{name=Arrow _, ...}) = doImplies()
	      | loop(f as Builtin{name=Darrow _, ...}) = doImplies()
	      | loop(f as Builtin{name=Not _, ...}) =
		let val E = exprDef e
		    val tp = getDefType E
		    val (Ct, Cf, is3) =
			  (case tp of
			       BoolType2 => (E, normNot E, false)
			     | BoolType _ => (normEq2(E, true3Def), normEq2(E, false3Def), true)
			     | _ => raise SympBug
				(funName^"/Not: type tp is not boolean: "^(pt2stringDebug tp)))
		in 
		    if is3 then normCond[(Ct, false3Def),
					 (Cf, true3Def),
					 (trueDef, botBoolDef)]
		    else Cf
		end
	      | loop(f as Builtin{name=Iff _, ...}) =
		let val E1 = exprDef(evalExpr(ExtractTuple(0, e)))
		    val E2 = exprDef(evalExpr(ExtractTuple(1, e)))
		    val (tp1, tp2) = (getDefType E1, getDefType E2)
		    fun getFor3() = (normEq2(E1, botBoolDef),
				     normEq2(E2, botBoolDef))
		    (* FIXME: kludge warning: `iff' is not monotonic, and we are approximating
		       3-valued args to (x=true). *)
		    fun getFor2() = ((case tp1 of
					  BoolType2 => E1
					| BoolType _ => normEq2(E1, true3Def)
					| _ => raise SympBug
					      (funName^"/Iff: type tp1 is not boolean: "
					       ^(pt2stringDebug tp1))),
				     (case tp2 of
					  BoolType2 => E2
					| BoolType _ => normEq2(E2, true3Def)
					| _ => raise SympBug
					      (funName^"/Iff: type tp2 is not boolean: "
					       ^(pt2stringDebug tp2))))
		in 
		    case (tp1, tp2) of
			(BoolType _, BoolType _) => 
			    let val (C1bot, C2bot) = getFor3()
				in normCond[(C1bot, botBoolDef),
					    (C2bot, botBoolDef),
					    (normEq2(E1, E2), true3Def),
					    (trueDef, false3Def)]
			    end
		      | _ => normIff(getFor2())
		end
	      (* The other operators preserve `undefined' value, and are handled uniformly *)
	      | loop(f as Builtin{name=Op, Type=tp}) =
		if isTemporal Op then doTemporal f
		else doOther(Op, tp, E)
	      | loop x = raise SympBug
		 ("TransGen.builtinDef: not a built-in function:\n  "^(pt2string x))
	    val res = loop f
	    val _ = popFunStackLazy(funName, fn()=>def2str res)
	in res
	end

    and quantifierDef(parms as (findObject, abs, _)) e = 
	let val funName = "quantifierDef"
	    val _ = pushFunStackLazy(funName, fn()=>pt2string e)
	    val options = getOptions()
	    val lim = get_limit options
	    val normCond = normalizeDefCond findObject
	    val normAnd = normAnd findObject
	    val normOr = normOr findObject
	    val exprDef = exprDef parms NONE
	    val normEq = normalizeDefEq true parms
	    val substVarValues = substVarValues(options, findObject, lim, abs)
	    (* Compute the lists C^t and C^f for `e' with every evaluation of variables `names' *)
	    fun makeConds(names, eds) = 
		let val Ct = List.map(fn e=>normEq(e, true3Def)) eds
		    val Cf = List.map(fn e=>normEq(e, false3Def)) eds
		in (Ct, Cf)
		end
	    (* Quantifiers are treated as expantions of `and' and
	       `or', with the same 3-valued semantics. *)

	    (* forall X: e => (\/ Cf, false), (/\ Ct, true), (true, bot) *)
	    fun loop(ForallClosure{names=names, body=e, ...}) =
		let val eds = List.map exprDef(case substVarValues(names, e) of
						   [] => [true3Val]
						 | lst => lst)
		    val tp = getDefType findObject (hd eds)
		    fun getFor3 () =
			let val (Ct, Cf) = makeConds(names, eds)
			    val falseCond = (normOr Cf, false3Def)
			    val trueCond = (normAnd Ct, true3Def)
			in (falseCond, trueCond)
			end
		in
		    case tp of
			BoolType2 => normAnd eds
		      | BoolType _ =>
			    let val (falseCond, trueCond) = getFor3()
			    in normCond[falseCond, trueCond, (trueDef, botBoolDef)]
			    end
		      | _ => raise SympBug
			    (funName^"/Forall: type tp is not boolean: "^(pt2stringDebug tp))
		end
	      (* exists X: e => (/\ Cf, false), (\/ Ct, true), (true, bot) *)
	      | loop(ExistsClosure{names=names, body=e, ...}) =
		let val eds = List.map exprDef (case substVarValues(names, e) of
						    [] => [true3Val]
						  | lst => lst)
		    val tp = getDefType findObject (hd eds)
		    fun getFor3 () =
			let val (Ct, Cf) = makeConds(names, eds)
			    val falseCond = (normAnd Cf, false3Def)
			    val trueCond = (normOr Ct, true3Def)
			in (falseCond, trueCond)
			end
		in
		    case tp of
			BoolType2 => normOr eds
		      | BoolType _ =>
			    let val (falseCond, trueCond) = getFor3()
			    in normCond[falseCond, trueCond, (trueDef, botBoolDef)]
			    end
		      | _ => raise SympBug
			    (funName^"/Forall: type tp is not boolean: "^(pt2stringDebug tp))
		end
	      | loop x = raise SympBug
		  ("TransGen.quantifierDef: not a quantifier:\n  "^(pt2string x))
	    val res = loop e
	    val _ = popFunStackLazy(funName, fn()=>def2str res)
	in
	    res
	end
		    

    (* Translation of an arbitrary expression into a def-structure.
       Sequent Omega; e ~> F *)
    and exprDef(parms as (findObject, abs, stateVarsOpt)) omegaOpt e =
	let val funName = "exprDef"
	    fun omega2str NONE = "NONE"
	      | omega2str (SOME lst) =
		"[["^(strlistmap2str "], [" (strlistmap2str ", " pt2string) lst)^"]]"
	    val _ = pushFunStackLazy(funName, fn()=>(omega2str omegaOpt)
				     ^(ptlistopt2str(", stateVars = ", "") stateVarsOpt)
				     ^", "^(pt2string e))
	    val exprDef = exprDef parms
	    val matches = matches parms
	    val options = getOptions()
	    val evalExpr = evaluateExpr options findObject
	    val normApp = normalizeDefAppl true parms
	    val normCond = normalizeDefCond findObject
	    val normArray = normalizeDefArray findObject
	    val normEq = normalizeDefEq true parms
	    fun normEqExpr(e1, e2) = normEq(DefExprAtomic(DefExprBodyAtomic e1),
					    DefExprAtomic(DefExprBodyAtomic e2))
	    val builtinDef = builtinDef parms
	    val quantifierDef = quantifierDef parms
	    val goodStateVar = goodStateVar(findObject, stateVarsOpt)
	    val lim = get_limit options
	    val typeVals = getTypeValuesAbs(options, findObject, abs)
	    val tryAbsMap = tryAbsMap(options, findObject, abs)
	    val exprType = getExprType findObject
	    val getDefType = getDefType findObject
	    fun isAbstractType(AbstractType _) = true
	      | isAbstractType t = isSome(abstract options abs t)
	    (* Compute the type of the original expression, for everyone else to use *)
	    val origTp = exprType e
            (* Trim omega to { Sigma | (Sigma, E) \in omega } *)
	    fun trimOmega(E, omegaOpt) =
		let fun eqE(ExtractTuple(i, _), ExtractTuple(j, _)) = (i = j)
		      | eqE(ExtractRecord(f1, _), ExtractRecord(f2, _)) = ptEq(f1, f2)
		      | eqE(ExtractIndex _, ExtractIndex _) = true
		      | eqE(ExtractAppl(c1, _), ExtractAppl(c2, _)) = ptEq(c1, c2)
		      (* Trim any application *)
		      | eqE(Appl _, Appl(_, _, Fake)) = true
		      (* Trim application of a particular constant *) 
		      | eqE(Appl(_, _, d1), Appl(_, _, d2)) = ptEq(d1, d2)
		      | eqE _ = false		         
		    fun doOne(E'::S) = if eqE(E', E) then SOME S else NONE
		      | doOne _ = NONE
		in Option.map(List.mapPartial doOne) omegaOpt
		end
	    (* Translate the functional type `t', with the arg type `t1'.
	       Take special care of state vars indexed by values of an abstracted type. *)
	    fun doFun(t, t1) =
		let val funName = funName^"/doFun"
		    val _ = pushFunStackLazy(funName, fn()=>pt2string t1)
		    fun filterIndices ds =
		        if isStateVar e andalso isAbstractType t1 then
			    List.filter (goodStateVar e) ds
			else ds
		    fun computeDs NONE =
		        (case typeVals t1 of
			     SOME ds => filterIndices ds
			   | NONE => raise TransError
			    ("Array index type is too large (or infinite) to list all its values: "
			     ^(pt2string t1)^"\n  in expression translation: "
			     ^(pt2string e)))
		      | computeDs(SOME omega) =
			let fun doOne((Appl(_, _, d))::_) = SOME d
			      | doOne _ = NONE
			in List.mapPartial doOne omega
			end
		    val ds = computeDs omegaOpt
		    val om = trimOmega(Appl(dp, Fake, Fake), omegaOpt)
		    fun doOne d = (DefExprAtomic(DefExprBodyAtomic d),
				   exprDef om (evalExpr(Appl(dp, e, d))))
		    val res = normArray(List.map doOne ds, t)
		    val _ = pushFunStackLazy(funName, fn()=>def2str res)
		in res
		end

	    (* Print either the first element of the ParseTree list,
	       or a message that it's empty *)
	    fun printFirstPt(pt::_) = pt2string pt
	      | printFirstPt [] = "<empty>"

	    (* Third stage: wrapping the extractors back and translating the application.
	       Sequent Sigma; D; e ~>_wrap F. *)
	    fun exprWrap((ExtractTuple(i, _))::S, D, e) = exprWrap(S, D, ExtractTuple(i, e))
	      | exprWrap((ExtractRecord(f, _))::S, D, e) = exprWrap(S, D, ExtractRecord(f, e))
	      | exprWrap((ExtractAppl(c, _))::S, D, e) = exprWrap(S, D, ExtractAppl(c, e))
	      | exprWrap((ExtractIndex _)::S, D, e) = exprWrap(S, D, ExtractIndex e)
	      | exprWrap((Appl(p, _, e'))::S, D, e) =
		let (* The result for the ``other'' elements *)
		    fun makeOther() = exprDef NONE (NondetExpr(dp, [anyvalue origTp, bot origTp]))
		    (* Handle the case when e' is not a constant *)
		    fun doNonConst() =
		        let fun getAllVals() =
		                let val tp = exprType e'
				in
				    case typeVals tp of
					SOME ds => ds
				      | NONE => raise TransError
					    ("Too large or infinite index type of the array:\n  "
					     ^(pt2string tp)^"\n  expr: "
					     ^(pt2string (Appl(p, e, e'))))
				end
			    val F' = exprDef omegaOpt e'
			    fun makePair d = (normEq(F', makeDefRaw(sharedDefsRaw(),
								    DefExprBodyAtomic d)),
					      exprWrap(S, D, Appl(p, e, d)))
			    val ds = if isStateVar e then
				      (case stateVarsOpt of
					   SOME V =>
					   (* Wrap vars into something to make `split...' happy *)
					     let val vars = List.map (vtWrap NormalFlag) V
					     in List.map #1 (splitByArrayIndex(vars, e))
					     end
					 | NONE => List.filter (goodStateVar e) (getAllVals()))
				     else getAllVals()
			    val pairs = List.map makePair ds
			in normCond(pairs@[(trueDef, makeOther())])
			end
		    (* Handle the case when e' is a constant *)
		    fun doConst d =
			if isStateVar e then
			    if goodStateVar e d then exprWrap(S, D, Appl(p, e, d))
			    else makeOther()
			else exprWrap(S, D, Appl(p, e, d))
		in
		    if isConst e' then doConst e' else doNonConst()
		end
	      | exprWrap(x::_, _, e) = raise SympBug
		 ("TransGen.exprDef/exprWrap: bad extractor: "^(pt2string x)
		  ^"\n  for expression: "^(pt2string e))

		 (* Wrap up the final expression.  Check if it is
		    abstracted, and if it is, get its abstract value. *)
	      | exprWrap([], D, e) = makeDefRaw(D, DefExprBodyAtomic(tryAbsMap e))

	    (* Second stage: translate the expression structure, rip the extractors.
	       Sequent Sigma; e ~> F. *)

	    fun exprSplit(S, e) =
		let val funName = funName^"/exprSplit"
		    val _ = pushFunStackLazy(funName, fn()=>"["
					     ^(strlistmap2str ", " pt2string S)
					     ^"], "^(pt2string e))
		    val res = exprSplit'(S, e)
		    val _ = popFunStackLazy(funName, fn()=>def2str res)
		in res
		end
	    (* First, rip the extractors *)
	    and exprSplit'(S, ExtractTuple(i, e)) = exprSplit((ExtractTuple(i, Fake))::S, e)
	      | exprSplit'(S, ExtractRecord(f, e)) = exprSplit((ExtractRecord(f, Fake))::S, e)
	      | exprSplit'(S, ExtractAppl(c, e)) = exprSplit((ExtractAppl(c, Fake))::S, e)
	      | exprSplit'(S, ExtractIndex e) = exprSplit((ExtractIndex Fake)::S, e)
	      (* Postpone the application till after dealing with datatype extractors *)
	      (* Now, try to annihilate extractors when possible *)
	      | exprSplit'((ExtractTuple(i, _))::S, e as TupleExpr(_, lst)) =
		  (exprSplit(S, List.nth(lst, i)) handle subscript => raise SympBug
		      ("TransGen.exprDef/exprSplit: TupleExpr is shorter than the "
		       ^"index in ExtractTuple:\n  "
		       ^(pt2string(ExtractTuple(i, e)))))
	      | exprSplit'(S, e as TupleExpr _) = raise SympBug
		 ("TransGen.exprDef/exprSplit: wrong extractor for TupleExpr:\n  extractor: "
		  ^(printFirstPt S)^"\n  expression: "^(pt2string e))
	      | exprSplit'((ExtractRecord(f, _))::S, e as RecordExpr(_, lst)) =
		let fun ff(RecordAsst(_, f', _)) = ptEq(f, f')
		      | ff x = raise SympBug
		         ("TransGen.exprDef/exprSplit: field in RecordExpr is not RecordAsst:\n  "
			  ^(pt2string x)^"\n  in expression: "^(pt2string e))
		    val e' = (case List.find ff lst of
				  SOME e => e
				| NONE => raise SympBug
				  ("TransGen.exprDef/exprSplit: field not found in RecordExpr:\n  "
				   ^"field: "^(pt2string f)^"\n  expression: "^(pt2string e)))
		in exprSplit(S, e')
		end
	      | exprSplit'(S, e as RecordExpr _) = raise SympBug
		 ("TransGen.exprDef/exprSplit: wrong extractor for RecordExpr:\n  extractor: "
		  ^(printFirstPt S)^"\n  expression: "^(pt2string e))
	      | exprSplit'(S as (ExtractRecord(f, _))::S', e as WithExpr(_, e', lst)) =
		let fun ff(WithAsst(_, f', _)) = ptEq(f, f')
		      | ff x = raise SympBug
		         ("TransGen.exprDef/exprSplit: field in WithExpr is not WithAsst:\n  "
			  ^(pt2string x)^"\n  in expression: "^(pt2string e))
		in
		    case List.find ff lst of
			SOME e => exprSplit(S', e)
		        (* `with' doesn't modify this field, remove it *)
		      | NONE => exprSplit(S, e')
		end
	      | exprSplit'(S, e as WithExpr _) = raise SympBug
		 ("TransGen.exprDef/exprSplit: wrong extractor for WithExpr:\n  extractor: "
		  ^(printFirstPt S)^"\n  expression: "^(pt2string e))
	      | exprSplit'((ExtractAppl(c1, _))::S, expr as Appl(_, c2 as TypeConstr _, e)) =
		 if ptEq(c1, c2) then exprSplit(S, e)
		 else raise TransError 
		     ("Applied a wrong datatype extractor:\n  "
		      ^(pt2string(ExtractAppl(c1, expr))))
	      | exprSplit'((ExtractIndex _)::S, expr as Appl(_, c as TypeConstr _, _)) =
		 exprSplit(S, c)
	      | exprSplit'(S, e as Appl(_, TypeConstr _, _)) = raise SympBug
		 ("TransGen.exprDef/exprSplit: wrong extractor for TypeConstr application:\n"
		  ^"  extractor: "^(printFirstPt S)^"\n  expression: "^(pt2string e))
	       (* And annihilate it with a `fn' expression *)
	      | exprSplit'((Appl(p, e1, _))::S, e as FunClosure _) =
		 exprSplit(S, evalExpr(Appl(p, e1, e)))
	      | exprSplit'(S, e as FunClosure _) = raise SympBug
		 ("TransGen.exprDef/exprSplit: wrong extractor for FunClosure:\n"
		  ^"  extractor: "^(printFirstPt S)^"\n  expression: "^(pt2string e))
	      (* Then, handle the compound expressions *)
	      | exprSplit'(S, expr as IfExpr(_, conds, last)) =
		let val tp = exprType last
		    val botDef = DefExprAtomic(DefExprBodyAtomic(bot tp))
		    fun doCond(CondExpr(_, c, e)) =
		        let val cDef = exprDef omegaOpt c
			    val C = normEq(cDef, true3Def)
			    val Cbot = normEq(cDef, botBoolDef)
			    val F = exprSplit(S, e)
			in [(C, F), (Cbot, botDef)]
			end
		      | doCond x = raise SympBug
			 ("TransGen.exprDef/exprSplit: not a CondExpr in IfExpr:\n  "
			  ^(pt2string x))
		    val lastPair = (trueDef, exprSplit(S, last))
		    val pairs = List.foldr(op @) [lastPair] (List.map doCond conds)
		in 
		    normCond pairs
		end
	      | exprSplit'(S, expr as CaseExpr(_, sel, conds)) =
		let val E = exprDef omegaOpt sel
		    val selType = exprType sel
		    val (chi, Edefs) = 
			  (case getBody E of
			       DefExprBodyAtomic e => (e, getDefsRaw E)
			     | _ => let val chi = DefVar{name=newName(), Type=selType}
				    in (chi, addDefs(getDefsRaw E, [(chi, [getBody E])]))
				    end)
		    val Cbot = normEqExpr(chi, bot selType)
		    val botDef = makeDefRaw(sharedDefsRaw(), DefExprBodyAtomic(bot(exprType expr)))
		    fun doCond(ChoiceClosure{pattern=pat, uname=sel, body=e, ...}) =
			let val C = matches(chi, pat)
			    val F = exprSplit(S, subst(sel, chi) e)
			in (C, F)
			end
		      | doCond x = raise SympBug
			("TransGen.exprDef/exprSplit: not a ChoiceClosure in CaseExpr:\n  "
			 ^(pt2string x)^"\n  expression: "^(pt2string expr))
		    val pairs = List.map doCond conds
		    val F = normCond((Cbot, botDef)::(pairs@[(trueDef, botDef)]))
		in
		    makeDefRaw(mergeDefs[Edefs, getDefsRaw F], getBody F)
		end
	      | exprSplit'(S, expr as LetExpr(_, _, e)) = exprSplit(S, e)
		(* There are only built-in expressions remaining (arithmetic and boolean) *)
	      | exprSplit'([], expr as Appl(_, f as Builtin _, e)) = builtinDef(f, e)
	      | exprSplit'([], expr as ForallClosure _) = quantifierDef expr
	      | exprSplit'([], expr as ExistsClosure _) = quantifierDef expr
	      | exprSplit'(S, expr as Builtin{name=Anyvalue p, Type=tp}) = 
		let val Ds = 
		        (case typeVals tp of
			     SOME lst => List.map (exprDef NONE) lst
			   | NONE => raise TransError
				 ((pos2string p)
				  ^": cannot translate AnyValue of too large or infinite type:\n  "
				  ^(pt2string tp)))
		    val v = DefVar{name=newName(), Type=tp}
		    val defs = addDefs(mergeDefs(List.map getDefsRaw Ds),
				       [(v, List.map getBody Ds)])
		in
		    exprWrap(S, defs, v)
		end
	      | exprSplit'(S, expr as NondetExpr(_, [])) = raise SympBug
		   ("TransGen.exprDef/exprSplit: empty list in NondetExpr")
	      | exprSplit'(S, expr as NondetExpr(_, [x])) = raise SympBug
		   ("TransGen.exprDef/exprSplit: single element in NondetExpr:\n  "
		    ^(pt2string expr))
	      | exprSplit'(S, expr as NondetExpr(_, lst)) =
		let val Es = List.map (exprDef omegaOpt) lst
		    val tp = getDefType(hd Es)
		    val v = DefVar{name=newName(), Type=tp}
		    val defs = addDefs(mergeDefs(List.map getDefsRaw Es),
				       [(v, List.map getBody Es)])
		in 
		    exprWrap(S, defs, v)
		end
	      | exprSplit'(S, expr as Appl(_, Builtin _, _)) = raise SympBug
		   ("TransGen.exprDef/exprSplit: non-empty list of extractors "
		    ^"for a built-in expression:\n  ["
		    ^(strlistmap2str ", " pt2string S)^"]\n  expression: "^(pt2string e))
	       (* The other type of application (not built-in) is an extractor,
		  not a constructor, rip it *)
	      | exprSplit'(S, e as Appl(p, e1, e2)) = exprSplit((Appl(p, Fake, e2))::S, e1)
		(* The expression must be atomic or a built-in operator *)
	      | exprSplit'(S, e) =
		if isAtomic e then exprWrap(S, sharedDefsRaw(), e)
		else raise SympBug
		    ("TransGen.exprDef/exprSplit: expression is not atomic:\n  : "^(pt2string e))

	    (* First stage: split the type of the expression (the main sequent e ~> F) *)
	    fun split tp =
		let val funName = funName^"/split"
		    val _ = pushFunStackLazy(funName, fn()=>pt2string tp)
		    val res = split' tp
		    val _ = popFunStackLazy(funName, fn()=>def2str res)
		in res
		end
	    and split'(TupleType(_, lst)) =
		let fun doOne i =
		        let val om = trimOmega(ExtractTuple(i, Fake), omegaOpt)
			in exprDef om (evalExpr(ExtractTuple(i, e)))
			end
		in normalizeDefTuple(List.map doOne (List.tabulate(List.length lst, fn x=>x)))
		end
	      | split'(tp as RecordType(_, lst)) =
		let fun doOne(RecordField{name=f, ...}) =
		        let val om = trimOmega(ExtractRecord(f, Fake), omegaOpt)
			in (f, exprDef om (evalExpr(ExtractRecord(f, e))))
			end
		      | doOne x = raise SympBug
		         ("TransGen.exprDef/split: RecordType field is not RecordField:\n  "
			  ^(pt2string x)^"\n  in type: "^(pt2string tp))
		in normalizeDefRecord(List.map doOne lst)
		end
	      | split'(tp as EnumType(_, lst)) = 
		let fun doOne(c as TypeConstr{Type=FunType _, ...}) =
			let val om = trimOmega(ExtractAppl(c, Fake), omegaOpt)
			in  (SOME(c, exprDef om (evalExpr(ExtractAppl(c, e))))
			     handle EvalError _ => NONE)
			end
		      | doOne(c as TypeConstr _) = NONE
		      | doOne x = raise SympBug
			  ("TransGen.exprDef/split: EnumType element is not TypeConstr:\n  "
			   ^(pt2string x)
			   ^"\n  in type: "^(pt2string tp))
		    val indDef = exprDef(trimOmega(ExtractIndex Fake, omegaOpt))
			                (evalExpr(ExtractIndex e))
		in normalizeDefEnum(indDef, List.mapPartial doOne lst)
		end
	      | split'(t as FunType(_, t1, t2)) = doFun(t, t1)
	      | split'(t as ArrayType(_, t1, t2)) = doFun(t, t1)
	      | split'(StaticFormalType{value=SOME t,...}) = split t
	      | split'(t as StaticFormalType{value=NONE,...}) =
		(case abstract options abs t of
		     SOME _ => exprSplit([], e)
		   | NONE => raise SympBug
			 ("TransGen/exprDef/split: cannot translate an uninterpreted type:\n  "
			  ^(pt2string t)
			  ^"Abstraction:\n"^(abs2str abs)))
	      | split'(tp as TypeClosure{recursive=false, def=t,...}) = split t
	      | split'(tp as TypeClosure{recursive=true, def=t,...}) = raise SympBug
		  ("TransGen/exprDef/split: cannot translate an (infinite) recursive type:\n  "
		   ^(pt2string tp)^" = "^(pt2string t))
	      | split'(t as Uid _) =
		(case findObject t of
		     SOME tp => split tp
		   | NONE => raise SympBug
		      ("TransGen/exprDef/split:  Uid type is not in object hash: "
		       ^(pt2string t)))
	      (* The type must be primitive *)
	      | split' tp = exprSplit([], e)
	    val res = (case omegaOpt of
			   SOME[] => DefExprAtomic(DefExprBodyAtomic(bot origTp))
			 | _ => split origTp)
	    val _ = popFunStackLazy(funName, fn()=>def2str res)
	in 
	    res
	end

    (* Generate a formula representing the transition relation given
       by the AsstVarsTree. *)

    (* Assumptions: all the variables in AsstVars are primitive, all
       the irrelevant vars are removed by the COI reduction (they
       don't hurt if finite, but we don't check the cone here). *)

    fun generateTrans(parms as (findObject, pvars, abs)) tree =
	let val funName = "generateTrans"
	    val options = getOptions()
	    val lim = get_limit options
	    val debug = lazyVerbDebug options
	    val debugStr = lazyVerbDebugStr options
	    (* Compute the list of all relevant variables,
	       we'll need them in boolExpr, etc. *)
	    val {init=initvars, norm=normvars, next=nextvars} = getAsstVars tree
	    val allVars = union ptEq (List.map (List.map vtName) [normvars, nextvars, initvars])
	    val _ = debug funName (fn()=>funName^"\n: allVars = ["^(ptlist2str ", " allVars)^"]\n")
	    val matches = matches(findObject, abs, SOME allVars)
	    val evalExpr = evaluateExpr options findObject
	    val eqFormula = eqFormula findObject
	    val normAnd = normAnd findObject
	    val normOr = normOr findObject
	    val normNot = normNot findObject
	    val normCond = normalizeDefCond findObject
	    val normApp = normalizeDefAppl true (findObject, abs, SOME allVars)
	    val normET = normalizeDefET findObject
	    val normER = normalizeDefER findObject
	    val normEC = normalizeDefEC findObject
	    val normEI = normalizeDefEI findObject
	    val normAsstAnd = normAsstAnd findObject
	    val normAsstOr = normAsstOr findObject
	    val normAsstCond = normalizeDefAsstCond findObject
	    val normAsstGuard = normalizeDefAsstGuard findObject
	    val exprDef = exprDef(findObject, abs, SOME allVars)
	    fun exprType tp = abstractType options (findObject, abs) (getExprType findObject tp)
	    val typeVals = getTypeValuesAbs(options, findObject, abs)
	    val normEq = normalizeDefEq true (findObject, abs, SOME allVars)
	    (* Same as normEq, only the abstraction is turned off *)
	    val normEqNoAbs = normalizeDefEq false (findObject, abs, SOME allVars)
	    val substAVT = substAVT(options, findObject, pvars, lim)
	    fun normAtom(W, F, T) =
		makeAsstDefRaw(mergeDefs[getDefsRaw W, getDefsRaw F, getDefsRaw T],
			       DefAsstBodyAtomic{norm=getBody W, next=getBody F, init=getBody T})
	    (* Keep the set of all vars around *)
	    (* val allVars = getAsstVars tree *)

	    (* Sequent: V |-_Omega g ~> Omega
	       Compute the set of extractor lists that define the parts of `stvar'
	       present in the cone `vars'. *)
	    fun getOmega(vars, stvar) =
		let fun doOne S v =
		        if svSimilar(v, stvar) then S
			else (case v of 
				  ExtractTuple(_, v') => doOne(v::S) v'
				| ExtractRecord(_, v') => doOne(v::S) v'
				| ExtractIndex v' => doOne(v::S) v'
				| ExtractAppl(_, v') => doOne(v::S) v'
				| Appl(_, v', _) => doOne(v::S) v'
				| _ => raise SympBug
				   ("TransGen.generateTrans/getOmega: "
				    ^"bad extractor on a state var:\n  "
				    ^(pt2string v)^"\n  Original state var:  "
				    ^(pt2string stvar)))
		in 
		    List.map (doOne []) vars
		end

	    (* Split by the type of `stvar' into primitive
	       assignments, check that we only assign relevant vars,
	       and generate the final def-structures.

	       Sequent V |-_t g := E. *)

	    fun splitPrim flag (vars: VarType list, stvar, E) =
		let val funName = "splitPrim"
		    val splitPrim = splitPrim flag
		    val tp = exprType stvar
		    fun getRelevant s = List.filter (fn v => svPartOf(vtName v, s)) vars
		    fun doFun() =
			let val pairs = splitByArrayIndex(vars, stvar)
			    fun doOne(d, V) =
				let val D = makeDefRaw(sharedDefsRaw(), DefExprBodyAtomic d)
				in splitPrim(V, Appl(dp, stvar, d), normApp(E, D))
				end
			in normAnd(List.map doOne pairs)
			end
		    fun split(TupleType(_, lst)) =
			let fun doOne i =
			        let val s = ExtractTuple(i, stvar)
				in 
				    case getRelevant s of
					[] => NONE
				      | V => SOME(splitPrim(V, s, normET(i, E)))
				end
			    val indices = List.tabulate(List.length lst, fn i=>i)
			in normAnd(List.mapPartial doOne indices)
			end
		      | split(RecordType(_, lst)) =
			let fun doOne(RecordField{name=f, ...}) =
			        let val s = ExtractRecord(f, stvar)
				in 
				    case getRelevant s of
					[] => NONE
				      | V => SOME(splitPrim(V, s, normER(f, E)))
				end
			      | doOne x = raise SympBug
				("TransGen.splitPrim/split: RecordType's field "
				 ^"is not RecordField:\n  "^(pt2string x))
			in normAnd(List.mapPartial doOne lst)
			end
		      | split(EnumType(_, lst)) =
			let fun doOne(c as TypeConstr{Type=FunType _,...}) =
			        let val s = ExtractAppl(c, stvar)
				in 
				    case getRelevant s of
					[] => NONE
				      | V => SOME(splitPrim(V, s, normEC(c, E)))
				end
			      | doOne(c as TypeConstr _) = NONE
			      | doOne x = raise SympBug
				("TransGen.splitPrim/split: EnumType's field is not TypeConstr:\n  "
				 ^(pt2string x))
			    val si = ExtractIndex stvar
			    val C = (case getRelevant si of
					 [] => []
				       | V => [splitPrim(V, si, normEI E)])
			in normAnd(C@(List.mapPartial doOne lst))
			end
		      | split(FunType(_, t1, t2)) = doFun()
		      | split(ArrayType(_, t1, t2)) = doFun()
		      | split(StaticFormalType{value=SOME t,...}) = split t
		      | split(t as StaticFormalType{value=NONE,...}) = raise SympBug
			("TransGen.splitPrim/split: cannot translate "
			 ^"an uninterpreted type:\n  "^(pt2string t)
			 ^"\nAbstraction:\n"^(abs2str abs))
		      | split(tp as TypeClosure{recursive=false, def=t,...}) = split t
		      | split(tp as TypeClosure{recursive=true, def=t,...}) = raise SympBug
			("TransGen/normalizeDefEq/split: cannot translate"
			 ^" an (infinite) recursive type:\n  "
			 ^(pt2string tp)^" = "^(pt2string t)
			 ^"\nAbstraction:\n"^(abs2str abs))
		      | split(t as Uid _) =
			(case findObject t of
			     SOME tp => split tp
			   | NONE => raise SympBug
				 ("TransGen/normalizeDefEq/split:  Uid type is not in object hash: "
				  ^(pt2string t)))
		      (* The type must be primitive *)
		      | split tp =
			let val s = (case flag of
					 NextFlag => Next stvar
				       | _ => stvar)
			in normEqNoAbs(makeDefRaw(sharedDefsRaw(), DefExprBodyAtomic s), E)
			end
		in
		    split tp
		end (* splitPrim *)

	    (* Translating the atomic assignments.
	       Sequent:  V; Sigma |- g << E ~> F *)
	    fun asstGen1 flag (vars: VarType list, stvar, E) =
		let val funName = "asstGen1"
		    val exprDef = exprDef NONE
		    val _ = pushFunStackLazy(funName, 
					     fn()=>((vtlist2str "," vars)^"; [] |- "
						    ^(vt2str(vtWrap flag stvar))
						    ^" << "^(def2str E)))
		    val botDef = makeDefRaw(sharedDefsRaw(),
					    DefExprBodyAtomic(bot(exprType stvar)))
		    fun wrapDef([], g) = g
		      | wrapDef((ExtractTuple(i, _))::S, g) = wrapDef(S, ExtractTuple(i, g))
		      | wrapDef((ExtractRecord(f, _))::S, g) = wrapDef(S, ExtractRecord(f, g))
		      | wrapDef((ExtractAppl(c, _))::S, g) = wrapDef(S, ExtractAppl(c, g))
		      | wrapDef((ExtractIndex _)::S, g) = wrapDef(S, ExtractIndex g)
		      | wrapDef((Appl(p, _, d))::S, g) = wrapDef(S, Appl(p, g, d))
		      | wrapDef(x::S, g) = raise SympBug
			 ("TransGen.generateTrans/getOmega: "
			  ^"bad extractor on a state var:\n  "
			  ^(pt2string x)^"\n  Original state var:  "
			  ^(pt2string stvar))
		    fun loop(V, S, e) =
			let val funName = funName^"/loop"
			    fun argFn() = ((vtlist2str "," vars)^"; ["
					   ^(ptlist2str ", " S)^"] |- "
					   ^(pt2string e)^" <<  <same E>")
			    val _ = pushFunStackLazy(funName, argFn)
			    val res = loop'(V, S, e)
			    val _ = popFunStackLazy(funName, fn()=>def2str res)
			in res
			end
		    and loop'(V, S, e as ExtractTuple(_, g)) = loop(V, e::S, g)
		      | loop'(V, S, e as ExtractRecord(_, g)) = loop(V, e::S, g)
		      | loop'(V, S, e as ExtractAppl(_, g)) = loop(V, e::S, g)
		      | loop'(V, S, e as ExtractIndex g) = loop(V, e::S, g)
		      | loop'(V, S, e as Appl(p, g, e')) =
		        let val E' = exprDef e'
			    val Ebot = (case flag of
					    NextFlag => E
					  | _ => botDef)
			    fun doOne(d, Vd) =
			        let val Fd = loop(Vd, (Appl(p, Fake, d))::S, g)
				    val Cd = normEq(exprDef d, E')
				    val Ud = splitPrim flag (vars, wrapDef(S, Appl(p, g, d)), Ebot)
				in ((Cd, Fd), Ud)
				end
			    (* fold(triple, Us, usedUs):
			       Process a pair (Cd, Fd) and move a U from left to right.
			       For each pair generate (Cd, Fd & (/\_{d!=d'} U_d')).*)
			    val errorStr = "TransGen.generateTrans/asstGen1: this can't happen"
			    fun fold([], [], Us) = [(trueDef, normAnd Us)]
			      | fold((Cd, Fd)::pairs, u::U, U') =
				  (Cd, normAnd(Fd::(U@U')))::(fold(pairs, U, u::U'))
			      | fold([], _, _) = raise SympBug errorStr
			      | fold(_, [], _) = raise SympBug errorStr
			    fun splitE() =
				let val Dpairs = splitByArrayIndex(V, g)
				    val triples = List.map doOne Dpairs
				    val Us = List.map #2 triples
				    val pairs = List.map #1 triples
				in normCond(fold(pairs, Us, []))
				end
			in
			    if isConst e' then loop(V, (Appl(p, Fake, e'))::S, g)
			    else splitE()
			end
				
		      (* g is atomic *)
		      | loop'(V, S, g) = splitPrim flag (vars, wrapDef(S, g), E)
		    val F = loop(vars, [], stvar)
		    val _ = popFunStackLazy(funName, fn()=> def2str F)
		in F
		end

	    (* Translate the atomic assignments.
	       Sequent V |- g := e ~> F *)
	    fun asstGen (flag: VarTypeFlag)
		        (vars: VarType list, stvar: ParseTree, expr: ParseTree) = 
		(* Strip ObjectInst from `stvar' and `expr', so we have type constructors
		   "naked" for easier comparison. *)
		let val funName = "asstGen"
		    val _ = pushFunStackLazy(funName, 
		      fn()=>((vtlist2str "," vars)^" |- "
			     ^(vt2str(vtWrap flag stvar))^" := "^(pt2string expr)))
		    val _ = debug "asstGenDebug" 
		      (fn()=>"debug: asstGen("^(vtlist2str "," vars)^" |- "
		       ^(vt2str(vtWrap flag stvar))^" := "^(pt2stringDebug expr)^")\n")
		    val V = List.map vtName vars
		    val omega = getOmega(V, stvar)
		    val E = exprDef(SOME omega) expr
		    val F = asstGen1 flag (vars, stvar, E)
		    val _ = debug "asstGenDebug" (fn()=>"debug: asstGen => "
						  ^(def2str F)^"\n")
		    val _ = popFunStackLazy(funName, fn()=>(def2str F))
		in F
		end


	    (* All vars in the list get assigned Undefined *)
	    fun AssignUndef(vars,flag) =
		let fun doAsst v =
			let val vname = vtName v
			    val tp = getExprType findObject vname
			in asstGen flag ([v], vname, bot tp)
			end
		in normAnd(List.map doAsst vars)
		end
	    (* All (vars: AsstVars) get assigned Undefined *)
	    fun AssignUndefAll{norm=norm,next=next,init=init} =
		normAtom(AssignUndef(norm,NormalFlag),
			 AssignUndef(next,NextFlag),
			 AssignUndef(init,InitFlag))

            (* All next vars get assigned their current values, the others get `undefined'. *)
	    fun AssignNop{norm=norm, next=next, init=init} =
		let fun oneNext v =
		        let val vname = vtName v
			in asstGen NextFlag ([v], vname, vname)
			end
		in normAtom(AssignUndef(norm, NormalFlag),
			    normAnd(List.map oneNext next),
			    AssignUndef(init,InitFlag))
		end

	    (* Walk over the AsstVarsTree, recurse into the compound assignments.
	       Sequent N; L; I |- A ~> D *)
	    fun walkTree avt = 
		let val funName = "walkTree"
		    val _ = pushFunStackLazy(funName, fn()=>(avt2str avt))
		    val D = walkTree0 avt
		    val _ = popFunStackLazy(funName, fn()=>defasst2str D)
		in D
		end
	    and walkTree0 (NopAsstTree vars) = AssignNop vars
	      | walkTree0 (NormalAsstTree(avars,stvar,expr)) =
		let val vars = (case avars of (* Sanity check *)
				    {norm=norm,next=[],init=[]} => norm
				  | {next=lst, init=[], ...} => raise SympBug
					("generateTrans/walkTree/normalAsst: "
					 ^"next vars non-empty:\n  "
					 ^(vtlist2str ", " lst))
				  | {init=lst, ...} => raise SympBug
				        ("generateTrans/walkTree/normalAsst: "
					 ^"init vars non-empty:\n  "
					 ^(vtlist2str ", " lst)))
		    val e = evalExpr expr
		    val F = asstGen NormalFlag (vars, stvar, e)
		in normAtom(F, trueDef, trueDef)
		end
	      | walkTree0 (NextAsstTree(avars,stvar,expr)) =
		let val vars = (case avars of (* Sanity check *)
				    {norm=[],next=next,init=[]} => next
				  | {norm=lst, init=[], ...} => raise SympBug
					("generateTrans/walkTree/nextAsst: "
					 ^"normal vars non-empty:\n  "
					 ^(vtlist2str ", " lst))
				  | {init=lst, ...} => raise SympBug
				        ("generateTrans/walkTree/nextAsst: "
					 ^"init vars non-empty:\n  "
					 ^(vtlist2str ", " lst)))
		    val e = evalExpr expr
		    val F = asstGen NextFlag (vars, stvar, e)
		in normAtom(trueDef, F, trueDef)
		end
	      | walkTree0 (InitAsstTree(avars, stvar, expr)) =
		let val vars = (case avars of (* Sanity check *)
				    {norm=[],next=[],init=init} => init
				  | {next=lst, norm=[], ...} => raise SympBug
					("generateTrans/walkTree/initAsst: "
					 ^"next vars non-empty:\n  "
					 ^(vtlist2str ", " lst))
				  | {norm=lst, ...} => raise SympBug
				        ("generateTrans/walkTree/initAsst: "
					 ^"norm vars non-empty:\n  "
					 ^(vtlist2str ", " lst)))
		    val e = evalExpr expr
		    val F = asstGen InitFlag (vars, stvar, e)
		in normAtom(trueDef, trueDef, F)
		end
	      | walkTree0 (ListAsstTree(vars, tlist)) = normAsstAnd(List.map walkTree tlist)
	      | walkTree0 (LetAsstTree(vars, defs, t)) = walkTree t
	      | walkTree0 (LabeledAsstTree(vars, label, t)) = walkTree t
	      | walkTree0 (IfAsstTree(vars, pairs, elsePart)) =
		let val Fbot = AssignUndefAll vars
		    fun doCond(c, t) =
			let val C = exprDef NONE c
			    val Ct = normEq(true3Def, C)
			    val Cbot = normEq(botBoolDef, C)
			    val F = walkTree t
			in [(Ct, F), (Cbot, Fbot)]
			end
		    val last = walkTree elsePart
		    val Fpairs = List.foldr(op @) [(trueDef, last)] (List.map doCond pairs)
		in normAsstCond Fpairs
		end
	      | walkTree0 (CaseAsstTree(vars, sel, conds)) =
		let val Fbot = AssignUndefAll vars
		    val F0 = AssignNop vars
		    val selType = getExprType findObject sel
		    val E = exprDef NONE sel
		    val (chi, Edefs) = 
			  (case getBody E of
			       DefExprBodyAtomic e => (e, getDefsRaw E)
			     | _ => let val chi = DefVar{name=newName(), Type=selType}
				    in (chi, addDefs(getDefsRaw E, [(chi, [getBody E])]))
				    end)
		    (* Do one choice in the case statement *)
		    fun doCond(u, pat, t) =
			let val D = walkTree(substAsstTree(u,chi) t)
			    val C = matches(chi, pat)
			in (C, D)
			end
		    val Cbot = exprDef NONE (eqFormula (chi, bot selType))
		    val pairs = ((Cbot, Fbot)::(List.map doCond conds))@[(trueDef, F0)]
		    val D = normAsstCond pairs
		in makeAsstDefRaw(mergeDefs[Edefs, getAsstDefsRaw D], getAsstBody D)
		end
	      | walkTree0 (ChooseAsstTree(vars, paramsOpt, pairs)) =
		let fun def2pt E =
		        let val defs = getDefsRaw E
			in
			    (case getBody E of
				 DefExprBodyAtomic e => (defs, e)
			       | b => let val n = newName()
					  val v = DefVar{name=n, Type=getDefType findObject E}
				      in (addDefs(defs, [(v, [b])]), v)
				      end)
			end
		    (* Compute the list of mapping param->DefVar, and
		       the list of new definitions for the parameters *)
		    fun handleParams NONE = ([], sharedDefsRaw())
		      | handleParams (SOME params) =
		        let (* Create a new def. var for a parameter. *)
			    fun param2def pt =
				let val tp = exprType pt
				    val V = exprDef NONE (anyvalue tp)
				    val (defs, v) = def2pt V
				in
				    ((pt, v), defs)
				end
			    val tmp = List.map param2def params
			    val substList = List.map #1 tmp
			    val defs = mergeDefs(List.map #2 tmp)
			in
			    (substList, defs)
			end
		    val Dbot = AssignNop vars
		    val (substList, defs) = handleParams paramsOpt
		    fun substTree t =
			(case substList of
			     [] => t
			   | lst => List.foldr (fn((old, new), t) => substAVT(t, old, new)) t lst)
		    fun substPt e =
			(case substList of
			     [] => e
			   | lst => List.foldr (fn((old, new), e) => subst(old, new) e) e lst)
		    fun doPair (p, t) =
			let val C = normEq(exprDef NONE (substPt p), true3Def)
			    val notC = normNot C
			    val F = walkTree(substTree t)
			in ((C, F), notC)
			end
		    val tmp = List.map doPair pairs
		    val notCs = normAnd(List.map #2 tmp)
		    val conds = (notCs, Dbot)::(List.map #1 tmp)
		    fun finishChoose([], _) = normAsstGuard conds
		      | finishChoose(substList, defs) = 
			let (* val invar = normAtom(normOr(notCs::(List.map(fn((c, _), _)=>c) tmp)),
						 trueDef, trueDef) *)
			    (* val asst = normAsstAnd[invar, normAsstGuard conds] *)
			    val asst = normAsstGuard conds
			    val defs = mergeDefs[getAsstDefsRaw asst, defs]
			in makeAsstDefRaw(defs, getAsstBody asst)
			end
		in
		    finishChoose(substList, defs)
		end
(*	      | walkTree0 (ChooseAsstTree(vars, NONE, pairs)) =
		let val debug = debug "walkTree"
		    val Dbot = AssignNop vars
		    fun doPair (p, t) =
			let val C = normEq(exprDef NONE p, true3Def)
			    val notC = normNot C
			    val F = walkTree t
			in ((C, F), notC)
			end
		    val tmp = List.map doPair pairs
		    val notCs = normAnd(List.map #2 tmp)
		    val conds = (notCs, Dbot)::(List.map #1 tmp)
		in 
		    normAsstGuard conds
		end
*)
	      | walkTree0 (ForeachAsstTree(vars, params, tree)) =
		let val ds = allVarValues(options, findObject, lim, abs) params
		    fun doOne dvec =
			let val pairs = zip(params, dvec)
			    val t = List.foldr(fn((p,d), t)=>substAVT(t,p,d)) tree pairs
			in walkTree t
			end
		in normAsstAnd(List.map doOne ds)
		end
	in
	    walkTree tree
	end (* generateTrans *)

    (* Clean up the mess.  Rebuild the defsHash for the given
       defs-structure to use only those defs that are mentioned in the
       body. *)

    (* Checking in a var appears in the body *)

    fun findVarExprBody(DefExprBodyAtomic e) v =
	let fun checkVar e = ptEq(v, e) orelse (List.exists checkVar (ptSubtrees e))
	in checkVar e
	end
      | findVarExprBody(DefSeqList lst) v =
	let fun checkPair(C, F) = findVarExprBody C v orelse findVarExprBody F v
	in List.exists checkPair lst
	end
      | findVarExprBody(DefArray{pairs=pairs,...}) v =
	  List.exists(fn (_, F)=> findVarExprBody F v) pairs
      | findVarExprBody(DefTuple lst) v = List.exists (fn F=>findVarExprBody F v) lst
      | findVarExprBody(DefRecord pairs) v = List.exists(fn (_, F)=> findVarExprBody F v) pairs
      | findVarExprBody(DefEnum(I, pairs)) v =
	  findVarExprBody I v orelse List.exists(fn (_, F)=> findVarExprBody F v) pairs

    fun findVarExpr(DefExprDecl{body=body, defs=defs}) v = findVarExprBody body v
      | findVarExpr(DefExprAtomic body) v = findVarExprBody body v

    (* Traverse def-structure and add the DefVar's used in it DESTRUCTIVELY to the `hash'.
       Copy the expressions associated with this var from `defHash' if it not there yet,
       and traverse these expressions recursively. *)
    fun collectDefVarsExprBody (hash, defHash) F =
	let fun loop(DefExprBodyAtomic e) =
	        let fun doExpr(e as DefVar _) =
		          (case findHash(hash, e) of
			       SOME _ => ()
			     | NONE => (case findHash(defHash, e) of
					    SOME Es => (insertHashDestructive(hash, e, Es);
							List.app loop Es)
					  | NONE => raise SympBug
				   ("TransGen.collectDefVarsExprBody: DefVar is not in defHash: "
				    ^(pt2string e))))
		      | doExpr e = List.app doExpr (ptSubtrees e)
		in doExpr e
		end
	      | loop(DefSeqList lst) = List.app(fn(C, F) => (loop C; loop F)) lst
	      | loop(DefArray{pairs=pairs,...}) = List.app(fn(_, F) => loop F) pairs
	      | loop(DefTuple lst) = List.app loop lst
	      | loop(DefRecord pairs) = List.app(fn(_, F) => loop F) pairs
	      | loop(DefEnum(I, pairs)) = (loop I; List.app(fn(_, F) => loop F) pairs)
	in loop F
	end

    (* Same thing for assignments *)
    fun collectDefVarsAsstBody(hash, defHash) A =
	let val EBody = collectDefVarsExprBody(hash, defHash)
	    fun loop(DefAsstBodyAtomic{norm=norm, next=next, init=init}) =
	          (EBody norm; EBody next; EBody init)
	      | loop(DefAsstSeqList pairs) = List.app(fn(C, F) => (EBody C; loop F)) pairs
	      | loop(DefAsstGuardList pairs) = List.app(fn(C, F) => (EBody C; loop F)) pairs
	      | loop(DefAsstAnd lst) = List.app loop lst
	      | loop(DefAsstOr lst) = List.app loop lst
	in loop A
	end

    fun minimizeDefVarsExpr(F as DefExprAtomic _) =  (* No defs here *)
	(verbDebug(getOptions()) "minimizeDefVarsExpr" 
	 ("\nminimizeDefVarsExpr: Atomic case (size = 0)\n");
	 F)
      | minimizeDefVarsExpr(DefExprDecl{defs=defs, body=body}) =
	let val funName = "minimizeDefVarsExpr"
	    val timer = startTimer funName
	    val _ = pushFunStackLazy(funName, fn()=>"initial size = "
				     ^(Int.toString(hashSize defs)))
	    val hash = newDefsRaw()
	    val _ = collectDefVarsExprBody(hash, defs) body
	    val res = makeDefRaw(hash, body)
	    val _ = updateStat timer
	    val _ = popFunStackLazy(funName, fn()=>"reduced size = "
				    ^(Int.toString(hashSize hash))
				    ^", "^(statTimer2string timer))
	in res
	end

    fun minimizeDefVarsAsst(F as DefAsstAtomic _) =  (* No defs here *)
	(verbDebug(getOptions()) "minimizeDefVarsAsst" 
	 ("\nminimizeDefAsst: Atomic case (size = 0)\n");
	 F)
      | minimizeDefVarsAsst(DefAsstDecl{defs=defs, body=body}) =
	let val funName = "minimizeDefVarsAsst"
	    val timer = startTimer funName
	    val _ = pushFunStackLazy(funName, fn()=>"initial size = "
				     ^(Int.toString(hashSize defs)))
	    val hash = newDefsRaw()
	    val _ = collectDefVarsAsstBody(hash, defs) body
	    val res = makeAsstDefRaw(hash, body)
	    val _ = updateStat timer
	    val _ = popFunStackLazy(funName, fn()=>"reduced size = "
				    ^(Int.toString(hashSize hash))
				    ^", "^(statTimer2string timer))
	in res
	end

    (* Translate a theorem into a primitive boolean formula.
       Currently, only theorems of the form `self |= e' are handled *)

(*    fun generateTheorem(parms as (findObject, abs)) (Theorem(_,_,Models(_, _, e))) =
	let val options = getOptions()
	    val evalExpr = evaluateExpr options findObject
	in 
	    normalizeDefEq parms (exprDef parms NONE e, true3Def)
	      (* before newResultHash() *)
	end
      | generateTheorem _ _ x = raise SympBug
	  ("generateTheorem: sorry, this theorem statement is not implemented yet:\n  "
	   ^(pt2string x))
*)
    (* Translate `e = true' into a primitive 2-valued boolean formula *)
    fun boolExpr(parms as (findObject, abs, stateVarsOpt)) e =
	let val funName = "boolExpr"
	    val _ = resetStat()
	    val timer = startTimer funName
	    val _ = pushFunStackLazy(funName, fn()=>pt2string e
				     ^(ptlistopt2str(", stateVars = ", "") stateVarsOpt))
	    val options = getOptions()
	    (* Create a fresh defs hash *)
	    val _ = defsHash := newDefsRaw()
	    val tp = getExprType findObject e
	    val res = 
		if hasTemporal e then exprDef parms NONE e (* This will return bool2 *)
		else normalizeDefEq true parms (exprDef parms NONE e, true3Def)
	    val res = minimizeDefVarsExpr res
	    val _ = updateStat timer
	    (* Clean up the defs, so they can be GC'ed later *)
	    val _ = defsHash := newDefsRaw()
	    val _ = popFunStackLazy(funName, fn()=>(def2str res)^"\n"^(stat2string()))
	in 
	    res
	      (* before newResultHash() *)
	end

    (* Gerenate a formula for keeping all the values of `next'
       variables unchanged in the next state.  Only one formula is
       generated and returned. *)

    fun generateStallRel(parms as (findObject, pvars, abs)) trans =
	let val options = getOptions()
	    val lim = get_limit options
	    val normAsstAnd = normAsstAnd findObject
	    fun loop (TransAtomic(AtomicModel model)) =
		let val { assts=assts,
			  cone=cone,
			  pvars=pvars, ... } = model
		    val generateTrans = generateTrans parms
		    val vars as {next=nextvars, ...} = getAsstVars assts
		    (* Generate assignments `next(x) := x' for all `next' vars *)
		    fun doVars trees ((v as NextVar n)::vars) =
			  doVars ((NextAsstTree({norm=[], next=[v], init=[]}, n, n))::trees) vars
		      | doVars trees [] = ListAsstTree(vars, trees)
		      | doVars _ (v::_) = raise SympBug
			  ("TransGen/generateStallRel: not a `next' var: "
			   ^(vt2str v))
		    val newAssts = doVars [] nextvars
		    val nextTr = generateTrans newAssts
		in
		    nextTr: DefAsst
		end
	      | loop (TransSync2(tr1,tr2)) = (normAsstAnd[loop tr1, loop tr2]): DefAsst
	      | loop (TransAsync2(tr1, tr2)) = (normAsstAnd[loop tr1, loop tr2]): DefAsst
	      | loop  _ = raise SympBug
		  ("TransGen/generateStallRel: sorry, closed form composition is not implemented yet")
	in
	    loop trans
	end
	
    (* Generate transition relation for a composition of models represented by TransRel *)
    fun generateTransRel(parms as (findObject, pvars, abs)) trans =
	let val options = getOptions()
	    val { limit=lim,...} = options
	    val normAsstAnd = normAsstAnd findObject
	    val normAsstOr = normAsstOr findObject
	    val generateTransRel = generateTransRel parms
	    val generateStallRel = generateStallRel parms
	    fun loop (TransAtomic(AtomicModel model)) =
		 let val { assts=assts,
			   cone=cone,
			   pvars=pvars, ... } = model
		 in  (* Need the model's own pvars, don't use `parms' *)
		     generateTrans(findObject, pvars, abs) assts
		 end
	      (* Synchronous parallel composition: symply conjunct all the formulas *)
	      | loop (TransSync2(tr1, tr2)) = 
		 let val F1 = generateTransRel tr1
		     val F2 = generateTransRel tr2
		 in
		     normAsstAnd[F1, F2]
		 end

	      (* Asynchronous composition is trickier.  First, we
	         generate an extra transition relation tr{1,2}' for
	         stalling each module (all latched vars remain the
	         same), and then put them together as follows:

                 (tr1 & tr2') | (tr2 & tr1')

		 That is, no selector variable is generated.  The
		 invariants and initial states are conjuncted as in
		 the synchronous case (??? Is this a reasonable
		 semantics ???) *)

	      | loop (TransAsync2(tr1, tr2)) =
		 let val F1 = generateTransRel tr1
		     val F2 = generateTransRel tr2
		     val stall1 = generateStallRel tr1
		     val stall2 = generateStallRel tr2
		 in
		      normAsstOr[normAsstAnd[F1, stall2], normAsstAnd[stall1, F2]]
		 end
		      
              | loop _ = raise SympBug
		 ("TransGen/generateTransRel: sorry, closed form parallel composition is not implemented yet")
	in
	    loop trans
	end

    fun generateModelTrans(model: Model) =
	let val funName = "generateModelTrans"
	    val _ = resetStat()
	    val timer = startTimer funName
	    (* Create a fresh defs hash *)
	    val _ = defsHash := newDefsRaw()
	    val { trans=trans,
		  objectHash=hash,
		  pvars=pvars,
		  abs=abs,... } = model
	    val _ = pushFunStack(funName, "<model>")
	    val res = minimizeDefVarsAsst(generateTransRel(makeFindObject hash, pvars, abs) trans)
	    val _ = updateStat timer
	    (* Clean up the defs, so they can be GC'ed later *)
	    val _ = defsHash := newDefsRaw()
	    val _ = popFunStackLazy(funName, fn()=>"Returning...\n"^(stat2string()))
	in 
	    res
	end

    (* Export exprDef in a sane way *)
    fun expressionDef parms e = exprDef parms NONE e

  end
