functor ProofTreeFun(structure ProofSystem: PROOF_SYSTEM
		     structure Hash: HASH
		     structure Interface: INTERFACE): PROOF_TREE =
  struct
    structure ProofSystem = ProofSystem
    structure Sequent = ProofSystem.Sequent
    structure Interface = Interface
    structure Hash = Hash

    open ProofSystem
    open Sequent
    open SequentCommon
    open SequentBase
    open Hash
    open Interface
    open SympBug
    open Str
    open ListOps

    (* ProofRule is an inference rule or tactic name with specific
       arguments as the user gave them *)
    type ProofRule = string * (UserInputArgument list)

    (* The status of a proof *)
    datatype ProofStatus = Incomplete | Complete | Untried | Unchecked

    (* ID is a pair (pid, subgoal #), where pid is a unique ID of the
       proof tree, and the subgoal # is a numbering of subgoals inside
       that proof tree. *)

    type SubgoalID = string * (int list)
    type SubgoalUID = int

    datatype Subgoal = Subgoal of { id: SubgoalID ref,

       (* When true, any child of this subgoal will get an ID by
	  extending it with one more number, even if it's the only child.  When
	  false, the only child will increment the last number in the list. *)
				    extend: bool ref,
				    (* Unique id - everything internally should use it,
				       not the id, which may change with time *)
				    uid: SubgoalUID,
				    proofRule: ProofRule option ref,
				    children: (SubgoalUID * (SequentHints option)) list ref,
				    parents: SubgoalUID list ref,
				    prooftree: ProofTree option ref }

    and ProofTree = ProofTree of { root: Subgoal,
				   hash: (SubgoalUID, Subgoal) Hash,
				   (* Hash of leaf subgoals with no rule *)
				   incomplete: (SubgoalUID, Subgoal) Hash }

    type ProofRun = { prooftree: ProofTree,

		     (* Associate subgoal IDs with sequents and a bool
		        flag of whether the proof rule at that sequent
		        has been applied or not. *)

		      assocHash: (SubgoalUID, Sequent * bool) Hash,
		      (* Set of subgoals in the prooftree that do not have sequents *)
		      incomplete: (SubgoalUID, unit) Hash }

    (* Compare subgoals by UIDs *)
    fun SubgoalEq(Subgoal{uid=uid1,...}, Subgoal{uid=uid2,...}) = (uid1 = uid2)

    (* Return the number of subgoals in the tree *)
    fun ProofTreeSize(ProofTree{hash=hash, ...}) = hashSize hash

    (* If `old' is a prefix of `lst', replace it by `new' and return
       the new list; otherwise return NONE *)
    fun replacePrefix(old, new) lst = 
	let fun loop([], new) lst = SOME(new@lst)
	      | loop(x::old, new) (y::lst) =
	    if x = y then loop(old, new) lst
	    else NONE
	      | loop (x::old, new) [] = NONE
	in
	     loop(old, new) lst
	end

    fun Option_app ff (SOME x) = ff x
      | Option_app ff NONE = ()

    fun subgoalID2string (s,lst) = strlist2str "." (s::(List.map Int.toString lst))
    fun subgoalUID2string uid = Int.toString uid
    fun subgoal2string(Subgoal{id=ref id,...}) = subgoalID2string id
    fun subgoal2stringDebug(Subgoal{id=ref id, uid=uid, extend=ref ex, ...}) =
	  (subgoalID2string id)^"["^(subgoalUID2string uid)^", extend="
	  ^(case ex of true => "true" | false => "false")^"]"

    (* Reverse translation of string to subgoal ID *)
    fun string2subgoalID str =
	let fun isDigit ch = List.exists(fn x=>x = ch)(String.explode "0123456789")
	    (* Collect all elements in lst till the first `.' with a 
	       digit after it, and return (elements, rest-after-dot) *)
	    fun nextChunk [] acc = (String.implode(List.rev acc), [])
	      | nextChunk (#"."::ch::lst) acc =
		if isDigit ch then (String.implode(List.rev acc), ch::lst)
		else nextChunk(ch::lst)(#"."::acc)
	      | nextChunk (ch::lst) acc = nextChunk lst (ch::acc)
	    fun splitString [] acc = List.rev acc
	      | splitString lst acc =
		let val (str, rest) = nextChunk lst []
		in splitString rest (str::acc)
		end
	    fun strings2nums [] acc = SOME(List.rev acc)
	      | strings2nums (str::lst) acc = 
		  Option.mapPartial(fn n=>strings2nums lst (n::acc)) (Int.fromString str)
	    fun isName str =
		let val lst = String.explode str
		in
		    case lst of
			ch::_ => not(isDigit ch)
		      | _ => false
		end
	in
	    case splitString(String.explode str) [] of
		[] => NONE
	      | name::nums =>
		    if isName name then
			Option.map(fn nums=> (name, nums): SubgoalID)(strings2nums nums [])
		    else NONE
	end

    fun getProofRule(Subgoal{proofRule=rule,...}) = !rule
    fun getRoot(ProofTree{root=root,...}) = root
    fun getProofTree(s as Subgoal{prooftree=tree,...}) = !tree
    fun getChildren(Subgoal{children=ch, prooftree=ref(SOME(ProofTree{hash=hash, ...})), ...}) =
	  List.mapPartial (fn (x,_) => findHash(hash,x)) (!ch)
      | getChildren x = raise ProverError
	  ("Can't list children of a subgoal without a proof tree: "
	   ^(subgoal2string x))
    fun getChildrenHints(Subgoal{children=ch,
				 prooftree=ref(SOME(ProofTree{hash=hash, ...})), ...}) =
	  List.mapPartial (fn (id,hints) => Option.map (fn s=>(s,hints))(findHash(hash,id))) (!ch)
      | getChildrenHints x = raise ProverError
	  ("Can't list children with hints of a subgoal without a proof tree: "
	   ^(subgoal2string x))
    fun getParents(Subgoal{parents=ref parents,
			   prooftree=ref(SOME(ProofTree{hash=hash, ...})),...}) =
	  List.mapPartial (fn x=>findHash(hash,x)) parents
      | getParents x = raise ProverError
	  ("Can't list parents of a subgoal without a proof tree: "
	   ^(subgoal2string x))
    fun getSubgoalID(Subgoal{id=ref id,...}) = id
    fun getSubgoalUID(Subgoal{uid=uid,...}) = uid
    fun getSubgoalExtend(Subgoal{extend=ref x, ...}) = x
    fun getProofTreeID p = #1 (getSubgoalID(getRoot p))
    (* Proof tree is uniquely identified by the UID of its root subgoal *)
    fun getProofTreeUID p = getSubgoalUID(getRoot p)

    fun ProofRun2ProofTree({prooftree=tree,...}: ProofRun) = tree

    (* List all the subgoals in the tree in a DFS order, starting from the root. *)
    fun listSubgoals tree =
	let val root = getRoot tree
	    val markHash = makeHashDefault(op =, subgoalUID2string)
	    fun mark id = insertHashDestructive(markHash, id, ())
	    fun loop sub acc =
		let val (uid, ch) = (getSubgoalUID sub, getChildren sub)
		in (case findHash(markHash, uid) of
			SOME _ => acc
		      | NONE => (mark uid; looplist (List.rev ch) (sub::acc)))
		end
	    and looplist [] acc = acc
	      | looplist (sub::lst) acc =
		  loop sub (looplist lst acc)
	in 
	    List.rev(loop root [])
	end

    fun ProofTree2strCommon (verbose, runOpt) tree = 
	let fun op *(x,y) = Conc(x,y)
	    val subgoals = listSubgoals tree
	    val ProofTree{ incomplete = treeInc, ...} = tree
	    fun showSeq NONE uid = NONE
	      | showSeq(SOME run) uid =
		let val {assocHash=hash, incomplete=runInc, ...}: ProofRun = run
		in
		    SOME(Str("\n"
			     ^(case findHash(runInc, uid) of
				   NONE => "[run-complete]\n"
				 | SOME _ => "[run-incomplete]\n")
			     ^(case findHash(hash, uid) of
				   NONE => "subgoal has no sequent"
				 | SOME (seq, _) => sequent2string seq)
			     ^"\n"))
		end

	    fun doSubgoal sub = 
		 let val id = getSubgoalID sub
		     val uid = getSubgoalUID sub
		     val incomplete = isSome(findHash(treeInc, uid))
		     fun idStr(id, uid) = 
			 Str((if verbose then
				  (subgoalID2string id)^"["^(subgoalUID2string uid)^"]"
				  ^(if incomplete then ", incomplete" else ", complete")
			      else (subgoalID2string id))
			     ^" ")
		     fun subStr sub = 
			 Str((if verbose then subgoal2stringDebug sub
			      else (subgoal2string sub))
			     ^" ")
		     fun rule2str (r, args) =
			  (Str (r^"("))
			  *(Strlist2Str ", " (List.map (fn x=>Str (UIArg2string x)) args))
			  *(Str ")")
		     val ruleStr = Option.map rule2str (getProofRule sub)
		     fun child2str (ch,hints) =
			 (subStr ch)
			  *(Str (case hints of
				     NONE => ""
				   | SOME h => "[\""^(hints2string h)^"\"]"))
		     val children = List.map child2str (getChildrenHints sub)
		     val childrenStr =
			  Option.map (fn _=> (Str "(" )*(Strlist2Str ", " children)*(Str ")"))
			             ruleStr
		 in Strlist2Str " " (List.mapPartial (fn x=>x) 
				     [SOME(Str "("),SOME(idStr(id,uid)),
				      showSeq runOpt uid,
				      ruleStr,childrenStr,SOME(Str ")\n"),
				      Option.map(fn _ => Str "\n") runOpt])
		 end
	in 
	    Strlist2Str "" (List.map doSubgoal subgoals)
	end

    val ProofTree2str = ProofTree2strCommon(false,NONE)
    fun ProofTree2string tree = Str2string(ProofTree2str tree)

    val ProofTree2strDebug = ProofTree2strCommon(true,NONE)
    fun ProofTree2stringDebug tree = Str2string(ProofTree2strDebug tree)

    fun ProofRun2str run = ProofTree2strCommon(false, SOME run) (ProofRun2ProofTree run)
    fun ProofRun2strDebug run = ProofTree2strCommon(true, SOME run) (ProofRun2ProofTree run)
    fun ProofRun2string run = Str2string(ProofRun2str run)
    fun ProofRun2stringDebug run = Str2string(ProofRun2strDebug run)

    (* Check if the first id is the desendant of the second *)
    fun isDescendant _ [] = true
      | isDescendant [] _ = false
      | isDescendant (x1::lst1) [x2] = (x1 >= x2)
      | isDescendant (x1::lst1) (x2::lst2) = x1 = x2 andalso isDescendant lst1 lst2

    (* Exported function; check if the second subgoal is a child of the first *)
    fun isChild parent child =
	let val (pstr, pid) = getSubgoalID parent
	    val (cstr, cid) = getSubgoalID child
	    val (ptree, ctree) = (getProofTree parent, getProofTree child)
	in
	    (case (ptree, ctree) of
		 (SOME ptree, SOME ctree) => 
		     (getProofTreeUID ptree) = (getProofTreeUID ctree)
		     andalso pstr = cstr andalso isDescendant pid cid
	       | _ => false)
	end

    (* Find subgoal in the proof tree by its UID *)
    fun UID2Subgoal (ProofTree{hash=hash,...}) uid = findHash(hash,uid)

    (* For now, just search the entire tree for the subgoal ID;
       inefficient, but simple and correct.  Besides, we do not expect
       to have behemoth proof trees here. *)

    fun ID2Subgoal (ProofTree{hash=hash,...}) id =
	let val subgoals = hash2any(fn (_, x) => x) hash
	in 
	    List.find (fn s=> id = (getSubgoalID s)) subgoals
	end

    (* Same, only the probable ancestor is given as a UID *)
    fun isDescendantUID proofTree sid parentUID =
	let val (_,parentSid) = getSubgoalID
	    (case UID2Subgoal proofTree parentUID of
		 SOME s => s
	       | NONE => raise SympBug
		     ("isDescendantUID: parent ["
		      ^(subgoalUID2string parentUID)^"] not found"))
	in isDescendant sid parentSid
	end

    (* If debugging for this function is turned on, check that the
       proof tree is valid and raise a SympBug if not. *)
    fun checkProofTree(tree) =
	let val funName = "checkProofTree"
	    val ptUid = getProofTreeUID tree
	    fun check() =
	        let val subs = listSubgoals tree
		    val root = getRoot tree
		    val idHash = makeHashDefault(op =, subgoalID2string)
		    fun markID sub = insertHashDestructive(idHash, getSubgoalID sub, sub)
		    (* Check for repeated subgoal IDs *)
		    fun checkID sub =
			(case findHash(idHash, getSubgoalID sub) of
			     SOME s => raise SympBug
			       (funName^": the proof tree has two subgoals with the same ID: "
				^(subgoal2string sub)^"\n\nProof Tree:\n\n"
				^(ProofTree2stringDebug tree))
			   | NONE => (insertHashDestructive(idHash, getSubgoalID sub, sub); ()))
		    (* Check for bad children/parents links *)
		    fun checkLinks sub =
			let val uid = getSubgoalUID sub
			    fun checkChild(ch as Subgoal{parents=ref p,...}) =
			        if List.exists(fn s => uid = s) p then ()
				else raise SympBug
				    (funName^": subgoal "^(subgoal2string sub)^" has a child "
				     ^(subgoal2string ch)
				     ^"\nthat doesn't list it as its parent.\n\nProof Tree:\n\n"
				     ^(ProofTree2stringDebug tree))
			    fun checkParent(p as Subgoal{children=ref ch,...}) =
			        if List.exists(fn (s,_) => uid = s) ch then ()
				else raise SympBug
				    (funName^": subgoal "^(subgoal2string sub)^" has a parent "
				     ^(subgoal2string p)
				     ^"\nthat doesn't list it as its child.\n\nProof Tree:\n\n"
				     ^(ProofTree2stringDebug tree))
			in 
			    (List.app checkChild (getChildren sub);
			     List.app checkParent (getParents sub))
			end
		    (* Check that the 'extend' flag is set correctly *)
		    fun checkExtend sub =
			let val ch = getChildren sub
			    val extend = (List.length ch) >= 2
			    fun checkChild(ch as Subgoal{extend=ref ex, ...}) =
				if not extend orelse ex then ()
				else raise SympBug
				  (funName^": subgoal "^(subgoal2string sub)^" has a child "
				   ^(subgoal2stringDebug ch)
				   ^" with the wrong 'extend' field.\n\nProof Tree:\n\n"
				   ^(ProofTree2stringDebug tree))
			in List.app checkChild ch
			end
		    fun checkTreeRef(sub as Subgoal{prooftree=ref tOpt, ...}) =
			(case tOpt of
			     SOME t => if (getProofTreeUID t) = ptUid then ()
				       else raise SympBug
			        (funName^": subgoal "^(subgoal2string sub)
				 ^" has a wrong proof tree reference:\n\n"
				 ^(ProofTree2stringDebug t)
				 ^"\n\nCorrect Proof Tree:\n\n"
				 ^(ProofTree2stringDebug tree))
			   | NONE => raise SympBug
			        (funName^": subgoal "^(subgoal2string sub)
				 ^" doesn't have a proof tree reference.\n\nProof Tree:\n\n"
				 ^(ProofTree2stringDebug tree)))
		    fun checkTrueParent sub =
			if (getSubgoalUID sub) = (getSubgoalUID root) then ()
			else 
			    let val parents = getParents sub
				val (_, id) = getSubgoalID sub
				fun isTrueParent p =
				    let val (_, parentID) = getSubgoalID p
				    in isDescendant id parentID
				    end
			    in if List.exists isTrueParent parents then ()
			       else raise SympBug
				(funName^": subgoal "^(subgoal2string sub)
				 ^" doesn't have a true parent.\n\nProof Tree:\n\n"
				 ^(ProofTree2stringDebug tree))
			    end
		    fun checkIncomplete sub =
			let val ProofTree{ incomplete=hash, ... } = tree
			    val Subgoal{children=ref ch, proofRule = ref rule, ...} = sub
			in
			    case (ch, rule) of
				(_::_, NONE) => raise SympBug
				  (funName^": subgoal "^(subgoal2string sub)
				   ^" has children but no rule.\n\nProof Tree:\n\n"
				   ^(ProofTree2stringDebug tree))
			      | (_, SOME _) =>
				  (case findHash(hash, getSubgoalUID sub) of
				       NONE => ()
				     | SOME _ => raise SympBug
					(funName^": subgoal "^(subgoal2string sub)
					 ^" has a rule, but is in incomplete set."
					 ^"\n\nProof Tree:\n\n"
					 ^(ProofTree2stringDebug tree)))
			      | (_, NONE) =>
				  (case findHash(hash, getSubgoalUID sub) of
				       SOME _ => ()
				     | NONE => raise SympBug
					(funName^": subgoal "^(subgoal2string sub)
					 ^" has no rule, but is not in incomplete set."
					 ^"\n\nProof Tree:\n\n"
					 ^(ProofTree2stringDebug tree)))
			end
		    (* See if there is garbage in the incomplete set (subgoals not in the tree) *)
		    fun checkIncompleteGarbage() =
			let val ProofTree{incomplete=inc, hash=hash, ...} = tree
			    val incSubs = hash2any #1 hash
			in
			    case List.find(fn id=>not(isSome(findHash(hash, id)))) incSubs of
				NONE => ()
			      | SOME s => raise SympBug
				 (funName^": subgoal "^(subgoalUID2string s)
					 ^" is in incomplete set, but not in the tree."
					 ^"\n\nProof Tree:\n\n"
					 ^(ProofTree2stringDebug tree))
			end
		in 
		    (List.app checkID subs;
		     List.app checkTreeRef subs;
		     List.app checkLinks subs;
		     List.app checkExtend subs;
		     List.app checkTrueParent subs;
		     List.app checkIncomplete subs;
		     checkIncompleteGarbage();
		     funName^": check successful\n")
		end
	in
	    lazyVerbDebug(getOptions()) funName check
	end

    (* Save the proof tree to a file. *)
    fun ProofTree2file (tree, file) =
	let val outs = TextIO.openOut file
	    fun dump (Str s) = TextIO.output(outs,s)
	      | dump (Conc(x,y)) = ((dump x);(dump y))
	in 
	    (dump(ProofTree2str tree);
	     TextIO.closeOut outs)
	end handle _ => raise ProverError ("Can't save proof to file "^file)

    (* For now, we do not check that the subgoal is from the same
       proof as the run, just use the subgoal's ID *)
    fun getSequent ({assocHash=hash,...}: ProofRun) (Subgoal{uid=uid,...}) = 
	 Option.map(fn(seq,_)=>seq)(findHash(hash,uid))

    fun getSequentByUid ({assocHash=hash,...}: ProofRun) uid =
	 Option.map(fn(seq,_)=> seq)(findHash(hash,uid))

    fun getRuleAppliedFlag ({assocHash=hash,...}: ProofRun) (Subgoal{uid=uid,...}) = 
	 Option.map(fn(_,flag)=>flag)(findHash(hash,uid))

    fun getRuleAppliedFlagByUid ({assocHash=hash,...}: ProofRun) uid =
	 Option.map(fn(_,flag)=>flag)(findHash(hash,uid))

    (* DESTRUCTIVELY Recompute and update the incomplete hash of the
       proof tree.  A proof tree is considered "complete" when all
       leaves have an inference rule but no children *)

    fun updateProofTreeIncomplete tree = 
	let val ProofTree{root=root, hash=hash, incomplete=incomplete} = tree
	    (* Add all the subgoals to incomplete *)
	    val _ = wipeHashDestructive incomplete
	    val _ = List.app(fn(k,v)=>(insertHashDestructive(incomplete, k,v); ()))
		            (hash2any(fn x=>x) hash)
	    fun notVisited s = isSome(findHash(incomplete, getSubgoalUID s))
	    (* Then remove them as we go *)
	    fun mark s = (removeHashDestructive(incomplete, getSubgoalUID s); ())
	    fun loop (s as Subgoal{children=ref [], proofRule=ref(SOME _), ...}) = mark s
	      | loop (Subgoal{children=ref [], proofRule=ref NONE, ...}) = ()
	      | loop (s as Subgoal{children=ref lst, proofRule=ref NONE, ...}) = raise SympBug
		  ("updateProofTreeIncomplete: subgoal has children but no proof rule: "
		   ^(subgoal2string s))
	      | loop s = (mark s; List.app loop (List.filter notVisited (getChildren s)))
	in (loop(getRoot tree); tree)
	end

    fun isProofTreeComplete(ProofTree{incomplete=hash, ...}) = (hashSize hash) = 0

    (* DESTRUCTIVELY remove all the subgoals from the proof run that
       are no longer present in the proof tree, and update the set of
       incomplete subgoals in the proof run *)
    fun updateProofRun(run as {prooftree=ProofTree{hash=treeHash,...},
			       assocHash=hash,
			       incomplete=incomplete}: ProofRun) = 
	let val funName = "updateProofRun"
	    val _ = pushFunStackLazy(funName, fn()=>ProofRun2stringDebug run)
	    val _ = wipeHashDestructive incomplete
	    val ids = hash2any #1 treeHash
	    val _ = List.app(fn s =>(insertHashDestructive(incomplete, s, ()); ())) ids
	    fun checkSequent(id, (seq, _)) =
		(case findHash(treeHash, id) of
		     SOME s =>
			 (case getProofRule s of
			      SOME _ => ()
			    (* Fix the `ruleApplied' flag in the run *)
			    | NONE => (insertHashDestructive(hash, id, (seq, false)); ()))
		   | NONE => (removeHashDestructive(hash, id); ()))
	    (* See if the rule is present in the tree and is applied in the run *)
	    fun hasRule uid =
		(case findHash(treeHash, uid) of
		     SOME s => isSome(getProofRule s) andalso
		       (case findHash(hash, uid) of
			    SOME(_, ruleApplied) => ruleApplied
			  | NONE => false)
		   | NONE => raise SympBug
		      (funName^": subgoal "^(subgoalUID2string uid)^" is not in the tree"))

	    fun checkSequentIncomplete uid =
		if hasRule uid then (removeHashDestructive(incomplete, uid); ())
		else ()
	    val runPairs = hash2any(fn x=>x) hash
	in (List.app checkSequent runPairs;
	    (* Check only remaining elements in the run's hash *)
	    List.app checkSequentIncomplete (hash2any #1 hash);
	    popFunStackLazy(funName, fn()=>ProofRun2stringDebug run);
	    run)
	end

    (* A proof run is complete if the corresponding proof tree is
       complete, and every subgoal has an associated sequent *)
    fun isProofRunComplete({prooftree=tree, incomplete=hash, ...}: ProofRun) =
	  isProofTreeComplete tree andalso (hashSize hash) = 0

    fun getProofRunStatus proofRun =
	let val tree = ProofRun2ProofTree proofRun
	    val root = getRoot tree
	in
	    if isSome(getProofRule root) then 
		if isProofRunComplete proofRun then Complete
		else if isProofTreeComplete tree then Unchecked else Incomplete
	    else Untried
	end

    fun ProofStatus2string Complete = "proven"
      | ProofStatus2string Incomplete = "incomplete"
      | ProofStatus2string Untried = "untried"
      | ProofStatus2string Unchecked = "unchecked"

    fun getProofRunStatusString proofRun =
	ProofStatus2string(getProofRunStatus proofRun)

    (* Check if the proof step at this subgoal is complete.  That is,
       the subgoal has a sequent and a proof rule, and all of its
       children have a sequent. *)

    fun isProofStepComplete proofRun s =
	let val Subgoal{uid=uid, proofRule=ref ruleOpt, children=ref children, ...} = s
	    val applied = (case getRuleAppliedFlag proofRun s of
			       SOME x => x
			     | NONE => false)
	    fun isComplete(uid, _) = isSome(getSequentByUid proofRun uid)
	in
	    isSome ruleOpt andalso applied andalso List.all isComplete ((uid,NONE)::children)
	end

    (* Total order on subgoal IDs for sorting *)
    fun subgoalIDless((s1, n1): SubgoalID, (s2, n2): SubgoalID) =
	let fun numLess([], []) = false (* equal *)
	      | numLess([], _) = true
	      | numLess(_, []) = false
	      | numLess(i1::l1, i2::l2) = i1 < i2 orelse (i1 = i2 andalso numLess(l1, l2))
	in
	    String.<(s1, s2) orelse (s1 = s2 andalso numLess(n1, n2))
	end

    (* Total order on subgoals (by IDs) for sorting *)
    fun subgoalLess(s1, s2) =
	let val (id1, id2) = (getSubgoalID s1, getSubgoalID s2)
	in
	    subgoalIDless(id1, id2)
	end

    (* Find all subgoals without a proof and return them in a list.  A
       subgoal is a leaf if it doesn't have a proof rule. *)
    fun getLeafSubgoals(ProofTree{incomplete=hash, ...}) =
	  sortList subgoalLess (hash2any #2 hash)

    fun getUnprovenSubgoals(proofRun: ProofRun) =
	let val funName = "getUnprovenSubgoals"
	    val _ = pushFunStack(funName, "")
	    val { prooftree = ProofTree{ incomplete=treeIncomplete, hash=treeHash, ...},
		  incomplete = runHash, ...} = proofRun
	    (* Hash of all incomplete subgoals *)
	    val hash = makeHashDefault(op =, subgoalUID2string)
	    (* Add incomplete subgoals from the tree *)
	    val _ = List.app(fn (uid, s) => (insertHashDestructive(hash, uid, s); ()))
		            (hash2any(fn x=>x) treeIncomplete)
	    (* Add incomplete subgoals from the proof run *)
	    fun addSubgoal uid =
		(case findHash(treeHash, uid) of
		     SOME s => (insertHashDestructive(hash, uid, s); ())
		   | NONE => raise SympBug
		      (funName^": subgoal UID from proof run is not in the tree: "
		       ^(subgoalUID2string uid)^"\n\n Proof run: \n\n"
		       ^(ProofRun2stringDebug proofRun)))
	    val _ = List.app addSubgoal (hash2any #1 runHash)
	    val res = sortList subgoalLess (hash2any #2 hash)
	    val _ = popFunStackLazy(funName, fn()=>
				    "["^(strlist2str ", " (List.map subgoal2string res))^"]")
	in
	    res
	end

    (* Handling the subgoal's IDs *)
    local val pid = ref 0
    in 
	
	(* Generate a new unique proof ID based on the old one *)
	fun newProofTreeID s = 
	    ((option2string "" s)^"<"^(Int.toString(!pid))^">") before (pid := (!pid) + 1)
	(* Generate an ID for a new subgoal *)
	fun initID s = (s, [1])
    end

    (* Find the next unproven subgoal after the given subgoal, if
       given.  Currently, we get the entire list of the unproven
       subgoals, and if the given subgoal is also unproven (and thus,
       occur in the list), get the next subgoal after it in the list
       order (or the first one, if the current subgoal is the last in
       the list).  In any other case, just give the first subgoal in
       the list if it's not empty. *)
    fun nextUnprovenSubgoal(proofRun, sub: Subgoal option) = 
	let val funName = "nextUnprovenSubgoal"
	    val _ = pushFunStackLazy(funName, fn()=> 
				     option2string "NONE" (Option.map subgoal2string sub))
	    fun nextSub hd [] = hd
	      | nextSub _ (hd::_) = hd
	    fun loop (hd,s) lst =
		let val _ = pushFunStackLazy
		    (funName^"/loop",
		     fn()=> "hd="^(subgoal2string hd)
		     ^", s="^(subgoal2string s)
		     ^", lst=["^(strlist2str "," (List.map subgoal2string lst))^"]")
		    val res = loop' (hd, s) lst
		    val _ = popFunStackLazy(funName^"/loop", fn()=> subgoal2string res)
		in res
		end
	    and loop' (hd,s) [] = hd
	      | loop' (hd,s) (s'::lst) = 
	          if SubgoalEq(s, s') then nextSub hd lst
		  else loop (hd,s) lst
	    val res = (case (getUnprovenSubgoals proofRun, sub) of
			   ([],_) => NONE
			 | (hd::lst, NONE) => SOME hd
			 | (hd::lst, SOME s) =>
			       if SubgoalEq(s, hd) then SOME(nextSub hd lst)
			       else SOME(loop (hd,s) lst))
	    val _ = popFunStackLazy(funName, fn()=> 
				    option2string "NONE" (Option.map subgoal2string res))
	in res
	end

    (* Generate new subgoal *)
    local 
	val nextUID = ref 0
	fun newUID() = (!nextUID) before (nextUID := (!nextUID) + 1)
    in
	fun makeSubgoal(proofTreeIDopt) =
	    Subgoal{ id=ref(initID(case proofTreeIDopt of
				       NONE => newProofTreeID NONE
				     | SOME s => s)),
		     extend = ref false,
		     uid=newUID(),
		     proofRule = ref(NONE),
		     children = ref([]),
		     parents=ref [],
		     prooftree=ref(NONE) }
    end

    fun copySubgoal pid (Subgoal{id=ref (_,id),
				 extend=ref ex,
				 proofRule=proofRule,
				 children=ch,
				 parents=parents,
				 prooftree=tree, ... }) =
	let val s as Subgoal{ id=newID,
			      extend=newExtend,
			      proofRule=newProofRule,
			      children=newCh,
			      parents=newP,
			      prooftree=newProofTree, ...} = makeSubgoal(SOME pid)
	in (newID :=(pid, id);
	    newExtend := ex;
	    newCh := !ch;
	    newP := !parents;
	    newProofRule := !proofRule;
	    newProofTree := !tree;
	    s)
	end

    fun makeProofTree pid =
	let val ptref = ref(NONE)
	    val newPid = 
		 (case pid of
		      NONE => newProofTreeID NONE
		    | SOME s => s)
	    val s as Subgoal{ id=id, prooftree=ptref, ...} = makeSubgoal(SOME newPid)
	    val hash = makeHashDefault(op =, subgoalUID2string)
	    val incomplete = makeHashDefault(op =, subgoalUID2string)
	    val pt = ProofTree{root=s, hash=hash,
			       incomplete=incomplete}
	in (id := initID newPid;
	    ptref := SOME(pt);
	    insertHashDestructive(hash,getSubgoalUID s, s);
	    insertHashDestructive(incomplete,getSubgoalUID s, s);
	    checkProofTree pt;
	    pt)
	end

    (* Build a copy of the proof subtree rooted at `s' (if there is one),
        and give it an optional `pid' *)
    fun extractSubtree pid (s as Subgoal{proofRule=ref(SOME rule), ...}) =
	let val funName = "extractSubtree"
	    val _ = pushFunStackLazy(funName,
				     fn()=>"pid="^(case pid of SOME s => s | NONE => "NONE")
				     ^", "^(subgoal2string s))
	    val ch = getChildrenHints s
	    val hash = makeHashDefault(op =, subgoalUID2string)
	    val incomplete = makeHashDefault(op =, subgoalUID2string)
	    val idHash = makeHashDefault(op =, subgoalID2string)
	    val oldPid = Option.map getProofTreeID (getProofTree s)
	    val newPid = 
		 (case pid of
		      NONE => newProofTreeID oldPid
		    | SOME s => s)
	    val newS = copySubgoal newPid s
	    val Subgoal{id=ref id, uid=uid, children=newCh, proofRule=newRule,
			parents=newP, prooftree=newTree, ...} = newS
	    val tree = ProofTree{root=newS, hash=hash, incomplete=incomplete}
	    fun loop parent (s, hints: SequentHints option) =
		let fun hOpt2str NONE = "NONE"
		      | hOpt2str (SOME h) = hints2string h
		    val _ = pushFunStackLazy(funName^"/loop", 
					     fn()=>(subgoal2string s)
					     ^", "^(hOpt2str hints))
		    val res as (uid, hints) = loop' parent (s, hints)
		    (* val _ = checkProofTree tree *)
		    val _ = popFunStackLazy(funName^"/loop",
					    fn()=>"("^(subgoalUID2string uid)
					    ^", hints="^(hOpt2str hints)^")")
		in res
		end
	    and loop' parent (s, hints: SequentHints option) =
		let val (uid, id as (_, idNum), ch) =
		           (getSubgoalUID s, getSubgoalID s, getChildrenHints s)
		    val newID = (newPid, idNum)
		    val Subgoal{children=parentCh, ...} = parent
		    val newSub as Subgoal{uid=newUID, ...} =
			(case findHash(idHash, newID) of
			     SOME newS =>
				 let val Subgoal{parents=p, ...} = newS
				 in (p := ((getSubgoalUID parent)::(!p));
				     newS)
				 end
			   | NONE => 
				 let val newS = copySubgoal newPid s
				     val Subgoal{children=newCh,
						 uid=uid,
						 parents=newP,
						 prooftree=newTree, ...} = newS
				     val id = getSubgoalID newS
				 in (newP := [getSubgoalUID parent]; newTree := SOME tree;
				     insertHashDestructive(hash,uid,newS);
				     insertHashDestructive(idHash,id,newS);
				     newCh := List.map(loop newS) ch;
				     newS)
				 end)
		in (parentCh := (!parentCh)@[(newUID, hints)];
		    (newUID, hints))
		end
	in (newP := []; newCh := []; newTree := SOME tree;
	    insertHashDestructive(hash,uid,newS);
	    insertHashDestructive(idHash, getSubgoalID newS, newS);
	    updateProofTreeIncomplete tree;
	    checkProofTree tree;
	    newCh := List.map(loop newS) ch;
	    popFunStackLazy(funName, fn()=> ProofTree2string tree);
	    SOME tree)
	end
      (* There is no proof tree at this subgoal *)
      | extractSubtree _ (Subgoal{proofRule=ref NONE, ...}) = NONE

    (* Copy the entire proof tree and give it an optional `pid' *)
    fun copyProofTree pid tree = 
	let val _ = pushFunStackLazy("copyProofTree", fn()=>ProofTree2string tree)
	    val res = (case extractSubtree pid (getRoot tree) of
			   SOME t => t
			 | NONE => makeProofTree pid)
	    val _ = popFunStackLazy("copyProofTree", fn()=>ProofTree2string res)
	in res
	end

    fun makeProofRun pt =
	let val prooftree as ProofTree{hash=hash, ...} =
	          copyProofTree (SOME(getProofTreeID pt)) pt
	    val incomplete = makeHashDefault(op =, subgoalUID2string)
	    val _ = List.app(fn s => (insertHashDestructive(incomplete, s, ()); ()))
		            (hash2any #1 hash)
	in
	    
	    { prooftree=prooftree,
	      assocHash=makeHashDefault(op =, subgoalUID2string),
	      incomplete = incomplete } : ProofRun
	end

    (* Look at the list of children of `parent' and figure out the new
       name of the subgoal when `parent' becomes its true parent.  The
       new id is guaranteed to be unique in the tree.
       Return (newID, extendFlag) for the child.  No side effects. *)

    fun computeNewID parent (id,uid) =
	let val funName = "computeNewID"
	    val Subgoal{id=ref parID, 
			extend=ref extend,
			children=ref ch,...} = parent
	    val tree = (case getProofTree parent of
			    SOME t => t
			  | NONE => raise SympBug
				(funName^": parent subgoal doesn't have a proof tree: "
				 ^(subgoal2string parent)))
	    val (pid, sid) = parID
	    fun loop [] _ = raise SympBug
		(funName^": new parent "^(subgoal2string parent)
		 ^" doesn't have the subgoal "^(subgoalID2string id)
		 ^"["^(subgoalUID2string uid)^"]"
		 ^" in its list of children")
	      | loop ((uid',_)::lst) n =
		if uid = uid' then (pid, sid@[n]) else loop lst (n+1)
	    (* Increment the last integer in the list and return the new list *)
	    fun incLast [n] = [n+1]
	      | incLast (x::lst) = x::(incLast lst)
	      | incLast [] = raise SympBug("changeParent/incLast: the SID list is empty")
	    (* Make sure the new name of the subgoal is unique,
	       unless the new name is the same as the old one. *)
	    fun makeUnique (pid, newID) =
		let val (_, oldID) = id
		    fun loop id' = (case ID2Subgoal tree (pid, id') of
					SOME _ => loop(incLast id')
				      | NONE => id')
		in
		    if oldID = newID then (pid, newID)
		    else (pid, loop newID)
		end
	in
	    case ch of
		lst as [(uid',_)] => 
		    if uid = uid' then
			if extend then (makeUnique(loop lst 1), false)
			else (makeUnique(pid, incLast sid), false)
		    else raise SympBug
			("changeParent: new parent "^(subgoal2string parent)
			 ^" doesn't have the subgoal "^(subgoalID2string id)
			 ^"["^(subgoalUID2string uid)^"]"
			 ^" in its list of children (the only child ["
			 ^(subgoalUID2string uid')^"] differs)")
	      | lst => (makeUnique(loop lst 1), true)
	end

    (* Change the true parent of the subgoal s to `newParent'.
       Destructively change the IDs of the subgoal `s' and its
       descendants, and update the proof tree of `s' (not the
       newParent's tree).

       This might potentially invalidate the set of incomplete subgoals. *)

    fun changeParent (newParentUID: SubgoalUID) s =
	let val Subgoal{prooftree=ref prooftreeOpt, ...} = s
	    val proofTree as ProofTree{hash=hash, ...} =
		  (case prooftreeOpt of
		       SOME pt => pt
		     | NONE => raise SympBug
			   ("changeParent: the subgoal "^(subgoal2string s)
			    ^" doesn't have a proof tree"))
	    val visited = makeHashDefault(op =, subgoalUID2string)
	    val newParent = (case UID2Subgoal proofTree newParentUID of
				 SOME p => p
			       | NONE => raise SympBug
				   ("changeParent: the new parent ["
				    ^(subgoalUID2string newParentUID)
				    ^"] not found"))
	    val (newPid, _) = getSubgoalID newParent
	    val (oldPid, _) = getSubgoalID s
	    val updateAll = not(oldPid = newPid)

	    (* updateID(oldID, newParentUID) id
	       If `id' matches the `oldID' in SID, rename it to be a child of `newParent',
	       and update PID if needed. Return NONE otherwise. *)
	    fun updateID((_,oldSid),newParentUID) (id as (_, sid), uid) = 
		let val parent = (case findHash(hash,newParentUID) of
				      SOME p => p
				    | NONE => raise SympBug
					  ("changeParent/updateID: parent not found: "
					   ^(subgoalUID2string newParentUID)))
		    val ((newPid,newSid),newExtend) = computeNewID parent (id,uid)
		in 
		    if oldSid = sid then SOME((newPid, newSid), newExtend)
		    else if updateAll then SOME((newPid, sid), newExtend)
		    else NONE
		end

	    fun updateSubgoal (oldID, parentUID) (s as Subgoal{id=id, uid=uid,
							       extend=exRef, ...}) =
	        Option.map(fn (newID, newExtend) =>
			   (id := newID;
			    exRef := newExtend;
			    newID)) (updateID (oldID, parentUID) (!id, uid))
	    (* The main loop; p is the current new parent for s *)
	    fun loop p s =
		(case findHash(visited, getSubgoalUID s) of
		     SOME _ => ()
		   | NONE => 
			 let val Subgoal{id=id, uid=uid,
					 children=chRef,
					 parents=parentsRef, ...} = s
			     val myOldID = !id
			     val newIdOpt = updateSubgoal (myOldID, p) s
			     (* Insert the *new* ID to the set of visited IDs *)
			     val _ = insertHashDestructive(visited, uid, s)
			 in
			     case newIdOpt of
				 (* If it didn't change, do nothing further *)
				 NONE => ()
			       (* Recurse on children with `s' as a new parent *)
			       | SOME newID => List.app (loop uid) (getChildren s)
			 end)
	in loop newParentUID s
	end

    (* Delete the parent from the subgoal's list of parents.
       The parent must be present.

       This might invalidate the set of incomplete subgoals. *)
    fun delParent parent (s as Subgoal{parents=plRef as ref pUIDlist,...}) =
	let val funName = "delParent"
	    val pl = getParents s
	    val _ = pushFunStackLazy(funName, fn()=>
				     (subgoal2stringDebug parent)^", "
				     ^(subgoal2stringDebug s)^", parents=["
				     ^(strlist2str ", " (List.map subgoalUID2string pUIDlist))
				     ^"], pl=["^(strlist2str ", " 
						 (List.map subgoal2stringDebug pl))^"]")
	    fun loop [] = raise SympBug
	           (funName^": parent subgoal is not "
		    ^"in the parents list:\n  "
		    ^"Parent: "^(subgoal2string parent)
		    ^"\n  Subgoal: "^(subgoal2string s))
	      | loop (s1::tl) =
		   (if SubgoalEq(parent,s1) then tl
		    else s1::(loop tl))
	in 
	    (plRef := List.map getSubgoalUID (loop pl);
	     popFunStack(funName, ""); ())
	end

    (* Garbage-collect the subgoal and all its children that apply.
       When removing the true parent of a child, renumber the child's
       subtree to be a descendant of some other remaining parent.

       This might invalidate the set of incomplete subgoals. *)
    fun garbageCollect s =
	let val funName = "garbageCollect"
	    val Subgoal{prooftree=ref proofTreeOpt, ...} = s
	    fun ptOpt2str ptOpt =
		(option2string "NONE"
		 (Option.map (fn p=>"[\n"^(ProofTree2stringDebug p)^"]") ptOpt))
	    fun argsFn(Subgoal{uid=uid, parents=ref plist, prooftree=ref ptOpt, ...}) () =
		("Subgoal{uid="^(subgoalUID2string uid)
		 ^",\n   parents=["^(strlist2str ", " (List.map subgoalUID2string plist))
		 ^"],\n   prooftree="^(ptOpt2str ptOpt)^"}")
	    val _ = pushFunStackLazy(funName, argsFn s)
	    fun loop s =
		let val funName = "garbageCollect/loop"
		    val _ = pushFunStackLazy(funName, argsFn s)
		    val res = loop' s
		    val _ = popFunStackLazy(funName, fn()=>ptOpt2str proofTreeOpt)
		in res
		end
	    (* If the parent list is empty, remove the subgoal. *)
	    and loop' (s as Subgoal{uid=uid, parents=ref [],
						 prooftree=ref(SOME pt), ...}) =
		let val ProofTree{hash=hash,...} = pt
		    val ch = getChildren s
		in (List.app (delParent s) ch;
		    removeHashDestructive(hash,uid);
		    List.app loop ch)
		end
	      (* If the parent list is non-empty, renumber the subgoal and its
	         subtree, if necessary, to remove all the traces of the old subgoal. *)
	      | loop' (s as Subgoal{id=idRef as ref id,
				   uid=uid,
				   parents=ref(plist as newParentUID::_),
				   prooftree=ref(SOME pt), ...}) = 
		let val (pid, sid) = id
		    fun getID uid = 
			  (case UID2Subgoal pt uid of
			       SOME s => getSubgoalID s
			     | NONE => raise SympBug
				   ("garbageCollect/getID: subgoal ["
				    ^(subgoalUID2string uid)^"] not found"))
		in
		    (* If the true parent is present, do nothing further *)
		    if List.exists(isDescendantUID pt sid) plist then ()
		    (* Otherwise change the parent *)
		    else changeParent newParentUID s
		end
	      (* If there is no proof tree, don't garbage collect *)
	      | loop' _ = ()
	in 
	    (loop s;
	     popFunStackLazy(funName, fn()=>ptOpt2str proofTreeOpt); ())
	end

    (* Delete the subtree originating at the subgoal, but leave the subgoal. *)
    fun deleteSubtree(s as Subgoal{children=chRef, proofRule=rule, ...}) =
	let val funName = "deleteSubtree"
	    val _ = pushFunStackLazy(funName, fn()=>(subgoal2string s)^",\n Proof Tree:\n"
				     ^(case getProofTree s of
					   SOME t => ProofTree2stringDebug t
					 | NONE => "   NONE"))
	    val ch = getChildren s
	in (List.app (delParent s) ch;
	    List.app garbageCollect ch;
	    rule := NONE; chRef := [];
	    Option.map updateProofTreeIncomplete (getProofTree s);
	    popFunStackLazy(funName, fn()=>"New Proof Tree:\n"
			    ^(case getProofTree s of
				  SOME t => ProofTree2stringDebug t
				| NONE => "   NONE")))
	end

    (* Check if the parent has the subgoal as its predecessor *)
    fun checkCycle parent subgoal =
	let val hash = makeHashDefault(op =, subgoalUID2string)
	    fun loop s =
		  if isSome(findHash(hash,getSubgoalUID s)) then false
		  else if SubgoalEq(subgoal,s) then true
		       else (insertHashDestructive(hash,getSubgoalUID s, s);
			     List.exists loop (getParents s))
	in loop parent
	end
		  
    (* Verify that the subgoal is from the same proof tree, and only
       then associate the sequent with it.  Add it to the incomplete
       set, since the sequent is assumed unproven. *)

    fun addSequent (proofrun as {prooftree=pt,
				 assocHash=hash,
				 incomplete=incomplete, ...}: ProofRun) (s, seq) =
	((case getProofTree s of
	      NONE => raise ProofTreeError
		  ("Subgoal doesn't have a proof tree: "^(subgoal2string s))
	    | SOME p => if (getProofTreeUID p) = (getProofTreeUID pt) then ()
			else raise ProofTreeError
			    ("Subgoal is not from the current proof tree: "
			     ^(subgoal2string s)));
	 lazyVerbDebug (getOptions()) "addSequent"
	   (fn()=>"addSequent: subgoal "
	    ^(subgoal2stringDebug s)
	    ^" is assigned sequent:\n"
	    ^(sequent2string seq)^"\n");
	 insertHashDestructive(hash, getSubgoalUID s, (seq, false));
	 insertHashDestructive(incomplete, getSubgoalUID s, ()); ())

    (* Mark the rule's flag with the `applied' bit for the subgoal in the proof run.
       Remove the subgoal from the incomplete set. *)
    fun setRuleAppliedFlag(proofrun as {prooftree=pt, assocHash=hash,
					incomplete=incomplete}: ProofRun) (s, applied) =
	let val _ = (case getProofTree s of
			 NONE => raise ProofTreeError
			     ("Subgoal doesn't have a proof tree: "^(subgoal2string s))
		       | SOME p => if (getProofTreeUID p) = (getProofTreeUID pt) then
			     if isSome(getProofRule s) then ()
			     else raise ProofTreeError("Subgoal doesn't have a rule: "
						       ^(subgoal2string s))
				   else raise ProofTreeError
				       ("Subgoal is not from the current proof tree: "
					^(subgoal2string s)))
	    val seq = (case findHash(hash, getSubgoalUID s) of
			   SOME (seq, _) => seq
			 | NONE => raise ProofTreeError
			       ("Subgoal doesn't have an associated sequent: "
				^(subgoal2string s)))
	in (insertHashDestructive(hash, getSubgoalUID s, (seq, applied));
	    removeHashDestructive(incomplete, getSubgoalUID s); ())
	end
	      

    (* Install a clean COPY of the ProofTree (if `copy' is true) at
       the Subgoal in the current proof tree.  The subgoal replaces
       the root subgoal of the new ProofTree.  The old subtree rooted
       at the subgoal is deleted.  It is an error for the subgoal to
       not have a proof tree. *)

    fun installProofTree copy (s, tree) =
	let val funName = "installProofTree"
	    val debug = lazyVerbDebug (getOptions()) funName
	    val Subgoal{id=ref id,
			parents=parents,
			proofRule=oldProofRuleRef,
			children = children, ...} = s
	    val oldTree as ProofTree{hash=hash,...} = 
		  (case getProofTree s of
		       SOME t => t
		     | NONE => raise SympBug
			   ("\ninstallProofTree: the subgoal doesn't have a proof tree: "
			    ^(subgoal2string s)))
	    val _ = pushFunStackLazy(funName,
				     fn()=>(subgoal2stringDebug s)
				     ^"\n Entire old tree:\n"
				     ^(ProofTree2stringDebug oldTree)
				     ^"\n New subtree:\n"
				     ^(ProofTree2stringDebug tree))
	    val oldCh = getChildren s
			
	    (* Proof tree ID and Subgoal ID of `s' *)
	    val (pid, sid) = id
	    (* The true parent of `s', if any *)
	    fun findTrueParent s = 
		let val Subgoal{parents=ref parents,
				id=ref(pid,sid),
				prooftree=tree, ...} = s
		in (Option.map(fn uid=> 
			       (case UID2Subgoal oldTree uid of
				    SOME p => p
				  | NONE => raise SympBug
					("installProofTree: parent ["
					 ^(subgoalUID2string uid)^"] of s = "
					 ^(subgoal2stringDebug s)^" not found in the proof tree")))
		    (List.find(isDescendantUID oldTree sid) parents))
		end
	    (* The new subgoal replacing `s' contents *)
	    val rootProofTree = (if copy then copyProofTree (SOME pid) tree 
				 else tree)
	    val root = getRoot rootProofTree
	    val Subgoal{children=ref rootCh,
			proofRule= ref newProofRule, ...} = root
	    (* The list of children subgoals (not IDs) *)
	    val rootChildren = getChildren root
	    (* Add all the new subgoals to the proof tree hash *)
	    fun addToHash s =
		if isSome(findHash(hash,getSubgoalUID s)) then ()
		else let val ch = getChildren s
			 val Subgoal{prooftree=pt,...} = s
		     in (insertHashDestructive(hash,getSubgoalUID s, s);
			 pt := SOME oldTree;
			 List.app addToHash ch)
		     end
	    (* Replace parent with oldUID by newUID *)
	    fun replaceParent(oldUID, newUID) (Subgoal{parents=p,...}) =
		let fun loop [] = raise SympBug
		          ("insatllProofTree/replaceParent: parent ["
			   ^(subgoalUID2string oldUID)^"] is not in the list:\n ["
			   ^(strlist2str "," (List.map subgoalUID2string (!p)))^"]")
		      | loop (uid::lst) =
			  if uid = oldUID then newUID::lst
			  else uid::(loop lst)
		in p := loop(!p)
		end
	    fun changeChildsParent newParent child =
		let val parentUID = getSubgoalUID newParent
		in
		    case findTrueParent child of
			SOME _ => () (* It's not our child, leave it alone *)
		      | NONE => changeParent parentUID child
		end
	in ((* First, fix the numbering in the new subtree.
	       If `s' is a root, then no renumbering is needed. *)
	    (* Option_app changeRootParent trueParentOpt; *)
	    debug(fn()=>" New tree renumbered:\n"
		  ^(ProofTree2stringDebug tree)^"\n");
	    (* Detach children of the old subgoal *)
	    List.app (delParent s) oldCh;
	    (* And garbage collect.  This should also renumber
	       children to detach them from this subtree completely *)
	    List.app garbageCollect oldCh;
	    debug(fn()=>" Old tree after removing the old subtree and GC'ing:\n"
		  ^(ProofTree2stringDebug oldTree)^"\n");
	    (* Copy the new root configuration to the old root (the
	       old root remains - we can't always replace it) *)
	    children := rootCh;
	    oldProofRuleRef := newProofRule;
	    List.app(replaceParent(getSubgoalUID root, getSubgoalUID s)) rootChildren;
	    (* And add its children to the tree hash recursively *)
	    List.app addToHash rootChildren;
	    (* Now renumber the new descendants *)
	    List.app (changeChildsParent s) rootChildren;
	    updateProofTreeIncomplete oldTree;
	    checkProofTree oldTree;
	    popFunStackLazy(funName,fn()=> ProofTree2stringDebug oldTree))
	end

    (* Add a ProofRule to the Subgoal that produces a Subgoal list.
       The existing subtree rooted at the Subgoal is deleted. Subgoals
       are NOT COPIED, but rather MODIFIED IN PLACE, so you may later
       safely associate them with sequents.

       The new subgoals must either belong to the current proof tree,
       or be completely new. *)

    fun addProofRule(s,proofRule,lst) =
	let val funName = "addProofRule"
	    val debug = lazyVerbDebug (getOptions()) funName
	    val _ = pushFunStackLazy(funName,
				     fn()=>(subgoal2stringDebug s)
				     ^", "^(#1 proofRule)^", ["
				     ^(strlist2str "," (List.map (subgoal2stringDebug o #1) lst))
				     ^"]")
	    val tree as ProofTree{ hash=hash, incomplete=incomplete, ...} = 
		(case getProofTree s of
		     SOME x => x
		   | NONE => raise ProofTreeError
			 ("Base subgoal "^(subgoal2string s)^"doesn't have a proof tree"))
	    val pid = getProofTreeID tree
	    val pUid = getProofTreeUID tree
	    val (_,sid) = getSubgoalID s
	    val extend = getSubgoalExtend s
	    fun wrongPid (ss,_) = 
		 (case Option.map(fn pt => pUid = (getProofTreeUID pt)) (getProofTree ss) of
		      SOME false => true
		    | _ => false)
	    val _ = (case List.find wrongPid lst of
			 SOME (ss,_) => raise ProofTreeError
			     ("Can't add subgoal "^(subgoal2string ss)
			      ^" from a different proof tree")
		       | _ => ())
	    (* Increment the last integer in the list and return the new list *)
	    fun incLast [n] = [n+1]
	      | incLast (x::lst) = x::(incLast lst)
	      | incLast [] = raise SympBug
		("addProofRule/incLast: the SID list is empty")
	    (* Renumber the new subgoals (without a proof tree) to be proper children of `s',
	       and add them to the tree hash.
	       Assume that `lst' are the only children of the parent. *)
	    fun renumber(hash,parentUID, parentID, proofTree) lst =
		let fun renameSub(s, sid, ex) =
		         let val Subgoal{id=id, uid=uid,
					 extend=extendRef,
					 parents=parents,
					 prooftree=pt, ...} = s
			 in  (* Only rename if the subgoal is brand new *)
			     case getProofTree s of
				 NONE => (id := (pid,sid);
					  extendRef := ex;
					  debug(fn()=>"addProofRule/renumber: renaming subgoal to "
						^(subgoalID2string(!id))^"\n");
					  parents := [parentUID];
					  pt := SOME proofTree;
					  insertHashDestructive(hash, uid, s); ())
			       | SOME _ => ()
			 end
		    fun loop _ [] _ = ()
		      | loop ex ((ch,_)::lst) n =
			  (renameSub(ch, sid@[n], ex);
			   loop ex lst (n+1))
		in
		    case lst of 
			lst as [(ch,_)] => 
			    if extend then loop false lst 1
			    else renameSub(ch, incLast sid, false)
		      | _ => loop true lst 1
		end
	    (* Detatch current children, if any, and rebuild the tree info if necessary *)
	    fun detachChildren [] = ()
	      | detachChildren oldCh =
		(List.app (delParent s) oldCh;
		 (* And garbage collect.  This should also renumber
		    children to detach them from this subtree completely *)
		 List.app garbageCollect oldCh;
		 updateProofTreeIncomplete tree; ())

	    fun addIncomplete(s as Subgoal{proofRule=ref NONE, ...}, _) =
		  (insertHashDestructive(incomplete, getSubgoalUID s, s); ())
	      | addIncomplete _ = ()		

	    (* Actually add the damn rule and update the tree info. *)
	    fun addRule(s, proofRule, lst) =
		let val _ = detachChildren(getChildren s)
		    val Subgoal{id=ref id, uid=uid,
				children=ch,
				proofRule=r, ...} = s
		in (r := SOME proofRule;
		    renumber(hash, uid, id, tree) lst;
		    ch := List.map (fn(s,h)=>(getSubgoalUID s,h)) lst;
		    removeHashDestructive(incomplete, uid);
		    List.app addIncomplete lst)
		end
	in 
	    ((case List.find (fn (x,_)=>checkCycle s x) lst of
		  SOME (ss,_) => raise ProofTreeError
		      ("Can't add the subgoal "^(subgoal2string ss)
		       ^" because it creates a cycle in the proof")
		| NONE => addRule(s, proofRule, lst)
		      (* (* Build a new proof tree and install it *)
		      let val pt as ProofTree{hash=hash,...} = makeProofTree(SOME pid)
			  val Subgoal{id=ref id, uid=uid,
				      children=ch,
				      proofRule=r, ...} = getRoot pt
		      in (r := SOME proofRule;
			  renumber(hash, uid, id, pt) lst;
			  ch := List.map (fn(s,h)=>(getSubgoalUID s,h)) lst;
			  installProofTree false (s,pt))
		      end *) );
	     checkProofTree tree;
	     popFunStackLazy(funName,fn()=>"new proof tree:\n  "
			     ^(ProofTree2stringDebug tree)))
	end

    (* Add proof rule producing the list of sequent/hint pairs to the proof run.

       This is the only EXPORTED function that builds a proof tree
       structure.  You cannot change the inference rule or the subgoal
       without changing the entire subtree, nor you can change any
       information at the subgoal. *)
    fun addProofRuleSequents proofRun (subgoal, proofRule, seqs) =
	let val funName = "addProofRuleSequents"
	    val _ = pushFunStackLazy(funName, fn()=>ProofRun2stringDebug proofRun)
	    val debug = lazyVerbDebug(getOptions()) funName
	    val proofTreeID = getProofTreeID(ProofRun2ProofTree proofRun)
	    (* Create a new subgoal for each sequent *)
	    fun newSubgoal(seq, hints) = 
		let val subgoal = makeSubgoal(SOME proofTreeID)
		in 
		    (subgoal, seq, hints)
		end
	    val _ = debug(fn()=>funName^": generating new subgoals...")
	    val triples = List.map newSubgoal seqs
	    val _ = debug(fn()=>"done\n")
	    fun triple2string(sub, seq, hints) =
		"("^(subgoal2stringDebug sub)
		^", sequent=[\n"^(sequent2string seq)
		^"]"
		^(case hints of
		      SOME h => ", hints=\""^(hints2string h)^"\""
		    | NONE => "")
		^")"
	    val subgoals = List.map(fn (s,_,h)=>(s,h)) triples
	    val _ = debug(fn()=>"\ninstallRule: triples:\n"
			  ^(strlist2str "\n" (List.map triple2string triples))
			  ^"\n")
	in ((* First add the subgoals to the proof tree *)
	    addProofRule(subgoal, proofRule, subgoals);
	    setRuleAppliedFlag proofRun (subgoal, true);
	    List.app(fn (sub,seq,_)=>addSequent proofRun (sub, seq)) triples;
	    setRuleAppliedFlag proofRun (subgoal, true);
	    popFunStackLazy(funName, fn()=>ProofRun2stringDebug proofRun);
	    subgoals)
	end

    (* Make the subtree rooted at the child (including the child) be
       the parent's own subtree, so that any destructive modifications
       to that subtree would not affect any other part of the tree.
       Return the new child.
       *)

    fun separateSubtree (parent, child) =
	let val funName = "separateSubtree"
	    val _ = pushFunStackLazy(funName, fn()=>(subgoal2stringDebug parent)^", "
				     ^(subgoal2stringDebug child))
	    val Subgoal{children=chRef,
			extend=ref parentExtend, ...} = parent
	    val tree = (case getProofTree parent of
			    SOME t => t
			  | NONE => raise SympBug
				(funName^": parent has no proof tree: "
				 ^(subgoal2stringDebug parent)))
	    val ProofTree{hash=treeHash, ...} = tree
	    val childUID = getSubgoalUID child
	    val subtree = (case extractSubtree NONE child of
			       SOME t => t
			     | NONE => raise SympBug
				(funName^": child has no proof tree: "
				 ^(subgoal2stringDebug child)))
	    val (newID, childExtend) = computeNewID parent (getSubgoalID child, childUID)
	    fun replaceChild(lst, acc) =
		let val funName = funName^"/replaceChild"
		    val _ = pushFunStackLazy
			(funName, fn()=>
			 ("["^(strlist2str ", "
			       (List.map (fn (uid,_)=>subgoalUID2string uid) lst))^"], ["
			  ^(strlist2str ", " (List.map (fn (uid,_)=>subgoalUID2string uid) acc))
			  ^"]"))
		    val res = replaceChild'(lst, acc)
		    val _ = popFunStackLazy(funName, fn()=>(subgoal2stringDebug res))
		in res
		end
	    and replaceChild'([], _) = raise SympBug
		(funName^": no child in the parent's children list:\n  "
		 ^"parent = "^(subgoal2stringDebug parent)
		 ^"\n  child = "^(subgoal2stringDebug child))
	      | replaceChild'((pair as (ch, hints))::lst, acc) =
		 if ch = childUID then
		     let val new as Subgoal{id=idRef,
					    uid=newUID,
					    parents=pRef,
					    prooftree=ptRef,
					    extend=exRef, ...} = makeSubgoal NONE
			 val newChildren = (List.rev acc)@((newUID, hints)::lst)
	       
		     in (idRef := newID;
			 ptRef := SOME tree;
			 pRef := [getSubgoalUID parent];
			 exRef := childExtend;
			 (* Update children in the parent before calling external functions *)
			 chRef := newChildren;
			 insertHashDestructive(treeHash, newUID, new);
			 installProofTree false (new, subtree);
			 new)
		     end
		 else replaceChild(lst, pair::acc)
	in (delParent parent child;
	    garbageCollect child;
	    let val newChild = replaceChild(!chRef, [])
	    in (updateProofTreeIncomplete tree;
		checkProofTree tree;
		popFunStackLazy(funName, fn()=>(subgoal2stringDebug newChild));
		newChild)
	    end)
	end

    (* Proof tree I/O functions must be one-to-one, so that a proof
       tree could be completely restored from the string or the
       `UserInputArgument' value *)

    (* For more efficient communication with GUI/Emacs/whatever *)
    fun ProofTree2UI tree = 
	let val subgoals = listSubgoals tree
	    fun doSubgoal sub = 
		let val id = getSubgoalID sub
		    fun id2UI id = UIstring(subgoalID2string id)
		    fun rule2UI (r, args) = UIlist((UIstring r)::args)
		    val ruleUI = Option.map rule2UI (getProofRule sub)
		    fun child2UI (ch,hints) =
			let val id = id2UI(getSubgoalID ch)
			in
			    case hints of
				NONE => id
			      | SOME h => UIlist[id, UIstring(hints2string h)]
			end
		    val children = UIlist(List.map child2UI (getChildrenHints sub))
		    val ruleChildren = Option.map(fn r => [r, children]) ruleUI
		in
		    UIlist(case ruleUI of
			       NONE => [id2UI id]
			     | SOME r => [id2UI id, r, children])
		end
	in
	    UIlist(List.map doSubgoal subgoals)
	end

    (* And the reverse translation: build a proof tree from the
       `string' or `UserInputArgument'.  This, of course, may raise an
       exception if the input is invalid. *)

    structure ProofTreeStruct: PROOF_TREE_STRUCT =
      struct
	datatype ProofTreeType =
	    PTsymbol of string
	  | PTstring of string
	  | PTnumber of int
	  | PTlist of ProofTreeType list
	  (* For named rule's arguments *)
	  | PTassoc of string * ProofTreeType
	  (* The name of the rule and the list of arguments *)
	  | PTrule of string * (ProofTreeType list)
	  | PTsubgoalID of string * (int list)
	  (* PTchild(subgoalID, optional-hints) *)
	  | PTchild of ProofTreeType * (string option)
	  (* The entire subgoal entry with rule and children; the last two are optional *)
	  | PTsubgoal of ProofTreeType * (ProofTreeType option) * (ProofTreeType list option)
	  (* The entire proof *)
	  | PTproof of ProofTreeType list
      end

    open ProofTreeStruct

    fun ptt2str(PTsymbol s) = s
      | ptt2str(PTstring s) = "\""^s^"\""
      | ptt2str(PTnumber n) = Int.toString n
      | ptt2str(PTlist lst) = "["^(strlist2str ", " (List.map ptt2str lst))^"]"
      | ptt2str(PTassoc(n, t)) = "("^n^" = "^(ptt2str t)^")"
      | ptt2str(PTrule(n, lst)) = "rule("^n^", ["^(strlist2str "," (List.map ptt2str lst))^"])"
      | ptt2str(PTsubgoalID sub) = "subgoalID("^(subgoalID2string sub)^")"
      | ptt2str(PTchild(t, sOpt)) =
	 ("child("^(ptt2str t)
	  ^(option2string "" (Option.map(fn s=>", "^s) sOpt))
	  ^")")
      | ptt2str(PTsubgoal(t,tOpt,tlstOpt)) =
	 ("subgoal("
	  ^(ptt2str t)
	  ^(option2string "" (Option.map(fn t => ", "^(ptt2str t)) tOpt))
	  ^(option2string "" 
	    (Option.map(fn lst => ", ["^(strlist2str "," (List.map ptt2str lst))^"]")
	               tlstOpt))
	  ^")")
      | ptt2str(PTproof lst) = "proof[\n"^(strlist2str "\n" (List.map ptt2str lst))^"]\n"

    structure ParseError: PARSE_ERROR =
      struct
        exception ParseError of string 
	val wasError = ref false
      end

    structure ProofTreeLrVals =
	ProofTreeLrValsFun(structure Token = LrParser.Token
			     structure ProofTreeStruct = ProofTreeStruct
			     structure ParseError = ParseError)

    structure ProofTreeLex =
	ProofTreeLexFun(structure Tokens = ProofTreeLrVals.Tokens
			  structure ParseError = ParseError)

    structure ProofTreeParser =
	ProofTreeJoin(structure LrParser = LrParser
		 structure ProofTreeStruct = ProofTreeStruct
		 structure ParserData = ProofTreeLrVals.ParserData
		 structure Lex = ProofTreeLex)

    type result = ProofTreeLrVals.ParserData.result
    open ProofTreeLrVals.ParserData

    fun ProofTreeParse nameOpt ins =
	let open ParseError
	    val name = (case nameOpt of
			    SOME n => n^":"
			  | NONE => "")
	    val print_error = fn (s,l,c) =>
		(wasError := true;
		 raise ProofTreeError("\n"^name^(Int.toString l) ^"."
					^ (Int.toString c) ^ ": " ^ s ^ "\n"))
		  (* Use 15 tokens for error correction *)
	    val invoke = fn lexstream => 
		(ProofTreeParser.parse(15,lexstream,print_error,())
		 handle ProofTreeParser.ParseError => 
		     raise ProofTreeError("Fatal ProofTree parser error")
		      | ParseError str => raise ProofTreeError str)
	    val lexer = ProofTreeParser.makeLexer 
		           (fn (i: int) => TextIO.inputLine ins) print_error
	    val (result,_) = (ProofTreeLex.UserDeclarations.pos := 1;
			      ProofTreeLex.UserDeclarations.yypos_bol := 0;
			      wasError := false;
			      invoke lexer)
			      
	in result
	end

    fun ProofTreeType2ProofTree (ptt as PTproof(root::lst)) =
	let val funName = "ProofTreeType2ProofTree"
	    val _ = pushFunStackLazy(funName, fn()=>ptt2str ptt)
	    val ProofTree{hash=hash, root=oldRoot, incomplete=incomplete } = makeProofTree NONE
	    (* Remember what subgoals we've already created *)
	    val idHash = makeHashDefault(op =, subgoalID2string)
	    fun ptt2arg (PTsymbol s) = UIstring s
	      | ptt2arg (PTstring s) = UIstring s
	      | ptt2arg (PTnumber n) = UInumber n
	      | ptt2arg (PTlist lst) = UIlist(List.map ptt2arg lst)
	      | ptt2arg (PTassoc(s, x)) = UIassoc(UIstring s, ptt2arg x)
	      | ptt2arg x = raise SympBug
		   ("ProofTreeType2ProofTree/ptt2arg: bad rule argument:\n "
		    ^(ptt2str x))
	    fun ptt2rule (PTrule(name,args)) = (name, List.map ptt2arg args)
	      | ptt2rule x = raise SympBug
		  ("ProofTreeType2ProofTree/ptt2ptt2rule: bad rule spec:\n "
		   ^(ptt2str x))
	    (* Recreate the tree with the same hash but different root *)
	    fun newRoot sub = 
		let val tree = ProofTree{hash=hash, root=sub, incomplete=incomplete}
		    val Subgoal{prooftree=treeRef,...} = sub
		in (removeHashDestructive(hash, getSubgoalUID oldRoot);
		    insertHashDestructive(hash, getSubgoalUID sub, sub);
		    treeRef := SOME tree;
		    tree)
		end
	    fun ptt2subgoal tree (PTsubgoal(PTsubgoalID id, rule, children)) =
		let val s as Subgoal{id=idRef,
				     proofRule=ruleRef,
				     prooftree=ptRef, ... } =
			(case findHash(idHash, id) of
				 SOME _ => raise ProofTreeError
				     ("The proof has a multiply defined subgoal: "
				      ^(subgoalID2string id))
			       | NONE => let val s = makeSubgoal(NONE)
					     val _ = insertHashDestructive(idHash, id, s)
					 in s
					 end)
		in (idRef := id;
		    ruleRef := Option.map ptt2rule rule;
		    ptRef := tree;
		    s)
		end
	      | ptt2subgoal _ x = raise ProverError
		  ("Error in the proof text: bad subgoal entry: "^(ptt2str x))

	    val tree = newRoot(ptt2subgoal NONE root)
	    val _ = let val Subgoal{prooftree=ptRef, ...} = getRoot tree
		    in ptRef := SOME tree
		    end

	    fun insertSubgoal sub =
		let val subgoal = ptt2subgoal(SOME tree) sub
		in 
		    (insertHashDestructive(hash, getSubgoalUID subgoal, subgoal);
		     subgoal)
		end

	    fun ID2Sub id = (case findHash(idHash, id) of
				 SOME s => s
			       | NONE => raise SympBug
				   ("ProofTreeType2ProofTree/ID2Sub: "
				    ^"subgoal is not in idHash: "^(subgoalID2string id)))

	    fun ptt2children NONE = []
	      | ptt2children (SOME lst) =
		let val extend = (List.length lst) >= 2
		    fun ptt2child(PTchild(PTsubgoalID id, hints)) =
			let val ch as Subgoal{extend=ex, ...} = ID2Sub id
			in ((* If extend is already set, don't erase it *)
			    ex := ((!ex) orelse extend);
			    (getSubgoalUID ch, Option.map string2hints hints))
			end
		      | ptt2child x = raise SympBug
			  ("ProofTreeType2ProofTree/ptt2child: bad child spec:\n "
			   ^(ptt2str x))
		in List.map ptt2child lst
		end

	    (* After all subgoals are in the hash, fix the children/parent links *)
	    fun fixChildren(PTsubgoal(PTsubgoalID id, rule, children)) =
		let val s as Subgoal{children=chRef,...} = ID2Sub id
		in 
		    chRef := ptt2children children
		end
	      | fixChildren x = raise SympBug
		  ("ProofTreeType2ProofTree/fixChildren: wrong subgoal spec:\n "
		   ^(ptt2str x))

	    fun fixParents sub =
		let val ch = getChildren sub
		    val uid = getSubgoalUID sub
		    (* Add `id' to the list of parents of the given subgoal *)
		    fun addParent(Subgoal{parents=p,...}) = (p := uid::(!p))
		in List.app addParent ch
		end
	    (* Check for cycles in the proof tree (which should actually be a DAG).
	       Returns a cycle if found, or NONE. *)
	    fun findCycle tree =
		let val root = getRoot tree
		    val hash = makeHashDefault(op =, subgoalUID2string)
		    fun mark uid = insertHashDestructive(hash, uid, ())
		    fun inStack uid [] acc = NONE
		      | inStack uid (uid'::lst) acc =
			 if uid = uid' then SOME(uid::acc) else inStack uid lst (uid::acc)
		    fun loop stack sub =
			let val (uid, ch) = (getSubgoalUID sub, getChildren sub)
			in 
			   case inStack uid stack [] of
			       SOME lst => SOME lst
			     | NONE =>
				   (case findHash(hash, uid) of
					SOME _ => NONE
				      | NONE => (mark uid; looplist (uid::stack) ch))
			end
		    and looplist stack [] = NONE
		      | looplist stack (s::lst) =
			  (case loop stack s of
			       SOME x => SOME x
			     | NONE => looplist stack lst)
		in 
		    loop [] root
		end
	    fun getID tree uid = 
		(case UID2Subgoal tree uid of
		     SOME s => getSubgoalID s
		   | NONE => raise SympBug
			 ("ProofTreeType2ProofTree/getID: subgoal ["
			  ^(subgoalUID2string uid)^"] not found"))

	    (* See if each subgoal has its true parent (by the
	       numbering), and if not, renumber the relevant subgoals. *)
	    fun checkTrueParent sub =
		(case UID2Subgoal tree (getSubgoalUID sub) of
		     (* If the subgoal is still in the tree *)
		     SOME _ =>
			 let val parents = getParents sub
			     val (_, id) = getSubgoalID sub
			     fun isTrueParent p =
				 let val (_, parentID) = getSubgoalID p
				 in isDescendant id parentID
				 end
			 in 
			     (case parents of
				  (* no parents, this subgoal is garbage, unless it's the root *)
				  [] => if (getSubgoalUID sub) = (getSubgoalUID(getRoot tree))
					    then ()
					else garbageCollect sub
				| p::_ => if List.exists isTrueParent parents then ()
					  else changeParent (getSubgoalUID p) sub)
			 end
		     (* Subgoal has been garbage collected, forget it. *)
		   | NONE => ())
	    val subgoals = (getRoot tree)::(List.map insertSubgoal lst)
	    val _ = List.app fixChildren (root::lst)
	    val _ = List.app fixParents subgoals
	    val _ = List.app checkTrueParent subgoals
	    val res = 
		(case findCycle tree of
		     NONE => tree
		   | SOME lst =>  raise ProverError
			 ("The proof tree contains a cycle:\n  "
			  ^(strlist2str " -> " (List.map (subgoalID2string o (getID tree)) lst))))
	    val _ = updateProofTreeIncomplete res
	    val _ = checkProofTree res
	    val _ = popFunStackLazy(funName, fn()=>ProofTree2string res)	    
	in res
	end
      | ProofTreeType2ProofTree x = raise SympBug
	  ("ProofTreeType2ProofTree: not PTproof:\n "
	   ^(ptt2str x))

    fun stream2ProofTree nameOpt ins =
	let val ptt = (ProofTreeParse nameOpt ins)
	              handle ProofTreeError msg => raise ProverError msg
	in 
	    ProofTreeType2ProofTree ptt
	end

    fun string2ProofTree str = stream2ProofTree NONE (TextIO.openString str)

    fun file2ProofTree filename =
	let val funName = "file2ProofTree"
	    val _ = pushFunStack(funName, filename)
	    val ins = (TextIO.openIn filename)
	                handle _ => raise SympError("Cannot open input file: "^filename)
	    val res = stream2ProofTree (SOME filename) ins
	    val _ = popFunStackLazy(funName, fn()=> ProofTree2stringDebug res)
	in 
	    res
	end

    fun UI2ProofTree arg = raise SympBug("UI2ProofTree: not implemented")

  end
