(* Abstract types of sequents, inference rules, and strategies,
   together with functions for applying them.

   There is no file `sequent.sml', because each implementation is
   specific to the problem domain, and should be placed in a separate
   appropriate module. *)

signature SEQUENT =
  sig
    structure SequentCommon: SEQUENT_COMMON

    type Sequent = SequentCommon.Sequent
    type InferenceRule

    type ProofSystemCommand
    type ProofSystemCommandResult
    (* An abstract type for instantiating a rule.  It preserves some
       useful information obtained at the time of matching to be
       passed to the applyRule function. *)
    type Substitution

    (* Some hints of what a sequent should look like, so that subgoals
       in a proof tree can be matched better when the theorem changes.
       Currently, the hints are added only by the internal functions
       like rule application, etc. *)

    type SequentHints = SequentCommon.SequentHints

    (* The result of the application of an inference rule.  *)
    type InferenceRuleResult = SequentCommon.InferenceRuleResult
    type SequentPart = SequentCommon.SequentPart
    type InferenceRuleArgumentValue = SequentCommon.InferenceRuleArgumentValue
    type InferenceRuleArgument = SequentCommon.InferenceRuleArgument
    type InferenceRuleArgumentType = SequentCommon.InferenceRuleArgumentType

    type ProgramObject = SequentCommon.SequentBase.InputLanguage.ProgramObject
    type ProgramContext = SequentCommon.SequentBase.InputLanguage.ProgramContext

    type InferenceRuleArgumentSpec

    type UserInputArgument = SequentCommon.Interface.UserInputArgument

    val IRArgSpecName: InferenceRuleArgumentSpec -> string
    val IRArgSpecType: InferenceRuleArgumentSpec -> InferenceRuleArgumentType

    (* Create a new sequent from a theorem (which may include a model) *)
    val makeSequent: ProgramContext -> ProgramObject -> Sequent
    val sequentEq: Sequent * Sequent -> bool
    (* Check if the first sequent subsumes the second.
       That is, proving the first one also proves the second one.
       This function can be soundly approximated with sequentEq. *)
    val sequentSubsumes: Sequent * Sequent -> bool
(*      val getHints: Sequent -> SequentHints option *)
    val hintsEq: SequentHints * SequentHints -> bool

    val UIArg2FormulaIndex: UserInputArgument -> SequentPart * (int option)
    val FormulaIndex2UIArg: SequentPart * (int option) -> UserInputArgument

    (* Operations on inference rules *)

    val string2hints: string -> SequentHints

    (* Getting extra information about a sequent.  The type of request
       is specified by a string, the output is also a string which the
       user will see. *)

    val getSequentInfo: Sequent -> string -> string

  end
